/***
 * Copyright (c) 2005, SkeTo Project
 * All rights reserved.
 */
/**
 * @file list_skeletons.h
 * @brief The header for list skeletons.
 * @author Yoshiki AKASHI (yoshiki@ipl.cs.uec.ac.jp)
 * @author Kiminori MATSUZAKI (kmatsu@ipl.t.u-tokyo.ac.jp)
 */

#ifndef __LIST_SKELETONS_H__
#define __LIST_SKELETONS_H__

#include <iostream>
#include <utility>
#include "dist_list.h"

/**
 * list_skeletons are skeletons for dist_list.
 */
class list_skeletons
{
  //------------------------------------------------------------------------
  // Map

  /**
   * @brief Map skeleton applies the function to each element.
   *
   * map f [a0, a1, ..., an-1] = [ f a0, f a1, ... , f an-1 ].\n
   * This skeleton uses the function in usual function call style.
   *
   * @param f : The function to apply to elements. [ in ]
   * @param as : The input list to be applied by function f. [ in ]
   */
public:
  template< typename F, typename A >
  static dist_list< typename F::result_type > *
  map( const F &f,
       const dist_list< A > *as );

  /**
   * @brief Map skeleton applies the function to each element.
   *
   * map f [a0, a1, ..., an-1] = [ f a0, f a1, ... , f an-1 ].\n
   * This skeleton uses the function, which returns the results by
   * putting it where the pointer indicates.
   *
   * @param f : The function to apply to elements, and this accepts the pointer
   *            to the return value for the first argument, and the constant
   *            pointer to the input value for the second argument. [ in ]
   * @param as : The input list to be applied by function f. [ in ]
   * @param bs : The result list, and the area must be allocated
   *             before calling this function. [ in, out ]
   * @attention If @a bs is pointing the same dist_list object as @a this,
   *            the pointers passed to @a f will be the same. 
   */
public:
  template< typename F, typename A >
  static void map( const F &f,
		   const dist_list< A > *as,
		   dist_list< typename F::first_argument_type > *bs );
  /**
   * @brief Map skeleton applies the function to each element.
   *
   * map f [a0, a1, ..., an-1] = [ f a0, f a1, ... , f an-1 ].\n
   * This skeleton uses the function, which overwrites the results on
   * the inputs.
   *
   * @param f : The function to apply to elements. [ in ]
   * @param as : The input list to be applied by function f. [ in ]
   */
public:
  template< typename F, typename A >
  static void map_ow( const F &f,
		      dist_list< A > *as);

  //------------------------------------------------------------------------
  // Mapindex

  /**
   * @brief Mapindex skeleton applies the function to each element.
   *        The function takes the index of the element for an argument.
   *
   * mapindex f [a0, a1, ..., an-1] =
   *      [f 0 a0, f 1 a1, ..., f (n-1) an-1]
   * @param f : The function to apply to the list. [ in ]
   * @param as : The input list to be applied by function f. [ in ]
   */
public:
  template< typename F, typename A >
  static dist_list< typename F::result_type > *
  mapindex( const F &f,
	    const dist_list< A > *as );

  /**
   * @brief Mapindex skeleton applies the function to each element.
   *        The function takes the index of the element for an argument.
   *
   * mapindex f [a0, a1, ..., an-1] =
   *      [f 0 a0, f 1 a1, ..., f (n-1) an-1]
   * This function uses the function, which handles the pointers for the
   * arguments and the return value.
   *
   * @attention If @a retlist is same as @e this object, the two pointers
   *            passed for @a f will be same.
   * @param f : The function to apply to the list. [ in ]
   * @param as : The input list to be applied by function f. [ in ]
   * @param bs : The list for result, and must be allocated
   *                   before calling function. [ in, out ]
   */
public:
  template< typename F, typename A >
  static void mapindex( const F &f,
			const dist_list< A > *as,
			dist_list< typename F::first_argument_type > *bs );

  /**
   * @brief Mapindex skeleton applies the function to each element.
   *        The function takes the index of the element for an argument.
   *
   * mapindex f [a0, a1, ..., an-1] =
   *      [f 0 a0, f 1 a1, ..., f (n-1) an-1]
   * This skeleton uses the function, which overwrites the results on
   * the inputs.
   *
   * @param f : The function to apply to elements. [ in ]
   * @param as : The input list to be applied by function f. [ in ]
   */
public:
  template< typename F, typename A >
  static void mapindex_ow( const F &f,
			   dist_list< A > *as);

  //------------------------------------------------------------------------
  // Reduce

  /**
   * @brief Reduce skeleton reduces the list into one element by applying
   *        a binary operator.
   *
   * reduce (oplus) e [a0, a1, ..., an-1] = e oplus a0 oplus a1 oplus ... oplus an-1.
   * @param oplus : The binary associative operator. [ in ]
   * @param e : The initial value. [ in ]
   * @param as : The input list to be applied by oplus. [ in ]
   */
public:
  template< typename OPLUS, typename A >
  static A reduce( const OPLUS &oplus,
		   const A &e,
		   const dist_list< A > *as );

  /**
   * @brief Reduce skeleton reduces the list into one element by applying
   *        a binary operator.
   *
   * reduce (oplus) e [a0, a1, ..., an-1] = e oplus a0 oplus a1 oplus ... oplus an-1.
   * This skeleton handles the pointers for the arguments. 
   * @param oplus : The binary associative operator. [ in ]
   * @param e : The initial value. [ in ]
   * @param as : The input list to be applied by oplus. [ in ]
   * @param retval : The pointer for the result value. [ out ]
   */
public:
  template< typename OPLUS, typename A >
  static void reduce( const OPLUS &oplus,
		      const A *e,
		      const dist_list< A > *as,
		      A *retval );

  //------------------------------------------------------------------------
  // Scan

  /**
   * @brief Scan skeleton performs the prefix-sum like computation by applying
   *        a binary operator.
   *
   * scan (oplus) e [a0, a1, ..., an-1] =
   *   [e, e oplus a0, e oplus a0 oplus a1, ..., e oplus a0 oplus a1 oplus ... oplus an-1]
   * @param oplus : The binary associative operator. [ in ]
   * @param e : The initial value. [ in ]
   * @param as : The input list to be applied by oplus. [ in ]
   */
public:
  template< typename A, typename OPLUS >
  static dist_list< A > *scan( const OPLUS &oplus,
			       const A &e,
			       const dist_list< A > *as );


  /**
   * @brief Scan skeleton performs the prefix-sum like computation by applying
   *        a binary operator.
   *
   * scan (oplus) e [a0, a1, ..., an-1] =
   *   [e, e oplus a0, e oplus a0 oplus a1, ..., e oplus a0 oplus a1 oplus ... oplus an-1]
   * This function handles the pointers for the arguments. 
   * @param oplus : The binary associative operator. [ in ]
   * @param e : The initial value. [ in ]
   * @param as : The input list to be applied by oplus. [ in ]
   * @param bs : The pointer for the result list. [ in, out ]
   * @attention The pointers passed to the operator @a oplus are guaranteed
   *            not to be same. 
   */
public:
  template< typename OPLUS, typename A >
  static void scan( const OPLUS &oplus,
		    const A *e,
		    const dist_list< A > *as,
		    dist_list< A > *bs );

  /**
   * @brief Scanr skeleton performs the postfix-sum like computation by
   *        applying a binary operator in the opposed way against @a scan.
   *
   * scanr (oplus) e [a0, ..., an-2, an-1] =
   *   [a0 oplus ... oplus an-1 oplus e, ..., an-2 oplus an-1 oplus e, an-1 oplus e, e]
   * @param oplus : The binary associative operator. [ in ]
   * @param e : The initial value. [ in ]
   * @param as : The input list to be applied by oplus. [ in ]
   */
public:
  template< typename A, typename OPLUS >
  static dist_list< A > *scanr( const OPLUS &oplus,
				const A &e,
				const dist_list< A > *as );

  /**
   * @brief Scanr skeleton performs the postfix-sum like computation by
   *        applying a binary operator in the opposed way against @a scan.
   *
   * scanr (oplus) e [a0, ..., an-2, an-1] =
   *   [a0 oplus ... oplus an-1 oplus e, ..., an-2 oplus an-1 oplus e, an-1 oplus e, e]
   * This function handles the pointers for the arguments. 
   * @param oplus : The binary associative operator. [ in ]
   * @param e : The initial value. [ in ]
   * @param as : The input list to be applied by oplus. [ in ]
   * @param bs : The pointer for the result list. [ in, out ]
   * @attention The pointers passed to the operator @a oplus are guaranteed
   *            not to be same. 
   */
public:
  template< typename OPLUS, typename A >
  static void scanr( const OPLUS &oplus,
		     const A *e,
		     const dist_list< A > *as,
		     dist_list< A > *bs );

  //------------------------------------------------------------------------
  // Prescan

  /**
   * @brief Prescan skeleton performs the prefix-sum like computation by applying
   *        a binary operator.
   *
   * prescan (oplus) e [a0, a1, ..., an-1] =
   *   [e, e oplus a0, e oplus a0 oplus a1, ..., e oplus a0 oplus a1 oplus ... oplus an-2]
   * @param oplus : The binary associative operator. [ in ]
   * @param e : The initial value. [ in ]
   * @param as : The input list to be applied by oplus. [ in ]
   */
public:
  template< typename A, typename OPLUS >
  static dist_list< A > *
  prescan( const OPLUS &oplus,
           const A &e,
           const dist_list< A > *as );

  /**
   * @brief Prescan skeleton performs the prefix-sum like computation by applying
   *        a binary operator.
   *
   * prescan (oplus) e [a0, a1, ..., an-1] =
   *   [e, e oplus a0, e oplus a0 oplus a1, ..., e oplus a0 oplus a1 oplus ... oplus an-2]
   * This function handles the pointers for the arguments.
   * @param oplus : The binary associative operator. [ in ]
   * @param e : The initial value. [ in ]
   * @param as : The input list to be applied by oplus. [ in ]
   * @param bs : The pointer for the result list. [ in, out ]
   * @attention The pointers passed to the operator @a oplus are guaranteed
   *            not to be same.
   */
public:
  template< typename OPLUS, typename A >
  static void
  prescan( const OPLUS &oplus,
           const A *e,
           const dist_list< A > *as,
           dist_list< A > *bs );

  /**
   * @brief Prescan skeleton performs the prefix-sum like computation by applying
   *        a binary operator.
   *
   * prescan (oplus) e [a0, a1, ..., an-1] =
   *   [e, e oplus a0, e oplus a0 oplus a1, ..., e oplus a0 oplus a1 oplus ... oplus an-2]
   * @param oplus : The binary associative operator. [ in ]
   * @param e : The initial value. [ in ]
   * @param as : The input list to be applied by oplus. [ in ]
   */
public:
  template< typename A, typename OPLUS >
  static void
  prescan_ow( const OPLUS &oplus,
              const A &e,
              dist_list< A > *as );
  /**
   * @brief Prescanr skeleton performs the prefix-sum like computation by applying
   *        a binary operator in the oppsed way ageinst @a prescan.
   *
   * prescanr (oplus) e [a0, a1, ..., an-1] =
   *   [e oplus a0 oplus a1 oplus ... oplus an-2, ..., e oplus a0, e]
   * @param oplus : The binary associative operator. [ in ]
   * @param e : The initial value. [ in ]
   * @param as : The input list to be applied by oplus. [ in ]
   */
public:
  template< typename A, typename OPLUS >
  static dist_list< A > *
  prescanr( const OPLUS &oplus,
	    const A &e,
	    const dist_list< A > *as );

  /**
   * @brief Prescanr skeleton performs the prefix-sum like computation by applying
   *        a binary operator in the oppsed way ageinst @a prescan.
   *
   * prescanr (oplus) e [a0, a1, ..., an-1] =
   *   [e oplus a0 oplus a1 oplus ... oplus an-2, ..., e oplus a0, e]
   * This function handles the pointers for the arguments.
   * @param oplus : The binary associative operator. [ in ]
   * @param e : The initial value. [ in ]
   * @param as : The input list to be applied by oplus. [ in ]
   * @param bs : The pointer for the result list. [ in, out ]
   * @attention The pointers passed to the operator @a oplus are guaranteed
   *            not to be same.
   */
public:
  template< typename OPLUS, typename A >
  static void
  prescanr( const OPLUS &oplus,
	    const A *e,
	    const dist_list< A > *as,
	    dist_list< A > *bs );

  /**
   * @brief Prescanr skeleton performs the prefix-sum like computation by applying
   *        a binary operator in the oppsed way ageinst @a prescan.
   *
   * prescanr (oplus) e [a0, a1, ..., an-1] =
   *   [e oplus a0 oplus a1 oplus ... oplus an-2, ..., e oplus a0, e]
   * @param oplus : The binary associative operator. [ in ]
   * @param e : The initial value. [ in ]
   * @param as : The input list to be applied by oplus. [ in ]
   */
public:
  template< typename A, typename OPLUS >
  static void
  prescanr_ow( const OPLUS &oplus,
	       const A &e,
	       dist_list< A > *as );

  //------------------------------------------------------------------------
  // Zip

  /**
   * @brief Zip skeleton accepts two list and returns an list
   *        whose elements are pairs.
   *
   * zip [a0, a1, ..., an-1] [b0, b1, ..., bn-1]
   *    = [(a0, b0), (a1, b1), ..., (an-1, bn-1)]
   * @param as : The first list to zip up. [ in ]
   * @param bs : The second list to zip up. [ in ]
   */
public:
  template< typename A, typename B >
  static dist_list< std::pair< A, B > > *
  zip( const dist_list< A > *as,
       const dist_list< B > *bs );
  /**
   * @brief Zip skeleton accepts two list and returns an list
   *        whose elements are pairs.
   *
   * zip [a0, a1, ..., an-1] [b0, b1, ..., bn-1]
   *    = [(a0, b0), (a1, b1), ..., (an-1, bn-1)]
   * @param as : The first list to zip up. [ in ]
   * @param bs : The second list to zip up. [ in ]
   * @param cs : The pointer for the result list. [ in, out ]
   */
public:
  template< typename A, typename B >
  static void
  zip( const dist_list< A > *as,
       const dist_list< B > *bs,
       dist_list< std::pair< A, B > > *cs );
  
  //------------------------------------------------------------------------
  // Zipwith

  /**
   * @brief Zipwith skeleton applies the function to each pair of two lists,
   *        and is equivalent to the map skeleton after zip skeleton.
   *
   * zipwith f [a0, a1, ..., an-1] [b0, b1, ..., bn-1]
   *    = [f a0 b0, f a1 b1, ..., f an-1 bn-1]
   * @param f : The function to apply to lists. [ in ]
   * @param as : The first list. [ in ]
   * @param bs : The second list. [ in ]
   */
public:
  template< typename F, typename A, typename B >
  static dist_list< typename F::result_type > *
  zipwith( const F &f,
	   const dist_list< A > *as,
	   const dist_list< B > *bs );

  /**
   * @brief Zipwith skeleton applies the function to each pair of two lists,
   *        and is equivalent to the map skeleton after zip skeleton.
   *
   * zipwith f [a0, a1, ..., an-1] [b0, b1, ..., bn-1]
   *    = [f a0 b0, f a1 b1, ..., f an-1 bn-1]
   *
   * This skeleton uses the function, which handles the pointers for
   * arguments.
   * @attention If you passed the same dist_list for this, @a as,
   *            or @a cs, the pointers passed to function @a f
   *            might become the same. 
   * 
   * @param f : The function to apply to lists. [ in ]
   * @param as : The first list. [ in ]
   * @param bs : The second list. [ in ]
   * @param cs : The result list, and must be allocated before
   *             calling this function. [ in, out ]
   */
public:
  template< typename F, typename A, typename B >
  static void zipwith( const F &f,
		       const dist_list< A > *as,
		       const dist_list< B > *bs,
		       dist_list< typename F::first_argument_type > *cs );

  //------------------------------------------------------------------------
  // Zipwithindex

  /**
   * @brief Zipwithindex skeleton applies the function to each corresponding
   *        pair of two lists, furthermore, the function takes the index
   *        of the pair for the first argument.
   *
   * This skeleton is equivalent to the successive call of mapindex skeleton
   * after zip skeleton.
   *
   * zipwithindex f [a0, a1, ..., an-1] [b0, b1, ..., bn-1]
   *   = [f 0 a0 b0, f 1 a1 b1, ..., f n-1 an-1 bn-1]. 
   * @param f : The function to apply to the each pair. [ in ]
   * @param as : The first list. [ in ]
   * @param bs : The second list. [ in ]
   */
public:
  template< typename F, typename A, typename B >
  static dist_list< typename F::result_type > *
  zipwithindex( const F &f,
		const dist_list< A > *as,
		const dist_list< B > *bs );
  
  /**
   * @brief Zipwithindex skeleton applies the function to each corresponding
   *        pair of two lists, furthermore, the function takes the index
   *        of the pair for the first argument.
   *
   * This skeleton is equivalent to the successive call of mapindex skeleton
   * after zip skeleton.
   *
   * zipwithindex f [a0, a1, ..., an-1] [b0, b1, ..., bn-1]
   *   = [f 0 a0 b0, f 1 a1 b1, ..., f n-1 an-1 bn-1].
   *
   * The function @a f used in this skeleton, accepts the index of pair,
   * and and two pointers for the elements of pair.
   *
   * @param f : The function to apply to the each pair. [ in ]
   * @param as : The first list. [ in ]
   * @param bs : The second list. [ in ]
   * @param cs: The result list, and must be allocated before
   *            calling this skeleton. [ out ]
   */
public:
  template< typename F, typename A, typename B >
  static void zipwithindex( const F &f,
			    const dist_list< A > *as,
			    const dist_list< B > *bs,
			    dist_list< typename F::first_argument_type > *cs );
  
  //------------------------------------------------------------------------
  // Shift

  /**
   * @brief Shiftr skeleton shifts the input list to the right.
   *
   * @param firstval : The value inserted into the leftmost position. [ in ]
   * @param as : The input list to be shifted to the right. [ in ]
   */
public:
  template< typename A >
  static dist_list< A > *shiftr( const A& firstval,
				 const dist_list< A > *as );

  /**
   * @brief Shiftr skeleton shifts the input list to the right. This function
   *        overwrites the input list. 
   *
   * @param firstval : The value inserted into the leftmost position. [ in ]
   * @param as : The input list to be shifted to the right. [ in ]
   */
public:
  template< typename A >
  static void shiftr_ow( const A& firstval,
			 dist_list< A > *as);
  
  /**
   * @brief Shiftl skeleton shifts the input list to the left.
   *
   * @param lastval : The value inserted into the rightmost position. [ in ]
   * @param as : The input list to be shifted to the left. [ in ]
   */
public:
  template< typename A >
  static dist_list< A > *shiftl( const A& lastval,
				 const dist_list< A > *as);

  /**
   * @brief Shiftl skeleton shifts the input list to the left. This function
   *        overwrites the input list. 
   *
   * @param lastval : The value inserted into the rightmost position. [ in ]
   * @param as : The input list to be shifted to the left. [ in ]
   */
public:
  template< typename A >
  static void shiftl_ow( const A& lastval,
			 dist_list< A > *as );
  
  //------------------------------------------------------------------------
  // Accumulate 

  /**
   * @brief Accumulate skeleton is an implementation for the parallelizable
   *        recursive function.
   *
   * Accumulate is a higher order function defined as follows. 
   * h = acc [[ g, (p, oplus), (q, otimes) ]]
   * h [] e       = g e
   * h (x : xs) e = p (x, e) oplus h xs (e otimes q x)
   *
   * To guarantee the efficient parallelization,
   *  - the associativity of oplus and otimes respectively,
   * are required.
   */
public:
  template< typename G, typename P, typename OPLUS, typename Q, typename OTIMES, typename A >
  static typename P::result_type
  accumulate( const G &g,
	      const P &p,
	      const OPLUS &oplus,
	      const Q &q,
	      const OTIMES &otimes,
	      const typename G::argument_type &e,
	      const dist_list< A > *as );
  
  /**
   * @brief Accumulate skeleton is an implementation for the parallelizable
   *        recursive function.
   *
   * Accumulate is a higher order function defined as follows. 
   * h = acc [ g, (p, oplus), (q, otimes) ]
   * h [] e       = g e
   * h (x : xs) e = p (x, e) oplus h xs (e otimes q x)
   *
   * To guarantee the efficient parallelization,
   *  - the associativity of oplus and otimes respectively,
   * are required.
   *
   * This skeleton uses the functions which handles the arguments with their
   * pointers. 
   */
public:
  template< typename G, typename P, typename OPLUS, typename Q, typename OTIMES, typename A >
  static void accumulate( const G &g,
			  const P &p,
			  const OPLUS &oplus,
			  const Q &q,
			  const OTIMES &otimes,
			  const typename G::second_argument_type *e,
			  const dist_list< A > *as,
			  typename P::first_argument_type *retval );

  //------------------------------------------------------------------------
  // Gather
  
  /**
   * @brief Gathers the distributed list to rank 0 process.
   *
   * @param as : The list to be gathered. [ in ]
   * @param array : The array to be stored result.
   */
public:
  template< typename A >
  static void
  gather( const dist_list< A > *as, A *array );

  //------------------------------------------------------------------------
  // Utilities

  /**
   * @brief Prints out the whole list.
   *
   * This function moves the elements to the rank 0 process,
   * and print out to @a out.
   * To use this function, the class @c A must implement the operator<<. 
   *
   * @param as : The list to be printed out. [ in ]
   * @param stream : The output stream to print out to. [ in ]
   */
  template< typename A >
  static void print( const dist_list< A > *as,
		     std::ostream &stream = std::cout );
  
  /**
   * @brief Prints out the index-th element.
   *
   * This function moves the @a index -th element to the rank 0 process,
   * and print out to @a out.
   * To use this function, the class @c A must implement the operator<<. 
   *
   * @param as : The list to be printed out. [ in ]
   * @param index : The global index to print out. [ in ]
   * @param stream : The output stream to print out to. [ in ]
   * @todo destination check is not efficient....
   */
  template< typename A >
  static void print( const dist_list< A > *as,
		     int index,
		     std::ostream &stream = std::cout );

  //------------------------------------------------------------------------
  // for optimize

public:
  template< typename OPLUS, typename P, typename A >
  static typename P::result_type
  cataj( const OPLUS &oplus,
         const P &p,
         const typename P::result_type &e,
         const dist_list< A > *as );

public:
  template< typename OPLUS, typename P, typename A >
  static void
  cataj( const OPLUS &oplus,
         const P &p,
         const typename P::first_argument_type *e,
         const dist_list< A > *as,
         typename P::first_argument_type *retval );

public:
  template< typename OPLUS, typename A, typename B >
  static std::pair< A, B >
  zip_reduce( const OPLUS &oplus,
              const std::pair< A, B > &e,
              const dist_list< A > *as,
              const dist_list< B > *bs );

public:
  template< typename OPLUS, typename A, typename B >
  static void
  zip_reduce( const OPLUS &oplus,
              const std::pair< A, B > *e,
              const dist_list< A > *as,
              const dist_list< B > *bs,
              std::pair< A, B > *result );

public:
  template< typename F, typename OPLUS, typename A, typename B >
  static typename F::result_type
  zipwith_reduce( const F &f,
                  const OPLUS &oplus,
                  const typename F::result_type &e,
                  const dist_list< A > *as,
                  const dist_list< B > *bs );
public:
  template< typename F, typename OPLUS, typename A, typename B >
  static void
  zipwith_reduce( const F &f,
                  const OPLUS &oplus,
                  const typename F::result_type *e,
                  const dist_list< A > *as,
                  const dist_list< B > *bs,
                  typename F::result_type *result );

};

#include "list_skeletons.tpp"
#endif // __LIST_SKELETONS_H__
