/***
 * Copyright (c) 2005, SkeTo Project
 * All rights reserved.
 */
/**
 * @file list_skeletons.tpp
 * @brief The template implementation of list_skeletons class.
 * @author Yoshiki AKASHI (yoshiki@ipl.cs.uec.ac.jp)
 * @author Kiminori MATSUZAKI (kmatsu@ipl.t.u-tokyo.ac.jp)
 */

#ifndef __LIST_SKELETONS_TPP__
#define __LIST_SKELETONS_TPP__
#ifdef __SKETO_DEBUG__
#define debug(x) x
#else // __SKETO_DEBUG__
#define debug(x)
#endif // __SKETO_DEBUG__

//------------------------------------------------------------------------
// Map

template< typename F, typename A >
inline dist_list< typename F::result_type > *
list_skeletons::map( const F &f,
		     const dist_list< A > *as )
{
  dist_list< typename F::result_type > *bs = new dist_list< typename F::result_type >( as->global_size );
  for ( int i = 0; i < as->local_size; i++ ) {
    bs->at( i ) = f( as->at( i ) );
  }

  debug(if(skeleton::rank == 0) std::cout << "map\n");

  return bs;
}

template< typename F, typename A >
inline void
list_skeletons::map( const F &f,
		     const dist_list< A > *as,
		     dist_list< typename F::first_argument_type > *bs )
{
  // check size
  if ( as->get_global_size( ) != bs->get_global_size( ) ) {
    throw skeleton::bad_size( as->get_global_size( ), bs->get_global_size( ) );
  }
  
  for ( int i = 0; i < as->local_size; i++ ) {
    f( &bs->at( i ), &as->at( i ) );
  }
  
  debug(if(skeleton::rank == 0) std::cout << "map\n");
}

template< typename F, typename A >
inline void
list_skeletons::map_ow( const F &f,
			dist_list< A > *as )
{
  for ( int i = 0; i < as->local_size; i++ ) {
    as->at( i ) = f( as->at( i ) );
  }

  debug(if(skeleton::rank == 0) std::cout << "map_ow\n");
}

//------------------------------------------------------------------------
// Mapindex

template< typename F, typename A >
inline dist_list< typename F::result_type > *
list_skeletons::mapindex( const F &f,
			  const dist_list< A > *as )
{
  dist_list< typename F::result_type > *bs = new dist_list< typename F::result_type >( as->global_size );
  for ( int lindex = 0, gindex = as->begin; lindex < as->local_size; lindex++, gindex++ ) {
    bs->at( lindex ) = f( gindex, as->at( lindex ) );
  }

  debug(if(skeleton::rank == 0) std::cout << "mapindex\n");

  return bs;
}

template< typename F, typename A >
inline void
list_skeletons::mapindex( const F &f,
			  const dist_list< A > *as,
			  dist_list< typename F::first_argument_type > *bs)
{
  // check size
  if ( as->get_global_size( ) != bs->get_global_size( ) ) {
    throw skeleton::bad_size( as->get_global_size( ), bs->get_global_size( ) );
  }

  for ( int lindex = 0, gindex = as->begin; lindex < as->local_size; lindex++, gindex++ ) {
    f( &bs->at( lindex ), &gindex, &as->at( lindex ) );
  }

  debug(if(skeleton::rank == 0) std::cout << "mapindex\n");
}

template< typename F, typename A >
inline void
list_skeletons::mapindex_ow( const F &f,
			     dist_list< A > *as )
{
  for ( int lindex = 0, gindex = as->begin; lindex < as->local_size; lindex++, gindex++ ) {
    as->at( lindex ) = f( gindex, as->at( lindex ) );
  }
  
  debug(if(skeleton::rank == 0) std::cout << "mapindex_ow\n");
}

//------------------------------------------------------------------------
// Reduce

template< typename OPLUS, typename A >
inline A
list_skeletons::reduce( const OPLUS &oplus,
			const A &e,
			const dist_list< A > *as )
{
  // local reduce
  A result = as->at( 0 );
  for ( int i = 1; i < as->local_size; i++ ) {
    result = oplus( result, as->at( i ) );
  }

  // tree reduction
  for ( int stage = 1; stage < skeleton::procs; stage <<= 1 ) {
    int target = skeleton::rank ^ stage;
    if ( target < skeleton::rank ) {
      MPI_Send( &result, 1, as->mpitype, target, TAG_REDUCE1, MPI_COMM_WORLD );
      break;
    }
    else if ( target < skeleton::procs ) {
      A received_data;
      MPI_Status st;
      MPI_Recv( &received_data, 1, as->mpitype, target, TAG_REDUCE1, MPI_COMM_WORLD, &st );
      result = oplus( result, received_data );
    }
  }
  result = oplus( e, result );
  MPI_Bcast( &result, 1, as->mpitype, 0, MPI_COMM_WORLD );

  debug(if(skeleton::rank == 0) std::cout << "reduce\n");
  
  return result;
}

template< typename OPLUS, typename A >
inline void
list_skeletons::reduce( const OPLUS &oplus,
			const A *e,
			const dist_list< A > *as,
			A *retval )
{
  // local reduce
  (*retval) = as->at( 0 );
  for ( int i = 1; i < as->local_size; i++ ) {
    A argument = (*retval);
    oplus( retval, &argument, &as->at( i ) );
  }

  // tree reduction
  for ( int stage = 1; stage < skeleton::procs; stage <<= 1 ) {
    int target = skeleton::rank ^ stage;
    if ( target < skeleton::rank ) {
      MPI_Send( retval, 1, as->mpitype,	target, TAG_REDUCE2, MPI_COMM_WORLD );
      break;
    }
    else if ( target < skeleton::procs ) {
      A received_data;
      MPI_Status st;
      MPI_Recv( &received_data, 1, as->mpitype, target, TAG_REDUCE2, MPI_COMM_WORLD, &st );
      A argument = (*retval);
      oplus( retval, &argument, &received_data );
    }
  }
  A argument = (*retval);
  oplus( retval, e, &argument);
  MPI_Bcast( retval, 1, as->mpitype, 0, MPI_COMM_WORLD );
  
  debug(if(skeleton::rank == 0) std::cout << "reduce\n");

}

//------------------------------------------------------------------------
// Scan

template< typename A, typename OPLUS >
inline dist_list< A > *
list_skeletons::scan( const OPLUS &oplus,
		      const A &e,
		      const dist_list< A > *as )
{
  // local reduce
  A total_val, first_val;
  if ( skeleton::rank == 0 ) {
    total_val = oplus( e, as->at( 0 ) );
    first_val = e;
  }
  else{
    total_val = as->at( 0 );
  }
  for ( int i = 1; i < as->local_size; i++ ) {
    total_val = oplus( total_val, as->at( i ) );
  }

  // butterfly computation
  bool flag = false;
  for ( int stage = 1; stage < skeleton::procs; stage <<= 1 ) {
    const int target = skeleton::rank ^ stage;
    if ( target >= skeleton::procs ) continue;
    A received_val;
    MPI_Status st;
    MPI_Sendrecv( &total_val, 1, as->mpitype, target, TAG_SCAN1,
		  &received_val, 1, as->mpitype, target, TAG_SCAN1, MPI_COMM_WORLD, &st );
    if ( target > skeleton::rank ) {
      total_val = oplus( total_val, received_val );
    }
    else {
      if( flag == false ) {
	first_val = received_val;
	flag = true;
      }
      else {
	first_val = oplus( received_val, first_val );
      }	
      total_val = oplus( received_val, total_val );
    }
  }

  // local scan
  int mod = as->global_size % skeleton::procs;
  dist_list< A > *result = new dist_list< A >( as->global_size + 1 );
  if ( skeleton::rank <= mod ) {
    result->at( 0 ) = first_val;
    int result_size = ( skeleton::rank == mod ) ? as->local_size + 1 : as->local_size;
    for ( int i = 1; i < result_size; i++ ) {
      result->at( i ) = oplus( result->at( i - 1 ), as->at( i - 1 ) );
    }
  }
  else {
    result->at( 0 ) = oplus( first_val, as->at( 0 ) );
    for ( int i = 1; i < as->local_size; i++ ) {
      result->at( i ) = oplus( result->at( i - 1 ), as->at( i ) );
    }
  }

  debug(if(skeleton::rank == 0) std::cout << "scan\n");

  return result;
}

template< typename OPLUS, typename A >
inline void
list_skeletons::scan( const OPLUS &oplus,
		      const A *e,
		      const dist_list< A > *as,
		      dist_list< A > *bs )
{
  // check size
  if ( (as->get_global_size( ) + 1) != bs->get_global_size( ) ) {
    throw skeleton::bad_size( as->get_global_size( ), bs->get_global_size( ) );
  }

  // local reduce
  A total_val, first_val;
  if( skeleton::rank == 0 ){
    total_val = first_val = *e;
    for ( int i = 0; i < as->local_size; i++ ) {
      { A tmp = total_val; oplus( &total_val, &tmp, &as->at( i ) ); }
    }
  }
  else {
    total_val = as->at( 0 );
    for ( int i = 1; i < as->local_size; i++ ) {
      { A tmp = total_val; oplus( &total_val, &tmp, &as->at( i ) ); }
    }
  }

  // butterfly computation
  bool flag = false;
  for( int stage = 1; stage < skeleton::procs; stage <<= 1 ) {
    const int target = skeleton::rank ^ stage;
    if ( target >= skeleton::procs ) continue;
    A received_val;
    MPI_Status st;
    MPI_Sendrecv( &total_val, 1, as->mpitype, target, TAG_SCAN2,
		  &received_val, 1, as->mpitype, target, TAG_SCAN2, MPI_COMM_WORLD, &st );

    if( target > skeleton::rank ) {
      { A tmp = total_val; oplus( &total_val, &tmp, &received_val ); }
    }
    else {
      if( flag == false ){
	first_val = received_val;
	flag = true;
      }
      else {
	{ A tmp = first_val; oplus( &first_val, &received_val, &tmp ); }
      }	
      { A tmp = total_val; oplus( &total_val, &received_val, &tmp ); }
    }
  }

  // local scan
  int mod = as->global_size % skeleton::procs;
  if ( skeleton::rank <= mod ) {
    bs->at( 0 ) = first_val;
    int result_size = ( skeleton::rank == mod ) ? as->local_size + 1 : as->local_size;
    for ( int i = 1; i < result_size; i++ ) {
      oplus( &bs->at( i ), &bs->at( i - 1 ), &as->at( i - 1 ) );
    }
  }
  else {
    oplus( &bs->at( 0 ), &first_val, &as->at( 0 ) );
    for ( int i = 1; i < as->local_size; i++ ) {
      oplus( &bs->at( i ), &bs->at( i - 1 ), &as->at( i ) );
    }
  }

  debug(if(skeleton::rank == 0) std::cout << "scan\n");

}

template< typename A, typename OPLUS >
inline dist_list< A > *
list_skeletons::scanr( const OPLUS &oplus,
		       const A &e,
		       const dist_list< A > *as)
{
  const int vrank = skeleton::procs - skeleton::rank - 1;
  A _e = e;
  
  // send e to last rank
  if ( skeleton::rank == 0 ) {
    MPI_Send( &_e, 1, as->mpitype, skeleton::procs-1, TAG_SCANR1, MPI_COMM_WORLD );
  }
  else if ( vrank == 0 ) {
    MPI_Status st;
    MPI_Recv( &_e, 1, as->mpitype, 0, TAG_SCANR1, MPI_COMM_WORLD, &st );
  }

  // local reduce
  A last_val = _e;
  A total_val = (vrank == 0) ? oplus( _e, as->at( 0 ) ) : as->at( 0 );
  for ( int i = 1; i < as->local_size; i++ ) {
    total_val = oplus( total_val, as->at( i ) );
  }

  // butterfly computation
  bool flag = false;
  for ( int stage = 1; stage < skeleton::procs; stage <<= 1 ) {
    const int vtarget = vrank ^ stage;
    const int target = skeleton::procs - vtarget - 1;
    if ( target < 0 ) continue;

    A received_val;
    MPI_Status st;
    MPI_Sendrecv( &total_val, 1, as->mpitype, target, TAG_SCANR1,
		  &received_val, 1, as->mpitype, target, TAG_SCANR1, MPI_COMM_WORLD, &st );
    
    if ( vtarget > vrank ) {
      total_val = oplus( received_val, total_val );
    }
    else {
      if ( flag == false ) {
	last_val = received_val;
	flag = true;
      }
      else {
	last_val = oplus( last_val, received_val );
      }
      total_val = oplus( total_val, received_val );
    }
  }

  // local scan
  int mod = as->global_size % skeleton::procs;
  dist_list< A > *result = new dist_list< A >( as->global_size + 1 );
  if ( skeleton::rank > mod ) {
    result->at( as->local_size - 1 ) = last_val;
    for ( int i = as->local_size - 1; i > 0; i-- ) {
      result->at( i - 1 ) = oplus( as->at( i ), result->at( i ) );
    }
  }
  else if ( skeleton::rank == mod ) {
    result->at( as->local_size ) = last_val;
    for ( int i = as->local_size - 1; i >= 0; i-- ) {
      result->at( i ) = oplus( as->at( i ), result->at( i + 1 ) );
    }
  }
  else { // rank < mod
    result->at( as->local_size - 1 ) = oplus( as->at( as->local_size - 1 ), last_val );
    for ( int i = as->local_size - 1; i > 0; i-- ) {
      result->at( i - 1 ) = oplus( as->at( i - 1 ), result->at( i ) );
    }
  }

  debug(if(skeleton::rank == 0) std::cout << "scanr\n");

  return result;
}

template< typename OPLUS, typename A >
inline void
list_skeletons::scanr( const OPLUS &oplus,
		       const A *e,
		       const dist_list< A > *as,
		       dist_list< A > *bs )
{
  // check size
  if ( (as->get_global_size( ) + 1) != bs->get_global_size( ) ) {
    throw skeleton::bad_size( as->get_global_size( ), bs->get_global_size( ) );
  }

  const int vrank = skeleton::procs - skeleton::rank - 1;
  A _e = *e;
  
  // send e to last rank
  if ( skeleton::rank == 0  ) {
    MPI_Send( &_e, 1, as->mpitype, skeleton::procs-1, TAG_SCANR2, MPI_COMM_WORLD );
  }
  else if ( vrank == 0 ) {
    MPI_Status st;
    MPI_Recv( &_e, 1, as->mpitype, 0, TAG_SCANR2, MPI_COMM_WORLD, &st );
  }

  // local reduce
  A last_val = _e;
  A total_val;
  if( vrank == 0 ){
    oplus( &total_val, &_e, &as->at( 0 ) );
  }
  else{
    total_val = as->at( 0 );
  }
  for ( int i = 1; i < as->local_size; i++ ) {
    { A tmp = total_val; oplus( &total_val, &as->at( i ), &tmp ); }
  }

  // butterfly computation
  bool flag = false;
  for ( int stage = 1; stage < skeleton::procs; stage <<= 1 ) {
    const int vtarget = vrank ^ stage;
    const int target = skeleton::procs - vtarget - 1;
    if ( target < 0 ) continue;

    A received_val;
    MPI_Status st;
    MPI_Sendrecv( &total_val, 1, as->mpitype, target, TAG_SCANR2,
		  &received_val, 1, as->mpitype, target, TAG_SCANR2, MPI_COMM_WORLD, &st );
    
    if ( vtarget > vrank ) {
      { A tmp = total_val; oplus( &total_val, &received_val, &tmp ); }
    }
    else {
      if ( flag == false ) {
	last_val = received_val;
	flag = true;
      }
      else {
	{ A tmp = last_val; oplus( &last_val, &tmp, &received_val ); }
      }
      { A tmp = total_val; oplus( &total_val, &tmp, &received_val ); }
    }
  }

  // local scan
  int mod = as->global_size % skeleton::procs;
  if ( skeleton::rank > mod ) {
    bs->at( as->local_size - 1 ) = last_val;
    for ( int i = as->local_size - 1; i > 0; i-- ) {
      oplus( &bs->at( i - 1 ), &as->at( i ), &bs->at( i ) );
    }
  }
  else if ( skeleton::rank == mod ) {
    bs->at( as->local_size ) = last_val;
    for ( int i = as->local_size - 1; i >= 0; i-- ) {
      oplus( &bs->at( i ), &as->at( i ), &bs->at( i + 1 ) );
    }
  }
  else {
    oplus( &bs->at( as->local_size - 1 ), &as->at( as->local_size - 1 ), &last_val );
    for ( int i = as->local_size - 1; i > 0; i-- ) {
      oplus( &bs->at( i - 1 ), &as->at( i - 1 ), &bs->at( i ) );
    }
  }

  debug(if(skeleton::rank == 0) std::cout << "scanr\n");
}

//------------------------------------------------------------------------
// Precan

template< typename A, typename OPLUS >
inline dist_list< A > *
list_skeletons::prescan( const OPLUS &oplus,
                         const A &e,
                         const dist_list< A > *as )
{
  // local reduce
  A total_val, first_val = e;
  total_val = ( skeleton::rank == 0 ) ? oplus( e, as->at( 0 ) ) : as->at( 0 );

  for ( int i = 1; i < as->local_size; i++ ) {
    total_val = oplus( total_val, as->at( i ) );
  }

  // butterfly computation
  bool flag = false;
  for ( int stage = 1; stage < skeleton::procs; stage <<= 1 ) {
    const int target = skeleton::rank ^ stage;
    if ( target >= skeleton::procs ) continue;
    A received_val;
    MPI_Status st;
    MPI_Sendrecv( &total_val, 1, as->mpitype, target, TAG_SCAN1,
                  &received_val, 1, as->mpitype, target, TAG_SCAN1, MPI_COMM_WORLD, &st );
    if ( target > skeleton::rank ) {
      total_val = oplus( total_val, received_val );
    }
    else {
      if( flag == false ) {
        first_val = received_val;
        flag = true;
      }
      else {
        first_val = oplus( received_val, first_val );
      }
      total_val = oplus( received_val, total_val );
    }
  }

  // local scan
  dist_list< A > *result = new dist_list< A >( as->global_size );
  result->at( 0 ) = first_val;
  for ( int i = 1; i < as->local_size; i++ ) {
    result->at( i ) = oplus( result->at( i - 1 ), as->at( i - 1 ) );
  }

  debug(if(skeleton::rank == 0) std::cout << "prescan\n");

  return result;
}

template< typename OPLUS, typename A >
inline void
list_skeletons::prescan( const OPLUS &oplus,
                         const A *e,
                         const dist_list< A > *as,
                         dist_list< A > *bs )
{
  // check size
  if ( (as->get_global_size( )) != bs->get_global_size( ) ) {
    throw skeleton::bad_size( as->get_global_size( ), bs->get_global_size( ) );
  }

  // local reduce
  A total_val, first_val = *e;
  if ( skeleton::rank == 0 ) {
    oplus( &total_val, e, &as->at( 0 ) );
  }
  else {
    total_val = as->at( 0 );
  }

  for ( int i = 1; i < as->local_size; i++ ) {
    { A tmp = total_val; oplus( &total_val, &tmp, &as->at( i ) ); }
  }

  // butterfly computation
  bool flag = false;
  for( int stage = 1; stage < skeleton::procs; stage <<= 1 ) {
    const int target = skeleton::rank ^ stage;
    if ( target >= skeleton::procs ) continue;
    A received_val;
    MPI_Status st;
    MPI_Sendrecv( &total_val, 1, as->mpitype, target, TAG_SCAN2,
                  &received_val, 1, as->mpitype, target, TAG_SCAN2, MPI_COMM_WORLD, &st );

    if( target > skeleton::rank ) {
      { A tmp = total_val; oplus( &total_val, &tmp, &received_val ); }
    }
    else {
      if( flag == false ){
        first_val = received_val;
        flag = true;
      }
      else {
        { A tmp = first_val; oplus( &first_val, &received_val, &tmp ); }
      }
      { A tmp = total_val; oplus( &total_val, &received_val, &tmp ); }
    }
  }

  // local scan
  bs->at( 0 ) = first_val;
  for ( int i = 1; i < as->local_size; i++ ) {
    oplus( &bs->at( i ), &bs->at( i - 1 ), &as->at( i - 1 ) );
  }

  debug(if(skeleton::rank == 0) std::cout << "prescan\n");
}

template< typename A, typename OPLUS >
inline void
list_skeletons::prescan_ow( const OPLUS &oplus,
                            const A &e,
                            dist_list< A > *as )
{
  // local reduce
  A total_val, first_val = e;
  total_val = ( skeleton::rank == 0 ) ? oplus( e, as->at( 0 ) ) : as->at( 0 );

  for ( int i = 1; i < as->local_size; i++ ) {
    total_val = oplus( total_val, as->at( i ) );
  }

  // butterfly computation
  bool flag = false;
  for ( int stage = 1; stage < skeleton::procs; stage <<= 1 ) {
    const int target = skeleton::rank ^ stage;
    if ( target >= skeleton::procs ) continue;
    A received_val;
    MPI_Status st;
    MPI_Sendrecv( &total_val, 1, as->mpitype, target, TAG_SCAN1,
                  &received_val, 1, as->mpitype, target, TAG_SCAN1, MPI_COMM_WORLD, &st );
    if ( target > skeleton::rank ) {
      total_val = oplus( total_val, received_val );
    }
    else {
      if( flag == false ) {
        first_val = received_val;
        flag = true;
      }
      else {
        first_val = oplus( received_val, first_val );
      }
      total_val = oplus( received_val, total_val );
    }
  }

  // local scan
  A tmp = as->at( 0 );
  as->at( 0 ) = first_val;
  for ( int i = 1; i < as->local_size; i++ ) {
    A result = oplus( as->at( i - 1 ), tmp );
    tmp = as->at( i );
    as->at( i ) = result;
  }

  debug(if(skeleton::rank == 0) std::cout << "prescan\n");
}

template< typename A, typename OPLUS >
inline dist_list< A > *
list_skeletons::prescanr( const OPLUS &oplus,
			  const A &e,
			  const dist_list< A > *as )
{
  const int vrank = skeleton::procs - skeleton::rank - 1;
  A _e = e;

  // send e to last rank
  if ( skeleton::rank == 0 ) {
    MPI_Send( &_e, 1, as->mpitype, skeleton::procs-1, TAG_SCANR1, MPI_COMM_WORLD );
  }
  else if( vrank == 0 ){
    MPI_Status st;
    MPI_Recv( &_e, 1, as->mpitype, 0, TAG_SCANR1, MPI_COMM_WORLD, &st );
  }

  // local reduce
  A last_val = _e;
  A total_val = ( vrank == 0 ) ? oplus( _e, as->at( 0 ) ) : as->at( 0 );
  for ( int i = 1; i < as->local_size; i++ ) {
    total_val = oplus( total_val, as->at( i ) );
  }
  
  // butterfly computation
  bool flag = false;
  for ( int stage = 1; stage < skeleton::procs; stage <<= 1 ) {
    const int vtarget = vrank ^ stage;
    const int target = skeleton::procs - vtarget - 1;
    if ( target < 0 ) continue;

    A received_val;
    MPI_Status st;
    MPI_Sendrecv( &total_val, 1, as->mpitype, target, TAG_SCANR1,
		  &received_val, 1, as->mpitype, target, TAG_SCANR1, MPI_COMM_WORLD, &st );
    
    if ( vtarget > vrank ) {
      total_val = oplus( received_val, total_val );
    }
    else {
      if ( flag == false ) {
	last_val = received_val;
	flag = true;
      }
      else {
	last_val = oplus( last_val, received_val );
      }
      total_val = oplus( total_val, received_val );
    }
  }
  
  // local scan
  dist_list< A > *result = new dist_list< A >( as->global_size );
  result->at( as->local_size - 1 ) = last_val;
  for ( int i = as->local_size - 1; i > 0; i-- ) {
    result->at( i - 1 ) = oplus( result->at( i ), as->at( i ) );
  }

  debug(if(skeleton::rank == 0) std::cout << "prescanr\n");

  return result;
}

template< typename OPLUS, typename A >
inline void
list_skeletons::prescanr( const OPLUS &oplus,
			  const A *e,
			  const dist_list< A > *as,
			  dist_list< A > *bs )
{
  // check size
  if ( (as->get_global_size( )) != bs->get_global_size( ) ) {
    throw skeleton::bad_size( as->get_global_size( ), bs->get_global_size( ) );
  }

  const int vrank = skeleton::procs - skeleton::rank - 1;
  A _e = *e;

  // send e to last rank
  if ( skeleton::rank == 0 ) {
    MPI_Send( &_e, 1, as->mpitype, skeleton::procs-1, TAG_SCANR1, MPI_COMM_WORLD );
  }
  else if( vrank == 0 ){
    MPI_Status st;
    MPI_Recv( &_e, 1, as->mpitype, 0, TAG_SCANR1, MPI_COMM_WORLD, &st );
  }

  // local reduce
  A last_val = _e;
  A total_val;
  if( vrank == 0 ) {
    oplus( &total_val, &_e, &as->at( 0 ) );
  }
  else {
    total_val = as->at( 0 );
  }
  for ( int i = 1; i < as->local_size; i++ ) {
    { A tmp = total_val; oplus( &total_val, &as->at( i ), &tmp ); }
  }

  // butterfly computation
  bool flag = false;
  for ( int stage = 1; stage < skeleton::procs; stage <<= 1 ) {
    const int vtarget = vrank ^ stage;
    const int target = skeleton::procs - vtarget - 1;
    if ( target < 0 ) continue;

    A received_val;
    MPI_Status st;
    MPI_Sendrecv( &total_val, 1, as->mpitype, target, TAG_SCANR1,
		  &received_val, 1, as->mpitype, target, TAG_SCANR1, MPI_COMM_WORLD, &st );
    
    if ( vtarget > vrank ) {
      { A tmp = total_val; oplus( &total_val, &received_val, &tmp ); }
    }
    else {
      if ( flag == false ) {
	last_val = received_val;
	flag = true;
      }
      else {
	{ A tmp = last_val; oplus( &last_val, &tmp, &received_val ); }
      }
      { A tmp = total_val; oplus( &total_val, &tmp, &received_val ); }
    }
  }
  
  // local scan
  bs->at( as->local_size - 1 ) = last_val;
  for ( int i = as->local_size - 1; i > 0; i-- ) {
    oplus( &bs->at( i - 1 ), &bs->at( i ), &as->at( i ) );
  }

  debug(if(skeleton::rank == 0) std::cout << "prescanr\n");
}

template< typename A, typename OPLUS >
inline void
list_skeletons::prescanr_ow( const OPLUS &oplus,
			     const A &e,
			     dist_list< A > *as )
{
  const int vrank = skeleton::procs - skeleton::rank - 1;
  A _e = e;

  // send e to last rank
  if ( skeleton::rank == 0 ) {
    MPI_Send( &_e, 1, as->mpitype, skeleton::procs-1, TAG_SCANR1, MPI_COMM_WORLD );
  }
  else if( vrank == 0 ){
    MPI_Status st;
    MPI_Recv( &_e, 1, as->mpitype, 0, TAG_SCANR1, MPI_COMM_WORLD, &st );
  }

  // local reduce
  A last_val = _e;
  A total_val = ( vrank == 0 ) ? oplus( _e, as->at( 0 ) ) : as->at( 0 );
  for ( int i = 1; i < as->local_size; i++ ) {
    total_val = oplus( total_val, as->at( i ) );
  }
  
  // butterfly computation
  bool flag = false;
  for ( int stage = 1; stage < skeleton::procs; stage <<= 1 ) {
    const int vtarget = vrank ^ stage;
    const int target = skeleton::procs - vtarget - 1;
    if ( target < 0 ) continue;

    A received_val;
    MPI_Status st;
    MPI_Sendrecv( &total_val, 1, as->mpitype, target, TAG_SCANR1,
		  &received_val, 1, as->mpitype, target, TAG_SCANR1, MPI_COMM_WORLD, &st );
    
    if ( vtarget > vrank ) {
      total_val = oplus( received_val, total_val );
    }
    else {
      if ( flag == false ) {
	last_val = received_val;
	flag = true;
      }
      else {
	last_val = oplus( last_val, received_val );
      }
      total_val = oplus( total_val, received_val );
    }
  }

  // local scan
  A tmp = as->at( as->local_size - 1 );
  as->at( as->local_size - 1 ) = last_val;
  for ( int i = as->local_size - 2; i >= 0; i-- ) {
    A result = oplus( as->at( i + 1 ), tmp );
    tmp = as->at( i );
    as->at( i ) = result;
  }

  debug(if(skeleton::rank == 0) std::cout << "prescanr_ow\n");
}

//------------------------------------------------------------------------
// Zip

template< typename A, typename B >
inline dist_list< std::pair< A, B > > *
list_skeletons::zip( const dist_list< A > *as,
		     const dist_list< B > *bs )
{
  // check size
  if ( as->get_global_size( ) != bs->get_global_size( ) ) {
    throw skeleton::bad_size( as->get_global_size( ), bs->get_global_size( ) );
  }

  dist_list< std::pair< A, B > > *retval 
    = new dist_list< std::pair< A, B > >( as->global_size );
  for ( int i = 0; i < as->local_size; i++ ) {
    retval->at( i ).first = as->at( i );
    retval->at( i ).second = bs->at( i );
  }

  debug(if(skeleton::rank == 0) std::cout << "zip\n");

  return retval;
}

template< typename A, typename B >
inline void
list_skeletons::zip( const dist_list< A > *as,
                     const dist_list< B > *bs,
                     dist_list< std::pair< A, B > > *cs )
{
  // check size
  if ( as->get_global_size( ) != bs->get_global_size( ) ) {
    throw skeleton::bad_size( as->get_global_size( ), bs->get_global_size( ) );
  }
  if ( as->get_global_size( ) != cs->get_global_size( ) ) {
    throw skeleton::bad_size( as->get_global_size( ), cs->get_global_size( ) );
  }

  for ( int i = 0; i < as->local_size; i++ ) {
    cs->at( i ).first  = as->at( i );
    cs->at( i ).second = bs->at( i );
  }

  debug(if(skeleton::rank == 0) std::cout << "zip\n");
}

//------------------------------------------------------------------------
// Zipwith

template< typename F, typename A, typename B >
inline dist_list< typename F::result_type > *
list_skeletons::zipwith( const F &f,
			 const dist_list< A > *as,
			 const dist_list< B > *bs )
{
  // check size
  if ( as->get_global_size( ) != bs->get_global_size( ) ) {
    throw skeleton::bad_size( as->get_global_size( ), bs->get_global_size( ) );
  }

  dist_list< typename F::result_type > *cs = new dist_list< typename F::result_type >( as->global_size );
  for ( int i = 0; i < as->local_size; i++ ) {
    cs->at( i ) = f( as->at( i ), bs->at( i ) );
  }

  debug(if(skeleton::rank == 0) std::cout << "zipwith\n");

  return cs;
}

template< typename F, typename A, typename B >
inline void
list_skeletons::zipwith( const F &f,
			 const dist_list< A > *as,
			 const dist_list< B > *bs,
			 dist_list< typename F::first_argument_type >* cs )
{
  // check size for bs and cs
  if ( as->get_global_size( ) != bs->get_global_size( ) ) {
    throw skeleton::bad_size( as->get_global_size( ), bs->get_global_size( ) );
  }
  if ( as->get_global_size( ) != cs->get_global_size( ) ) {
    throw skeleton::bad_size( as->get_global_size( ), cs->get_global_size( ) );
  }

  for ( int i = 0; i < as->local_size; i++ ) {
    f( &cs->at( i ), &as->at( i ), &bs->at( i ) );
  }

  debug(if(skeleton::rank == 0) std::cout << "zipwith\n");
}

//------------------------------------------------------------------------
// Zipwithindex

template< typename F, typename A, typename B >
inline dist_list< typename F::result_type > *
list_skeletons::zipwithindex( const F &f,
			      const dist_list< A > *as,
			      const dist_list< B > *bs )
{
  // check size
  if ( as->get_global_size( ) != bs->get_global_size( ) ) {
    throw skeleton::bad_size( as->get_global_size( ), bs->get_global_size( ) );
  }

  dist_list< typename F::result_type > *cs = new dist_list< typename F::result_type >( as->global_size );
  for ( int i = 0, index = as->begin; i < as->local_size; i++, index++ ) {
    cs->at( i ) = f( index, as->at( i ), bs->at( i ) );
  }

  debug(if(skeleton::rank == 0) std::cout << "zipwithindex\n");

  return cs;
}

template< typename F, typename A, typename B >
inline void
list_skeletons::zipwithindex( const F &f,
			      const dist_list< A > *as,
			      const dist_list< B > *bs,
			      dist_list< typename F::first_argument_type > *cs )
{
  // check size
  if ( as->get_global_size( ) != bs->get_global_size( ) ) {
    throw skeleton::bad_size( as->get_global_size( ), bs->get_global_size( ) );
  }
  if ( as->get_global_size( ) != cs->get_global_size( ) ) {
    throw skeleton::bad_size( as->get_global_size( ), cs->get_global_size( ) );
  }

  for ( int i = 0, index = as->begin; i < as->local_size; i++, index++ ) {
    f( &cs->at( i ), index, &as->at( i ), &bs->at( i ) );
  }

  debug(if(skeleton::rank == 0) std::cout << "zipwithindex\n");
}

//------------------------------------------------------------------------
// Shift

template< typename A >
inline dist_list< A > *
list_skeletons::shiftr( const A& firstval,
			const dist_list< A > *as )
{
  dist_list< A > *bs = new dist_list< A >( as->global_size );

  if ( skeleton::rank != skeleton::procs - 1 ) {
    A next_val = as->at( as->local_size - 1 );
    MPI_Send( &next_val, 1, as->mpitype, skeleton::rank + 1, TAG_SHIFTR1, MPI_COMM_WORLD );
  }
  if ( skeleton::rank == 0 ) {
    bs->at( 0 ) = firstval;
  }
  else {
    MPI_Status st;
    MPI_Recv( &bs->at( 0 ), 1, as->mpitype, skeleton::rank - 1, TAG_SHIFTR1, MPI_COMM_WORLD, &st );
  }
  for ( int i = 1; i < as->local_size; i++ ) {
    bs->at( i ) = as->at( i - 1 );
  }      

  debug(if(skeleton::rank == 0) std::cout << "shiftr\n");
  return bs;
}

template< typename A >
inline void
list_skeletons::shiftr_ow( const A& firstval,
			   dist_list< A > *as )
{
  if ( skeleton::rank != skeleton::procs - 1 ) {
    MPI_Send( &as->at( as->local_size - 1 ), 1, as->mpitype, skeleton::rank + 1, TAG_SHIFTR2, MPI_COMM_WORLD );
  }
  for ( int i = as->local_size - 1; i >= 1; i-- ) {
    as->at( i ) = as->at( i - 1 );
  }
  if ( skeleton::rank == 0 ) {
    as->at( 0 ) = firstval;
  }
  else {
    MPI_Status st;
    MPI_Recv( &as->at( 0 ), 1, as->mpitype, skeleton::rank - 1, TAG_SHIFTR2, MPI_COMM_WORLD, &st );
  }

  debug(if(skeleton::rank == 0) std::cout << "shiftr_ow\n");
}

template< typename A >
inline dist_list< A > *
list_skeletons::shiftl( const A& lastval,
			const dist_list< A > *as )
{
  dist_list< A > *bs = new dist_list< A >( as->global_size );

  if ( skeleton::rank != 0 ) {
    A prev_val = as->at( 0 );
    MPI_Send( &prev_val, 1, as->mpitype, skeleton::rank - 1, TAG_SHIFTL1, MPI_COMM_WORLD );
  }
  if ( skeleton::rank == skeleton::procs - 1 ) {
    bs->at( as->local_size - 1 ) = lastval;
  }
  else {
    MPI_Status st;
    MPI_Recv( &bs->at( as->local_size - 1 ), 1, as->mpitype, skeleton::rank + 1, TAG_SHIFTL1, MPI_COMM_WORLD, &st );
  }
  for ( int i = 0; i < as->local_size - 1; i++ ) {
    bs->at( i ) = as->at( i + 1 );
  }

  debug(if(skeleton::rank == 0) std::cout << "shiftl\n");

  return bs;
}

template< typename A >
inline void
list_skeletons::shiftl_ow( const A& lastval,
			   dist_list< A > *as )
{
  if ( skeleton::rank != 0 ) {
    MPI_Send( &as->at( 0 ), 1, as->mpitype, skeleton::rank - 1, TAG_SHIFTL2, MPI_COMM_WORLD );
  }
  for ( int i = 0; i < as->local_size - 1; i++ ) {
    as->at( i ) = as->at( i + 1 );
  }
  if ( skeleton::rank == skeleton::procs - 1 ) {
    as->at( as->local_size - 1 ) = lastval;
  }
  else {
    MPI_Status st;
    MPI_Recv( &as->at( as->local_size - 1 ), 1, as->mpitype, skeleton::rank + 1, TAG_SHIFTL2, MPI_COMM_WORLD, &st );
  }

  debug(if(skeleton::rank == 0) std::cout << "shiftl_ow\n");
}

//------------------------------------------------------------------------
// Accumulate

template< typename G, typename P, typename OPLUS, typename Q, typename OTIMES, typename A >
inline typename P::result_type
list_skeletons::accumulate( const G &g,
			    const P &p,
			    const OPLUS &oplus,
			    const Q &q,
			    const OTIMES &otimes,
			    const typename G::argument_type &e,
			    const dist_list< A > *as )
{
  MPI_Datatype mpitype_b = template_mpitype< typename G::argument_type >::type( );
  MPI_Datatype mpitype_c = template_mpitype< typename P::result_type >::type( );  

  // map and local reduce
  typename G::argument_type totalval, firstval;
  typename G::argument_type *bs = new typename G::argument_type[ as->local_size ];
  if ( skeleton::rank == 0 ) {
    totalval = firstval = e;
    for ( int i = 1; i < as->local_size; i++ ) {
      bs[ i ] = q( as->at( i - 1 ) );
      totalval = otimes( totalval, bs[ i ] );
    }
  }
  else {
    bs[ 1 ] =  q( as->at( 0 ) );
    totalval = bs[ 1 ];
    for ( int i = 2; i < as->local_size; i++ ) {
      bs[ i ] = q( as->at( i - 1 ) );
      totalval = otimes( totalval, bs[ i ] );
    }
  }
  totalval = otimes( totalval, q( as->at( as->local_size - 1 ) ) );
  
  // butterfly computation
  bool flag = false;
  for ( int stage = 1; stage < skeleton::procs; stage <<= 1 ) {
    const int target = skeleton::rank ^ stage;
    if ( target >= skeleton::procs ) continue;
    typename G::argument_type recvval;
    MPI_Status st;
    MPI_Sendrecv( &totalval, 1, mpitype_b, target, TAG_ACCUMULATE1_1,
		  &recvval, 1, mpitype_b, target, TAG_ACCUMULATE1_1, MPI_COMM_WORLD, &st );
    if ( target > skeleton::rank ) {
      totalval = otimes( totalval, recvval );
    }
    else {
      if ( flag == false ) {
	firstval = recvval;
	flag = true;
      }
      else{
 	firstval = otimes( recvval, firstval );
      }
      totalval = otimes( recvval, totalval );
    }
  }

  // local scan (making bs complete), zipwith and local reduce
  bs[ 0 ] = firstval;
  typename P::result_type reduceval;
  reduceval = p( as->at( 0 ), bs[ 0 ] );
  for ( int i = 1; i < as->local_size; i++ ) {
    bs[ i ] = otimes( bs[ i - 1 ], bs[ i ] );
    reduceval = oplus( reduceval, p( as->at( i ), bs[ i ] ) );
  }

  delete[] bs;
  
  // tree computation
  for ( int stage = 1; stage < skeleton::procs; stage <<= 1 ) {
    const int target = skeleton::rank ^ stage;
    if ( target < skeleton::rank ) {
      MPI_Send( &reduceval, 1, mpitype_c,
		target, TAG_ACCUMULATE1_2,
		MPI_COMM_WORLD );
      break;
    }
    else if ( target < skeleton::procs ) {
      typename P::result_type recvval;
      MPI_Status st;
      MPI_Recv( &recvval, 1, mpitype_c,
		target, TAG_ACCUMULATE1_2, MPI_COMM_WORLD, &st );
      reduceval = oplus( reduceval, recvval );
    }
  }

  typename P::result_type retval = oplus( reduceval, g( totalval ) );
  MPI_Bcast( &retval, 1, mpitype_c, 0, MPI_COMM_WORLD );

  debug(if(skeleton::rank == 0) std::cout << "acc\n");

  return retval;
}

template< typename G, typename P, typename OPLUS, typename Q, typename OTIMES, typename A >
inline void
list_skeletons::accumulate( const G &g,
			    const P &p,
			    const OPLUS &oplus,
			    const Q &q,
			    const OTIMES &otimes,
			    const typename G::second_argument_type *e,
			    const dist_list< A > *as,
			    typename P::first_argument_type *retval )
{
  MPI_Datatype mpitype_b = template_mpitype< typename G::second_argument_type >::type( );
  MPI_Datatype mpitype_c = template_mpitype< typename P::first_argument_type >::type( );  
  
  // map and local reduce
  typename G::second_argument_type totalval, firstval;
  typename G::second_argument_type *bs = new typename G::second_argument_type[ as->local_size ];
  if ( skeleton::rank == 0 ) {
    totalval = firstval = *e;
    for ( int i = 1; i < as->local_size; i++ ) {
      q( &bs[ i ], &as->at( i - 1 ) );
      { typename G::second_argument_type tmp = totalval; otimes( &totalval, &tmp, &bs[ i ] ); }
    }
  }
  else {
    q( &totalval, &as->at( 0 ) );
    bs[ 1 ] = totalval;
    for ( int i = 2; i < as->local_size; i++ ) {
      q( &bs[ i ], &as->at( i - 1 ) );
      { typename G::second_argument_type tmp = totalval; otimes( &totalval, &tmp, &bs[ i ] ); }
    }
  }
  typename G::second_argument_type qval;
  q( &qval, &as->at( as->local_size - 1 ) );
  { typename G::second_argument_type tmp = totalval; otimes( &totalval, &tmp, &qval ); }
  
  // butterfly computation
  bool flag = false;
  for ( int stage = 1; stage < skeleton::procs; stage <<= 1 ) {
    const int target = skeleton::rank ^ stage;
    if ( target >= skeleton::procs ) continue;
    typename G::second_argument_type recvval;
    MPI_Status st;
    MPI_Sendrecv( &totalval, 1, mpitype_b, target, TAG_ACCUMULATE2_1,
		  &recvval, 1, mpitype_b, target, TAG_ACCUMULATE2_1, MPI_COMM_WORLD, &st );
    if ( target > skeleton::rank ) {
      { typename G::second_argument_type tmp = totalval; otimes( &totalval, &tmp, &recvval ); }
    }
    else {
      if ( flag == false ) {
	firstval = recvval;
	flag = true;
      }
      else {
	{ typename G::second_argument_type tmp = firstval; otimes( &firstval, &recvval, &tmp ); }
      }
      { typename G::second_argument_type tmp = totalval; otimes( &totalval, &recvval, &tmp ); }
    }
  }

  // local scan (making bs complete), zipwith and local reduce
  bs[ 0 ] = firstval;
  typename P::first_argument_type reduceval;
  p( &reduceval, &as->at( 0 ), &bs[ 0 ] );
  for ( int i = 1; i < as->local_size; i++ ) {
    typename P::first_argument_type zipval;
    { typename G::second_argument_type tmp = bs[ i ]; otimes( &bs[ i ], &bs[ i - 1 ], &tmp ); }
    p( &zipval, &as->at( i ), &bs[ i ] );
    { typename P::first_argument_type tmp = reduceval; oplus( &reduceval, &tmp, &zipval ); }
  }

  delete[] bs;
  
  // tree computation
  for ( int stage = 1; stage < skeleton::procs; stage <<= 1 ) {
    const int target = skeleton::rank ^ stage;
    if ( target < skeleton::rank ) {
      MPI_Send( &reduceval, 1, mpitype_c,
		target, TAG_ACCUMULATE2_2,
		MPI_COMM_WORLD );
      break;
    }
    else if ( target < skeleton::procs ) {
      typename P::first_argument_type recvval;
      MPI_Status st;
      MPI_Recv( &recvval, 1, mpitype_c,
		target, TAG_ACCUMULATE2_2, MPI_COMM_WORLD, &st );
      { typename P::first_argument_type tmp = reduceval; oplus( &reduceval, &tmp, &recvval ); }
    }
  }

  typename P::first_argument_type gval;
  g( &gval, &totalval );
  oplus( retval, &reduceval, &gval );
  MPI_Bcast( retval, 1, mpitype_c, 0, MPI_COMM_WORLD );
  
  debug(if(skeleton::rank == 0) std::cout << "acc\n");
}

//------------------------------------------------------------------------
// Gather

template< typename A >
inline void
list_skeletons::gather( const dist_list< A > *as, A *array )
{
  if( skeleton::rank == 0 ) {
    // for master process: local copy
    std::memcpy( array, as->data, sizeof( A ) * as->local_size );

    // for the other processes
    for( int src = 1; src < skeleton::procs; src++ ) {
      MPI_Status st;
      MPI_Recv( &array[ as->get_begin_index( src ) ],
		as->get_local_size( src ), as->mpitype,
		src, TAG_GATHER, MPI_COMM_WORLD, &st );
    }
  }
  else {
    MPI_Send( as->data, as->local_size, as->mpitype,
	      0, TAG_GATHER,
	      MPI_COMM_WORLD );
  }

  debug(if(skeleton::rank == 0) std::cout << "gather\n");
}

//------------------------------------------------------------------------
// Utilities

template< typename A >
inline void
list_skeletons::print( const dist_list< A > *as,
		       std::ostream &stream )
{
  if ( skeleton::rank == 0 ) {
    for ( int i = 0; i < as->local_size; i++ ) {
      stream << as->at( i ) << ' ';
    }

    A *tmp = new A[ as->local_size ];
    for ( int dest = 1; dest < skeleton::procs; dest++ ) {
      int dest_size = as->get_local_size( dest );
      MPI_Status st;
      MPI_Recv( tmp, dest_size, as->mpitype, dest, TAG_PRINT1, MPI_COMM_WORLD, &st );
      for( int i = 0; i < dest_size; i++) {
	stream << tmp[i] << ' ';
      }
    }
    delete[] tmp;
    stream << '\n';
  }
  else {
    MPI_Send( as->data, as->local_size, as->mpitype, 0, TAG_PRINT1, MPI_COMM_WORLD );
  }
}

template< typename A >
inline void
list_skeletons::print( const dist_list< A > *as,
		       int index, std::ostream &stream )
{
  A value;
  if ( skeleton::rank == 0 ) {
    if ( as->begin <= index && index < as->end ) {
      value = as->at( index - as->begin );
    }
    else {
      // this destination check is not efficient...
      for ( int dest = 1; dest < skeleton::procs; dest++ ) {
	if ( as->get_begin_index( dest ) <= index &&
	     index < as->get_end_index( dest ) ) {
	  MPI_Status st;
	  MPI_Recv( &value, 1, as->mpitype, dest, TAG_PRINT2, MPI_COMM_WORLD, &st );
	}
      }
    }
    stream << value;
  }
  else {
    if ( as->begin <= index && index < as->end ) {
      MPI_Send( &as->at( index - as->begin ), 1, as->mpitype, 0, TAG_PRINT2, MPI_COMM_WORLD );
    }
  }
}

//------------------------------------------------------------------------
// for optimize

template< typename OPLUS, typename P, typename A >
inline typename P::result_type
list_skeletons::cataj( const OPLUS &oplus,
		       const P &p,
		       const typename P::result_type &e,
		       const dist_list< A > *as )
{
  MPI_Datatype mpitype_b = template_mpitype< typename P::result_type >::type( );

  // map and local reduce
  typename P::result_type retval = p( as->at( 0 ) );
  for ( int i = 1; i < as->local_size; i++ ) {
    retval = oplus( retval, p( as->at( i ) ) );
  }

  // tree computation
  for ( int stage = 1; stage < skeleton::procs; stage <<= 1 ) {
    const int target = skeleton::rank ^ stage;
    if ( target < skeleton::rank ) {
      MPI_Send( &retval, 1, mpitype_b, target, TAG_CATAJ1, MPI_COMM_WORLD );
      break;
    }
    else if( target < skeleton::procs ) {
      typename P::result_type recvval;
      MPI_Status st;
      MPI_Recv( &recvval, 1, mpitype_b, target, TAG_CATAJ1, MPI_COMM_WORLD, &st );
      retval = oplus( retval, recvval );
    }
  }
  retval = oplus( retval, e );
  MPI_Bcast( &retval, 1, mpitype_b, 0, MPI_COMM_WORLD );

  debug(if(skeleton::rank == 0) std::cout << "cataj\n");

  return retval;
}

template< typename OPLUS, typename P, typename A >
inline void
list_skeletons::cataj( const OPLUS &oplus,
		       const P &p,
		       const typename P::first_argument_type *e,
		       const dist_list< A > *as,
		       typename P::first_argument_type *retval )
{
  MPI_Datatype mpitype_b = template_mpitype< typename P::first_argument_type >::type( );

  // map and local reduce
  p( retval, &as->at( 0 ) );
  for ( int i = 1; i < as->local_size; i++ ) {
    typename P::first_argument_type pval;
    p( &pval, &as->at( i ) );
    { typename P::first_argument_type tmp = (*retval); oplus( retval, &tmp, &pval ); }
  }

  // tree computation
  for ( int stage = 1; stage < skeleton::procs; stage <<= 1 ) {
    const int target = skeleton::rank ^ stage;
    if ( target < skeleton::rank ) {
      MPI_Send( retval, 1, mpitype_b, target, TAG_CATAJ2, MPI_COMM_WORLD );
      break;
    }
    else if( target < skeleton::procs ) {
      typename P::first_argument_type recvval;
      MPI_Status st;
      MPI_Recv( &recvval, 1, mpitype_b, target, TAG_CATAJ2, MPI_COMM_WORLD, &st );
      { typename P::first_argument_type tmp = (*retval); oplus( retval, &tmp, &recvval ); }
    }
  }
  { typename P::first_argument_type tmp = (*retval); oplus( retval, &tmp, e ); }
  MPI_Bcast( retval, 1, mpitype_b, 0, MPI_COMM_WORLD );
  
  debug(if(skeleton::rank == 0) std::cout << "cataj\n");
}

template< typename OPLUS, typename A, typename B >
inline std::pair< A, B >
list_skeletons::zip_reduce( const OPLUS &oplus,
                            const std::pair< A, B > &e,
                            const dist_list< A > *as,
                            const dist_list< B > *bs )
{
  // check size
  if ( as->get_global_size( ) != bs->get_global_size( ) ) {
    throw skeleton::bad_size( as->get_global_size( ), bs->get_global_size( ) );
  }

  MPI_Datatype mpitype = template_mpitype< A >::type( );

  // local reduce
  std::pair< A, B > result = make_pair( as->at( 0 ), bs->at( 0 ) );
  for ( int i = 1; i < as->local_size; i++ ) {
    result = oplus( result, make_pair( as->at( i ), bs->at( i ) ) );
  }

  // tree reduction
  for ( int stage = 1; stage < skeleton::procs; stage <<= 1 ) {
    int target = skeleton::rank ^ stage;
    if ( target < skeleton::rank ) {
      MPI_Send( &result, 1, mpitype, target, TAG_REDUCE1, MPI_COMM_WORLD );
      break;
    }
    else if ( target < skeleton::procs ) {
      std::pair< A, B > received_data;
      MPI_Status st;
      MPI_Recv( &received_data, 1, mpitype, target, TAG_REDUCE1, MPI_COMM_WORLD, &st );
      result = oplus( result, received_data );
    }
  }
  result = oplus( e, result );
  MPI_Bcast( &result, 1, mpitype, 0, MPI_COMM_WORLD );

  debug(if(skeleton::rank == 0) std::cout << "zip_reduce\n");

  return result;
}

template< typename OPLUS, typename A, typename B >
inline void
list_skeletons::zip_reduce( const OPLUS &oplus,
                            const std::pair< A, B > *e,
                            const dist_list< A > *as,
                            const dist_list< B > *bs,
                            std::pair< A, B > *result )
{
  // check size
  if ( as->get_global_size( ) != bs->get_global_size( ) ) {
    throw skeleton::bad_size( as->get_global_size( ), bs->get_global_size( ) );
  }

  MPI_Datatype mpitype = template_mpitype< A >::type( );

  // local reduce
  (*result) = make_pair( as->at( 0 ), bs->at( 0 ) );
  for ( int i = 1; i < as->local_size; i++ ) {
    std::pair< A, B > argument = (*result);
    oplus( result, &argument, &make_pair( as->at( i ), bs->at( i ) ) );
  }

  // tree reduction
  for ( int stage = 1; stage < skeleton::procs; stage <<= 1 ) {
    int target = skeleton::rank ^ stage;
    if ( target < skeleton::rank ) {
      MPI_Send( result, 1, mpitype, target, TAG_REDUCE2, MPI_COMM_WORLD );
      break;
    }
    else if ( target < skeleton::procs ) {
      std::pair< A, B > received_data;
      MPI_Status st;
      MPI_Recv( &received_data, 1, mpitype, target, TAG_REDUCE2, MPI_COMM_WORLD, &st );
      std::pair< A, B > argument = (*result);
      oplus( result, &argument, &received_data );
    }
  }
  std::pair< A, B > argument = (*result);
  oplus( result, e, &argument );
  MPI_Bcast( result, 1, mpitype, 0, MPI_COMM_WORLD );

  debug(if(skeleton::rank == 0) std::cout << "zip_reduce\n");
}

template< typename F, typename OPLUS, typename A, typename B >
typename F::result_type
list_skeletons::zipwith_reduce( const F &f,
                                const OPLUS &oplus,
                                const typename F::result_type &e,
                                const dist_list< A > *as,
                                const dist_list< B > *bs )
{
  // check size
  if ( as->get_global_size( ) != bs->get_global_size( ) ) {
    throw skeleton::bad_size( as->get_global_size( ),
                              bs->get_global_size( ) );
  }

  MPI_Datatype mpitype = template_mpitype< typename F::result_type >::type( );

  // local reduce
  typename F::result_type result = f( as->at( 0 ), bs->at( 0 ) );
  for ( int i = 1; i < as->local_size; i++ ) {
    result = oplus( result, f( as->at( i ), bs->at( i ) ) );
  }

  // tree reduction
  for ( int stage = 1; stage < skeleton::procs; stage <<= 1 ) {
    int target = skeleton::rank ^ stage;
    if ( target < skeleton::rank ) {
      MPI_Send( &result, 1, mpitype, target, TAG_REDUCE1, MPI_COMM_WORLD );
      break;
    }
    else if ( target < skeleton::procs ) {
      typename F::result_type received_data;
      MPI_Status st;
      MPI_Recv( &received_data, 1, mpitype, target, TAG_REDUCE1, MPI_COMM_WORLD, &st );
      result = oplus( result, received_data );
    }
  }
  result = oplus( e, result );
  MPI_Bcast( &result, 1, mpitype, 0, MPI_COMM_WORLD );

  debug(if(skeleton::rank == 0) std::cout << "zipwith_reduce\n");

  return result;
}

template< typename F, typename OPLUS, typename A, typename B >
void
list_skeletons::zipwith_reduce( const F &f,
                                const OPLUS &oplus,
                                const typename F::result_type *e,
                                const dist_list< A > *as,
                                const dist_list< B > *bs,
                                typename F::result_type *result )
{
  // check size
  if ( as->get_global_size( ) != bs->get_global_size( ) ) {
    throw skeleton::bad_size( as->get_global_size( ), bs->get_global_size( ) );
  }

  MPI_Datatype mpitype = template_mpitype< typename F::result_type >::type( );

  // local reduce
  f( result, &as->at( 0 ), &bs->at( 0 ) );
  for ( int i = 1; i < as->local_size; i++ ) {
    typename F::result_type f_result;
    typename F::result_type argument = (*result);
    f( &f_result, &as->at( i ), &bs->at( i ) );
    oplus( result, &argument, &f_result );
  }

  // tree reduction
  for ( int stage = 1; stage < skeleton::procs; stage <<= 1 ) {
    int target = skeleton::rank ^ stage;
    if ( target < skeleton::rank ) {
      MPI_Send( result, 1, mpitype, target, TAG_REDUCE2, MPI_COMM_WORLD );
      break;
    }
    else if ( target < skeleton::procs ) {
      typename F::result_type received_data;
      MPI_Status st;
      MPI_Recv( &received_data, 1, mpitype, target, TAG_REDUCE2, MPI_COMM_WORLD, &st );
      typename F::result_type argument = (*result);
      oplus( result, &argument, &received_data );
    }
  }
  typename F::result_type argument = (*result);
  oplus( result, e, &argument );
  MPI_Bcast( result, 1, mpitype, 0, MPI_COMM_WORLD );

  debug(if(skeleton::rank == 0) std::cout << "zipwith_reduce\n");
}

#endif // __LIST_SKELETONS_TPP__
