/***
 * Copyright (c) 2005, SkeTo Project
 * All rights reserved.
 */
/**
 * @file test_list_skeletons.cpp
 * @brief Simple test code for class list_skeletons
 * @author Kiminori MATSUZAKI
 *
 * @attention The following skeletons are not tested.
 * - zipwithindex
 * - zipwithindex (pointer)
 * - print
 * - print (with index)
 * - scan_reduce
 * - scan_cataj
 * - snan_map
 * - map_accumulate
 * - scan_accumulate
 * - map_scan
 * - scan_scan
 */

#include <iostream>
#include <cstdlib>
#include <ctime>
#include <utility>

#include "list_skeletons.h"

#include "matrix2.h"

////////////////////////////////////////////////////////////////////////
// common variables

static int size;
static int number;
static int count;

////////////////////////////////////////////////////////////////////////
// common check routine

template< typename A >
static bool
list_compare( const A* list1, int list1_size,
	      const A* list2, int list2_size )
{
  if ( list1_size != list2_size ) {
    std::cout << "size different" << std::endl;
    return false;
  }

  for ( int i = 0; i < list1_size; i++ ) {
    if ( !(list1[ i ] == list2[ i ]) ) {
      std::cout << "element at index " << i << " is different" << std::endl;
      return false;
    }
  }
  return true;
}

template< typename A, typename B >
void check_routine( void (*gen)( A** ), 
		    void (*f1)( const A*, B** ),
		    void (*f2)( const A*, B** ),
		    const char* message, int size )
{
  A* input = NULL;
  gen( &input );

  B* output1; f1( input, &output1 );
  B* output2; f2( input, &output2 );

  if ( !skeleton::is_master( ) ) return;
  
  if ( list_compare( output1, size, output2, size ) ) {
    std::cout << "Check " << number << ": "
	      << message << " --> OK" << std::endl;
    number++; count++;
  } else {
    std::cout << "Check " << number << ": "
	      << message << " ---> Failure" << std::endl;
    number++;
  }
  return;
}

////////////////////////////////////////////////////////////////////////
// Basic functions

struct func_square_char_double : skeleton::unary_function< char, double >
{
  double operator()( const char& x ) const {
    return static_cast< double >( x ) * static_cast< double >( x );
  }
};

struct func_square_char_double_ptr : skeleton::unary_ptr_function< double, char >
{
  void operator()( double* res, const char* x ) const {
    *res = static_cast< double >( *x ) * static_cast< double >( *x );
  }
};

struct func_square_char_char : skeleton::unary_function< char, char >
{
  char operator()( const char& x ) const {
    return x * x;
  }
};

struct func_shift_or_int_char : skeleton::binary_function< int, char, char >
{
  char operator()( const int& index, const char& val ) const {
    return ( index & 0x7f ) | ( val << ( index % 4 ) );
  }
};

struct func_shift_or_int_char_pair : skeleton::unary_function< std::pair< int, char >, char >
{
  char operator()( const std::pair< int, char >& x ) const {
    return func_shift_or_int_char( )( x.first, x.second );
  }
};

struct func_shift_or_int_char_ptr : skeleton::binary_ptr_function< char, int, char >
{
  void operator()( char* res, const int* index, const char* val ) const {
    *res = func_shift_or_int_char( )( *index, *val );
  }
};

template< typename A >
struct func_multiply_matrix
  : skeleton::binary_function< Matrix2< A >, Matrix2< A >, Matrix2< A > >
{
  Matrix2< A > operator()( const Matrix2< A >& a, const Matrix2< A >& b ) const {
    return a * b;
  }
};

template< typename A >
struct func_multiply_matrix_ptr
  : skeleton::binary_function< Matrix2< A >, Matrix2< A >, Matrix2< A > >
{
  void operator()( Matrix2< A >* res, const Matrix2< A >* a, const Matrix2< A >* b ) const {
    *res = *a * *b;
  }
};

struct func_char_to_matrix_char : skeleton::unary_function< char, Matrix2< char > >
{
  Matrix2< char > operator()( const char& x ) const {
    return Matrix2< char >( x, x ^ (x << 1), x ^ (x << 2), x ^ (x << 3) );
  }
};

struct func_char_to_matrix_char_ptr
  : skeleton::unary_ptr_function< Matrix2< char >, char >
{
  void operator()( Matrix2< char >* res, const char* x ) const {
    *res = func_char_to_matrix_char( )( *x );
  }
};

struct func_matrix_char_to_int
  : skeleton::unary_function< Matrix2< char >, Matrix2< int > >
{
  Matrix2< int > operator()( const Matrix2< char >& x ) const {
    return Matrix2< int >( x.at( 0, 0 ), x.at( 0, 1 ),
			   x.at( 1, 0 ), x.at( 1, 1 ) );
  }
};

struct func_matrix_char_to_int_ptr
  : skeleton::unary_ptr_function< Matrix2< int >, Matrix2< char > >
{
  void operator()( Matrix2< int >* res, const Matrix2< char >* x ) const {
    *res = func_matrix_char_to_int( )( *x );
  }
};

struct func_matrix_shift_square
  : skeleton::binary_function< char, Matrix2< char >, Matrix2< int > >
{
  Matrix2< int > operator()( const char& x, const Matrix2< char >& e ) const {
    Matrix2< int > y = Matrix2< int >( e.at( 0, 0 ) << ( x % 4 ),
				       e.at( 0, 1 ) << ( x % 4 ),
				       e.at( 1, 0 ) << ( x % 4 ),
				       e.at( 1, 1 ) << ( x % 4 ) );
    return y * y;
  }
};

struct func_matrix_shift_square_ptr
  : skeleton::binary_ptr_function< Matrix2< int >, char, Matrix2< char > >
{
  void operator()( Matrix2< int >* res,
		   const char* x, const Matrix2< char >* e ) const {
    *res = func_matrix_shift_square( )( *x, *e );
  }
};

////////////////////////////////////////////////////////////////////////
// generator functions

void generator_char( char** list ) {
  *list = NULL;
  if ( skeleton::is_master( ) ) {
    char *listp = new char[ size ];
    for ( int i = 0; i < size; i++ ) {
      listp[ i ] = rand( ) % 128;
    }
    *list = listp;
  }
}

void generator_matrix_int( Matrix2< int >** list )
{
  *list = NULL;
  if ( skeleton::is_master( ) ) {
    Matrix2< int >* listp = new Matrix2< int >[ size ];
    for ( int i = 0; i < size; i++ ) {
      listp[ i ].at( 0, 0 ) = rand( );
      listp[ i ].at( 0, 1 ) = rand( );
      listp[ i ].at( 1, 0 ) = rand( );
      listp[ i ].at( 1, 1 ) = rand( );
    }
    *list = listp;
  }
}

////////////////////////////////////////////////////////////////////////
// check function

void dist_gather_skel_char( const char* input, char** output ) {
  dist_list< char > *as = NULL;
  *output = NULL;

  if ( skeleton::is_master( ) ) {
    *output = new char[ size ];
  }
  as = new dist_list< char >( size, input );
  list_skeletons::gather( as, *output );

  skeleton::safe_delete( as );
}

template< typename A >
void do_nothing( const A* input, A** output ) {
  *output = NULL;
  if ( skeleton::is_master( ) ) {
    char* outputp = new char[ size ];
    for ( int i = 0; i < size; i++ ) {
      outputp[ i ] = input[ i ];
    }
    *output = outputp;
  }
}

void map_square_double_char_skel1( const char* input, double **output ) {
  dist_list< char > *as = new dist_list< char >( size, input );
  dist_list< double > *bs
    = list_skeletons::map( func_square_char_double( ), as );

  *output = NULL;
  if ( skeleton::is_master( ) ) *output = new double[ size ];
  list_skeletons::gather( bs, *output );

  skeleton::safe_delete( bs );
  skeleton::safe_delete( as );
}

void map_square_double_char_seq( const char* input, double** output ) {
  *output = NULL;
  if ( skeleton::is_master( ) ) {
    double* outputp = new double[ size ];
    for ( int i = 0; i < size; i++ ) {
      outputp[ i ] = func_square_char_double( )( input[ i ] );
    }
    *output = outputp;
  }
}

void map_square_double_char_skel2( const char* input, double** output ) {
  dist_list< char > *as = new dist_list< char >( size, input );
  dist_list< double > *bs = new dist_list< double >( size );

  list_skeletons::map( func_square_char_double_ptr( ), as, bs );

  *output = NULL;
  if ( skeleton::is_master( ) ) *output = new double[ size ];
  list_skeletons::gather( bs, *output );

  skeleton::safe_delete( bs );
  skeleton::safe_delete( as );
}

void map_square_char_char_seq( const char* input, char** output ) {
  *output = NULL;
  if ( skeleton::is_master( ) ) {
    char* outputp = new char[ size ];
    for ( int i = 0; i < size; i++ ) {
      outputp[ i ] = func_square_char_char( )( input[ i ] );
    }
    *output = outputp;
  }
}

void map_square_char_char_skel( const char* input, char** output ) {
  dist_list< char > *as = new dist_list< char >( size, input );

  list_skeletons::map_ow( func_square_char_char( ), as );

  *output = NULL;
  if ( skeleton::is_master( ) ) *output = new char[ size ];
  list_skeletons::gather( as, *output );

  skeleton::safe_delete( as );
}

void zip_shift_or_index( const char* input, char** output ) {
  dist_list< char > *as = new dist_list< char >( size, input );
  int* input2 = new int[ size ];
  for ( int i = 0; i < size; i++ ) input2[ i ] = i;
  dist_list< int > *bs = new dist_list< int >( size, input2 );
  delete[] input2;

  dist_list< std::pair< int, char > > *cs = list_skeletons::zip( bs, as );
  dist_list< char > *ds
    = list_skeletons::map( func_shift_or_int_char_pair( ), cs );

  *output = NULL;
  if ( skeleton::is_master( ) ) *output = new char[ size ];
  list_skeletons::gather( ds, *output );

  skeleton::safe_delete( ds );
  skeleton::safe_delete( cs );
  skeleton::safe_delete( bs );
  skeleton::safe_delete( as );
}

void zip_shift_or_index_seq( const char* input, char** output ) {
  *output = NULL;
  if ( skeleton::is_master( ) ) {
    char* outputp = new char[ size ];
    for ( int i = 0; i < size; i++ ) {
      outputp[ i ] = func_shift_or_int_char( )( i, input[ i ] );
    }
    *output = outputp;
  }
}

void mapindex_shift_or_index( const char* input, char** output ) {
  dist_list< char > *as = new dist_list< char >( size, input );
  dist_list< char > *bs
    = list_skeletons::mapindex( func_shift_or_int_char( ), as );

  *output = NULL;
  if ( skeleton::is_master( ) ) *output = new char[ size ];
  list_skeletons::gather( bs, *output );

  skeleton::safe_delete( bs );
  skeleton::safe_delete( as );
}

void mapindex_shift_or_index_ptr( const char* input, char** output ) {
  dist_list< char > *as = new dist_list< char >( size, input );
  dist_list< char > *bs = new dist_list< char >( size );
  list_skeletons::mapindex( func_shift_or_int_char_ptr( ), as, bs );

  *output = NULL;
  if ( skeleton::is_master( ) ) *output = new char[ size ];
  list_skeletons::gather( bs, *output );

  skeleton::safe_delete( bs );
  skeleton::safe_delete( as );
}

void mapindex_ow_shift_or_index( const char* input, char** output ) {
  dist_list< char > *as = new dist_list< char >( size, input );
  list_skeletons::mapindex_ow( func_shift_or_int_char( ), as );

  *output = NULL;
  if ( skeleton::is_master( ) ) *output = new char[ size ];
  list_skeletons::gather( as, *output );

  skeleton::safe_delete( as );
}

template < typename A >
void reduce_multiply_matrix_skel( const Matrix2< A >* input, Matrix2< A >** output ) {
  dist_list< Matrix2< A > > *as = new dist_list< Matrix2< A > >( size, input );
  Matrix2< A > e( 1, 2, 3, 4 );
  Matrix2< A > result = list_skeletons::reduce( func_multiply_matrix< A >( ),
						e, as );

  *output = NULL;
  if ( skeleton::is_master( ) ) {
    *output = new Matrix2< A >[ 1 ];
    **output = result;
  }

  skeleton::safe_delete( as );
}

template < typename A >
void reduce_multiply_matrix_skel_ptr( const Matrix2< A >* input,
				      Matrix2< A >** output )
{
  dist_list< Matrix2< A > > *as = new dist_list< Matrix2< A > >( size, input );
  Matrix2< A > e( 1, 2, 3, 4 );
  Matrix2< A > result;
  list_skeletons::reduce( func_multiply_matrix_ptr< A >( ),
			  &e, as, &result );

  *output = NULL;
  if ( skeleton::is_master( ) ) {
    *output = new Matrix2< A >[ 1 ];
    **output = result;
  }

  skeleton::safe_delete( as );
}

template < typename A >
void reduce_multiply_matrix_seq( const Matrix2< A >* input, Matrix2< A >** output ) {
  *output = NULL;
  if ( skeleton::is_master( ) ) {
    Matrix2< A > e( 1, 2, 3, 4 );
    for ( int i = 0; i < size; i++ ) {
      e = e * input[ i ];
    }
    *output = new Matrix2< A >[ 1 ];
    **output = e;
  }
}

template < typename A >
void scan_multiply_matrix_skel( const Matrix2< A >* input,
				Matrix2< A >** output )
{
  dist_list< Matrix2< A > > *as = new dist_list< Matrix2< A > >( size, input );
  Matrix2< A > e( 1, 2, 3, 4 );

  dist_list< Matrix2< A > > *bs
    = list_skeletons::scan( func_multiply_matrix< A >( ),
			    e, as );

  *output = NULL;
  if ( skeleton::is_master( ) ) *output = new Matrix2< A >[ size + 1 ];
  list_skeletons::gather( bs, *output );

  skeleton::safe_delete( bs );
  skeleton::safe_delete( as );
}

template < typename A >
void scan_multiply_matrix_skel_ptr( const Matrix2< A >* input,
				    Matrix2< A >** output )
{
  dist_list< Matrix2< A > > *as = new dist_list< Matrix2< A > >( size, input );
  Matrix2< A > e( 1, 2, 3, 4 );

  dist_list< Matrix2< A > > *bs = new dist_list< Matrix2< A > >( size + 1 );
  list_skeletons::scan( func_multiply_matrix_ptr< A >( ), &e, as, bs );

  *output = NULL;
  if ( skeleton::is_master( ) ) *output = new Matrix2< A >[ size + 1 ];
  list_skeletons::gather( bs, *output );

  skeleton::safe_delete( bs );
  skeleton::safe_delete( as );
}

template < typename A >
void scan_multiply_matrix_seq( const Matrix2< A >* input,
			       Matrix2< A >** output )
{
  *output = NULL;
  if ( skeleton::is_master( ) ) {
    Matrix2< A >* outputp = new Matrix2< A >[ size + 1 ];
    Matrix2< A > e( 1, 2, 3, 4 );
    outputp[ 0 ] = e;
    for ( int i = 0; i < size; i++ ) {
      outputp[ i + 1 ] = outputp[ i ] * input[ i ];
    }
    *output = outputp;
  }
}

template < typename A >
void scanr_multiply_matrix_skel( const Matrix2< A >* input,
				 Matrix2< A >** output )
{
  dist_list< Matrix2< A > > *as = new dist_list< Matrix2< A > >( size, input );
  Matrix2< A > e( 1, 2, 3, 4 );

  dist_list< Matrix2< A > > *bs
    = list_skeletons::scanr( func_multiply_matrix< A >( ),
			     e, as );

  *output = NULL;
  if ( skeleton::is_master( ) ) *output = new Matrix2< A >[ size + 1 ];
  list_skeletons::gather( bs, *output );

  skeleton::safe_delete( bs );
  skeleton::safe_delete( as );
}

template < typename A >
void scanr_multiply_matrix_skel_ptr( const Matrix2< A >* input,
				     Matrix2< A >** output )
{
  dist_list< Matrix2< A > > *as = new dist_list< Matrix2< A > >( size, input );
  Matrix2< A > e( 1, 2, 3, 4 );

  dist_list< Matrix2< A > > *bs = new dist_list< Matrix2< A > >( size + 1 );
  list_skeletons::scanr( func_multiply_matrix_ptr< A >( ), &e, as, bs );

  *output = NULL;
  if ( skeleton::is_master( ) ) *output = new Matrix2< A >[ size + 1 ];
  list_skeletons::gather( bs, *output );

  skeleton::safe_delete( bs );
  skeleton::safe_delete( as );
}

template < typename A >
void scanr_multiply_matrix_seq( const Matrix2< A >* input,
				Matrix2< A >** output )
{
  *output = NULL;
  if ( skeleton::is_master( ) ) {
    Matrix2< A >* outputp = new Matrix2< A >[ size + 1 ];
    Matrix2< A > e( 1, 2, 3, 4 );
    outputp[ size ] = e;
    for ( int i = size - 1; i >= 0; i-- ) {
      outputp[ i ] = input[ i ] * outputp[ i + 1 ];
    }
    *output = outputp;
  }
}

template < typename A >
void shiftl_matrix_skel( const Matrix2< A >* input,
			 Matrix2< A >** output )
{
  dist_list< Matrix2< A > > *as = new dist_list< Matrix2< A > >( size, input );
  Matrix2< A > e( 1, 2, 3, 4 );

  dist_list< Matrix2< A > > *bs = list_skeletons::shiftl( e, as );

  *output = NULL;
  if ( skeleton::is_master( ) ) *output = new Matrix2< A >[ size ];
  list_skeletons::gather( bs, *output );

  skeleton::safe_delete( bs );
  skeleton::safe_delete( as );
  
}

template< typename A >
void shiftl_ow_matrix_skel( const Matrix2< A >* input,
			    Matrix2< A >** output )
{
  dist_list< Matrix2< A > > *as = new dist_list< Matrix2< A > >( size, input );
  Matrix2< A > e( 1, 2, 3, 4 );

  list_skeletons::shiftl_ow( e, as );

  *output = NULL;
  if ( skeleton::is_master( ) ) *output = new Matrix2< A >[ size ];
  list_skeletons::gather( as, *output );

  skeleton::safe_delete( as );
}

template< typename A >
void shiftl_matrix_seq( const Matrix2< A >* input,
			Matrix2< A >** output )
{
  *output = NULL;
  if ( skeleton::is_master( ) ) {
    Matrix2< A >* outputp = new Matrix2< A >[ size ];
    Matrix2< A > e( 1, 2, 3, 4 );
    for ( int i = 0; i < size - 1; i++ ) {
      outputp[ i ] = input[ i + 1 ];
    }
    outputp[ size - 1 ] = e;
    *output = outputp;
  }
}

template < typename A >
void shiftr_matrix_skel( const Matrix2< A >* input,
			 Matrix2< A >** output )
{
  dist_list< Matrix2< A > > *as = new dist_list< Matrix2< A > >( size, input );
  Matrix2< A > e( 1, 2, 3, 4 );

  dist_list< Matrix2< A > > *bs = list_skeletons::shiftr( e, as );

  *output = NULL;
  if ( skeleton::is_master( ) ) *output = new Matrix2< A >[ size ];
  list_skeletons::gather( bs, *output );

  skeleton::safe_delete( bs );
  skeleton::safe_delete( as );
  
}

template< typename A >
void shiftr_ow_matrix_skel( const Matrix2< A >* input,
			    Matrix2< A >** output )
{
  dist_list< Matrix2< A > > *as = new dist_list< Matrix2< A > >( size, input );
  Matrix2< A > e( 1, 2, 3, 4 );

  list_skeletons::shiftr_ow( e, as );

  *output = NULL;
  if ( skeleton::is_master( ) ) *output = new Matrix2< A >[ size ];
  list_skeletons::gather( as, *output );

  skeleton::safe_delete( as );
}

template< typename A >
void shiftr_matrix_seq( const Matrix2< A >* input,
			Matrix2< A >** output )
{
  *output = NULL;
  if ( skeleton::is_master( ) ) {
    Matrix2< A >* outputp = new Matrix2< A >[ size ];
    Matrix2< A > e( 1, 2, 3, 4 );
    for ( int i = size - 1; i > 0; i-- ) {
      outputp[ i ] = input[ i - 1 ];
    }
    outputp[ 0 ] = e;
    *output = outputp;
  }
}

void cataj_char_matrix_skel( const char* input,
			     Matrix2< char >** output )
{
  dist_list< char > *as = new dist_list< char >( size, input );
  Matrix2< char > e( 1, 2, 3, 4 );
  Matrix2< char > result
    = list_skeletons::cataj( func_multiply_matrix< char >( ),
			     func_char_to_matrix_char( ),
			     e, as );

  *output = NULL;
  if ( skeleton::is_master( ) ) {
    *output = new Matrix2< char >[ 1 ];
    **output = result;
  }

  skeleton::safe_delete( as );
}

void cataj_char_matrix_skel_ptr( const char* input,
				 Matrix2< char >** output )
{
  dist_list< char > *as = new dist_list< char >( size, input );
  Matrix2< char > e( 1, 2, 3, 4 );
  Matrix2< char > result;

  list_skeletons::cataj( func_multiply_matrix_ptr< char >( ),
			 func_char_to_matrix_char_ptr( ),
			 &e, as, &result );

  *output = NULL;
  if ( skeleton::is_master( ) ) {
    *output = new Matrix2< char >[ 1 ];
    **output = result;
  }

  skeleton::safe_delete( as );
}

void cataj_char_matrix_map_reduce( const char* input,
				   Matrix2< char >** output )
{
  dist_list< char > *as = new dist_list< char >( size, input );
  Matrix2< char > e( 1, 2, 3, 4 );
  dist_list< Matrix2< char > > *bs
    = list_skeletons::map( func_char_to_matrix_char( ), as );
  Matrix2< char > result
    = list_skeletons::reduce( func_multiply_matrix< char >( ), e, bs );

  *output = NULL;
  if ( skeleton::is_master( ) ) {
    *output = new Matrix2< char >[ 1 ];
    **output = result;
  }

  skeleton::safe_delete( bs );
  skeleton::safe_delete( as );
}

void accumulate_char_matrix_skel( const char* input,
				  Matrix2< int >** output )
{
  dist_list< char > *as = new dist_list< char >( size, input );
  Matrix2< char > e( 1, 2, 3, 4 );
  Matrix2< int > result
    = list_skeletons::accumulate( func_matrix_char_to_int( ),
				  func_matrix_shift_square( ),
				  func_multiply_matrix< int >( ),
				  func_char_to_matrix_char( ), 
				  func_multiply_matrix< char >( ),
				  e, as );

  *output = NULL;
  if ( skeleton::is_master( ) ) {
    *output = new Matrix2< int >[ 1 ];
    **output = result;
  }

  skeleton::safe_delete( as );
}

void accumulate_char_matrix_skel_ptr( const char* input,
				      Matrix2< int >** output )
{
  dist_list< char > *as = new dist_list< char >( size, input );
  Matrix2< char > e( 1, 2, 3, 4 );
  Matrix2< int > result;
  list_skeletons::accumulate( func_matrix_char_to_int_ptr( ),
			      func_matrix_shift_square_ptr( ),
			      func_multiply_matrix_ptr< int >( ),
			      func_char_to_matrix_char_ptr( ), 
			      func_multiply_matrix_ptr< char >( ),
			      &e, as, &result );

  *output = NULL;
  if ( skeleton::is_master( ) ) {
    *output = new Matrix2< int >[ 1 ];
    **output = result;
  }

  skeleton::safe_delete( as );
}

void accumulate_char_matrix_diffusion( const char* input,
				       Matrix2< int >** output )
{
  dist_list< char > *as = new dist_list< char >( size, input );
  Matrix2< char > e( 1, 2, 3, 4 );
  dist_list< Matrix2< char > > *bs
    = list_skeletons::map( func_char_to_matrix_char( ), as );
  dist_list< Matrix2< char > > *cs
    = list_skeletons::scan( func_multiply_matrix< char >( ), e, bs );

  Matrix2< char >* cs_cp = NULL;
  if ( skeleton::is_master( ) ) cs_cp = new Matrix2< char >[ size + 1 ];
  list_skeletons::gather( cs, cs_cp );

  dist_list< Matrix2< char > > *ds
    = new dist_list< Matrix2< char > >( size, cs_cp );
  dist_list< Matrix2< int > > *es
    = list_skeletons::zipwith( func_matrix_shift_square( ), as, ds );
  Matrix2< int > cp;
  if ( skeleton::is_master( ) ) cp = func_matrix_char_to_int( )( cs_cp[ size ] );
  skeleton::safe_delete_array( cs_cp );

  Matrix2< int > unit( 1, 0, 0, 1 );
  Matrix2< int > sub_result
    = list_skeletons::reduce( func_multiply_matrix< int >( ), unit, es );
  Matrix2< int > result = func_multiply_matrix< int >( )( sub_result, cp );

  *output = NULL;
  if ( skeleton::is_master( ) ) {
    *output = new Matrix2< int >[ 1 ];
    **output = result;
  }

  skeleton::safe_delete( es );
  skeleton::safe_delete( ds );
  skeleton::safe_delete( cs );
  skeleton::safe_delete( bs );
  skeleton::safe_delete( as );
}  

////////////////////////////////////////////////////////////////////////
// Main routine
  
int SketoMain( int argc, char*argv[] )
{
  std::srand( std::time( NULL ) );

  if ( argc <= 1 + 0 ) {
    size = 25;
  } else {
    size = std::atoi( argv[ 1 ] );
  }
  number = 0; count = 0;
  
  check_routine( generator_char,
		 dist_gather_skel_char, do_nothing< char >, 
		 "Distribute and Gather skeleton for char", size );

  check_routine( generator_char,
		 map_square_double_char_skel1, map_square_double_char_seq, 
		 "map skeleton with square function", size );

  check_routine( generator_char,
		 map_square_double_char_skel2, map_square_double_char_seq,
		 "map skeleton with square function via pointer", size );

  check_routine( generator_char,
		 map_square_char_char_skel, map_square_char_char_seq,
		 "overiting map skeleton with square function", size );

  check_routine( generator_char,
		 zip_shift_or_index, zip_shift_or_index_seq,
		 "zip skeleton followed by map with shift-or", size );

  check_routine( generator_char,
		 mapindex_shift_or_index, zip_shift_or_index_seq,
		 "mapindex skeleton with shift-or", size );

  check_routine( generator_char,
		 mapindex_shift_or_index_ptr, zip_shift_or_index_seq,
		 "mapindex skeleton with shift-or pointer function", size );

  check_routine( generator_char,
		 mapindex_ow_shift_or_index, zip_shift_or_index_seq,
		 "mapindex_ow skeleton with shift-or", size );

  check_routine( generator_matrix_int,
		 reduce_multiply_matrix_skel< int >,
		 reduce_multiply_matrix_seq< int >,
		 "reduce skeleton with matrix multiplication", 1 );

  check_routine( generator_matrix_int,
		 reduce_multiply_matrix_skel_ptr< int >,
		 reduce_multiply_matrix_seq< int >,
		 "reduce skeleton with matrix multiplication with pointers",
		 1 );

  check_routine( generator_matrix_int,
		 scan_multiply_matrix_skel< int >,
		 scan_multiply_matrix_seq< int >,
		 "scan skeleton with matrix multiplication", size + 1 );

  check_routine( generator_matrix_int,
		 scan_multiply_matrix_skel_ptr< int >,
		 scan_multiply_matrix_seq< int >,
		 "scan skeleton with matrix multiplication with pointers",
		 size + 1 );

  check_routine( generator_matrix_int,
		 scanr_multiply_matrix_skel< int >,
		 scanr_multiply_matrix_seq< int >,
		 "scanr skeleton with matrix multiplication", size + 1 );

  check_routine( generator_matrix_int,
		 scanr_multiply_matrix_skel_ptr< int >,
		 scanr_multiply_matrix_seq< int >,
		 "scanr skeleton with matrix multiplication with pointers",
		 size + 1 );

  check_routine( generator_matrix_int,
		 shiftl_matrix_skel< int >, shiftl_matrix_seq< int >,
		 "shiftl skeleton over list of matrices", size );

  check_routine( generator_matrix_int,
		 shiftl_ow_matrix_skel< int >, shiftl_matrix_seq< int >,
		 "shiftl_ow skeleton over list of matrices", size );

  check_routine( generator_matrix_int,
		 shiftr_matrix_skel< int >, shiftr_matrix_seq< int >,
		 "shiftr skeleton over list of matrices", size );

  check_routine( generator_matrix_int,
		 shiftr_ow_matrix_skel< int >, shiftr_matrix_seq< int >,
		 "shiftr_ow skeleton over list of matrices", size );

  check_routine( generator_char,
		 cataj_char_matrix_skel, cataj_char_matrix_map_reduce,
		 "cataj skeleton with matrix multiplication", 1 );

  check_routine( generator_char,
		 cataj_char_matrix_skel_ptr, cataj_char_matrix_map_reduce,
		 "cataj skeleton with matrix multiplication with pointers", 1 );

  check_routine( generator_char,
		 accumulate_char_matrix_skel, accumulate_char_matrix_diffusion,
		 "accumulate skeleton with matrix multiplication", 1 );

  check_routine( generator_char,
		 accumulate_char_matrix_skel_ptr, accumulate_char_matrix_diffusion,
		 "accumulate skeleton with matrix multiplication with pointers", 1 );

  if ( skeleton::is_master( ) ) {
    std::cout << "Checked: " << number << ", success: " << count
	      << ", failure: " << ( number - count ) << std::endl;
    std::cout << "Check routine of list_skeleton ended" << std::endl;
  }
  return 0;
}
