/***
 * Copyright (c) 2005, SkeTo Project
 * All rights reserved.
 */
/**
 * @file Serializer.h
 * @brief Template utility for serialization (which is used in communication of skeletons.)
 * @author Kento EMOTO
 *
 */
#ifndef __SERIALIZER_H__
#define __SERIALIZER_H__

#include<utility>
/**
 * @brief A default template class to serialize an object of class A.
 *
 * This is default implementation and invokes @c A::serialize(buf), @c A::construct(buf),  @c A::getSerializedSize() and @c A::needsSerializing().
 *
 * Instances for some primitive types are written in this file.
 * If you use other primitive types, you should make instances for them.
 * For classes used in skeletons, i.e. dist_pair and matrix, 
 * you don't have to are also written in this file.
 * If you use classes except those used in skeletons, i.e. dist_pair and matrix, , you should do either of the following:
 *  1. add members @c A::serialize(buf), @c A::construct(buf),  @c A::getSerializedSize() and @c A::needsSerializing() which are used in default Shadower,
 *  2. write an instance of Shadower for the class.
 */
template <typename A>
class Serializer {
public:
	/**
	 * @brief Construct an object @a obj from the serialized data stream (@a buf).
	 *
	 * This function just invokes A::construct(buf).
	 * @param obj an object to be constructed.
	 * @param buf a serialized data stream from which the object is constructed.
	 * @return A pointer indicating the head of the rest of the stream.
	 * @see matrix<A>::construct()
	 */
	static unsigned char* construct(A *obj, unsigned char *buf){
		return obj->construct(buf);
	}
	/**
	 * @brief Serialize an object @a obj to the stream (@a buf).
	 *
	 * This function just invokes A::serialize(buf).
	 * @param obj an object to be serialized.
	 * @param buf a stream to which the object is serialized.
	 * @return A pointer indicating the tail of the stream.
	 * @see matrix<A>::serialize()
	 */
	static unsigned char* serialize(const A& obj, unsigned char *buf){
		return obj.serialize(buf);
	}
	/**
	 * @brief Return the size of serialized data of @a obj.
	 *
	 * This function just invokes A::getSerializedSize().
	 * @param obj an object to be serialized.
	 * @return The size of serialized data of @a obj.
	 * @see matrix<A>::getSerializedSize()
	 */
	static int getSerializedSize(const A& obj){
		return obj.getSerializedSize();
	}
	/**
	 * @brief Return whether this needs serialization in communication.
	 *
	 * This function just invokes A::needsSerializing().
	 * @see matrix<A>::needsSerializing()
	 */
	static bool needsSerializing(){
		return A::needsSerializing(); 
	}
};
/**
 * @brief Instance of Serializer for int
 */
template <>
class Serializer<int> {
public:
	/**
	 * @brief Just casting.
	 */
	static unsigned char* construct(int*obj, unsigned char *buf){
		*obj=*(int*)buf;		
		return buf+sizeof(int);
	}
	/**
	 * @brief Just casting.
	 */
	static unsigned char* serialize(const int& obj, unsigned char *buf){
		*((int*)buf) = obj;
		return buf+sizeof(int);
	}
	/**
	 * @brief Just sizeof.
	 */
	static int getSerializedSize(const int& obj){
		return sizeof(obj);
	}
	/**
	 * @brief Primitives don't need serialization.
	 */
	static bool needsSerializing(){
		return false; 
	}
};
/**
 * @brief Instance of Serializer for unsigned int
 */
template <>
class Serializer<unsigned int> {
public:
	/**
	 * @brief Just casting.
	 */
	static unsigned char* construct(unsigned int *obj, unsigned char *buf){
		*obj=*(unsigned int*)buf;		
		return buf+sizeof(unsigned int);
	}
	/**
	 * @brief Just casting.
	 */
	static unsigned char* serialize(const unsigned int& obj, unsigned char *buf){
		*((unsigned int*)buf) = obj;
		return buf+sizeof(unsigned int);
	}
	/**
	 * @brief Just sizeof.
	 */
	static int getSerializedSize(const unsigned int& obj){
		return sizeof(obj);
	}
	/**
	 * @brief Primitives don't need serialization.
	 */
	static bool needsSerializing(){
		return false; 
	}
};
/**
 * @brief Instance of Serializer for char
 */
template <>
class Serializer<char> {
public:
	/**
	 * @brief Just casting.
	 */
	static unsigned char* construct(char*obj, unsigned char *buf){
		*obj=*(char*)buf;		
		return buf+sizeof(char);
	}
	/**
	 * @brief Just casting.
	 */
	static unsigned char* serialize(const char& obj, unsigned char *buf){
		*((char*)buf) = obj;
		return buf+sizeof(char);
	}
	/**
	 * @brief Just sizeof.
	 */
	static int getSerializedSize(const char& obj){
		return sizeof(obj);
	}
	/**
	 * @brief Primitives don't need serialization.
	 */
	static bool needsSerializing(){
		return false; 
	}
};
/**
 * @brief Instance of Serializer for unsinged char
 */
template <>
class Serializer<unsigned char> {
public:
	/**
	 * @brief Just casting.
	 */
	static unsigned char* construct(unsigned char* obj, unsigned char *buf){
		*obj=*(unsigned char*)buf;		
		return buf+sizeof(unsigned char);
	}
	/**
	 * @brief Just casting.
	 */
	static unsigned char* serialize(const unsigned char& obj, unsigned char *buf){
		*((unsigned char*)buf) = obj;
		return buf+sizeof(unsigned char);
	}
	/**
	 * @brief Just sizeof.
	 */
	static int getSerializedSize(const unsigned char& obj){
		return sizeof(obj);
	}
	/**
	 * @brief Primitives don't need serialization.
	 */
	static bool needsSerializing(){
		return false; 
	}
};
/**
 * @brief Instance of Serializer for bool
 */
template <>
class Serializer<bool> {
public:
	/**
	 * @brief Just casting.
	 */
	static unsigned char* construct(bool* obj, unsigned char *buf){
		*obj=*(bool*)buf;		
		return buf+sizeof(bool);
	}
	/**
	 * @brief Just casting.
	 */
	static unsigned char* serialize(const bool& obj, unsigned char *buf){
		*((bool*)buf) = obj;
		return buf+sizeof(bool);
	}
	/**
	 * @brief Just sizeof.
	 */
	static int getSerializedSize(const bool& obj){
		return sizeof(obj);
	}
	/**
	 * @brief Primitives don't need serialization.
	 */
	static bool needsSerializing(){
		return false; 
	}
};

/**
 * @brief Instance of Serializer for double
 */
template <>
class Serializer<double> {
public:
	/**
	 * @brief Just casting.
	 */
	static unsigned char* construct(double* obj, unsigned char *buf){
		*obj=*(double*)buf;		
		return buf+sizeof(double);
	}
	/**
	 * @brief Just casting.
	 */
	static unsigned char* serialize(const double& obj, unsigned char *buf){
		*((double*)buf) = obj;
		return buf+sizeof(double);
	}
	/**
	 * @brief Just sizeof.
	 */
	static int getSerializedSize(const double& obj){
		return sizeof(obj);
	}
	/**
	 * @brief Primitives don't need serialization.
	 */
	static bool needsSerializing(){
		return false; 
	}
};

/**
 * @brief Instance of Serializer for double
 */
template <>
class Serializer<const double> {
public:
	/**
	 * @brief Just casting.
	 */
	static unsigned char* construct(double* obj, unsigned char *buf){
		*obj=*(double*)buf;		
		return buf+sizeof(double);
	}
	/**
	 * @brief Just casting.
	 */
	static unsigned char* serialize(const double& obj, unsigned char *buf){
		*((double*)buf) = obj;
		return buf+sizeof(double);
	}
	/**
	 * @brief Just sizeof.
	 */
	static int getSerializedSize(const double& obj){
		return sizeof(obj);
	}
	/**
	 * @brief Primitives don't need serialization.
	 */
	static bool needsSerializing(){
		return false; 
	}
};


/**
 * @brief Instance of Serializer for float
 */
template <>
class Serializer<float> {
public:
	/**
	 * @brief Just casting.
	 */
	static unsigned char* construct(float* obj, unsigned char *buf){
		*obj=*(float*)buf;		
		return buf+sizeof(float);
	}
	/**
	 * @brief Just casting.
	 */
	static unsigned char* serialize(const float& obj, unsigned char *buf){
		*((float*)buf) = obj;
		return buf+sizeof(float);
	}
	/**
	 * @brief Just sizeof.
	 */
	static int getSerializedSize(const float& obj){
		return sizeof(obj);
	}
	/**
	 * @brief Primitives don't need serialization.
	 */
	static bool needsSerializing(){
		return false; 
	}
};



template <typename A, typename B>
class Serializer<const std::pair<A,B> > {
public:
	/**
	 * @brief Construct this pair from the stream (@a buf).
	 *
	 * This function constructs this pair from a stream (@a buf) and
	 * returns a pointer indicating the head of rest of the stream.
	 * @param buf a stream of a serialized pair.
	 * @param obj the object to be constructed.
	 * @return a pointer indicating the head of rest of the stream.
	 */
 public:	
	static unsigned char* construct(std::pair<A,B>* obj, unsigned char *buf){
		buf=Serializer<A>::construct(&obj->first, buf);
		buf=Serializer<B>::construct(&obj->second, buf);
		return buf;
	}
	/**
	 * @brief Serialize this pair to the stream (@a buf).
	 *
	 * This function serializes this pair into the stream (@a buf) and
	 * returns a pointer indicating the tail of the stream.
	 * @param buf a stream to which the pair is serialized.
	 * @param obj the object to be serialized.
	 * @return A pointer indicating the tail of the stream.
	 */
 public:
	static unsigned char* serialize(const std::pair<A,B>& obj, unsigned char *buf){
		buf=Serializer<A>::serialize(obj.first, buf);
		buf=Serializer<B>::serialize(obj.second, buf);
		return buf;
	}
	/**
	 * @brief Return the size of serialized this pair.
	 *
	 * This function returns the size of serialized data of this pair.
	 * @return The size of serialized data of this pair.
	 * @param obj the object to be serialized.
	 */
 public:
	static int getSerializedSize(const std::pair<A,B>& obj){
		int size = 0;
		size+=Serializer<A>::getSerializedSize(obj.first);
		size+=Serializer<B>::getSerializedSize(obj.second);
		return size;
	}
	/**
	 * @brief pair needs serialization.
	 */
	static bool needsSerializing(){
		return true; 
	}
};

template <typename A, typename B>
class Serializer<std::pair<A,B> > {
public:
	/**
	 * @brief Construct this pair from the stream (@a buf).
	 *
	 * This function constructs this pair from a stream (@a buf) and
	 * returns a pointer indicating the head of rest of the stream.
	 * @param buf a stream of a serialized pair.
	 * @param obj the object to be constructed.
	 * @return a pointer indicating the head of rest of the stream.
	 */
 public:	
	static unsigned char* construct(std::pair<A,B>* obj, unsigned char *buf){
		buf=Serializer<A>::construct(&obj->first, buf);
		buf=Serializer<B>::construct(&obj->second, buf);
		return buf;
	}
	/**
	 * @brief Serialize this pair to the stream (@a buf).
	 *
	 * This function serializes this pair into the stream (@a buf) and
	 * returns a pointer indicating the tail of the stream.
	 * @param buf a stream to which the pair is serialized.
	 * @param obj the object to be serialized.
	 * @return A pointer indicating the tail of the stream.
	 */
 public:
	static unsigned char* serialize(const std::pair<A,B>& obj, unsigned char *buf){
		buf=Serializer<A>::serialize(obj.first, buf);
		buf=Serializer<B>::serialize(obj.second, buf);
		return buf;
	}
	/**
	 * @brief Return the size of serialized this pair.
	 *
	 * This function returns the size of serialized data of this pair.
	 * @param obj the object to be serialized.
	 * @return The size of serialized data of this pair.
	 */
 public:
	static int getSerializedSize(const std::pair<A,B>& obj){
		int size = 0;
		size+=Serializer<A>::getSerializedSize(obj.first);
		size+=Serializer<B>::getSerializedSize(obj.second);
		return size;
	}
	/**
	 * @brief pair needs serialization.
	 */
	static bool needsSerializing(){
		return true; 
	}
};



#endif // __SERIALIZER_H__
