/***
 * Copyright (c) 2005, SkeTo Project
 * All rights reserved.
 */
/**
 * @file matrix.h
 * @brief The header and implementation of class @a matrix and @a m_skeletons.
 *
 * @a matrix and @a m_skeleton can be regarded as 
 * sequential versions of dist_matrix and matrix_skeletons.
 * @author Kento EMOTO
 */
#ifndef _MATRIX_H_
#define _MATRIX_H_

#include "Serializer.h"
#include "Shadower.h"
#include "../util/functions.h"

#include <iostream>
#include <iomanip>
#include <assert.h>
#include <utility>




class m_skeletons;

/**
 * @brief Sequential Matrix which can be used with skeletons.
 *
 * This class is also used in dist_matrix.
 * The matrix can be stored in column-major, 
 * however, operations often need matrices to be row-major.
 * Please be careful to use column-major mode.
 *
 * The matrix is consists of three constructors 'abv'(above), 'bsd'(beside) and 'sing'(singleton).
 * So, in haskell notation, datatype matrix is as follows:
 * @code
 * data Matrix a = Sing a
 *         | Abv (Matrix a) (Matrix a)
 *         | Bsd (Matrix a) (Matrix a)
 * @endcode
 * A singleton matrix with an element E is 'Sing E' or '|E|' for brevity.
 * A matrix made by placing matrix B on the left of matrix C is 'Bsd B C' or 'B `Bsd` C'. The heights of B and C should be the same.
 * A matrix made by placing matrix B on the top of matrix C is 'Abv B C' or 'B `Abv` C'. The widths of B and C should be the same.
 * The constructors 'Bsd' and 'Abv' are associative and have abiding property:
 * '(B `Bsd` C) `Abv` (E `Bsd` F) = (B `Abv` E) `Bsd` (C `Abv` F)'.
 * Skeletons are defined as recursive functions on the structure.
 * Indices are starting from 0.
 */
template<typename A>
class matrix{
	friend class m_skeletons; /**< The skeletons. */
private:
	/** @brief Size of the matrix (@a n x @a m). */
	int n;				
	/** @brief Size of the matrix (@a n x @a m). */
	int m;
	/** @brief Elements of this matrix stored in 1D array. */
	A *ele;
	/** @brief Pointer to each row/column. */
	A **idxEle;
	/** @brief Whether column-wisely stored. */
	bool colwise;
	/** 
	 * @brief Whether ele and idxEle can be freed. 
	 *
	 * This flag is set to false when this matrix is constructed by 
	 * the constructor with element and index arrays 
	 * (i.e. it is used as a wrapper of dist_matrix).
	 * or this matrix is shadow of some other matrix.
	 */
	bool canfree;
	/** @brief Whether this is shadow of other matrix. */
	bool shadowed;
	/**
	 * @brief Pointer to the original item which this shadows.
	 */
	matrix <A> const * original;
	
	/**
	 * @brief Returns whether it is shadowable.
	 *
	 * @return Always true.
	 */
 public:
	static const bool isShadowable(){ return true; }
	/**
	 * @brief Returns whether it is shadow of other pair.
	 *
	 * @retval true it is shadow(i.e. its elements are shadow of other elements ).
	 * @retval false it has entity.
	 */
 public:
	bool isShadowed() const{
		return shadowed;
	}
	/**
	 * @brief The argument matrix is made to be shadow of this matrix.
	 *
	 * @param res the matrix to be shdow of this matrix.
	 * i.e. @a res doesn't have entity.
	 */
 public:
	void shadow(matrix <A> *res) const {
		res->freeMemory();
		res->colwise = colwise;
		res->n = n;
		res->m = m;
		res->idxEle = idxEle;
		res->ele = ele;
		res->canfree = false;
		res->shadowed = true;
		res->original = this;
	}
	/**
	 * @brief Reshadow the original.
	 */
 private:
	void reshadow() {
		assert(original!=NULL);
		colwise = original->colwise;
		n = original->n;
		m = original->m;
		idxEle = original->idxEle;
		ele = original->ele;
	}

	/**
	 * @brief Return whether this needs serialization in communication.
	 *
	 * This function always returns true.
	 * @return Always true (needs serialization.)
	 */
 public:
	static bool needsSerializing(){
		return true;
	}
	/**
	 * @brief Whether column-wisely stored.
	 */
	bool isColumnWise() const {return colwise;}
	/**
	 * @brief The number of rows of this matrix.
	 */
	int getRows() const {return n;}
	/**
	 * @brief The number of columns of this matrix.
	 */
	int getCols() const {return m;}
	/**
	 * @brief Return referece of element at (@a i, @a j) of this matrix.
	 *
	 * @param i row-index of the element.
	 * @param j column-index of the element.
	 * @return Referece of element at (@a i, @a j).
	 */
	A &get(int i, int j) const {return idxEle[i][j];}
	/**
	 * @brief Return referece of element at (@a i, @a j) of this matrix.
	 *
	 * @param i row-index of the element.
	 * @param j column-index of the element.
	 * @return Referece of element at (@a i, @a j).
	 */
	A &at(int i, int j) {return idxEle[i][j];}
	/**
	 * @brief Return const referece of element at (@a i, @a j) of this matrix.
	 *
	 * @param i row-index of the element.
	 * @param j column-index of the element.
	 * @return Const referece of element at (@a i, @a j).
	 */
	const A &at(int i, int j) const {return idxEle[i][j];}
	/**
	 * @brief Put value @a v at (@a i, @a j) of this matrix.
	 *
	 * @param i row-index of the element.
	 * @param j column-index of the element.
	 * @param v the value to be set.
	 */
	void set(int i, int j, A &v) {idxEle[i][j] = v;}

	/******************* constructors  ****************************/

	/**
	 * @brief Default constructor.
	 *
	 * row-major stored
	 */
 public:
	matrix(){
		colwise = false;
		canfree = true;
		shadowed = false;
		original = NULL;
		init0();
	}
	/**
	 * @brief Constructor with size specification.
	 *
	 * @param nn the number of rows of this matrix.
	 * @param mm the number of columns of this matrix.
	 * @param cflg column-wisely stored if this flag is true (default false).
	 */
 public:
	matrix(int nn, int mm, bool cflg = false){
		colwise = cflg;
		canfree = true;
		shadowed = false;
		original = NULL;
		init0();
		setSize(nn, mm);
	}

	/**
	 * @brief Constructor with element and index arrays.
	 *
	 * The arrays @a es and @a idx are used as this ele and idxEle.
	 * These are not freed in the destructor.
	 * This constructor is used to wrap submatrices of dist_matrix with matrix.
	 *
	 * @param nn the number of rows of this matrix.
	 * @param mm the number of columns of this matrix.
	 * @param es the array of elements.
	 * @param idx pointer array to each row/col.
	 * @param cflg column-wisely stored if this flag is true (default false).
	 */
 public:
	matrix(int nn, int mm, A* es, A** idx, bool cflg = false){
		colwise = cflg;
		n = nn;
		m = mm;
		idxEle = idx;
		ele = es;
		canfree = false;
		shadowed = false;
		original = NULL;
	}

	/**
	 * @brief Constructor with initial values.
	 *
	 * Initial values are copied from the 2D array @a es.
	 *
	 * @param nn the number of rows of this matrix.
	 * @param mm the number of columns of this matrix.
	 * @param es the 2D array of elements.
	 * @param cflg column-wisely stored if this flag is true (default false).
	 */
 public:
	matrix(int nn, int mm, A** es, bool cflg = false){
		colwise = cflg;
		canfree = true;
		shadowed = false;
		original = NULL;
		init0();
		setSize(nn, mm);
		if(isColumnWise()){
			for(int j = 0; j < m; j++){
				for(int i = 0; i < n; i++){
					at(j, i) = es[i][j];
				}
			}
		} else {
			for(int i = 0; i < n; i++){
				for(int j = 0; j < m; j++){
					at(i, j) = es[i][j];
				}
			}
		}
	}

	/**
	 * @brief Resize the matrix.
	 * @param nn the number of rows of this matrix.
	 * @param mm the number of columns of this matrix.
	 */
 public:
	void setSize(int nn, int mm){
		assert(canfree);
		if(n!=nn || m!= mm){
			freeMemory();
			n = nn;
			m = mm;
			init();
			init2();
		}
	}
	/**
	 * @brief Initialize some fields.
	 *
	 * The matrix has no elements.
	 */
 private:
	void init0(){
		n = -1;
		m = -1;
		ele = NULL;
		idxEle = NULL;
	}
	/**
	 * @brief Initialize some fields.
	 * 
	 * An array for elements is allocated.
	 * Index array is still NULL.
	 */
 private:
	void init(){
		ele = new A[n*m];
		idxEle = NULL;
	}
 	/**
	 * @brief Initialize some fields.
	 *
	 * Index array is allocated and initilized according to colwise flag.
	 */
 private:
	void init2(){
		if(!colwise){
			if(idxEle!=NULL) delete [] idxEle;
			idxEle = new A*[n];	
			for(int i = 0; i < n; i++){
				idxEle[i] = &ele[i*m];
			}
		} else {
			if(idxEle!=NULL) delete [] idxEle;
			idxEle = new A*[m];				
			for(int i = 0; i < m; i++){
				idxEle[i] = &ele[i*n];
			}
		}
	}
	/*
	 * @brief Free the memory.
	 */
 private:
	void freeMemory(){
		if(!canfree)	return;
		if(ele!=NULL){
			delete [] ele;
			ele = NULL;
		}
		if(idxEle!=NULL) {
			delete [] idxEle;
			idxEle = NULL;
		}
	}
 	/**
	 * @brief A destructor.
	 */
 public:
	~matrix(){
		freeMemory();
	}

	/**
	 * @brief A copy constructor.
	 *
	 * If this matrix is shadow or cannot be freed, 
	 * these are reset and it becomes normal matrix.
	 */
 public:
	matrix(const matrix &mat){
		if(!mat.shadowed && !mat.canfree){
			colwise = mat.colwise;
			canfree = false;
			n = mat.n;
			m = mat.m;
			idxEle = mat.idxEle;
			ele = mat.ele;
			return;
		}
		shadowed = false;
		original = NULL;
		canfree = true;
		init0();
		colwise = mat.isColumnWise();
		setSize(mat.getRows(), mat.getCols());
		if(colwise){
			for(int j = 0; j < m; j++){
				for(int i = 0; i < n; i++){
					at(j, i) = mat.at(j, i);
				}
			}
		} else {
			for(int i = 0; i < n; i++){
				for(int j = 0; j < m; j++){
					at(i,j) = mat.at(i, j);
				}
			}
		}
	}
	/**
	 * @brief A substitution operator.
	 *
	 * If this matrix is shadow, this reshadows the original.
	 */
 public:
	matrix& operator=(const matrix &mat){
		if(shadowed) {
			reshadow();
			return *this;
		}
		assert(canfree);
		colwise = mat.isColumnWise();
		setSize(mat.getRows(), mat.getCols());
		if(colwise){
			for(int j = 0; j < m; j++){
				for(int i = 0; i < n; i++){
					at(j, i) = mat.at(j, i);
				}
			}
		} else {
			for(int i = 0; i < n; i++){
				for(int j = 0; j < m; j++){
					at(i,j) = mat.at(i, j);
				}
			}
		}
		return *this;
	}
	/**
	 * @brief An equivalence operator.
	 */
 public:
	bool operator==(const matrix &mat) const{
   		if(isColumnWise() != mat.isColumnWise()) return false;
		if(getRows() != mat.getRows()) return false;
		if(getCols() != mat.getCols()) return false;
		if(colwise){
			for(int j = 0; j < m; j++){
				for(int i = 0; i < n; i++){
					if(!(at(j,i) == mat.at(j, i))) return false;
				}
			}
		} else {
			for(int i = 0; i < n; i++){
				for(int j = 0; j < m; j++){
					if(!(at(i, j) == mat.at(i, j))) return false;
				}
			}
		}
		return true;
	}
	/**
	 * @brief Initializing by generate function.
	 * @param f a generator function object.
	 *   f should have void operator()(A&, int, int) const.
	 * A sample of function object is as follows: 
	 * (the value of element at (i, j) is i + j)
	 * @code
	 * class Gen {
	 * public:
	 *     void operator()(int &r, int i, int j) const {
	 *         r = i + j;
	 *     }
	 * };
	 * @endcode
	 */
public:
	template <typename F>
	void generate(const F &f){
		if(shadowed) {
			reshadow();
			return;
		}
		if(isColumnWise()){
			for(int j = 0; j < m; j++){
				for(int i = 0; i < n; i++){
					f(i, j, &at(j, i));
				}
			}
		} else {
			for(int i = 0; i < n; i++){
				for(int j = 0; j < m; j++){
					f(i, j, &at(i, j));
				}
			}
		}
	}
	/**
	 * @brief Transposing copy.
	 *
	 * Matrices should be row-major.
	 * @param mat a matrix to which the copy of transposition of this matris is stored.
	 */
 public:
	void copyTr(const matrix &mat){
		assert(canfree && !shadowed && !mat.isColumnWise());
		setSize(mat.getRows(), mat.getCols());
		for(int i = 0; i < n; i++){
			for(int j = 0; j < m; j++){
				at(i, j) = mat.at(j, i);
			}
		}
		return;
	}

	/******************* for serialization  ****************************/
	/**
	 * @brief Construct this matrix from the stream (@a buf).
	 *
	 * This function constructs this matrix from a stream (@a buf) and
	 * returns a pointer indicating the head of rest of the stream.
	 * Shadowing is reset when this matrix is shadow.
	 * @param buf a stream of a serialized pair.
	 * @return a pointer indicating the head of rest of the stream.
	 */
 public:
	unsigned char* construct(unsigned char *buf){
 		if(shadowed){
			ele = NULL;
			idxEle = NULL;
		}
		int nn, mm;
		shadowed = false;
		original = NULL;
		canfree = true;
		buf=Serializer<int>::construct(&nn, buf);
		buf=Serializer<int>::construct(&mm, buf);
		buf=Serializer<bool>::construct(&colwise, buf);
		setSize(nn, mm);
		if(colwise){
			for(int j = 0; j < m; j++){
				for(int i = 0; i < n; i++){
					buf=Serializer<A>::construct(&at(j, i), buf);
				}
			}
		} else {
			for(int i = 0; i < n; i++){
				for(int j = 0; j < m; j++){
					buf=Serializer<A>::construct(&at(i, j), buf);
				}
			}
		}		
		return buf;
	}
	/**
	 * @brief Serialize this matrix to the stream (@a buf).
	 *
	 * This function serializes this matrix into the stream (@a buf) and
	 * returns a pointer indicating the tail of the stream.
	 * @param buf a stream to which the matrix is serialized.
	 * @return A pointer indicating the tail of the stream.
	 */
 public:
	unsigned char* serialize(unsigned char *buf) const{
		buf=Serializer<int>::serialize(n, buf);
		buf=Serializer<int>::serialize(m, buf);
		buf=Serializer<bool>::serialize(colwise, buf);
		if(colwise){
			for(int j = 0; j < m; j++){
				for(int i = 0; i < n; i++){
					buf=Serializer<A>::serialize(at(j, i), buf);
				}
			}
		} else {
			for(int i = 0; i < n; i++){
				for(int j = 0; j < m; j++){
					buf=Serializer<A>::serialize(at(i, j), buf);
				}
			}
		}		
		return buf;
	}
	/**
	 * @brief Return the size of serialized data of this matrix.
	 *
	 * This function returns the size of serialized data of this matrix.
	 * @return The size of serialized data of this matrix.
	 */
 public:
	int getSerializedSize() const{
		int size = 0;
		size+=Serializer<int>::getSerializedSize(n);
		size+=Serializer<int>::getSerializedSize(m);
		size+=Serializer<bool>::getSerializedSize(colwise);
		if(colwise){
			for(int j = 0; j < m; j++){
				for(int i = 0; i < n; i++){
					size+=Serializer<A>::getSerializedSize(at(j, i));
				}
			}
		} else {
			for(int i = 0; i < n; i++){
				for(int j = 0; j < m; j++){
					size+=Serializer<A>::getSerializedSize(at(i, j));
				}
			}
		}		
		return size;
	}
	
	/**
	 * @brief Make this matrix a singleton of value @a v.
	 * @param v an element of the singleton.
	 */
 public:
	void singleton(const A &v){
		if(shadowed) {
			reshadow();
			return;
		}
		setSize(1, 1);
		at(0, 0) = v;
	}


	/**
	 * @brief Make this matrix an identity matrix with @a one and @a zero.
	 * 
	 * This matrix should be row-major.
	 * @param one an element to be stored in the diagonal elements.
	 * @param zero an element to be stored in other elements.
	 */
 public:
	void eye(const A &one, const A &zero) {
		assert(!isColumnWise());
		int n = getCols();
		int m = getRows();
		for(int i = 0; i < m; i++){
			for(int j = 0; j < n; j++){
				if(i==j) at(i, j) = one;
				else     at(i, j) = zero;
			}
		}
	}
	/**
	 * @brief Fill up this matrix with an element @a zero.
	 * 
	 * This matrix should be row-major.
	 * @param zero an element to be stored in all elements.
	 */
 public:
	void fill(const A &zero) {
		assert(!isColumnWise());
		int n = getCols();
		int m = getRows();
		for(int i = 0; i < m; i++){
			for(int j = 0; j < n; j++){
				at(i, j) = zero;
			}
		}
	}
	
	/**
	 * @brief Transpose this matrix.
	 */
 public:
	void tr(){
		if(shadowed) {
			reshadow();
			return;
		}
		assert(canfree);
		if(getRows()==1){
			n = m;
			m = 1;
			init2();
		} else if(getCols()==1){
			m = n;
			n = 1;
			init2();
		} else {
			A *oele = ele;
			A **oidxEle = idxEle;
			ele = NULL;
			idxEle = NULL;
			int tmp = m;
			m = n;
			n = tmp;
			init();
			init2();
			if(isColumnWise()){
				for(int i = 0; i < n; i++){
					for(int j = 0; j < m; j++){
						idxEle[j][i] = oidxEle[i][j];
					}
				}
			} else {
				for(int i = 0; i < n; i++){
					for(int j = 0; j < m; j++){
						idxEle[i][j] = oidxEle[j][i];
					}
				}
			}
			delete [] oele;
			delete [] oidxEle;
		}
	}

	/**
	 * @brief Make this matrix columnwise.
	 */
 public:
	void makeColumnWise(){
		if(colwise) return;
		if(shadowed) {
			reshadow();
			return;
		}
		assert(canfree);
		colwise = true;
		A *nele = new A[n*m];
		A **nidxEle = new A*[m];
		int i;
		A *p;
		for(i = 0; i < m; i++){
			p = &nele[i*n];
			nidxEle[i] = p;
			for(int j = 0; j < n; j++){
				*p++ = idxEle[j][i];
			}
		}
		for(;i < m; i++){
			nidxEle[i] = &nele[i*n];
		}
		delete [] ele;
		delete [] idxEle;
		ele = nele;
		idxEle = nidxEle;		
	}
	/**
	 * @brief Make this matrix rowwise.
	 */
 public:
	void makeRowWise(){
		if(!colwise) return;
		if(shadowed) {
			reshadow();
			return;
		}
		assert(canfree);
		colwise = false;
		A *nele = new A[n*m];
		A **nidxEle = new A*[n];
		int i;
		A *p;
		for(i = 0; i < n; i++){
			p = &nele[i*m];
			nidxEle[i] = p;
			for(int j = 0; j < m; j++){
				*p++ = idxEle[j][i];
			}
		}
		for(;i < n; i++){
			nidxEle[i] = &nele[i*m];
		}
		delete [] ele;
		delete [] idxEle;
		ele = nele;
		idxEle = nidxEle;
	}
	
	/**
	 * @brief Print whole matrix to output-stream.
	 *
	 * @param os an output-stream.
	 */
 public:
	void print(std::ostream& os) const{
		if(!colwise) {
			for(int i = 0; i < n; i++){
				os << "| ";
				for(int j = 0; j < m; j++){
					os << at(i, j) << " ";
				}
				os << "|" << std::endl;
			}
		} else {
			for(int i = 0; i < n; i++){
				os << "| ";	
				for(int j = 0; j < m; j++){
					os << at(j, i) << " ";
				}
				os << "|" << std::endl;
			}
		}
	}
 public:
	void printStat(std::ostream& os) const{
		os << "n = " << n << std::endl;
		os << "m = " << m << std::endl;
		os << "canfree = " << canfree << std::endl;
		os << "shadowed = " << shadowed << std::endl;
	}
	/**
	 * @brief print whole matrix to output-stream.
	 *
	 * @param stream an output-stream.
	 * @param mat a matrix to be printed.
	 */
 public:
	friend std::ostream &operator<<(std::ostream& stream, const matrix &mat){
		mat.print(stream);
		return stream;
	}
	/**
	 * @brief Split this matrix horizontally.
	 *
	 * i.e. *this = @a matt `abv` @a matb, height @a matt = @a h.
	 *
	 * Matrices should be row-major.
	 * @param h the height of upper submatrix.
	 * @param matt to which the upper submatrix is stored.
	 * @param matb to which the lower submatrix is stored.
	 */
public:
	void splith(int h, matrix <A> *matt, matrix<A> *matb) const {
		assert(0<h && h < n && !colwise);
		matt->setSize(h, m);
		matb->setSize(n - h, m);
		for(int i = 0; i < h; i++){
			for(int j = 0; j < m; j++){
				matt->at(i, j) = at(i, j);
			}
		}
		for(int i = h, ii = 0; i < n; i++, ii++){
			for(int j = 0; j < m; j++){
				matb->at(ii, j) = at(i, j);
			}
		}
	}
	/**
	 * @brief Split this matrix vertically.
	 *
	 * i.e. *this = @a matl `bsd` @a matr, width @a matl = @a w.
	 *
	 * Matrices should be row-major.
	 * @param w the width of left submatrix.
	 * @param matl to which the left submatrix is stored.
	 * @param matr to which the right submatrix is stored.
	 */
public:
	void splitv(int w, matrix <A> *matl, matrix<A> *matr) const {
		assert(0<w && w < m && !colwise);
		matl->setSize(n, w);
		matr->setSize(n, m - w);
		for(int i = 0; i < n; i++){
			for(int j = 0; j < w; j++){
				matl->at(i, j) = at(i, j);
			}
			for(int j = w, jj = 0; j < m; j++, jj++){
				matr->at(i, jj) = at(i, j);
			}
		}
	}
	/**
	 * @brief Split this matrix into four parts.
	 * 
	 * i.e. *this = (@a mat1 `bsd` @a mat2) 'abv' (@a mat3 `bsd` @a mat4)
	 *           , height @a mat1 = @a h, width @a mat1 = @a w
	 *
	 * Matrices should be row-major.
	 *
	 * @param h the height of upper submatrix.
	 * @param w the width of left submatrix.
	 * @param mat1 to which the upper-left submatrix is stored.
	 * @param mat2 to which the upper-right submatrix is stored.
	 * @param mat3 to which the lower-left submatrix is stored.
	 * @param mat4 to which the lower-right submatrix is stored.
	 */
public:
	void splitQ(int h, int w, matrix <A> *mat1, matrix<A> *mat2, matrix <A> *mat3, matrix<A> *mat4) const {
		assert(0<w && w < m && 0<h && h < n && !colwise);
		mat1->setSize(h, w);
		mat2->setSize(h, m - w);
		mat3->setSize(n-h, w);
		mat4->setSize(n-h, m - w);
		for(int i = 0; i < h; i++){
			for(int j = 0; j < w; j++){
				mat1->at(i, j) = at(i, j);
			}
			for(int j = w, jj = 0; j < m; j++, jj++){
				mat2->at(i, jj) = at(i, j);
			}
		}
		for(int i = h, ii = 0; i < n; i++, ii++){
			for(int j = 0; j < w; j++){
				mat3->at(ii, j) = at(i, j);
			}
			for(int j = w, jj = 0; j < m; j++, jj++){
				mat4->at(ii, jj) = at(i, j);
			}
		}
	}
	/**
	 * @brief Embed a matrix at (@a is, @a js).
	 *
	 * The element of the matrix @a mat at (i, j) is copied into 
	 * (is + i, js + j) of this matrix.
	 * The size of the matrix @a mat should not exceed (n - is, m - js)
	 *  where (n, m) is the size of this matrix.
	 * Matrices should be row-major.
	 * @param is row-index of the upper-left element
	 * @param js column-index of the upper-left element
	 * @param mat a matrix to be embedded into this matrix at (@a is, @a js).
	 */
public:
	void embed(int is, int js, const matrix <A> *mat) {
		assert(0<=is && 0<= js && (is + mat->getRows()) <= n && (js + mat->getCols()) <= m );
		for(int i = is, ii = 0; i < n; i++, ii++){
			for(int j = js, jj = 0; j < m; j++, jj++){
				if(jj < mat->m && ii < mat->n) {
					at(i, j) = mat->at(ii, jj);
				}
			}
		}
	}
};

/**
 * @brief Skeletons for matrix.
 *
 * This implementation is sequencial.
 * Skeletons often need the matrices to be row-major.
 */
class m_skeletons {
	/**
	 * @brief The "beside" operator.
	 *
	 * This takes two matrices @a a and @a b and makes a new matrix in which
	 * the matrix @a a is placed on the left of the matrix @a b.
	 * The heights of the matrices should be the same.
	 * @param a a matrix.
	 * @param b a matrix.
	 * @param res resulting matrix (a `Bbv` b).
	 */
 public:
	template<typename A>
	static
	void bsd(const matrix <A> *a, const matrix <A> *b, matrix <A> *res){
		if((a->getRows()!=b->getRows()) || (a->isColumnWise() != b->isColumnWise())) return;
		int nn = a->getRows();
		int m1 = a->getCols();
		int m2 = b->getCols();
		res->setSize(nn, m1+m2);
		if(a->isColumnWise()){
			res->makeColumnWise();
			for(int j = 0; j < m1; j++){
				for(int i = 0; i < nn; i++){
					res->at(j, i) = a->at(j, i);
				}
			}
			int k = m1;
			for(int j = 0; j < m2; j++, k++){	
				for(int i = 0; i < nn; i++){
					res->at(k, i) = b->at(j, i);
				}
			}
		} else {
			res->makeRowWise();
			for(int i = 0; i < nn; i++){
				for(int j = 0; j < m1; j++){
					res->at(i, j) = a->at(i, j);
				}
			}
			for(int i = 0; i < nn; i++){
				int k = m1;
				for(int j = 0; j < m2; j++, k++){
					res->at(i, k) = b->at(i, j);
				}
			}
		}
	}
	/**
	 * @brief The "above" operator.
	 *
	 * This takes two matrices @a a and @a b and makes a new matrix in which
	 * the matrix @a a is placed on the top of the matrix @a b.
	 * The widths of the matrices should be the same.
	 * @param a a matrix.
	 * @param b a matrix.
	 * @param res resulting matrix (a `Abv` b).
	 */
 public:
	template<typename A>
	static void abv(const matrix <A> *a, const matrix <A> *b, matrix <A> *res){
		if((a->getCols()!=b->getCols()) || (a->isColumnWise() != b->isColumnWise())) return;
		int n1 = a->getRows();
		int n2 = b->getRows();
		int mm = a->getCols();
		res->setSize(n1+n2, mm);
		if(a->isColumnWise()){
			res->makeColumnWise();
			for(int j = 0; j < mm; j++){
				for(int i = 0; i < n1; i++){
					res->at(j, i) = a->at(j, i);
				}
			}
			for(int j = 0; j < mm; j++){	
				int k = n1;
				for(int i = 0; i < n2; i++, k++){
					res->at(j, k) = b->at(j, i);
				}
			}
		} else {
			res->makeRowWise();
			for(int i = 0; i < n1; i++){
				for(int j = 0; j < mm; j++){
					res->at(i, j) = a->at(i, j);
				}
			}
			int k = n1;
			for(int i = 0; i < n2; i++, k++){
				for(int j = 0; j < mm; j++){
					res->at(k, j) = b->at(i, j);
				}
			}
		}
	}


	/******************* data parallel skeletons  ****************************/

	/**
	 * @brief The map skeleton (inplace).
	 *
	 * This function applies the function to each element of the matrix.
	 *
	 * @param f a function object taking A, 
	 *    and overwriting the result on the argument.
	 *    @a f should have the member void operator()(A*) const.
	 * A sample of function object is as follows: 
	 * (the value *r is replaced by *r*2)
	 * @code
	 * class Double {
	 * public:
	 *     void operator()(int *r) const {
	 *         *r = *r*2;
	 *     }
	 * };
	 * @endcode
	 * @param dma matrix to which the skeleton is applied.
	 * @see m_skeletons::map(const F&, const matrix <A>*)
	 */
 public:
	template <typename A, typename F>
	static void map_i(const F &f, matrix<A> *dma){
		if(dma->shadowed) {
			dma->reshadow();
			return;
		}
		if(dma->isColumnWise()){
			for(int j = 0; j < dma->m; j++){
				for(int i = 0; i < dma->n; i++){
					f(&dma->at(j, i));
				}
			}
		} else {
			for(int i = 0; i < dma->n; i++){
				for(int j = 0; j < dma->m; j++){
					f(&dma->at(i,j));
				}
			}
		}
	}
	/**
	 * @brief The map skeleton.
	 *
	 * This function applies the function to each element of the matrix.
	 *
	 * @param f a function taking A, and returning result in B
	 *   @a f should have the member void operator()(const A&, B*) const.
	 * The function object f should inherit skeleton::unary_function.
	 * A sample of function object is as follows: 
	 * (square root of the value. type of result is double and type of argument is int)
	 * @code
	 * class Square : public skeleton::unary_function <int, double> {
	 * public:
	 *     void operator()(const int &i, double *r) const {
	 *         *r = sqrt((double)i);
	 *     }
	 * };
	 * @endcode
	 * @param dma matrix to which the skeleton is applied.
	 * @param dmb matrix to which the result is stored.
	 * @see m_skeletons::map(const F&, const matrix <A>*)
	 */
 public:
	template <typename B, typename A, typename F>
	static void map(const F &f, const matrix<A> *dma, matrix<B> *dmb) {
		dmb->setSize(dma->getRows(), dma->getCols());
		if(dma->isColumnWise()){
			dmb->makeColumnWise();
			for(int j = 0; j < dma->m; j++){
				for(int i = 0; i < dma->n; i++){
					f(dma->at(j, i), &dmb->at(j, i));
				}
			}
		} else {
			dmb->makeRowWise();
			for(int i = 0; i < dma->n; i++){
				for(int j = 0; j < dma->m; j++){
					f(dma->at(i, j), &dmb->at(i, j));
				}
			}
		}
	}
	/**
	 * @brief The map skeleton(crates a new matrix).
	 *
	 * This function applies the function to each element of the matrix.
	 *
	 * Definition (in Haskell)
	 * @code
	 * map :: (a -> b) -> Matrix a -> Matrix b
	 * map f (Sing a)  = Sing (f a)
	 * map f (Abv x y) = Abv (map f x) (map f y)
	 * map f (Bsd x y) = Bsd (map f x) (map f y)
	 * @endcode
	 *
	 * Sample: \n
	 * < Haskell code >
	 * @code
	 *   m2 = map f m1
	 * @endcode
	 * < C++ code >
	 * @code
	 *   m2 = m_skeletons::map( f, m1 );
	 * @endcode
	 * @param f a function taking A, and returning result in B
	 *   @a f should have the member void operator()(const A&, B*) const.
	 * The function object f should inherit skeleton::unary_function.
	 * A sample of function object is as follows: 
	 * (square root of the value. type of result is double and type of argument is int)
	 * @code
	 * class Square : public skeleton::unary_function <int, double> {
	 * public:
	 *     void operator()(const int &i, double *r) const {
	 *         *r = sqrt((double)i);
	 *     }
	 * };
	 * @endcode
	 * @param dma matrix to which the skeleton is applied.
	 * @return A matrix to which the result is stored.
	 */
 public:
	template <typename A, typename F>
	static matrix<typename F::result_type>* map(const F &f, const matrix <A> *dma) {
		matrix<typename F::result_type>* dmb = new matrix<typename F::result_type>(dma->getRows(), dma->getCols(), dma->isColumnWise());
		map(f, dma, dmb);
		return dmb;
	}
	/**
	 * @brief The zipwith skeleton(crates a new matrix).
	 *
	 * This function applies the function to the pair of 
	 * corresponding elements of the matrices.
	 *
	 * @param f a function taking A and B, writing result into C
	 *   @a f should have the member void operator()(const A&, const B&, C*) const.
	 * The function object f should inherit skeleton::binary_function.
	 * A sample of function object is as follows: 
	 * (ratio of two arguments)
	 * @code
	 * class Ratio : public skeleton::binary_function <int, int, double> {
	 * public:
	 *     void operator()(const int &i, const int &j, double *r) const {
	 *         *r = i / (double)j;
	 *     }
	 * };
	 * @endcode
	 * @param dma matrix to be zipped with.
	 * @param dmb matrix to be zipped with.
	 * @param dmc matrix to which result is stored
	 * @see m_skeletons::zipwith(const F &, const matrix<A>*, const matrix<B>*)
	 */
 public:
	template<typename C, typename A, typename B, typename F>
	static void zipwith(const F &f, const matrix<A>* dma, const matrix<B>* dmb, matrix<C>*dmc ) {
		assert(dma->isColumnWise()==dmb->isColumnWise());
		dmc->setSize(dma->getRows(), dma->getCols());
		if(dma->isColumnWise()){
			dmc->makeColumnWise();
			for(int j = 0; j < dma->m; j++){
				for(int i = 0; i < dma->n; i++){
					f(dma->at(j, i), dmb->at(j, i), &dmc->at(j, i));
				}
			}
		} else {
			dmc->makeRowWise();
			for(int i = 0; i < dma->n; i++){
				for(int j = 0; j < dma->m; j++){
					f( dma->at(i, j), dmb->at(i,j), &dmc->at(i, j));
				}
			}
		}
	}
	/**
	 * @brief The zipwith skeleton(creates a new matrix).
	 *
	 * This function applies the function to the pair of 
	 * corresponding elements of the matrices.
	 *
	 * Definition (in Haskell)
	 * @code
	 * zipwith :: (a->b->c) -> Matrix a -> Matrix b -> Matrix c
	 * zipwith f (Sing a) (Sing b)   = Sing (f a b)
	 * zipwith f (Abv x y) (Abv u v) = Abv (zipwith f x u) (zipwith f y v)
	 * zipwith f (Bsd x y) (Bsd u v) = Bsd (zipwith f x u) (zipwith f y v)
	 * @endcode
	 *
	 * Sample: \n
	 * < Haskell code >
	 * @code
	 *   m3 = zipwith f m1 m2
	 * @endcode
	 * < C++ code >
	 * @code
	 *   m3 = m_skeletons::zipwith( f, m1, m2 );
	 * @endcode
	 * @param f a function taking A and B, writing result into C
	 *   @a f should have the member void operator()(const A&, const B&, C*) const.
	 * The function object f should inherit skeleton::binary_function.
	 * A sample of function object is as follows: 
	 * (ratio of two arguments)
	 * @code
	 * class Ratio : public skeleton::binary_function <int, int, double> {
	 * public:
	 *     void operator()(const int &i, const int &j, double *r) const {
	 *         *r = i / (double)j;
	 *     }
	 * };
	 * @endcode
	 * @param dma matrix to be zipped with.
	 * @param dmb matrix to be zipped with.
	 * @return A matrix to which the result is stored.
	 */
 public:
	template< typename A, typename B, typename F >
	static matrix<typename F::result_type> *zipwith(const F &f, const matrix<A>* dma, const matrix<B>* dmb) {
		matrix<typename F::result_type>* dmc = new matrix<typename F::result_type>(dma->getRows(), dma->getCols(), dma->isColumnWise());	
		zipwith(f, dma, dmb, dmc);
		return dmc;
	
	}
	/**
	 * @brief Make a pair (fst, snd).
	 *
	 * This is used to make zip by specializing zipwith.
	 */
 private:
	template< typename B , typename C >
	class Mkpair : public skeleton::binary_function<B, C, std::pair<B, C> >{
	public:
	/** @brief Make a pair.*/
		void operator()(const B& fst, const C& snd, std::pair<B, C>* pair) const{
			pair->first = fst;
			pair->second = snd;
		}
	};
	/**
	 * @brief The zip skeleton which is a specialization of zipwith.
	 *
	 * Definition (in Haskell)
	 * @code
	 * zip :: (Matrix a, Matrix b) -> Matrix (a, b)
	 * zip (u, v) = zipwith (\ x y -> (x, y)) u v
	 * @endcode
	 * @param dma matrix to be zipped with.
	 * @param dmb matrix to be zipped with
	 * @result A matrix consisting of pairs of corresponding elements of the matrices.
	 * @see m_skeletons::zipwith(const F &, const matrix<A>*, const matrix<B>*)
	 */
 public:
	template< typename A, typename B>
	static matrix < std::pair<A,B> > *zip(const matrix<A>* dma, const matrix<B>* dmb) {
		return zipwith(Mkpair<A,B>(), dma, dmb);   
	}

	/**
	 * @brief The reduce skeleton (create new value).
	 *
	 * This skeleton reduces the matrix into a value with two binary operators.
	 *
	 * Definition (in Haskell)
	 * @code
	 * reduce :: (a->a->a, a->a->a) -> Matrix a -> a
	 * reduce (oplus, otimes) (Sing a)  = a
	 * reduce (oplus, otimes) (Abv x y) = oplus ( reduce (oplus, otimes) x) ( reduce (oplus, otimes) y)
	 * reduce (oplus, otimes) (Bsd x y) = otimes ( reduce (oplus, otimes) x) ( reduce (oplus, otimes) y)
	 * @endcode
	 * Two operators should have abiding property.
	 *
	 * Sample: \n
	 * < Haskell code >
	 * @code
	 *   r = reduce ((+), (+)) m
	 * @endcode
	 * < C++ code >
	 * @code
	 *   r = m_skeletons::reduce( Add(), Add(), m);
	 * @endcode
	 *
	 * The matrices must be row-wise.
	 * @param oplus a function taking two arguments x and y, returning result z,
	 * i.e. z = x @a oplus y, and must be abiding with otimes.
	 * @a oplus should have the member void operator()(const A&, const A&, A*) const.
	 * The function object @a oplus should inherit skeleton::binary_function.
	 * A sample of function object is as follows: 
	 * (Addition)
	 * @code
	 * class Add : public skeleton::binary_function <int, int, int> {
	 * public:
	 *     void operator()(const int &i, const int &j, int *r) const {
	 *         *r = i + j;
	 *     }
	 * };
	 * @endcode
	 * @a oplus should be associative.
	 * @param otimes a function taking two arguments x and y, returning result z,
	 * i.e. z = x @a otimes y, and must be abiding with oplus.
	 * The function @a otimes oplus should inherit skeleton::binary_function.
	 * @a otimes should be associative.
	 * Associative and commutative operator has abiding property with itself.
	 * @param dma a matrix to which the reduce is applied.
	 * @return Resulting reduced value. User should free it.
	 */
 public:
	template< typename A , typename F, typename G>
	static A *reduce(const F &oplus, const G &otimes, const matrix<A> *dma) {
		A* res = new A;
		reduce(oplus, otimes, dma, res);
		return res;		
	}
	/**
	 * @brief The reduce skeleton.
	 *
	 * This skeleton reduces the matrix into a value with two binary operators.
	 * The matrices must be row-wise.
	 * @param oplus a function taking two arguments x and y, returning result z,
	 * i.e. z = x @a oplus y, and must be abiding with @a otimes.
	 * @a oplus should have the member void operator()(const A&, const A&, *A) const.
	 * The function object @a oplus should inherit skeleton::binary_function.
	 * A sample of function object is as follows: 
	 * (Max)
	 * @code
	 * class Max : public skeleton::binary_function <int, int, int> {
	 * public:
	 *     void operator()(const int &i, const int &j, int *r) const {
	 *         *r = (i > j) ? i : j;
	 *     }
	 * };
	 * @endcode
	 * @a oplus should be associative.
	 * @param otimes a function taking two arguments x and y, returning result z,
	 * i.e. z = x @a otimes y, and must be abiding with @a oplus.
	 * The function @a otimes oplus should inherit skeleton::binary_function.
	 * @a otimes should be associative.
	 * Associative and commutative operator has abiding property with itself.
	 * @param dma a matrix to which the reduce is applied.
	 * @param res resulting reduced value.
	 * @see m_skeletons::reduce(const F &, const G &, const matrix<A> *)

	 */
 public:
	template< typename A , typename F, typename G>
	static void reduce(const F &oplus, const G &otimes, const matrix<A> *dma, A* res) {
		assert(!dma->isColumnWise());
		A *vals = new A[dma->n];
		for(int i = 0; i < dma->n; i++){
			A val = dma->at(i, 0);
			for(int j = 1; j < dma->m; j++){
				A tmp;
				otimes(val, dma->at(i, j), &tmp);
				val = tmp;
			}
			vals[i] = val;
		}
		A val = vals[0];
		for(int i = 1; i < dma->n; i++){
			A tmp;
			oplus(val, vals[i], &tmp);
			val = tmp;
		}
		delete [] vals;
		*res = val; 
	}

	/**
	 * @brief The reduce skeleton(divide and conquer).
	 *
	 * This skeleton reduces the matrix into a value with two binary operators.
	 * The matrices must be row-wise.
	 * @param oplus a function taking two arguments x and y, returning result z,
	 * i.e. z = x @a oplus y, and must be abiding with @a otimes.
	 * @a oplus should have the member void operator()(const A&, const A&, *A) const.
	 * The function object @a oplus should inherit skeleton::binary_function.
	 * A sample of function object is as follows: 
	 * (Max)
	 * @code
	 * class Max : public skeleton::binary_function <int, int, int> {
	 * public:
	 *     void operator()(const int &i, const int &j, int *r) const {
	 *         *r = (i > j) ? i : j;
	 *     }
	 * };
	 * @endcode
	 * @a oplus should be associative.
	 * @param otimes a function taking two arguments x and y, returning result z,
	 * i.e. z = x @a otimes y, and must be abiding with @a oplus.
	 * The function @a otimes oplus should inherit skeleton::binary_function.
	 * @a otimes should be associative.
	 * Associative and commutative operator has abiding property with itself.
	 * @param dma a matrix to which the reduce is applied.
	 * @param res resulting reduced value.
	 * @see m_skeletons::reduce(const F &, const G &, const matrix<A> *)

	 */
	 public:
	template< typename A , typename F, typename G>
	static void reduceDC(const F &oplus, const G &otimes, const matrix<A> *dma, A* res) {
		assert(!dma->isColumnWise());
		matrix<A> tmp = *dma;
		const int n = tmp.n;
		const int m = tmp.m;
		const int nm = n > m ? n : m;
		for(int stage = 1; stage < nm; stage <<=1){
			const int stage2 = stage << 1;
			for(int i = 0; i < n; i+=stage2){
				for(int j = 0; j < m; j+=stage2){
					const int i2 = i ^ stage;
					const int j2 = j ^ stage;
					if(i2 < n){
						A t = tmp.at(i, j);
						oplus(t, tmp.at(i2, j), &tmp.at(i,j));
						if(j2 < m){
							A u = tmp.at(i, j2);
							oplus(u, tmp.at(i2, j2), &tmp.at(i,j2));
						}
					}
					if(j2 < m){
						A t = tmp.at(i, j);
						otimes(t, tmp.at(i, j2), &tmp.at(i,j));
					}
				}
			}
		}
		*res = tmp.at(0,0); 
	}
	/**
	 * @brief The reduce skeleton (create new value, divide and conquer).
	 *
	 * This skeleton reduces the matrix into a value with two binary operators.
	 * The matrices must be row-wise.
	 * @param oplus a function taking two arguments x and y, returning result z,
	 * i.e. z = x @a oplus y, and must be abiding with @a otimes.
	 * @a oplus should have the member void operator()(const A&, const A&, *A) const.
	 * The function object @a oplus should inherit skeleton::binary_function.
	 * A sample of function object is as follows: 
	 * (Max)
	 * @code
	 * class Max : public skeleton::binary_function <int, int, int> {
	 * public:
	 *     void operator()(const int &i, const int &j, int *r) const {
	 *         *r = (i > j) ? i : j;
	 *     }
	 * };
	 * @endcode
	 * @a oplus should be associative.
	 * @param otimes a function taking two arguments x and y, returning result z,
	 * i.e. z = x @a otimes y, and must be abiding with @a oplus.
	 * The function @a otimes oplus should inherit skeleton::binary_function.
	 * @a otimes should be associative.
	 * Associative and commutative operator has abiding property with itself.
	 * @param dma a matrix to which the reduce is applied.
	 * @see m_skeletons::reduce(const F &, const G &, const matrix<A> *)
	 */
 public:
	template< typename A , typename F, typename G>
	static A *reduceDC(const F &oplus, const G &otimes, const matrix<A> *dma) {
		A* res = new A;
		reduceDC(oplus, otimes, dma, res);
		return res;		
	}
		/**
	 * @brief A specialization of reduce: allreduce_row.
	 *
	 * This reduces the matrix in row-direction only and
	 * broadcasts the result in each row.
	 * So almost all elements are shadows of first column.
	 * Please be careful with modifing the elements.
	 *
	 * Definition (in Haskell)
	 * @code
	 * allreduce_row :: (a -> a -> a) -> Matrix a -> Matrix a
	 * allreduce_row otimes = scanr (ll, rr) . scan (rr, otimes)
	 *   where ll a b = a
	 *         rr a b = b
	 * @endcode
	 *
	 * Sample: \n
	 * < Haskell code >
	 * @code
	 *   m2 = allreduce_row (+) m1
	 * @endcode
	 * < C++ code >
	 * @code
	 *   m2 = m_skeletons::allreduce_row( Add(), m1 );
	 * @endcode
	 * @param otimes a function taking two arguments x and y, returning result z,
	 * i.e. z = x otimes y.
	 * @a otimes should have the member void operator()(const A&, const A&. *A) const.
	 * The function object @a otimes should inherit skeleton::binary_function.
	 * @a otimes should be associative.
	 * @param dma a matrix to be reduced.
	 * @return The resulting matrix in which the elements in a row have the same value, the result of reducing the row with @a otimes.
	 * @see m_skeletons::reduce(const F &, const G &, const matrix<A> *)
	 * @see m_skeletons::scan(const F &, const G &, const matrix<A> *)
	 * @see m_skeletons::scanr(const F &, const G &, const matrix<A> *)
	 */
 public:
	template< typename A , typename F>
	static matrix<A> * allreduce_row(const F &otimes, const matrix<A> *dma) {
		matrix<A>* dmb = new matrix<A>(dma->getRows(), dma->getCols(), dma->isColumnWise());
		assert(!dma->isColumnWise());
		for(int i = 0; i < dma->n; i++){
			A val = dma->at(i, 0);
			for(int j = 1; j < dma->m; j++){
				A tmp;
				otimes(val, dma->at(i, j), &tmp);
				val = tmp;
			}
			dmb->at(i, 0) = val;
		}
		for(int i = 0; i < dma->n; i++){
			for(int j = 1; j < dma->m; j++){
				Shadower<A>::shadow(dmb->at(i, 0), &dmb->at(i, j));
			}
		}
		return dmb;
	}

	/**
	 * @brief A specialization of reduce: allreduce_col.
	 *
	 * This reduces the matrix in column-direction only and
	 * broadcasts the result in each column.
	 * So almost all elements are shadows of first row.
	 * Please be careful with modifing the elements.
	 *
	 * Definition (in Haskell)
	 * @code
	 * allreduce_col :: (a -> a -> a) -> Matrix a -> Matrix a
	 * allreduce_col oplus = scanr (rr, ll) . scan (oplus, rr)
	 *   where ll a b = a
	 *         rr a b = b
	 * @endcode
	 *
	 * Sample: \n
	 * < Haskell code >
	 * @code
	 *   m2 = allreduce_col (+) m1
	 * @endcode
	 * < C++ code >
	 * @code
	 *   m2 = m_skeletons::allreduce_col( Add(), m1 );
	 * @endcode
	 * @param oplus a function taking two arguments x and y, returning result z,
	 * i.e. z = x oplus y.
	 * @a oplus should have the member void operator()(const A&, const A&, A*) const.
	 * The function object @a oplus should inherit skeleton::binary_function.
	 * @a oplus should be associative.
	 * @param dma a matrix to be reduced.
	 * @return The resulting matrix in which the elements in a column have the same value, the result of reducing the column with @a oplus.
	 * @see m_skeletons::reduce(const F &, const G &, const matrix<A> *)
	 * @see m_skeletons::scan(const F &, const G &, const matrix<A> *)
	 * @see m_skeletons::scanr(const F &, const G &, const matrix<A> *)
	 */
 public:
	template< typename A , typename F>
	static matrix<A> * allreduce_col(const F &oplus, const matrix<A> *dma) {
		matrix<A>* dmb = new matrix<A>(dma->getRows(), dma->getCols(), dma->isColumnWise());
		assert(!dma->isColumnWise());
		for(int j = 0; j < dma->m; j++){
			A val = dma->at(0, j);
			for(int i = 1; i < dma->n; i++){
				A tmp;
				oplus(val, dma->at(i, j), &tmp);
				val = tmp;
			}
			dmb->at(0, j) = val;
		}
		for(int j = 0; j < dma->m; j++){
			for(int i = 1; i < dma->n; i++){
				Shadower<A>::shadow(dmb->at(0, j),&dmb->at(i, j));
			}
		}
		return dmb;
	}

	/**
	 * @brief The scan skeleton.
	 *
	 * This skeleton stores every process of reduction of a matrix into a value with two binary operators.
	 * The matrices must be row-wise.
	 * @param oplus a function taking two arguments x and y, returning result z,
	 * i.e. z = x @a oplus y, and must be abiding with @a otimes.
	 * @a oplus should have the member void operator()(const A&, const A&, A*) const.
	 * The function object @a oplus should inherit skeleton::binary_function.
	 * A sample of function object is as follows: 
	 * (Max)
	 * @code
	 * class Max : public skeleton::binary_function <int, int, int> {
	 * public:
	 *     void operator()(const int &i, const int &j, int *r) const {
	 *         *r = (i > j) ? i : j;
	 *     }
	 * };
	 * @endcode
	 * @a oplus should be associative.
	 * @param otimes a function taking two arguments x and y, returning result z,
	 * i.e. z = x @a otimes y, and must be abiding with @a oplus.
	 * The function @a otimes oplus should inherit skeleton::binary_function.
	 * @a otimes should be associative.
	 * Associative and commutative operator has abiding property with itself.
	 * @param dma a matrix to which the reduce is applied.
	 * @param dmb resulting matrix of all reduction values.
	 * @see m_skeletons::scan(const F &, const G &, const matrix<A> *)

	 */
 public:
	template< typename A, typename F, typename G>
	static void scan(const F &oplus, const G &otimes, const matrix<A> *dma, matrix<A>* dmb) {
		assert(!dma->isColumnWise());
		dmb->setSize(dma->getRows(), dma->getCols());
		for(int i = 0; i < dma->n; i++){
			dmb->at(i, 0) = dma->at(i, 0);
			for (int j = 1; j < dma->m; j++){
				otimes(dmb->at(i, j-1), dma->at(i, j), &dmb->at(i, j));
			}
		}
		for (int j = 0; j < dma->m; j++){
			for(int i = 1; i < dma->n; i++){
				A tmp;
				oplus(dmb->at(i-1, j), dmb->at(i, j), &tmp);
				dmb->at(i, j) = tmp;
			}
		}
	}
	/**
	 * @brief The scan skeleton.
	 *
	 * This skeleton stores every process of reduction of a matrix into a value with two binary operators.
	 * The matrices must be row-wise.
	 *
	 * Definition (in Haskell)
	 * @code
	 * scan :: (a->a->a, a->a->a) -> Matrix a -> Matrix a
	 * scan (oplus, otimes) (Sing a) = (Sing a)
	 * scan (oplus, otimes) (Abv x y) = oplus' ( scan (oplus, otimes) x) ( scan (oplus, otimes) y)
	 *     where oplus' sx sy = Abv sx sy'
	 *         where sy' = mapr (zipwith oplus (bottom sx)) sy
	 * scan (oplus, otimes) (Bsd x y) = otimes' ( scan (oplus, otimes) x) ( scan (oplus, otimes) y)
	 *     where otimes' sx sy = Bsd sx sy'
	 *         where sy' = mapc (zipwith otimes (right sx)) sy
	 * @endcode
	 * Two operators should have abiding property.
	 *
	 * Sample: \n
	 * < Haskell code >
	 * @code
	 *   m2 = scan ((+), (+)) m1
	 * @endcode
	 * < C++ code >
	 * @code
	 *   m2 = m_skeletons::scan( Add(), Add(), m1);
	 * @endcode
	 * @param oplus a function taking two arguments x and y, returning result z,
	 * i.e. z = x @a oplus y, and must be abiding with @a otimes.
	 * @a oplus should have the member void operator()(const A&, const A&, A*) const.
	 * The function object @a oplus should inherit skeleton::binary_function.
	 * A sample of function object is as follows: 
	 * (Addition)
	 * @code
	 * class Add : public skeleton::binary_function <int, int, int> {
	 * public:
	 *     void operator()(const int &i, const int &j, int *r) const {
	 *         *r = i+j;
	 *     }
	 * };
	 * @endcode
	 * @a oplus should be associative.
	 * @param otimes a function taking two arguments x and y, returning result z,
	 * i.e. z = x @a otimes y, and must be abiding with @a oplus.
	 * The function @a otimes oplus should inherit skeleton::binary_function.
	 * @a otimes should be associative.
	 * Associative and commutative operator has abiding property with itself.
	 * @param dma a matrix to which the reduce is applied.
	 * @return Resulting matrix of all reduction values. User should free it.
	 * @see m_skeletons::reduce(const F &, const G &, const matrix<A> *)
	 * @see m_skeletons::zipwith(const F &, const matrix<A>*, const matrix<B>*)
	 * @see m_skeletons::right(const matrix<A> *)
	 * @see m_skeletons::bottom(const matrix<A> *)
	 * @see m_skeletons::mapr(const F &, const matrix<A> *)
	 * @see m_skeletons::mapc(const F &, const matrix<A> *)
	 */
 public:
	template< typename A , typename F, typename G>
	static matrix<A>* scan(const F &oplus, const G &otimes, const matrix<A> *dma) {

		matrix<A>* dmb = new matrix<A>(dma->getRows(), dma->getCols(), dma->isColumnWise());
		scan(oplus, otimes, dma, dmb);
		return dmb;
	}

	/**
	 * @brief The reverse-scan skeleton.
	 *
	 * This skeleton stores every process of reverse reduction of a matrix into a value with two binary operators.
	 * The matrices must be row-wise.
	 *
	 * @param oplus a function taking two arguments x and y, returning result z,
	 * i.e. z = x @a oplus y, and must be abiding with @a otimes.
	 * @a oplus should have the member void operator()(const A&, const A&, *A) const.
	 * The function object @a oplus should inherit skeleton::binary_function.
	 * A sample of function object is as follows: 
	 * (Addition)
	 * @code
	 * class Add : public skeleton::binary_function <int, int, int> {
	 * public:
	 *     void operator()(const int &i, const int &j, int *r) const {
	 *         *r = i+j;
	 *     }
	 * };
	 * @endcode
	 * @a oplus should be associative.
	 * @param otimes a function taking two arguments x and y, returning result z,
	 * i.e. z = x @a otimes y, and must be abiding with @a oplus.
	 * The function @a otimes oplus should inherit skeleton::binary_function.
	 * @a otimes should be associative.
	 * Associative and commutative operator has abiding property with itself.
	 * @param dma a matrix to which the reduce is applied.
	 * @param dmb resulting matrix of all reduction values. User should free it.
	 * @see m_skeletons::scanr(const F &, const G &, const matrix<A> *)

	 */
 public:
	template< typename A , typename F, typename G>
	static void scanr(const F &oplus, const G &otimes, const matrix<A> *dma, matrix<A>* dmb) {
		assert(!dma->isColumnWise());
		dmb->setSize(dma->getRows(), dma->getCols());
		for(int i = 0; i < dma->n; i++){
			dmb->at(i, dma->m-1) = dma->at(i, dma->m-1);
			for (int j = dma->m-2; j >= 0; j--){
				otimes(dma->at(i, j), dmb->at(i, j+1), &dmb->at(i, j));
			}
		}
		for (int j = 0; j < dma->m; j++){
			for(int i = dma->n-2; i >=0; i--){
				A tmp;
				oplus(dmb->at(i, j), dmb->at(i+1, j), &tmp);
				dmb->at(i, j) = tmp;
			}
		}
	}
	/**
	 * @brief The reverse-scan skeleton.
	 *
	 * This skeleton stores every process of reverse reduction of a matrix into a value with two binary operators.
	 * The matrices must be row-wise.
	 *
	 * Definition (in Haskell)
	 * @code
	 * scanr :: (a->a->a, a->a->a) -> Matrix a -> Matrix a
	 * scanr (oplus, otimes) = rev . scan (binrev oplus, binrev otimes) . rev
	 *    where 
	 *    rev = reduce (binrev Abv, binrev Bsd) . map Sing
	 *    binrev f a b = f b a
	 * @endcode
	 * Two operators should have abiding property.
	 *
	 * Sample: \n
	 * < Haskell code >
	 * @code
	 *   m2 = scanr ((+), (+)) m1
	 * @endcode
	 * < C++ code >
	 * @code
	 *   m2 = m_skeletons::scanr( Add(), Add(), m1);
	 * @endcode
	 * @param oplus a function taking two arguments x and y, returning result z,
	 * i.e. z = x @a oplus y, and must be abiding with @a otimes.
	 * @a oplus should have the member void operator()(const A&, const A&, *A) const.
	 * The function object @a oplus should inherit skeleton::binary_function.
	 * A sample of function object is as follows: 
	 * (Addition)
	 * @code
	 * class Add : public skeleton::binary_function <int, int, int> {
	 * public:
	 *     void operator()(const int &i, const int &j, int *r) const {
	 *         *r = i+j;
	 *     }
	 * };
	 * @endcode
	 * @a oplus should be associative.
	 * @param otimes a function taking two arguments x and y, returning result z,
	 * i.e. z = x @a otimes y, and must be abiding with @a oplus.
	 * The function @a otimes oplus should inherit skeleton::binary_function.
	 * @a otimes should be associative.
	 * Associative and commutative operator has abiding property with itself.
	 * @param dma a matrix to which the reduce is applied.
	 * @return Resulting matrix of all reduction values. User should free it.
	 * @see m_skeletons::scan(const F &, const G &, const matrix<A> *)
	 * @see m_skeletons::reduce(const F &, const G &, const matrix<A> *)
	 * @see m_skeletons::map(const F &, const matrix<A> *)
	 */
 public:
	template< typename A , typename F, typename G>
	static matrix<A>* scanr(const F &oplus, const G &otimes, const matrix<A> *dma) {

		matrix<A>* dmb = new matrix<A>(dma->getRows(), dma->getCols(), dma->isColumnWise());
		scanr(oplus, otimes, dma, dmb);
		return dmb;
	}
	/******************* some utilities  ****************************/

	/**
	 * @brief The ternary zipwith skeleton (creates a new matrix).
	 *
	 * This function applies the function to three
	 * corresponding elements of the matrices.
	 * @param f a function taking A, B and C, writing result into D
	 *   @a f should have the member void operator()(const A&, const B&, const C&, D*) const.
	 * The function object f should inherit ternary_function.
	 * (ratio of two arguments plus third argument)
	 * A sample of function object is as follows: 
	 * @code
	 * class Ratio : public ternary_function <int, int, int, double> {
	 * public:
	 *     void operator()(const int &i, const int &j, const int &k, double *r) const {
	 *         *r = i / (double)j + k;
	 *     }
	 * };
	 * @endcode
	 * @param dma matrix to be zipped with.
	 * @param dmb matrix to be zipped with.
	 * @param dmc matrix to be zipped with.
	 * @param dmd matrix to which the result is stored.
	 * @see m_skeletons::zipwith(const F &, const matrix<A>*, const matrix<B>*, const matrix<C>*)
	 */
 public:
	template<typename D, typename A, typename B, typename C, typename F >
	static void zipwith(const F &f, const matrix<A>* dma, const matrix<B>* dmb, const matrix<C> *dmc, matrix<D>* dmd ) {
		assert((dma->isColumnWise()==dmb->isColumnWise()) && (dmb->isColumnWise()==dmc->isColumnWise()));
		dmd->setSize(dma->getRows(), dma->getCols());
		if(dma->isColumnWise()){
			dmd->makeColumnWise();
			for(int j = 0; j < dma->m; j++){
				for(int i = 0; i < dma->n; i++){
					f(dma->at(j, i), dmb->at(j, i), dmc->at(j, i), &dmd->at(j, i));
				}
			}
		} else {
			dmd->makeRowWise();
			for(int i = 0; i < dma->n; i++){
				for(int j = 0; j < dma->m; j++){
					f(dma->at(i, j), dmb->at(i,j), dmc->at(i, j), &dmd->at(i, j));
				}
			}
		}
	}
	/**
	 * @brief The ternary zipwith skeleton (creates a new matrix).
	 *
	 * This function applies the function to three
	 * corresponding elements of the matrices.
	 *
	 * Definition (in Haskell)
	 * @code
	 * zipwith3 :: (a->b->c->d) -> Matrix a -> Matrix b -> Matrix c -> Matrix d
	 * zipwith3 f (Sing a) (Sing b) (Sing c)   = Sing (f a b c)
	 * zipwith3 f (Abv x y) (Abv u v) (Abv w z) = Abv (zipwith f x u w) (zipwith f y v z)
	 * zipwith3 f (Bsd x y) (Bsd u v) (Bsd w z) = Bsd (zipwith f x u w) (zipwith f y v z)
	 * @endcode
	 *
	 * Sample: \n
	 * < Haskell code >
	 * @code
	 *   m4 = zipwith3 f m1 m2 m3
	 * @endcode
	 * < C++ code >
	 * @code
	 *   m4 = m_skeletons::zipwith( f, m1, m2, m3);
	 * @endcode
	 * @param f a function taking A, B and C, writing result into D
	 *   @a f should have the member void operator()(const A&, const B&, const C&, D*) const.
	 * The function object f should inherit ternary_function.
	 * A sample of function object is as follows: 
	 * (ratio of two arguments plus third argument)
	 * @code
	 * class Ratio : public ternary_function <int, int, int, double> {
	 * public:
	 *     void operator()(const int &i, const int &j, const int &k, double *r) const {
	 *         *r = i / (double)j + k;
	 *     }
	 * };
	 * @endcode
	 * @param dma matrix to be zipped with.
	 * @param dmb matrix to be zipped with.
	 * @param dmc matrix to be zipped with.
	 * @return A matrix to which the result is stored.
	 */
 public:
	template<typename A, typename B, typename C, typename F >
	static matrix<typename F::result_type> *zipwith(const F &f, const matrix<A>* dma, const matrix<B>* dmb, const matrix<C>* dmc) {
		matrix<typename F::result_type>* dmd = new matrix<typename F::result_type>(dma->getRows(), dma->getCols(), dma->isColumnWise());	
		zipwith(f, dma, dmb, dmc, dmd);
		return dmd;	
	}


	/**
	 * @brief The rowwise-map skeleton(crates a new matrix).
	 *
	 * This function applies the function to each row of the matrix.
	 *
	 * @param f a function taking a matrix, and returning result in a matrix where heights of the argument matrices are 1.
	 *   @a f should have the member void operator()(const matrix<A>&, matrix<B>*) const.
	 * The function object f should inherit skeleton::unary_function.
	 * @param dma a matrix to which the skeleton is applied.
	 * @param dmb a matrix to which the result is stored.
	 * @see m_skeletons::mapr(const F &, const matrix<A> *)
	 */
 public:
	template< typename B, typename A, typename F >
	static void mapr(const F &f, const matrix<A> *dma, matrix<B> *dmb) {
		assert(!dma->isColumnWise());
		dmb->setSize(dma->getRows(), dma->getCols());
		if(dmb->isColumnWise()){
			dmb->makeRowWise();
		}
		matrix <A> tmpA(1, dma->getCols());
		matrix <B> tmpB(1, dma->getCols());
		
		for(int i = 0; i < dma->n; i++){
			for(int j = 0; j < dma->m; j++){
				tmpA.at(0, j) = dma->at(i, j);
			}
			f(tmpA, &tmpB);
			for(int j = 0; j < dma->m; j++){
				dmb->at(i, j) = tmpB.at(0, j);
			}
		}
	}
	/**
	 * @brief The rowwise-map skeleton(crates a new matrix).
	 *
	 * This function applies the function to each row of the matrix.
	 *
	 * Definition (in Haskell)
	 * @code
	 * mapr :: (Matrix a -> Matrix b) -> Matrix a -> Matrix b
	 * mapr f = reduce (Abv, ll) . map f . rows
	 *   where rows   = reduce (Abv, zipwith Bsd) . map (\x -> Sing (Sing x))
	 *         ll a b = a
	 * @endcode
	 *
	 * Sample: \n
	 * < Haskell code >
	 * @code
	 *   m2 = mapr f m1
	 * @endcode
	 * < C++ code >
	 * @code
	 *   m2 = m_skeletons::mapr( f, m1 );
	 * @endcode
	 * @param f a function taking a matrix, and returning result in a matrix where heights of the argument matrices are 1.
	 *   @a f should have the member void operator()(const matrix<A>&, matrix<B>*) const.
	 * The function object f should inherit skeleton::unary_function.
	 * A sample of function object is as follows: 
	 * (divide each element by the summation of all elements)
	 * @code
	 * class Square : public skeleton::unary_function <matrix<int> , matrix<double> > {
	 * public:
	 *     void operator()(const matrix<int> &i, matrix<double> *r) const {
	 *         int a;
	 *         m_skeletons::reduce(Add(), Add(), &i, &a);
	 *         m_skeletons::map(Div(a), &i, r);
	 *     }
	 * };
	 * @endcode
	 * @param dma matrix to which the skeleton is applied.
	 * @return A matrix to which the result is stored.
	 * @see m_skeletons::map(const F &, const matrix<A> *)
	 * @see m_skeletons::zipwith(const F &, const matrix<A>*, const matrix<B>*)
	 * @see m_skeletons::reduce(const F &, const G &, const matrix< A > *)
	 * @see m_skeletons::mapc(const F &, const matrix<A> *)
	 */
 public:
	template< typename A, typename F >
	static typename F::result_type* mapr(const F &f, const matrix<A> *dma) {

		typename F::result_type* dmb = new typename F::result_type (dma->getRows(), dma->getCols(), dma->isColumnWise());
		mapr(f, dma, dmb);
		return dmb;
	}
	/**
	 * @brief The columnwise-map skeleton(crates a new matrix).
	 *
	 * This function applies the function to each column of the matrix.
	 *
	 * @param f a function taking a matrix, and returning result in a matrix where widths of the argument matrices are 1.
	 *   @a f should have the member void operator()(const matrix<A>&, matrix<B>*) const.
	 * The function object f should inherit skeleton::unary_function.
	 * @param dma a matrix to which the skeleton is applied.
	 * @param dmb a matrix to which the result is stored.
	 * @see m_skeletons::mapc(const F &, const matrix<A> *)
	 */
 public:
	template< typename B, typename A, typename F >
	static void mapc(const F &f, const matrix<A> *dma, matrix<B> *dmb) {
		assert(!dma->isColumnWise());
		dmb->setSize(dma->getRows(), dma->getCols());
		if(dmb->isColumnWise()){
			dmb->makeRowWise();
		}
		matrix <A> tmpA(dma->getRows(), 1);
		matrix <B> tmpB(dma->getRows(), 1);
		
		for(int j = 0; j < dma->m; j++){
			for(int i = 0; i < dma->n; i++){
				tmpA.at(i, 0) = dma->at(i, j);
			}
			f(tmpA, &tmpB);
			for(int i = 0; i < dma->n; i++){
				dmb->at(i, j) = tmpB.at(i, 0);
			}
		}
	}
	/**
	 * @brief The columnwise-map skeleton(crates a new matrix).
	 *
	 * This function applies the function to each column of the matrix.
	 *
	 * Definition (in Haskell)
	 * @code
	 * mapc :: (Matrix a -> Matrix b) -> Matrix a -> Matrix b
	 * mapc f = reduce (ll, Bsd) . map f . cols
	 *   where cols   = reduce (zipwith Abv, Bsd) . map (\x -> Sing (Sing x))
	 *         ll a b = a
	 * @endcode
	 *
	 * Sample: \n
	 * < Haskell code >
	 * @code
	 *   m2 = mapc f m1
	 * @endcode
	 * < C++ code >
	 * @code
	 *   m2 = m_skeletons::mapc( f, m1 );
	 * @endcode
	 * @param f a function taking a matrix, and returning result in a matrix where widths of the argument matrices are 1.
	 *   @a f should have the member void operator()(const matrix<A>&, matrix<B>*) const.
	 * The function object f should inherit skeleton::unary_function.
	 * A sample of function object is as follows: 
	 * (divide each element by the summation of all elements)
	 * @code
	 * class Square : public skeleton::unary_function < matrix<int>, matrix<double> > {
	 * public:
	 *     void operator()(const matrix<int> &i, matrix<double> *r) const {
	 *         int a;
	 *         m_skeletons::reduce(Add(), Add(), &i, &a);
	 *         m_skeletons::map(Div(a), &i, &r);
	 *     }
	 * };
	 * @endcode
	 * @param dma matrix to which the skeleton is applied.
	 * @return A matrix to which the result is stored.
	 * @see m_skeletons::map(const F &, const matrix<A> *)
	 * @see m_skeletons::reduce(const F &, const G &, const matrix< A > *)
	 * @see m_skeletons::zipwith(const F &, const matrix<A>*, const matrix<B>*)
	 * @see m_skeletons::mapr(const F &, const matrix<A> *)
	 */
 public:
	template< typename A, typename F >
	static typename F::result_type* mapc(const F &f, const matrix<A> *dma) {

		typename F::result_type* dmb = new typename F::result_type(dma->getRows(), dma->getCols(), dma->isColumnWise());
		mapc(f, dma, dmb);
		return dmb;
	}
	/**
	 * @brief The top function.
	 *
	 * This function returns the top row of the matrix.
	 *
	 * Definition (in Haskell)
	 * @code
	 * top :: Matrix a -> Matrix a
	 * top = reduce (ll, Bsd) . map Sing
	 *   where ll a b = a
	 * @endcode
	 *
	 * Sample: \n
	 * < Haskell code >
	 * @code
	 *   m2 = top m1
	 * @endcode
	 * < C++ code >
	 * @code
	 *   m2 = m_skeletons::top( m1 );
	 * @endcode
	 * @param dma matrix whose top is taken.
	 * @return The top of the matrix @a dma.
	 * @see m_skeletons::reduce(const F &, const G &, const matrix<A> *)
	 * @see m_skeletons::map(const F&, const matrix <A>*)
	 */
 public:
	template< typename A >
	static matrix<A>* top(const matrix<A> *dma)  {
		matrix<A> *res = new matrix<A>();
		top(dma, res);
		return res;
	}
	/**
	 * @brief The top function.
	 *
	 * This function returns the top row of the matrix.
	 *
	 * @param dma matrix whose top is taken.
	 * @param res the top of the matrix @a dma is stored in.
	 * @see m_skeletons::top(const matrix<A> *)
	 */
	template< typename A >
	static void top(const matrix<A> *dma, matrix <A> *res) {
		assert(!dma->isColumnWise());
		int n = dma->getCols();
		res->setSize(1, n);
		if(res->isColumnWise()){
			res->makeRowWise();
		}
		for(int i = 0; i < n; i++){
			res->at(0, i) = dma->at(0, i);
		}
	}
	/**
	 * @brief The bottom function.
	 *
	 * This function returns the bottom row of the matrix.
	 *
	 * Definition (in Haskell)
	 * @code
	 * bottom :: Matrix a -> Matrix a
	 * bottom = reduce (rr, Bsd) . map Sing
	 *   where rr a b = b
	 * @endcode
	 *
	 * Sample: \n
	 * < Haskell code >
	 * @code
	 *   m2 = bottom m1
	 * @endcode
	 * < C++ code >
	 * @code
	 *   m2 = m_skeletons::bottom( m1 );
	 * @endcode
	 * @param dma matrix whose bottom is taken.
	 * @return The bottom of the matrix @a dma.
	 * @see m_skeletons::reduce(const F &, const G &, const matrix<A> *)
	 * @see m_skeletons::map(const F&, const matrix <A>*)
	 */
	template< typename A >
	static matrix<A>* bottom(const matrix<A> *dma) {
		matrix<A> *res = new matrix<A>();
		bottom(dma, res);
		return res;
	}
	/**
	 * @brief The bottom function.
	 *
	 * This function returns the bottom row of the matrix.
	 *
	 * @param dma matrix whose bottom is taken.
	 * @param res the bottom of the matrix @a dma is stored in.
	 * @see m_skeletons::bottom(const matrix<A> *)
	 */
	template< typename A >
	static void bottom(const matrix<A> *dma, matrix <A> *res) {
		assert(!dma->isColumnWise());
		int n = dma->getCols();
		int m = dma->getRows();
		res->setSize(1, n);
		if(res->isColumnWise()){
			res->makeRowWise();
		}
		for(int i = 0; i < n; i++){
			res->at(0, i) = dma->at(m-1, i);
		}
	}
	/**
	 * @brief The left function.
	 *
	 * This function returns the left column of the matrix.
	 *
	 * Definition (in Haskell)
	 * @code
	 * left :: Matrix a -> Matrix a
	 * left = reduce (Abv, ll) . map Sing
	 *   where ll a b = a
	 * @endcode
	 *
	 * Sample: \n
	 * < Haskell code >
	 * @code
	 *   m2 = left m1
	 * @endcode
	 * < C++ code >
	 * @code
	 *   m2 = m_skeletons::left( m1 );
	 * @endcode
	 * @param dma matrix whose left is taken.
	 * @return The left of the matrix @a dma.
	 * @see m_skeletons::reduce(const F &, const G &, const matrix<A> *)
	 * @see m_skeletons::map(const F&, const matrix <A>*)
	 */
	template< typename A >
	static matrix<A>* left(const matrix<A> *dma) {
		matrix<A> *res = new matrix<A>();
		left(dma, res);
		return res;
	}
	/**
	 * @brief The left function.
	 *
	 * This function returns the left column of the matrix.
	 *
	 * @param dma matrix whose left is taken.
	 * @param res the left of the matrix @a dma is stored in.
	 * @see m_skeletons::left(const matrix<A> &)
	 */
	template< typename A >
	static void left(const matrix<A> *dma, matrix <A> *res) {
		assert(!dma->isColumnWise());
		int m = dma->getRows();
		res->setSize(m, 1);
		if(res->isColumnWise()){
			res->makeRowWise();
		}
		for(int i = 0; i < m; i++){
			res->at(i, 0) = dma->at(i, 0);
		}
	}
	/**
	 * @brief The right function.
	 *
	 * This function returns the right column of the matrix.
	 *
	 * Definition (in Haskell)
	 * @code
	 * right :: Matrix a -> Matrix a
	 * right = reduce (Abv, rr) . map Sing
	 *   where rr a b = b
	 * @endcode
	 *
	 * Sample: \n
	 * < Haskell code >
	 * @code
	 *   m2 = right m1
	 * @endcode
	 * < C++ code >
	 * @code
	 *   m2 = m_skeletons::right( m1 );
	 * @endcode
	 * @param dma matrix whose right is taken.
	 * @return The right of the matrix @a dma.
	 * @see m_skeletons::reduce(const F &, const G &, const matrix<A> *)
	 * @see m_skeletons::map(const F&, const matrix <A>*)
	 */
	template< typename A >
	static matrix<A>* right(const matrix<A> *dma) {
		matrix<A> *res = new matrix<A>();
		right(dma, res);
		return res;
	}
	/**
	 * @brief The right function.
	 *
	 * This function returns the right column of the matrix.
	 *
	 * @param dma matrix whose right is taken.
	 * @param res the right of the matrix @a dma is stored in.
	 * @see m_skeletons::right(const matrix<A> &)
	 */
	template< typename A >
	static void right(const matrix<A> *dma, matrix <A> *res) {
		assert(!dma->isColumnWise());
		int n = dma->getCols();
		int m = dma->getRows();
		res->setSize(m, 1);
		if(res->isColumnWise()){
			res->makeRowWise();
		}
		for(int i = 0; i < m; i++){
			res->at(i, 0) = dma->at(i, n-1);
		}
	}
	/**
	 * @brief Makes uppertriangular matrix 
	 *
	 * This function makes uppertriangular matrix  (@a a11 `bsd` @a a12) `abv` (zero `bsd` @a a22), where zero is a matrix whose elements are all @a z.
	 *
	 * @param a11 the matrix at (1,1)-block
	 * @param a12 the matrix at (1,2)-block
	 * @param a22 the matrix at (2,2)-block
	 * @param z   the element with which zero is filled.
	 * @param res resulting matrix is stored in.
	 *
	 */
	template< typename A >
	static void uppertriangular(const matrix<A> *a11, const matrix<A> *a12, const matrix<A> *a22, const A &z, matrix <A> *res) {
		int r1 = a11->getRows();
		int r2 = a22->getRows();
		int c1 = a11->getCols();
		int c2 = a22->getCols();
		assert(r1 == a12->getRows() && c2 == a12->getCols());
		int r = r1 + r2;
		int c = c1 + c2;
		res->setSize(r, c);
		for(int i = 0; i < r1; i++){
			for(int j = 0; j < c1; j++){
				res->at(i, j) = a11->at(i, j);
			}
			for(int j = c1, jj = 0; j < c; j++, jj++){
				res->at(i, j) = a12->at(i, jj);
			}
		}
		for(int i = r1, ii = 0; i < r; i++, ii++){
			for(int j = 0; j < c1; j++){
				res->at(i, j) = z;
			}
			for(int j = c1, jj = 0; j < c; j++, jj++){
				res->at(i, j) = a22->at(ii, jj);
			}
		}
	}
	
};



#endif // _MATRIX_H_
