/***
 * Copyright (c) 2005, SkeTo Project
 * All rights reserved.
 */
/**
 * @file matrix_skeletons.h
 * @brief The Header of class @a matrix_skeletons,
 *        consists of skeletons of matrices(two-dimensional arrays).
 * @author Kento EMOTO
 */

#ifndef __MATRIX_SKELETONS_H__
#define __MATRIX_SKELETONS_H__


#if 0   // for C++ bindings
#define SKEL_MPI_COMM_WORLD MPI::COMM_WORLD
#define SKEL_MPI_INTRACOMM MPI::Intracomm
#define SKEL_MPI_Request MPI::Request
#define SKEL_MPI_Wait(request) request.Wait()
#define SKEL_MPI_Send(buf, count, datatype, dest, tag, comm) comm.Send(buf, count, datatype, dest, tag)
#define SKEL_MPI_Recv(buf, count, datatype, source, tag, comm) comm.Recv(buf, count, datatype, source, tag)
#define SKEL_MPI_Isend( buf, count, datatype, dest, tag, comm, request) request = comm.Isend(buf, count, datatype, dest, tag)
#define SKEL_MPI_Irecv( buf, count, datatype, source, tag, comm, request) request = comm.Irecv(buf, count, datatype, source, tag)
#define	SKEL_MPI_Bcast( buf, count, datatype, source, comm) comm.Bcast(buf, count, datatype, source);
#define SKEL_MPI_Comm_split(oldcomm, color, key, newcomm) newcomm = oldcomm.Split(color, key)
#define SKEL_MPI_Comm_rank(comm, rank) rank = comm.Get_rank()
#define SKEL_MPI_Comm_size(comm, size) size = comm.Get_size()
#define SKEL_MPI_Comm_free(comm) comm.Free()

#else           // for C 
#define SKEL_MPI_COMM_WORLD MPI_COMM_WORLD
#define SKEL_MPI_INTRACOMM MPI_Comm
#define SKEL_MPI_Request MPI_Request
#define SKEL_MPI_Wait(request) MPI_Wait (&request, MPI_STATUS_IGNORE)
#define SKEL_MPI_Send(buf, count, datatype, dest, tag, comm) MPI_Send(buf, count, datatype, dest, tag, comm)
#define SKEL_MPI_Recv(buf, count, datatype, source, tag, comm) MPI_Recv(buf, count, datatype, source, tag, comm, MPI_STATUS_IGNORE)
#define SKEL_MPI_Isend( buf, count, datatype, dest, tag, comm, request) MPI_Isend( (void*)(buf), count, datatype, dest, tag, comm, &request )
#define SKEL_MPI_Irecv( buf, count, datatype, source, tag, comm, request) MPI_Irecv( (void*)(buf), count, datatype, source, tag, comm, &request )
#define	SKEL_MPI_Bcast( buf, count, datatype, source, comm) MPI_Bcast((void*)(buf), count, datatype, source, comm)
#define SKEL_MPI_Comm_split(oldcomm, color, key, newcomm) MPI_Comm_split(oldcomm, color, key, &newcomm)
#define SKEL_MPI_Comm_rank(comm, rank) MPI_Comm_rank(comm, &rank)
#define SKEL_MPI_Comm_size(comm, size) MPI_Comm_size(comm, &size)
#define SKEL_MPI_Comm_free(comm) MPI_Comm_free(&comm)

#endif


#include <utility>


#include <mpi.h>
#include <iostream>
#include <iomanip>
#include "Serializer.h"
#include "matrix.h"
#include "template_mpitype.h"
#include "../util/skeleton.h"


template< typename A >
class dist_matrix;


/**
 * @brief Collection of skeletons on matricies(two-dimensional arrays).
 *
 * This implementation is parallel, distributed.
 */
class matrix_skeletons
{
	////////////////////////////////////////////////////////////////////////
	// Didtribution Policies for dist_matrix
public:
	/**
	 * @brief Policy of implicit distribution by constructors.
	 *
	 * This policy is used when user uses constructors without specifying
	 * distribution (the number of division of the matrix) explicitly.
	 * Default policy is DPOLICY_RECTANGLE.
	 *
	 * @todo This variable should be a member of matrix_skeletons...
	 */
	static int distributionPolicy;
	/**
	 * @brief Distribute policy RECTANGLE. 
	 * Using all processors to distribute the matrix.
	 * Distributed matrix can be seen as a (r, c) matrix of matrices.
	 * r*c = #processors, 
	 * where r is #division in row-direction, 
	 * and c is #division in column-direction.
	 */
	static const int DPOLICY_RECTANGLE = 0; 
	/**
	 * @brief Distribute policy SQUARE. 
	 * The matrix is distributed symmetrically in row- and column-direction.
	 * Distributed matrix can be seen as a (r, r) square matrix of matrices.
	 * r*r =< #processors,
	 * where r is #division in each-direction
	 * and (r+1) * (r+1) > #processors
	 */
	static const int DPOLICY_SQUARE = 1;
	/**
	 * @brief Distribute policy ROW. 
	 * The matrix is distributed in row-direction.
	 * Distributed matrix can be seen as a (r, 1) matrix (vector) of matrices.
	 * r = #processors,
	 * where r is #division in row-direction.
	 */
	static const int DPOLICY_ROW = 2;
	/**
	 * @brief Distribute policy COLUMN. 
	 * The matrix is distributed in column-direction.
	 * Distributed matrix can be seen as a (1, c) matrix (vector) of matrices.
	 * c = #processors,
	 * where c is #division in column-direction.
	 */
	static const int DPOLICY_COLUMN = 3;
	/**
	 * @brief Set the distribution policy.
	 *
	 * @param p new policy.
	 * @return Old policy.
	 * @see DPOLICY_RECTANGLE
	 * @see DPOLICY_SQUARE
	 * @see DPOLICY_ROW
	 * @see DPOLICY_COLUMN
	 */
	static int setDistributionPolicy(int p);
	/**
	 * @brief Return the distribution policy.
	 *
	 * @return Current policy.
	 * @see DPOLICY_RECTANGLE
	 * @see DPOLICY_SQUARE
	 * @see DPOLICY_ROW
	 * @see DPOLICY_COLUMN
	 */
	static int getDistributionPolicy();
	/**
	 * @brief Calculate #divisions (r, c) according to the policy.
	 * @param r #row-division
	 * @param c #col-division
	 * @param com the communicator
	 */
	static void calcDivision(int &r, int &c, const SKEL_MPI_INTRACOMM &com);
	////////////////////////////////////////////////////////////////////////
	// Data parallel Skeletons
 public:
	/**
	 * @brief The map skeleton (inplace).
	 *
	 * This function applies the function to each element of the matrix.
	 *
	 * @param f a function object taking A, 
	 *    and overwriting the result on the argument.
	 *    @a f should have the member void operator()(A*) const.
	 * A sample of function object is as follows: 
	 * (the value *r is replaced by *r*2)
	 * @code
	 * class Double {
	 * public:
	 *     void operator()(int *r) const {
	 *         *r = *r*2;
	 *     }
	 * };
	 * @endcode
	 * @param mat matrix to which the skeleton is applied.
	 * @see matrix_skeletons::map(const F&, const dist_matrix <A>*)
	 */
	template <typename A, typename F>
	static void map_i(const F &f, dist_matrix<A> *mat);

	/**
	 * @brief The map skeleton.
	 *
	 * This function applies the function to each element of the matrix.
	 *
	 * @param f a function taking A, and returning result in B
	 *   @a f should have the member void operator()(const A&, B*) const.
	 * The function object f should inherit skeleton::unary_function.
	 * A sample of function object is as follows: 
	 * (square root of the value. type of result is double and type of argument is int)
	 * @code
	 * class Square : public skeleton::unary_function <int, double> {
	 * public:
	 *     void operator()(const int &i, double *r) const {
	 *         *r = sqrt((double)i);
	 *     }
	 * };
	 * @endcode
	 * @param mat matrix to which the skeleton is applied.
	 * @param dmb matrix to which the result is stored.
	 * @see matrix_skeletons::map(const F&, const dist_matrix <A>*)
	 */
	template< typename B, typename A, typename F>
	static void map(const F &f, const dist_matrix<A> *mat, dist_matrix<B> *dmb);
	
	/**
	 * @brief The map skeleton(crates a new matrix).
	 *
	 * This function applies the function to each element of the matrix.
	 *
	 * Definition (in Haskell)
	 * @code
	 * map :: (a -> b) -> Matrix a -> Matrix b
	 * map f (Sing a)  = Sing (f a)
	 * map f (Abv x y) = Abv (map f x) (map f y)
	 * map f (Bsd x y) = Bsd (map f x) (map f y)
	 * @endcode
	 *
	 * Sample: \n
	 * < Haskell code >
	 * @code
	 *   m2 = map f m1
	 * @endcode
	 * < C++ code >
	 * @code
	 *   m2 = matrix_skeletons::map( f, m1 );
	 * @endcode
	 * @param f a function taking A, and returning result in B
	 *   @a f should have the member void operator()(const A&, B*) const.
	 * The function object f should inherit skeleton::unary_function.
	 * A sample of function object is as follows: 
	 * (square root of the value. type of result is double and type of argument is int)
	 * @code
	 * class Square : public skeleton::unary_function <int, double> {
	 * public:
	 *     void operator()(const int &i, double *r) const {
	 *         *r = sqrt((double)i);
	 *     }
	 * };
	 * @endcode
	 * @param mat matrix to which the skeleton is applied.
	 * @return A matrix to which the result is stored.
	 * @see m_skeletons::map(const F&, const matrix <A>*)
	 */
	template<typename A, typename F>
	static dist_matrix<typename F::result_type>* map(const F &f, const dist_matrix<A> *mat);

	/**
	 * @brief The zipwith skeleton(crates a new matrix).
	 *
	 * This function applies the function to the pair of 
	 * corresponding elements of the matrices.
	 *
	 * @param f a function taking A and B, writing result into C
	 *   @a f should have the member void operator()(const A&, const B&, C*) const.
	 * The function object f should inherit skeleton::binary_function.
	 * A sample of function object is as follows: 
	 * (ratio of two arguments)
	 * @code
	 * class Ratio : public skeleton::binary_function <int, int, double> {
	 * public:
	 *     void operator()(const int &i, const int &j, double *r) const {
	 *         *r = i / (double)j;
	 *     }
	 * };
	 * @endcode
	 * @param dma matrix to be zipped with.
	 * @param dmb matrix to be zipped with.
	 * @param dmc matrix to which result is stored
	 * @see matrix_skeletons::zipwith(const F &, const dist_matrix<A>*, const dist_matrix<B>*)
	 */
	template< typename C, typename A, typename B, typename F>
	static void zipwith(const F &f, const dist_matrix<A>* dma, const dist_matrix<B>* dmb, dist_matrix<C>* dmc );
	
	/**
	 * @brief The zipwith skeleton(creates a new matrix).
	 *
	 * This function applies the function to the pair of 
	 * corresponding elements of the matrices.
	 *
	 * Definition (in Haskell)
	 * @code
	 * zipwith :: (a->b->c) -> Matrix a -> Matrix b -> Matrix c
	 * zipwith f (Sing a) (Sing b)   = Sing (f a b)
	 * zipwith f (Abv x y) (Abv u v) = Abv (zipwith f x u) (zipwith f y v)
	 * zipwith f (Bsd x y) (Bsd u v) = Bsd (zipwith f x u) (zipwith f y v)
	 * @endcode
	 *
	 * Sample: \n
	 * < Haskell code >
	 * @code
	 *   m3 = zipwith f m1 m2
	 * @endcode
	 * < C++ code >
	 * @code
	 *   m3 = matrix_skeletons::zipwith( f, m1, m2 );
	 * @endcode
	 * @param f a function taking A and B, writing result into C
	 *   @a f should have the member void operator()(const A*, const B*, C*) const.
	 * The function object f should inherit skeleton::binary_function.
	 * A sample of function object is as follows: 
	 * (ratio of two arguments)
	 * @code
	 * class Ratio : public skeleton::binary_function <int, int, double> {
	 * public:
	 *     void operator()(const int &i, const int &j, double *r) const {
	 *         *r = i / (double)j;
	 *     }
	 * };
	 * @endcode
	 * @param dma matrix to be zipped with.
	 * @param dmb matrix to be zipped with.
	 * @return A matrix to which the result is stored.
	 * @see m_skeletons::zipwith(const F &, const matrix<A>*, const matrix<B>*)
	 */
	template< typename A, typename B, typename F>
	static dist_matrix<typename F::result_type>*  zipwith(const F &f, const dist_matrix<A>* dma, const dist_matrix<B>* dmb); 
	
	/**
	 * @brief The zip skeleton which is a specialization of zipwith.
	 *
	 * Definition (in Haskell)
	 * @code
	 * zip :: (Matrix a, Matrix b) -> Matrix (a, b)
	 * zip (u, v) = zipwith (\ x y -> (x, y)) u v
	 * @endcode
	 * @param dma matrix to be zipped with.
	 * @param dmb matrix to be zipped with
	 * @result A matrix consisting of pairs of corresponding elements of the matrices.
	 * @see matrix_skeletons::zipwith(const F &, const dist_matrix<A>*, const dist_matrix<B>*)
	 * @see m_skeletons::zip(const matrix<A>*, const matrix<B>*)
	 */
	template< typename A, typename B >
	static dist_matrix < std::pair<A,B> > *zip(const dist_matrix<A>* dma, const dist_matrix<B>* dmb);

	/**
	 * @brief The reduce with broadcast.
	 *
	 * Broadcasting after reduction.
	 * @param oplus a function taking two arguments x and y, returning result z,
	 * i.e. z = x @a oplus y, and must be abiding with @a otimes.
	 * @a oplus should have the member void operator()(const A&, const A&, *A) const.
	 * The function object @a oplus should inherit skeleton::binary_function.
	 * A sample of function object is as follows: 
	 * (Max)
	 * @code
	 * class Max : public skeleton::binary_function <int, int, int> {
	 * public:
	 *     void operator()(const int &i, const int &j, int *r) const {
	 *         *r = (i > j) ? i : j;
	 *     }
	 * };
	 * @endcode
	 * @a oplus should be associative.
	 * @param otimes a function taking two arguments x and y, returning result z,
	 * i.e. z = x @a otimes y, and must be abiding with @a oplus.
	 * The function @a otimes oplus should inherit skeleton::binary_function.
	 * @a otimes should be associative.
	 * Associative and commutative operator has abiding property with itself.
	 * @param mat a matrix to which the reduce is applied.
	 * @param res resulting reduced value.
	 * @see matrix_skeletons::reduceB(const F &, const G &, const dist_matrix<A> *)
	 * @see matrix_skeletons::reduce(const F &, const G &, const dist_matrix<A> *)
	 * @see dist_matrix::bcast(B &, const dist_matrix< A > *)
	 */
 public:
	template <typename A, typename F, typename G>
	static void reduceB(const F &oplus, const G &otimes, const dist_matrix<A>* mat, A *res);

	/**
	 * @brief The reduce with broadcast.
	 *
	 * Broadcasting after reduction.
	 * @param oplus a function taking two arguments x and y, returning result z,
	 * i.e. z = x @a oplus y, and must be abiding with @a otimes.
	 * @a oplus should have the member void operator()(const A&, const A&, A*) const.
	 * The function object @a oplus should inherit skeleton::binary_function.
	 * A sample of function object is as follows: 
	 * (Max)
	 * @code
	 * class Max : public skeleton::binary_function <int, int, int> {
	 * public:
	 *     void operator()(const int &i, const int &j, int *r) const {
	 *         *r = (i > j) ? i : j;
	 *     }
	 * };
	 * @endcode
	 * @a oplus should be associative.
	 * @param otimes a function taking two arguments x and y, returning result z,
	 * i.e. z = x @a otimes y, and must be abiding with @a oplus.
	 * The function @a otimes oplus should inherit skeleton::binary_function.
	 * @a otimes should be associative.
	 * Associative and commutative operator has abiding property with itself.
	 * @param mat a matrix to which the reduce is applied.
	 * @return Resulting reduced value.
	 * @see matrix_skeletons::reduce(const F &, const G &, const dist_matrix<A> *)
	 * @see dist_matrix::bcast(B &, const dist_matrix< A > *)
	 */
 public:
	template <typename A, typename F, typename G>
	static A *reduceB(const F &oplus, const G &otimes, const dist_matrix<A>* mat);
	/**
	 * @brief The reduce skeleton.
	 *
	 * This skeleton reduces the matrix into a value with two binary operators.
	 * The matrices must be row-wise.
	 * @param oplus a function taking two arguments x and y, returning result z,
	 * i.e. z = x @a oplus y, and must be abiding with @a otimes.
	 * @a oplus should have the member void operator()(const A&, const A&, A*) const.
	 * The function object @a oplus should inherit skeleton::binary_function.
	 * A sample of function object is as follows: 
	 * (Max)
	 * @code
	 * class Max : public skeleton::binary_function <int, int, int> {
	 * public:
	 *     void operator()(const int &i, const int &j, int *r) const {
	 *         *r = (i > j) ? i : j;
	 *     }
	 * };
	 * @endcode
	 * @a oplus should be associative.
	 * @param otimes a function taking two arguments x and y, returning result z,
	 * i.e. z = x @a otimes y, and must be abiding with @a oplus.
	 * The function @a otimes oplus should inherit skeleton::binary_function.
	 * @a otimes should be associative.
	 * Associative and commutative operator has abiding property with itself.
	 * @param mat a matrix to which the reduce is applied.
	 * @param res resulting reduced value.
	 * @see matrix_skeletons::reduce(const F &, const G &, const dist_matrix<A> *)
	 */
 public:
	template <typename A, typename F, typename G>
	static void reduce(const F &oplus, const G &otimes, const dist_matrix<A>* mat, A *res);

	/**
	 * @brief The reduce skeleton (create new value).
	 *
	 * This skeleton reduces the matrix into a value with two binary operators.
	 *
	 * Definition (in Haskell)
	 * @code
	 * reduce :: (a->a->a, a->a->a) -> Matrix a -> a
	 * reduce (oplus, otimes) (Sing a)  = a
	 * reduce (oplus, otimes) (Abv x y) = oplus ( reduce (oplus, otimes) x) ( reduce (oplus, otimes) y)
	 * reduce (oplus, otimes) (Bsd x y) = otimes ( reduce (oplus, otimes) x) ( reduce (oplus, otimes) y)
	 * @endcode
	 * Two operators should have abiding property.
	 *
	 * Sample: \n
	 * < Haskell code >
	 * @code
	 *   r = reduce ((+), (+)) m
	 * @endcode
	 * < C++ code >
	 * @code
	 *   r = matrix_skeletons::reduce( Add(), Add(), m);
	 * @endcode
	 *
	 * The matrices must be row-wise.
	 * @param oplus a function taking two arguments x and y, returning result z,
	 * i.e. z = x @a oplus y, and must be abiding with otimes.
	 * @a oplus should have the member void operator()(const A&, const A&, A*) const.
	 * The function object @a oplus should inherit skeleton::binary_function.
	 * A sample of function object is as follows: 
	 * (Addition)
	 * @code
	 * class Add : public skeleton::binary_function <int, int, int> {
	 * public:
	 *     void operator()(const int &i, const int &j, int *r) const {
	 *         *r = i + j;
	 *     }
	 * };
	 * @endcode
	 * @a oplus should be associative.
	 * @param otimes a function taking two arguments x and y, returning result z,
	 * i.e. z = x @a otimes y, and must be abiding with oplus.
	 * The function @a otimes oplus should inherit skeleton::binary_function.
	 * @a otimes should be associative.
	 * Associative and commutative operator has abiding property with itself.
	 * @param mat a matrix to which the reduce is applied.
	 * @return Resulting reduced value. User should free it.
	 * @see m_skeletons::reduce(const F &, const G &, const matrix<A> *)
	 */
 public:
	template <typename A, typename F, typename G>
	static A *reduce(const F &oplus, const G &otimes, const dist_matrix<A>* mat);

	/**
	 * @brief The reduce skeleton(with divide and conquer local computation).
	 *
	 * This skeleton reduces the matrix into a value with two binary operators.
	 * The matrices must be row-wise.
	 * @param oplus a function taking two arguments x and y, returning result z,
	 * i.e. z = x @a oplus y, and must be abiding with @a otimes.
	 * @a oplus should have the member void operator()(const A&, const A&, A*) const.
	 * The function object @a oplus should inherit skeleton::binary_function.
	 * A sample of function object is as follows: 
	 * (Max)
	 * @code
	 * class Max : public skeleton::binary_function <int, int, int> {
	 * public:
	 *     void operator()(const int &i, const int &j, int *r) const {
	 *         *r = (i > j) ? i : j;
	 *     }
	 * };
	 * @endcode
	 * @a oplus should be associative.
	 * @param otimes a function taking two arguments x and y, returning result z,
	 * i.e. z = x @a otimes y, and must be abiding with @a oplus.
	 * The function @a otimes oplus should inherit skeleton::binary_function.
	 * @a otimes should be associative.
	 * Associative and commutative operator has abiding property with itself.
	 * @param mat a matrix to which the reduce is applied.
	 * @param res resulting reduced value.
	 * @see matrix_skeletons::reduce(const F &, const G &, const dist_matrix<A> *)
	 */
 public:
	template <typename A, typename F, typename G>
	static void reduceDC(const F &oplus, const G &otimes, const dist_matrix<A>* mat, A *res);

	/**
	 * @brief The reduce skeleton (create new value, with divide and conquer local computation).
	 *
	 * This skeleton reduces the matrix into a value with two binary operators.
	 *
	 * Definition (in Haskell)
	 * @code
	 * reduce :: (a->a->a, a->a->a) -> Matrix a -> a
	 * reduce (oplus, otimes) (Sing a)  = a
	 * reduce (oplus, otimes) (Abv x y) = oplus ( reduce (oplus, otimes) x) ( reduce (oplus, otimes) y)
	 * reduce (oplus, otimes) (Bsd x y) = otimes ( reduce (oplus, otimes) x) ( reduce (oplus, otimes) y)
	 * @endcode
	 * Two operators should have abiding property.
	 *
	 * Sample: \n
	 * < Haskell code >
	 * @code
	 *   r = reduce ((+), (+)) m
	 * @endcode
	 * < C++ code >
	 * @code
	 *   r = matrix_skeletons::reduce( Add(), Add(), m);
	 * @endcode
	 *
	 * The matrices must be row-wise.
	 * @param oplus a function taking two arguments x and y, returning result z,
	 * i.e. z = x @a oplus y, and must be abiding with otimes.
	 * @a oplus should have the member void operator()(const A&, const A&, A*) const.
	 * The function object @a oplus should inherit skeleton::binary_function.
	 * A sample of function object is as follows: 
	 * (Addition)
	 * @code
	 * class Add : public skeleton::binary_function <int, int, int> {
	 * public:
	 *     void operator()(const int &i, const int &j, int *r) const {
	 *         *r = i + j;
	 *     }
	 * };
	 * @endcode
	 * @a oplus should be associative.
	 * @param otimes a function taking two arguments x and y, returning result z,
	 * i.e. z = x @a otimes y, and must be abiding with oplus.
	 * The function @a otimes oplus should inherit skeleton::binary_function.
	 * @a otimes should be associative.
	 * Associative and commutative operator has abiding property with itself.
	 * @param mat a matrix to which the reduce is applied.
	 * @return Resulting reduced value. User should free it.
	 * @see m_skeletons::reduce(const F &, const G &, const matrix<A> *)
	 */
 public:
	template <typename A, typename F, typename G>
	static A *reduceDC(const F &oplus, const G &otimes, const dist_matrix<A>* mat);

	/**
	 * @brief The reduce skeleton(with divide and conquer local computation, nested parallelism).
	 *
	 * This skeleton reduces the matrix into a value with two binary operators.
	 * The matrices must be row-wise.
	 * @param oplus a function taking two arguments x and y, returning result z,
	 * i.e. z = x @a oplus y, and must be abiding with @a otimes.
	 * @a oplus should have the member void operator()(const A&, const A&, A*) const.
	 * The function object @a oplus should inherit skeleton::binary_function.
	 * A sample of function object is as follows: 
	 * (Max)
	 * @code
	 * class Max : public skeleton::binary_function <int, int, int> {
	 * public:
	 *     void operator()(const int &i, const int &j, int *r) const {
	 *         *r = (i > j) ? i : j;
	 *     }
	 * };
	 * @endcode
	 * @a oplus should be associative.
	 * @param otimes a function taking two arguments x and y, returning result z,
	 * i.e. z = x @a otimes y, and must be abiding with @a oplus.
	 * The function @a otimes oplus should inherit skeleton::binary_function.
	 * @a otimes should be associative.
	 * Associative and commutative operator has abiding property with itself.
	 * @param mat a matrix to which the reduce is applied.
	 * @param res resulting reduced value.
	 * @see matrix_skeletons::reduce(const F &, const G &, const dist_matrix<A> *)
	 */
 public:
	template <typename A, typename F, typename G>
	static void reduceDCN(const F &oplus, const G &otimes, const dist_matrix<A>* mat, A *res);

	/**
	 * @brief The reduce skeleton (create new value, with divide and conquer local computation, nested parallelism).
	 *
	 * This skeleton reduces the matrix into a value with two binary operators.
	 *
	 * Definition (in Haskell)
	 * @code
	 * reduce :: (a->a->a, a->a->a) -> Matrix a -> a
	 * reduce (oplus, otimes) (Sing a)  = a
	 * reduce (oplus, otimes) (Abv x y) = oplus ( reduce (oplus, otimes) x) ( reduce (oplus, otimes) y)
	 * reduce (oplus, otimes) (Bsd x y) = otimes ( reduce (oplus, otimes) x) ( reduce (oplus, otimes) y)
	 * @endcode
	 * Two operators should have abiding property.
	 *
	 * Sample: \n
	 * < Haskell code >
	 * @code
	 *   r = reduce ((+), (+)) m
	 * @endcode
	 * < C++ code >
	 * @code
	 *   r = matrix_skeletons::reduce( Add(), Add(), m);
	 * @endcode
	 *
	 * The matrices must be row-wise.
	 * @param oplus a function taking two arguments x and y, returning result z,
	 * i.e. z = x @a oplus y, and must be abiding with otimes.
	 * @a oplus should have the member void operator()(const A&, const A&, A*) const.
	 * The function object @a oplus should inherit skeleton::binary_function.
	 * A sample of function object is as follows: 
	 * (Addition)
	 * @code
	 * class Add : public skeleton::binary_function <int, int, int> {
	 * public:
	 *     void operator()(const int &i, const int &j, int *r) const {
	 *         *r = i + j;
	 *     }
	 * };
	 * @endcode
	 * @a oplus should be associative.
	 * @param otimes a function taking two arguments x and y, returning result z,
	 * i.e. z = x @a otimes y, and must be abiding with oplus.
	 * The function @a otimes oplus should inherit skeleton::binary_function.
	 * @a otimes should be associative.
	 * Associative and commutative operator has abiding property with itself.
	 * @param mat a matrix to which the reduce is applied.
	 * @return Resulting reduced value. User should free it.
	 * @see m_skeletons::reduce(const F &, const G &, const matrix<A> *)
	 */
 public:
	template <typename A, typename F, typename G>
	static A *reduceDCN(const F &oplus, const G &otimes, const dist_matrix<A>* mat);
	/**
	 * @brief The reduce skeleton(with divide and conquer local computation, nested parallelism, broadcasting operands for each operator).
	 *
	 * This skeleton reduces the matrix into a value with two binary operators.
	 * The matrices must be row-wise.
	 * @param oplus a function taking two arguments x and y, returning result z,
	 * i.e. z = x @a oplus y, and must be abiding with @a otimes.
	 * @a oplus should have the member void operator()(const A&, const A&, A*) const.
	 * The function object @a oplus should inherit skeleton::binary_function.
	 * A sample of function object is as follows: 
	 * (Max)
	 * @code
	 * class Max : public skeleton::binary_function <int, int, int> {
	 * public:
	 *     void operator()(const int &i, const int &j, int *r) const {
	 *         *r = (i > j) ? i : j;
	 *     }
	 * };
	 * @endcode
	 * @a oplus should be associative.
	 * @param otimes a function taking two arguments x and y, returning result z,
	 * i.e. z = x @a otimes y, and must be abiding with @a oplus.
	 * The function @a otimes oplus should inherit skeleton::binary_function.
	 * @a otimes should be associative.
	 * Associative and commutative operator has abiding property with itself.
	 * @param mat a matrix to which the reduce is applied.
	 * @param res resulting reduced value.
	 * @see matrix_skeletons::reduce(const F &, const G &, const dist_matrix<A> *)
	 */
 public:
	template <typename A, typename F, typename G>
	static void reduceDCN2(const F &oplus, const G &otimes, const dist_matrix<A>* mat, A *res);

	/**
	 * @brief The reduce skeleton (create new value, with divide and conquer local computation, nested parallelism, broadcasting operands for each operator).
	 *
	 * This skeleton reduces the matrix into a value with two binary operators.
	 *
	 * Definition (in Haskell)
	 * @code
	 * reduce :: (a->a->a, a->a->a) -> Matrix a -> a
	 * reduce (oplus, otimes) (Sing a)  = a
	 * reduce (oplus, otimes) (Abv x y) = oplus ( reduce (oplus, otimes) x) ( reduce (oplus, otimes) y)
	 * reduce (oplus, otimes) (Bsd x y) = otimes ( reduce (oplus, otimes) x) ( reduce (oplus, otimes) y)
	 * @endcode
	 * Two operators should have abiding property.
	 *
	 * Sample: \n
	 * < Haskell code >
	 * @code
	 *   r = reduce ((+), (+)) m
	 * @endcode
	 * < C++ code >
	 * @code
	 *   r = matrix_skeletons::reduce( Add(), Add(), m);
	 * @endcode
	 *
	 * The matrices must be row-wise.
	 * @param oplus a function taking two arguments x and y, returning result z,
	 * i.e. z = x @a oplus y, and must be abiding with otimes.
	 * @a oplus should have the member void operator()(const A&, const A&, A*) const.
	 * The function object @a oplus should inherit skeleton::binary_function.
	 * A sample of function object is as follows: 
	 * (Addition)
	 * @code
	 * class Add : public skeleton::binary_function <int, int, int> {
	 * public:
	 *     void operator()(const int &i, const int &j, int *r) const {
	 *         *r = i + j;
	 *     }
	 * };
	 * @endcode
	 * @a oplus should be associative.
	 * @param otimes a function taking two arguments x and y, returning result z,
	 * i.e. z = x @a otimes y, and must be abiding with oplus.
	 * The function @a otimes oplus should inherit skeleton::binary_function.
	 * @a otimes should be associative.
	 * Associative and commutative operator has abiding property with itself.
	 * @param mat a matrix to which the reduce is applied.
	 * @return Resulting reduced value. User should free it.
	 * @see m_skeletons::reduce(const F &, const G &, const matrix<A> *)
	 */
 public:
	template <typename A, typename F, typename G>
	static A *reduceDCN2(const F &oplus, const G &otimes, const dist_matrix<A>* mat);
	/**
	 * @brief The cata skeleton.
	 *
	 * This skeleton reduces the matrix into a value with two binary operators
	 * after applying a function to each element.
	 * The matrices must be row-wise.
	 * @param oplus a function taking two arguments x and y, returning result z,
	 * i.e. z = x @a oplus y, and must be abiding with @a otimes.
	 * @a oplus should have the member void operator()(const A&, const A&, A*) const.
	 * The function object @a oplus should inherit skeleton::binary_function.
	 * A sample of function object is as follows: 
	 * (Max)
	 * @code
	 * class Max : public skeleton::binary_function <int, int, int> {
	 * public:
	 *     void operator()(const int &i, const int &j, int *r) const {
	 *         *r = (i > j) ? i : j;
	 *     }
	 * };
	 * @endcode
	 * @a oplus should be associative.
	 * @param otimes a function taking two arguments x and y, returning result z,
	 * i.e. z = x @a otimes y, and must be abiding with @a oplus.
	 * The function @a otimes oplus should inherit skeleton::binary_function.
	 * @a otimes should be associative.
	 * Associative and commutative operator has abiding property with itself.
	 * @param f a function taking A, and returning result in B
	 *   @a f should have the member void operator()(const A&, B*) const.
	 * The function object f should inherit skeleton::unary_function.
	 * A sample of function object is as follows: 
	 * (square root of the value. type of result is double and type of argument is int)
	 * @code
	 * class Square : public skeleton::unary_function <int, double> {
	 * public:
	 *     void operator()(const int &i, double *r) const {
	 *         *r = sqrt((double)i);
	 *     }
	 * };
	 * @endcode
	 * @param mat a matrix to which the cata is applied.
	 * @param res resulting reduced value.
	 * @see matrix_skeletons::cata(const F &, const G &, const H &, const dist_matrix<A> *)
	 */
 public:
	template <typename A, typename F, typename G, typename H>
	static void cata(const F &oplus, const G &otimes, const H &f, const dist_matrix<A>* mat, typename H::result_type *res);

	/**
	 * @brief The cata skeleton (create new value).
	 *
	 * This skeleton reduces the matrix into a value with two binary operators
	 * after applying a function to each element.
	 *
	 * Definition (in Haskell)
	 * @code
	 * cata :: (a->a->a, a->a->a, b->a) -> Matrix b -> a
	 * cata (oplus, otimes, f) (Sing a)  = f a
	 * cata (oplus, otimes, f) (Abv x y) = oplus ( cata (oplus, otimes, f) x) ( cata (oplus, otimes, f) y)
	 * cata (oplus, otimes, f) (Bsd x y) = otimes ( cata (oplus, otimes, f) x) ( cata (oplus, otimes, f) y)
	 * @endcode
	 * Two operators should have abiding property.
	 *
	 * Sample: \n
	 * < Haskell code >
	 * @code
	 *   r = cata ((+), (+), square) m
	 * @endcode
	 * < C++ code >
	 * @code
	 *   r = matrix_skeletons::cata( Add(), Add(), Square(), m);
	 * @endcode
	 *
	 * The matrices must be row-wise.
	 * @param oplus a function taking two arguments x and y, returning result z,
	 * i.e. z = x @a oplus y, and must be abiding with otimes.
	 * @a oplus should have the member void operator()(const A&, const A&, A*) const.
	 * The function object @a oplus should inherit skeleton::binary_function.
	 * A sample of function object is as follows: 
	 * (Addition)
	 * @code
	 * class Add : public skeleton::binary_function <int, int, int> {
	 * public:
	 *     void operator()(const int &i, const int &j, int *r) const {
	 *         *r = i + j;
	 *     }
	 * };
	 * @endcode
	 * @a oplus should be associative.
	 * @param otimes a function taking two arguments x and y, returning result z,
	 * i.e. z = x @a otimes y, and must be abiding with oplus.
	 * The function @a otimes oplus should inherit skeleton::binary_function.
	 * @a otimes should be associative.
	 * Associative and commutative operator has abiding property with itself.
	 * @param f a function taking A, and returning result in B
	 *   @a f should have the member void operator()(const A&, B*) const.
	 * The function object f should inherit skeleton::unary_function.
	 * A sample of function object is as follows: 
	 * (square root of the value. type of result is double and type of argument is int)
	 * @code
	 * class Square : public skeleton::unary_function <int, double> {
	 * public:
	 *     void operator()(const int &i, double *r) const {
	 *         *r = sqrt((double)i);
	 *     }
	 * };
	 * @endcode
	 * @param mat a matrix to which the reduce is applied.
	 * @return Resulting reduced value. User should free it.
	 * @see m_skeletons::reduce(const F &, const G &, const matrix<A> *)
	 */
 public:
	template <typename A, typename F, typename G, typename H>
	static typename H::result_type *cata(const F &oplus, const G &otimes, const H &f, const dist_matrix<A>* mat);

	/**
	 * @brief The cata skeleton(with divide and conquer local computation, nested parallelism, broadcasting operands for each operator).
	 *
	 * This skeleton reduces the matrix into a value with two binary operators
	 * after applying a function to each element.
	 * The matrices must be row-wise.
	 * @param oplus a function taking two arguments x and y, returning result z,
	 * i.e. z = x @a oplus y, and must be abiding with @a otimes.
	 * @a oplus should have the member void operator()(const A&, const A&, A*) const.
	 * The function object @a oplus should inherit skeleton::binary_function.
	 * A sample of function object is as follows: 
	 * (Max)
	 * @code
	 * class Max : public skeleton::binary_function <int, int, int> {
	 * public:
	 *     void operator()(const int &i, const int &j, int *r) const {
	 *         *r = (i > j) ? i : j;
	 *     }
	 * };
	 * @endcode
	 * @a oplus should be associative.
	 * @param otimes a function taking two arguments x and y, returning result z,
	 * i.e. z = x @a otimes y, and must be abiding with @a oplus.
	 * The function @a otimes oplus should inherit skeleton::binary_function.
	 * @a otimes should be associative.
	 * Associative and commutative operator has abiding property with itself.
	 * @param f a function taking A, and returning result in B
	 *   @a f should have the member void operator()(const A&, B*) const.
	 * The function object f should inherit skeleton::unary_function.
	 * A sample of function object is as follows: 
	 * (square root of the value. type of result is double and type of argument is int)
	 * @code
	 * class Square : public skeleton::unary_function <int, double> {
	 * public:
	 *     void operator()(const int &i, double *r) const {
	 *         *r = sqrt((double)i);
	 *     }
	 * };
	 * @endcode
	 * @param mat a matrix to which the cata is applied.
	 * @param res resulting reduced value.
	 * @see matrix_skeletons::cataDCN2(const F &, const G &, const H &, const dist_matrix<A> *)
	 */
 public:
	template <typename A, typename F, typename G, typename H>
	static void cataDCN2(const F &oplus, const G &otimes, const H &f, const dist_matrix<A>* mat, typename H::result_type *res);

	/**
	 * @brief The cata skeleton (create new value, with divide and conquer local computation, nested parallelism, broadcasting operands for each operator).
	 *
	 * This skeleton reduces the matrix into a value with two binary operators
	 * after applying a function to each element.
	 *
	 * Definition (in Haskell)
	 * @code
	 * cata :: (a->a->a, a->a->a, b->a) -> Matrix b -> a
	 * cata (oplus, otimes, f) (Sing a)  = f a
	 * cata (oplus, otimes, f) (Abv x y) = oplus ( cata (oplus, otimes, f) x) ( cata (oplus, otimes, f) y)
	 * cata (oplus, otimes, f) (Bsd x y) = otimes ( cata (oplus, otimes, f) x) ( cata (oplus, otimes, f) y)
	 * @endcode
	 * Two operators should have abiding property.
	 *
	 * Sample: \n
	 * < Haskell code >
	 * @code
	 *   r = cata ((+), (+), square) m
	 * @endcode
	 * < C++ code >
	 * @code
	 *   r = matrix_skeletons::cata( Add(), Add(), Square(), m);
	 * @endcode
	 *
	 * The matrices must be row-wise.
	 * @param oplus a function taking two arguments x and y, returning result z,
	 * i.e. z = x @a oplus y, and must be abiding with otimes.
	 * @a oplus should have the member void operator()(const A&, const A&, A*) const.
	 * The function object @a oplus should inherit skeleton::binary_function.
	 * A sample of function object is as follows: 
	 * (Addition)
	 * @code
	 * class Add : public skeleton::binary_function <int, int, int> {
	 * public:
	 *     void operator()(const int &i, const int &j, int *r) const {
	 *         *r = i + j;
	 *     }
	 * };
	 * @endcode
	 * @a oplus should be associative.
	 * @param otimes a function taking two arguments x and y, returning result z,
	 * i.e. z = x @a otimes y, and must be abiding with oplus.
	 * The function @a otimes oplus should inherit skeleton::binary_function.
	 * @a otimes should be associative.
	 * Associative and commutative operator has abiding property with itself.
	 * @param f a function taking A, and returning result in B
	 *   @a f should have the member void operator()(const A&, B*) const.
	 * The function object f should inherit skeleton::unary_function.
	 * A sample of function object is as follows: 
	 * (square root of the value. type of result is double and type of argument is int)
	 * @code
	 * class Square : public skeleton::unary_function <int, double> {
	 * public:
	 *     void operator()(const int &i, double *r) const {
	 *         *r = sqrt((double)i);
	 *     }
	 * };
	 * @endcode
	 * @param mat a matrix to which the reduce is applied.
	 * @return Resulting reduced value. User should free it.
	 * @see m_skeletons::reduce(const F &, const G &, const matrix<A> *)
	 */
 public:
	template <typename A, typename F, typename G, typename H>
	static typename H::result_type *cataDCN2(const F &oplus, const G &otimes, const H &f, const dist_matrix<A>* mat);



	/**
	 * @brief The scan skeleton.
	 *
	 * This skeleton stores every process of reduction of a matrix into a value with two binary operators.
	 * The matrices must be row-wise.
	 * @param oplus a function taking two arguments x and y, returning result z,
	 * i.e. z = x @a oplus y, and must be abiding with @a otimes.
	 * @a oplus should have the member void operator()(const A&, const A&, A*) const.
	 * The function object @a oplus should inherit skeleton::binary_function.
	 * A sample of function object is as follows: 
	 * (Max)
	 * @code
	 * class Max : public skeleton::binary_function <int, int, int> {
	 * public:
	 *     void operator()(const int &i, const int &j, int *r) const {
	 *         *r = (i > j) ? i : j;
	 *     }
	 * };
	 * @endcode
	 * @a oplus should be associative.
	 * @param otimes a function taking two arguments x and y, returning result z,
	 * i.e. z = x @a otimes y, and must be abiding with @a oplus.
	 * The function @a otimes oplus should inherit skeleton::binary_function.
	 * @a otimes should be associative.
	 * Associative and commutative operator has abiding property with itself.
	 * @param dma a matrix to which the reduce is applied.
	 * @param dmb resulting matrix of all reduction values.
	 * @see matrix_skeletons::scan(const F &, const G &, const dist_matrix<A> *)
	 */
	template <typename A, typename F, typename G>
	static void scan(const F &oplus, const G &otimes, const dist_matrix<A>* dma, dist_matrix<A>* dmb);

	/**
	 * @brief The scan skeleton.
	 *
	 * This skeleton stores every process of reduction of a matrix into a value with two binary operators.
	 * The matrices must be row-wise.
	 *
	 * Definition (in Haskell)
	 * @code
	 * scan :: (a->a->a, a->a->a) -> Matrix a -> Matrix a
	 * scan (oplus, otimes) (Sing a) = (Sing a)
	 * scan (oplus, otimes) (Abv x y) = oplus' ( scan (oplus, otimes) x) ( scan (oplus, otimes) y)
	 *     where oplus' sx sy = Abv sx sy'
	 *         where sy' = mapr (zipwith oplus (bottom sx)) sy
	 * scan (oplus, otimes) (Bsd x y) = otimes' ( scan (oplus, otimes) x) ( scan (oplus, otimes) y)
	 *     where otimes' sx sy = Bsd sx sy'
	 *         where sy' = mapc (zipwith otimes (right sx)) sy
	 * @endcode
	 * Two operators should have abiding property.
	 *
	 * Sample: \n
	 * < Haskell code >
	 * @code
	 *   m2 = scan ((+), (+)) m1
	 * @endcode
	 * < C++ code >
	 * @code
	 *   m2 = matrix_skeletons::scan( Add(), Add(), m1);
	 * @endcode
	 * @param oplus a function taking two arguments x and y, returning result z,
	 * i.e. z = x @a oplus y, and must be abiding with @a otimes.
	 * @a oplus should have the member void operator()(const A&, const A&, A*) const.
	 * The function object @a oplus should inherit skeleton::binary_function.
	 * A sample of function object is as follows: 
	 * (Addition)
	 * @code
	 * class Add : public skeleton::binary_function <int, int, int> {
	 * public:
	 *     void operator()(const int &i, const int &j, int *r) const {
	 *         *r = i+j;
	 *     }
	 * };
	 * @endcode
	 * @a oplus should be associative.
	 * @param otimes a function taking two arguments x and y, returning result z,
	 * i.e. z = x @a otimes y, and must be abiding with @a oplus.
	 * The function @a otimes oplus should inherit skeleton::binary_function.
	 * @a otimes should be associative.
	 * Associative and commutative operator has abiding property with itself.
	 * @param dma a matrix to which the reduce is applied.
	 * @return Resulting matrix of all reduction values. User should free it.
	 * @see matrix_skeletons::reduce(const F &, const G &, const dist_matrix<A> *)
	 * @see matrix_skeletons::zipwith(const F &, const dist_matrix<A>*, const dist_matrix<B>*)
	 * @see m_skeletons::scan(const F &, const G &, const matrix<A> *)
	 * @see m_skeletons::right(const matrix<A> *)
	 * @see m_skeletons::bottom(const matrix<A> *)
	 * @see m_skeletons::mapr(const F &, const matrix<A> *)
	 * @see m_skeletons::mapc(const F &, const matrix<A> *)
	 */
	template <typename A, typename F, typename G>
	static dist_matrix<A>* scan(const F &oplus, const G &otimes, const dist_matrix<A>* dma);

	/**
	 * @brief The reverse-scan skeleton.
	 *
	 * This skeleton stores every process of reverse reduction of a matrix into a value with two binary operators.
	 * The matrices must be row-wise.
	 *
	 * @param oplus a function taking two arguments x and y, returning result z,
	 * i.e. z = x @a oplus y, and must be abiding with @a otimes.
	 * @a oplus should have the member void operator()(const A&, const A&, A*) const.
	 * The function object @a oplus should inherit skeleton::binary_function.
	 * A sample of function object is as follows: 
	 * (Addition)
	 * @code
	 * class Add : public skeleton::binary_function <int, int, int> {
	 * public:
	 *     void operator()(const int &i, const int &j, int *r) const {
	 *         *r = i+j;
	 *     }
	 * };
	 * @endcode
	 * @a oplus should be associative.
	 * @param otimes a function taking two arguments x and y, returning result z,
	 * i.e. z = x @a otimes y, and must be abiding with @a oplus.
	 * The function @a otimes oplus should inherit skeleton::binary_function.
	 * @a otimes should be associative.
	 * Associative and commutative operator has abiding property with itself.
	 * @param dma a matrix to which the reduce is applied.
	 * @param dmb resulting matrix of all reduction values. User should free it.
	 * @see matrix_skeletons::scanr(const F &, const G &, const dist_matrix<A> *)

	 */
	template <typename A, typename F, typename G>
	static void scanr(const F &oplus, const G &otimes, const dist_matrix<A>* dma, dist_matrix<A>* dmb);

	/**
	 * @brief The reverse-scan skeleton.
	 *
	 * This skeleton stores every process of reverse reduction of a matrix into a value with two binary operators.
	 * The matrices must be row-wise.
	 *
	 * Definition (in Haskell)
	 * @code
	 * scanr :: (a->a->a, a->a->a) -> Matrix a -> Matrix a
	 * scanr (oplus, otimes) = rev . scan (binrev oplus, binrev otimes) . rev
	 *    where 
	 *    rev = reduce (binrev Abv, binrev Bsd) . map Sing
	 *    binrev f a b = f b a
	 * @endcode
	 * Two operators should have abiding property.
	 *
	 * Sample: \n
	 * < Haskell code >
	 * @code
	 *   m2 = scanr ((+), (+)) m1
	 * @endcode
	 * < C++ code >
	 * @code
	 *   m2 = matrix_skeletons::scanr( Add(), Add(), m1);
	 * @endcode
	 * @param oplus a function taking two arguments x and y, returning result z,
	 * i.e. z = x @a oplus y, and must be abiding with @a otimes.
	 * @a oplus should have the member void operator()(const A&, const A&, A*) const.
	 * The function object @a oplus should inherit skeleton::binary_function.
	 * A sample of function object is as follows: 
	 * (Addition)
	 * @code
	 * class Add : public skeleton::binary_function <int, int, int> {
	 * public:
	 *     void operator()(const int &i, const int &j, int *r) const {
	 *         *r = i+j;
	 *     }
	 * };
	 * @endcode
	 * @a oplus should be associative.
	 * @param otimes a function taking two arguments x and y, returning result z,
	 * i.e. z = x @a otimes y, and must be abiding with @a oplus.
	 * The function @a otimes oplus should inherit skeleton::binary_function.
	 * @a otimes should be associative.
	 * Associative and commutative operator has abiding property with itself.
	 * @param dma a matrix to which the reduce is applied.
	 * @return Resulting matrix of all reduction values. User should free it.
	 * @see matrix_skeletons::scan(const F &, const G &, const dist_matrix<A> *)
	 * @see matrix_skeletons::reduce(const F &, const G &, const dist_matrix<A> *)
	 * @see matrix_skeletons::map(const F &, const dist_matrix<A> *)
	 * @see m_skeletons::scanr(const F &, const G &, const matrix<A> *)
	 */
	template <typename A, typename F, typename G>
	static dist_matrix<A>* scanr(const F &oplus, const G &otimes, const dist_matrix<A>* dma);


	/**
	 * @brief A specialization of allreduce_row: all_rows.
	 * 
	 * Resulting matrix consists of rows of the matrix @a mat. i.e.
	 * each element in a row of the matrix is corresponding row of this matrix.
	 * Definition (in Haskell)
	 * @code
	 * allrows :: Matrix a -> Matrix (Matrix a)
	 * allrows = allreduce_row (Bsd) . map Sing
	 * @endcode
	 *
	 * Sample: \n
	 * < Haskell code >
	 * @code
	 *   m2 = allrows m1
	 * @endcode
	 * < C++ code >
	 * @code
	 *   m2 = matrix_skeletons::allrows( m1 );
	 * @endcode
	 * @param mat a matrix.
	 * @return a matrix consisting of rows of the matrix @a mat.
	 * @see matrix_skeletons::allreduce_row(const F &, const dist_matrix<A> *)
	 */
	template<typename A>
	static dist_matrix< matrix< A > >* all_rows(const dist_matrix<A> *mat);
	/**
	 * @brief A specialization of allreduce_col: all_cols.
	 * 
	 * Resulting matrix consists of columns of the matrix @a mat. i.e.
	 * each element in a col of the matrix is corresponding col of this matrix.
	 * Definition (in Haskell)
	 * @code
	 * allcols :: Matrix a -> Matrix (Matrix a)
	 * allcols = allreduce_col (Abv) . map Sing
	 * @endcode
	 *
	 * Sample: \n
	 * < Haskell code >
	 * @code
	 *   m2 = allcols m1
	 * @endcode
	 * < C++ code >
	 * @code
	 *   m2 = matrix_skeletons::allcols( m1 );
	 * @endcode
	 * @param mat a matrix.
	 * @return a matrix consisting of columns of the matrix @a mat.
	 * @see matrix_skeletons::allreduce_col(const F &, const dist_matrix<A> *)
	 */
	template<typename A>
	static dist_matrix< matrix< A > >* all_cols(const dist_matrix<A> *mat);

	/**
	 * @brief A specialization of reduce: allreduce_row.
	 *
	 * This reduces the matrix in row-direction only and
	 * broadcasts the result in each row.
	 * So almost all elements are shadows of first column.
	 * Please be careful with modifing the elements.
	 *
	 * Definition (in Haskell)
	 * @code
	 * allreduce_row :: (a -> a -> a) -> Matrix a -> Matrix a
	 * allreduce_row otimes = scanr (ll, rr) . scan (rr, otimes)
	 *   where ll a b = a
	 *         rr a b = b
	 * @endcode
	 *
	 * Sample: \n
	 * < Haskell code >
	 * @code
	 *   m2 = allreduce_row (+) m1
	 * @endcode
	 * < C++ code >
	 * @code
	 *   m2 = matrix_skeletons::allreduce_row( Add(), m1 );
	 * @endcode
	 * @param otimes a function taking two arguments x and y, returning result z,
	 * i.e. z = x otimes y.
	 * @a otimes should have the member void operator()(const A&, const A&, A*) const.
	 * The function object @a otimes should inherit skeleton::binary_function.
	 * @a otimes should be associative.
	 * @param mat a matrix to be reduced.
	 * @return The resulting matrix in which the elements in a row have the same value, the result of reducing the row with @a otimes.
	 * @see matrix_skeletons::reduce(const F &, const G &, const dist_matrix<A> *)
	 * @see matrix_skeletons::scan(const F &, const G &, const dist_matrix<A> *)
	 * @see matrix_skeletons::scanr(const F &, const G &, const dist_matrix<A> *)
	 * @see m_skeletons::allreduce_row(const F &, const matrix<A> *)

	 */
	template <typename A, typename F>
	static dist_matrix< A >* allreduce_row(const F &otimes, const dist_matrix<A> *mat);

	/**
	 * @brief A specialization of reduce: allreduce_col.
	 *
	 * This reduces the matrix in column-direction only and
	 * broadcasts the result in each column.
	 * So almost all elements are shadows of first row.
	 * Please be careful with modifing the elements.
	 *
	 * Definition (in Haskell)
	 * @code
	 * allreduce_col :: (a -> a -> a) -> Matrix a -> Matrix a
	 * allreduce_col oplus = scanr (rr, ll) . scan (oplus, rr)
	 *   where ll a b = a
	 *         rr a b = b
	 * @endcode
	 *
	 * Sample: \n
	 * < Haskell code >
	 * @code
	 *   m2 = allreduce_col (+) m1
	 * @endcode
	 * < C++ code >
	 * @code
	 *   m2 = matrix_skeletons::allreduce_col( Add(), m1 );
	 * @endcode
	 * @param oplus a function taking two arguments x and y, returning result z,
	 * i.e. z = x oplus y.
	 * @a oplus should have the member void operator()(const A&, const A&, A*) const.
	 * The function object @a oplus should inherit skeleton::binary_function.
	 * @a oplus should be associative.
	 * @param mat a matrix to be reduced.
	 * @return The resulting matrix in which the elements in a column have the same value, the result of reducing the column with @a oplus.
	 * @see matrix_skeletons::reduce(const F &, const G &, const dist_matrix<A> *)
	 * @see matrix_skeletons::scan(const F &, const G &, const dist_matrix<A> *)
	 * @see matrix_skeletons::scanr(const F &, const G &, const dist_matrix<A> *)
	 * @see m_skeletons::allreduce_col(const F &, const matrix<A> *)

	 */
	template <typename A, typename F>
	static dist_matrix< A >* allreduce_col(const F &oplus, const dist_matrix<A> *mat);


	/** accumulate skeleton prototypes **/
	template <typename A, typename OP2, typename OT2, typename P, typename OP, typename OT, typename Q>
	static void acc(const OP2 &oplus2, const OT2 &otimes2, const P &p, const OP &oplus, const OT &otimes, const Q&q, const dist_matrix<A>* dma, typename P::result_type *res);

	template <typename A, typename OP2, typename OT2, typename P, typename OP, typename OT, typename Q>
	static typename P::result_type * acc(const OP2 &oplus2, const OT2 &otimes2, const P &p, const OP &oplus, const OT &otimes, const Q&q, const dist_matrix<A>* dma);


	template <typename A, typename F, typename G>
	static void scan2(const F &oplus, const G &otimes, const dist_matrix<A>* dma, dist_matrix<A>* dmb);
	template <typename A, typename F, typename G>
	static dist_matrix<A>* scan2(const F &oplus, const G &otimes, const dist_matrix<A>* dma);



	////////////////////////////////////////////////////////////////////////
	// Data parallel Skeletons (shifting)
 public:
	/**
	 * @brief Rotate row-blocks by s.
	 * @param s shift.
	 * @param mat a dist_matrix to which the skeleton is applied.
	 */
	template<typename A>
	static void rotateRowsConst(int s, dist_matrix<A> *mat);
	/**
	 * @brief Rotate row-blocks by f(col-position).
	 * @param f a function object taking column position, returning shift-value
	 *     f should have int operator()(int a) const.
	 * A sample of function object is as follows: 
	 * (i-th row is shifted by i to the left)
	 * @code
	 * class ShiftLinear {
	 * public:
	 *     int operator()(int i) const {
	 *         return i;
	 *     }
	 * };
	 * @endcode
	 * @param mat a dist_matrix to which the skeleton is applied.
	 */
	template<typename A, typename F>
	static void rotateRows(const F &f, dist_matrix<A> *mat);
	/**
	 * @brief Rotate column-blocks by s.
	 * @param s shift.
	 * @param mat a dist_matrix to which the skeleton is applied.
	 */
	template<typename A>
	static void rotateColsConst(int s, dist_matrix<A> *mat);
	/**
	 * @brief Rotate column-blocks by f(row-position).
	 * @param f a function object taking row position, returning shift-value
	 *     f should have int operator()(int a) const.
	 * A sample of function object is as follows: 
	 * (i-th col is shifted by i to the bottom)
	 * @code
	 * class ShiftLinear {
	 * public:
	 *     int operator()(int i) const {
	 *         return i;
	 *     }
	 * };
	 * @endcode
	 * @param mat a dist_matrix to which the skeleton is applied.
	 */
	template<typename A, typename F>
	static void rotateCols(const F &f, dist_matrix<A> *mat);

	////////////////////////////////////////////////////////////////////////
	// Private Utilities
private:
	/**
	 * @brief Return row position of process to which this block-matrix is to be sent by row shifting.
	 * @param s shift.
	 * @apram y sender's row position.
	 * @param mat a dist_matrix to which the skeleton is applied.
	 * @return Receiver's row position.
	 */
	template<typename A>
	static int rotRow(int s, int y, dist_matrix<A> *mat);
	/**
	 * @brief Return column position of process to which x's block-matrix is to be sent by col shifting.
	 * @params s shift.
	 * @param x  sender's column position.
	 * @param mat a dist_matrix to which the skeleton is applied.
	 * @return Receiver's col position.
	 */
	template<typename A>
	static int rotCol(int s, int x, dist_matrix<A> *mat);
	/**
	 * @brief Send possessing block-matrix to receiver
	 *                and receive sent one from sender.
	 * @params sender rank of sender process.
	 * @params receiver rank of receiver process.
	 * @param mat a dist_matrix to which the skeleton is applied.
	 */
	template<typename A>
	static void permute(int sender, int receiver, dist_matrix<A> *mat);
	
	/**
	 * @brief A function object making a pair (fst, snd).
	 */
	template< typename B , typename C >
	class Mkpair: public skeleton::binary_function<B, C, std::pair< B, C > >{
		public:
		/** @brief Make a pair. */
		void operator()(const B& fst, const C& snd, std::pair< B, C >* pair) const {
			pair->first = fst;
			pair->second = snd;
		}
	};
};

//int matrix_skeletons::distributionPolicy = matrix_skeletons::DPOLICY_RECTANGLE;

#include "dist_matrix.h"

#include "matrix_skeletons.tpp"

#endif  // __MATRIX_SKELETONS_H__
