/***
 * Copyright (c) 2005, SkeTo Project
 * All rights reserved.
 */
/**
 * @file matrix_skeletons.tpp
 * @brief The implementation of class @a matrix_skeletons,
 *        consists of skeletons on matrices(two-dimensional arrays)
 *        that manipulating the distributed matrix.
 * @author Kento EMOTO
 */

#ifdef __MATRIX_SKELETONS_H__
#ifndef __MATRIX_SKELETONS_TPP__
#define __MATRIX_SKELETONS_TPP__


////////////////////////////////////////////////////////////////////////
// Didtribution Policies for dist_matrix
// void
// matrix_skeletons::calcDivision(int &r, int &c, const SKEL_MPI_INTRACOMM &com)
// {
// 	int np;
//  SKEL_MPI_Comm_size(com, np);
// 	switch(distributionPolicy){
// 		case DPOLICY_SQUARE:{
// 			int k;
// 			for(k = 1; k*k <= np; k++);
// 			k--;
// 			r = k;
// 			c = k;}
// 			break;
// 		case DPOLICY_ROW:
// 			r = np;
// 			c = 1;
// 			break;
// 		case DPOLICY_COLUMN:
// 			r = 1;
// 			c = np;
// 			break;
// 		case DPOLICY_RECTANGLE:{
// 			int k;
// 			for(k = 1; k*k <= np; k++);
// 			k--;
// 			int l;
// 			for(l = k; l > 0; l--){
// 				if((np % l) == 0){
// 					break;
// 				}
// 			}
// 			r = np / l;
// 			c = l;}
// 			break;
// 		default:
// 		break;
// 	}
// }

// int
// matrix_skeletons::setDistributionPolicy(int p)
// {
// 	int o = distributionPolicy;
// 	distributionPolicy = p;
// 	return o;
// }

// int
// matrix_skeletons::getDistributionPolicy()
// {
// 	return distributionPolicy;
// }

////////////////////////////////////////////////////////////////////////
// Data parallel Skeletons
template <typename A, typename F>
void
matrix_skeletons::map_i(const F &f, dist_matrix<A> *mat)
{
	if(mat->myrank >= mat->blocks) return;
	for(int i = 0; i < mat->localRows; i++){
		for(int j = 0; j < mat->localCols; j++){
			f(&mat->at(i, j));
		}
	}
}

template< typename B, typename A, typename F>
void
matrix_skeletons::map(const F &f, const dist_matrix<A> *mat, dist_matrix<B> *dmb)
{
	/***
	 * We should check
	 *  whether dma and dmb have the same communicators and shapes.
	 */
	if(mat->myrank >= mat->blocks) return;
	for(int i = 0; i < mat->localRows; i++){
		for(int j = 0; j < mat->localCols; j++){
			f(mat->at(i, j), &dmb->at(i, j));
		}
	}
}
	
template<typename A, typename F>
dist_matrix<typename F::result_type>*
matrix_skeletons::map(const F &f, const dist_matrix<A> *mat)
{
	dist_matrix<typename F::result_type>* dmb = new dist_matrix<typename F::result_type>();
	dmb->shapeCopy(mat);
	map(f, mat, dmb);
	return dmb;
}


template< typename C, typename A, typename B, typename F>
void
matrix_skeletons::zipwith(const F &f, const dist_matrix<A>* dma, const dist_matrix<B>* dmb, dist_matrix<C>* dmc )
{
	/***
	 * We should check
	 *  whether dma, dmb and dmc have the same communicators and shapes.
	 */
	if(dma->myrank >= dma->blocks) return;
	for(int i = 0; i < dma->localRows; i++){
		for(int j = 0; j < dma->localCols; j++){
			f(dma->at(i, j), dmb->at(i, j), &dmc->at(i, j));
		}
	}
}

template< typename A, typename B, typename F>
dist_matrix<typename F::result_type>*
matrix_skeletons::zipwith(const F &f, const dist_matrix<A>* dma, const dist_matrix<B>* dmb)
{
	dist_matrix<typename F::result_type>* dmc = new dist_matrix<typename F::result_type>();
	dmc->shapeCopy(dma);
	zipwith(f, dma, dmb, dmc);
	return dmc;
}
	
template< typename A, typename B >
dist_matrix < std::pair<A,B> > *
matrix_skeletons::zip(const dist_matrix<A>* dma, const dist_matrix<B>* dmb)
{
	return zipwith(Mkpair<A,B>(), dma, dmb);
}


template <typename A, typename F, typename G>
A*
matrix_skeletons::reduceB(const F &oplus, const G &otimes, const dist_matrix<A>*mat)
{
	A* res = new A;
	reduceB(oplus, otimes, mat, res);
	return res;		
}

template <typename A, typename F, typename G>
void
matrix_skeletons::reduceB(const F &oplus, const G &otimes, const dist_matrix<A>*mat, A *res)
{
	reduce(oplus, otimes, mat, res);
	dist_matrix<A>::bcast(res, mat);
}
template <typename A, typename F, typename G>
void
matrix_skeletons::reduce(const F &oplus, const G &otimes, const dist_matrix<A>* mat, A *res)
{
	if(mat->myrank >= mat->blocks) return;
	//if(mat->myrank==0) std::cout << "reduce 0\n";
	int bicfloor = (mat->myrank/mat->bic)*mat->bic;
	A *vals = new A[mat->localRows];
	for(int i = 0; i < mat->localRows; i++){
		A val = mat->at(i, 0);
		for(int j = 1; j < mat->localCols; j++){
			//if(mat->myrank==0) std::cout << "i = " << i << ", j = " << j << std::endl;
			A tmp = val;
			otimes(tmp, mat->at(i, j), &val);
		}
		vals[i] = val;
	}
	//if(mat->myrank==0) std::cout << "reduce 0.5\n";
	A val = vals[0];
	for(int i = 1; i < mat->localRows; i++){
		A tmp = val;
		oplus(tmp, vals[i], &val);
	}
	delete [] vals;
	//if(mat->myrank==0) std::cout << "reduce 1\n";
	// tree communication (row)
	for(int stage = 1; stage < mat->bic; stage <<= 1){
		int target = ((mat->myrank%mat->bic)^stage) + bicfloor;
		//cout << "rank = " << myrank << " target : " << target << endl;
		if( target < mat->myrank ) {
			SKEL_MPI_Request reqs;
			if(Serializer<A>::needsSerializing()){
				unsigned char* sbuf = mat->sendSerializingData(target, &val, 1, TAG_REDUCE3, TAG_REDUCE1, reqs);
				SKEL_MPI_Wait(reqs);	
				delete [] sbuf;
				//cout << "rank = " << myrank << " sent" << endl;
			} else {
				SKEL_MPI_Isend(&val, 1, mat->mpitype, target, TAG_REDUCE1, mat->comm, reqs);
				SKEL_MPI_Wait(reqs);	
			}
			break;
		} else if (((mat->myrank%mat->bic)^stage) < mat->bic ) {
			A rval;
			if(Serializer<A>::needsSerializing()){
				//cout << "rank = " << myrank << " ready to receive" << endl;
				mat->recvSerializingData(target, &rval, 1, TAG_REDUCE3, TAG_REDUCE1);
				//cout << "rank = " << myrank << " received" << endl;
			} else {
				SKEL_MPI_Request reqr;
				SKEL_MPI_Irecv(&rval, 1, mat->mpitype, target, TAG_REDUCE1, mat->comm, reqr);
				SKEL_MPI_Wait(reqr);
			}
			A tmp = val;
			otimes(tmp, rval, &val);
		}
	}
	//if(myrank==0) cout << "reduce 2\n";
	if(mat->myrank%mat->bic==0){
		for(int stage = 1; stage < mat->bir; stage <<= 1){
			int target = ((mat->myrank/mat->bic)^stage)*mat->bic;
			//cout << "rank = " << myrank << " target : " << target << endl;
			if( target < mat->myrank ) {	
				SKEL_MPI_Request reqs;
				if(Serializer<A>::needsSerializing()){
					unsigned char* sbuf = mat->sendSerializingData(target, &val, 1, TAG_REDUCE4, TAG_REDUCE2, reqs);
					SKEL_MPI_Wait(reqs);	
					delete [] sbuf;
					//cout << "rank = " << myrank << " sent" << endl;
				} else {
					SKEL_MPI_Isend(&val, 1, mat->mpitype, target, TAG_REDUCE2, mat->comm, reqs);
					SKEL_MPI_Wait(reqs);	
				}
				break;
				//} else if ( target < blocks ) {
			} else if (((mat->myrank/mat->bic)^stage) < mat->bir ) {
				A rval;
				if(Serializer<A>::needsSerializing()){
					//cout << "rank = " << myrank << " ready to receive" << endl;
					mat->recvSerializingData(target, &rval, 1, TAG_REDUCE4, TAG_REDUCE2);
					//cout << "rank = " << myrank << " received" << endl;
				} else {
					SKEL_MPI_Request reqr;
					SKEL_MPI_Irecv(&rval, 1, mat->mpitype, target, TAG_REDUCE2, mat->comm, reqr);
					SKEL_MPI_Wait(reqr);
				}
				A tmp = val;
				oplus(tmp, rval, &val);
			}
		}
	}
	*res = val; 
}


template <typename A, typename F, typename G>
A*
matrix_skeletons::reduce(const F &oplus, const G &otimes, const dist_matrix<A>* mat)
{
	A* res = new A;
	reduce(oplus, otimes, mat, res);
	return res;		
}

template <typename A, typename F, typename G>
void
matrix_skeletons::reduceDC(const F &oplus, const G &otimes, const dist_matrix<A>* mat, A *res)
{
	if(mat->myrank >= mat->blocks) return;
	dist_matrix<A> *tmp = new dist_matrix<A>(*mat);
	{
		const int n = tmp->localRows;
		const int m = tmp->localCols;
		const int nm = n > m ? n : m;
		for(int stage = 1; stage < nm; stage <<=1){
			const int stage2 = stage << 1;
			for(int i = 0; i < n; i+=stage2){
				for(int j = 0; j < m; j+=stage2){
					const int i2 = i ^ stage;
					const int j2 = j ^ stage;
					if(i2 < n){
						A t = tmp->at(i, j);
						oplus(t, tmp->at(i2, j), &tmp->at(i,j));
						if(j2 < m){
							A u = tmp->at(i, j2);
							oplus(u, tmp->at(i2, j2), &tmp->at(i,j2));
						}
					}
					if(j2 < m){
						A t = tmp->at(i, j);
						otimes(t, tmp->at(i, j2), &tmp->at(i,j));
					}
				}
			}
		}
	}
	A val = tmp->at(0,0);
	delete tmp;
	int bicfloor = (mat->myrank/mat->bic)*mat->bic;
	int maxstage = mat->bir;
	if(maxstage < mat->bic) maxstage = mat->bic;
	for(int stage = 1; stage < maxstage; stage <<= 1){
		if(stage < mat->bic){
			int target = ((mat->myrank%mat->bic)^stage) + bicfloor;
			if( target < mat->myrank ) {
				SKEL_MPI_Request reqs;
				if(Serializer<A>::needsSerializing()){
					unsigned char* sbuf = mat->sendSerializingData(target, &val, 1, TAG_REDUCE3, TAG_REDUCE1, reqs);
					SKEL_MPI_Wait(reqs);	
					delete [] sbuf;
				} else {
					SKEL_MPI_Isend(&val, 1, mat->mpitype, target, TAG_REDUCE1, mat->comm, reqs);
					SKEL_MPI_Wait(reqs);	
				}
				break;
			} else if (((mat->myrank%mat->bic)^stage) < mat->bic ) {
				A rval;
				if(Serializer<A>::needsSerializing()){
					mat->recvSerializingData(target, &rval, 1, TAG_REDUCE3, TAG_REDUCE1);
				} else {
					SKEL_MPI_Request reqr;
					SKEL_MPI_Irecv(&rval, 1, mat->mpitype, target, TAG_REDUCE1, mat->comm, reqr);
					SKEL_MPI_Wait(reqr);
				}
				A tmp = val;
				otimes(tmp, rval, &val);
			}
		}
		if(stage < mat->bir){
			int target = ((mat->myrank/mat->bic)^stage)*mat->bic+(mat->myrank%mat->bic);
			if( target < mat->myrank ) {	
				SKEL_MPI_Request reqs;
				if(Serializer<A>::needsSerializing()){
					unsigned char* sbuf = mat->sendSerializingData(target, &val, 1, TAG_REDUCE4, TAG_REDUCE2, reqs);
					SKEL_MPI_Wait(reqs);	
					delete [] sbuf;
				} else {
					SKEL_MPI_Isend(&val, 1, mat->mpitype, target, TAG_REDUCE2, mat->comm, reqs);
					SKEL_MPI_Wait(reqs);	
				}
				break;
			} else if (((mat->myrank/mat->bic)^stage) < mat->bir ) {
				A rval;
				if(Serializer<A>::needsSerializing()){
					mat->recvSerializingData(target, &rval, 1, TAG_REDUCE4, TAG_REDUCE2);
				} else {
					SKEL_MPI_Request reqr;
					SKEL_MPI_Irecv(&rval, 1, mat->mpitype, target, TAG_REDUCE2, mat->comm, reqr);
					SKEL_MPI_Wait(reqr);
				}
				A tmp = val;
				oplus(tmp, rval, &val);
			}
		}
	}
	*res = val; 
}


template <typename A, typename F, typename G>
A*
matrix_skeletons::reduceDC(const F &oplus, const G &otimes, const dist_matrix<A>* mat)
{
	A* res = new A;
	reduceDC(oplus, otimes, mat, res);
	return res;		
}


template <typename A, typename F, typename G>
void
matrix_skeletons::scan(const F &oplus, const G &otimes, const dist_matrix<A>* dma, dist_matrix<A>* dmb)
{
	/***
	 * We should check
	 *  whether dma and dmb have the same communicators and shapes.
	 */
	/***
	 * this function consists of two sets of three steps:
	 * 1. row-wise
	 *  1.1. local reduce
	 *  1.2. communication
	 *  1.3. local scan
	 * 2. col-wise
	 *  2.1. local reduce
	 *  2.2. communication
	 *  2.3. local scan
	 */
	if(dma->myrank >= dma->blocks) return;

	int bicfloor = (dma->myrank/dma->bic)*dma->bic;
	A *t_vals;
	A *f_vals;
	A *r_vals;
	
	/* 		for(int i = 0; i < localRows; i++){ */
	/* 			for (int j = 0; j < localCols; j++){ */
	/* 				dmb->at(i, j) = at(i, j); */
	/* 			} */
	/* 		} */
	/* 		goto LLABEL; */
	/***
	 * 1.1. local reduce(row)
	 */
	//if(myrank==0) cout << "scan 1.1\n";
	t_vals = new A[dma->localRows];
	for(int i = 0; i < dma->localRows; i++){
		A t_val = dma->at(i, 0);
		for(int j = 1; j < dma->localCols; j++){
			A tmp;
			otimes(t_val, dma->at(i, j), &tmp);
			t_val = tmp;
		}
		t_vals[i] = t_val;
	}
	/***
	 * 1.2. communication (row) (h-cube)
	 */
	//if(myrank==0) cout << "scan 1.2\n";
	f_vals = new A[dma->localRows];
	for(int i = 0; i < dma->localRows; i++){
		f_vals[i] = dma->at(i, 0);
	}
	r_vals = new A[dma->localRows];
	for(int stage = 1; stage < dma->bic; stage <<= 1){
		int target = ((dma->myrank%dma->bic)^stage) + bicfloor;
		if (((dma->myrank%dma->bic)^stage) >= dma->bic) continue;
		//if (target >= blocks ) continue;
		SKEL_MPI_Request reqs, reqr;
		
		if(Serializer<A>::needsSerializing()){
			unsigned char* sbuf = dma->sendSerializingData(target, t_vals, dma->localRows, TAG_SCAN12s, TAG_SCAN12, reqs);
			dma->recvSerializingData(target, r_vals, dma->localRows, TAG_SCAN12s, TAG_SCAN12);
			SKEL_MPI_Wait(reqs);	
			delete [] sbuf;
		} else {
			SKEL_MPI_Isend(t_vals, dma->localRows, dma->mpitype, target, TAG_SCAN12, dma->comm, reqs);
			SKEL_MPI_Irecv(r_vals, dma->localRows, dma->mpitype, target, TAG_SCAN12, dma->comm, reqr);
			SKEL_MPI_Wait(reqs);
			SKEL_MPI_Wait(reqr);
		}
		if( target > dma->myrank ) {
			for(int i = 0; i < dma->localRows; i++){
				A tmp;
				otimes(t_vals[i], r_vals[i], &tmp);
				t_vals[i] = tmp;
			}
		} else {
			for(int i = 0; i < dma->localRows; i++){
				A tmp;
				otimes(r_vals[i], f_vals[i], &tmp);
				f_vals[i] = tmp;
			}
			for(int i = 0; i < dma->localRows; i++){
				A tmp;
				otimes(r_vals[i], t_vals[i], &tmp);
				t_vals[i] = tmp;
			}
		}
	}
	/***
	 * 1.3. local scan(row)
	 */
	//if(myrank==0) cout << "scan 1.3\n";
	for(int i = 0; i < dma->localRows; i++){
		dmb->at(i, 0) = f_vals[i];
		for (int j = 1; j < dma->localCols; j++){
			otimes(dmb->at(i, j-1), dma->at(i, j), &dmb->at(i, j));
		}
	}
	delete [] r_vals;
	delete [] f_vals;
	delete [] t_vals;
	/* 		return; */
	/* 	LLABEL: */
	  
	/***
	 * 2.1. local reduce(col)
	 */
	//if(myrank==0) cout << "scan 2.1\n";
	t_vals = new A[dma->localCols];
	for(int j = 0; j < dma->localCols; j++){
		A t_val = dmb->at(0, j);
		for(int i = 1; i < dma->localRows; i++){
			A tmp;
			oplus(t_val, dmb->at(i, j), &tmp);
			t_val = tmp;
		}
		t_vals[j] = t_val;
	}
	/***
	 * 2.2. communication (row) (h-cube)
	 */
	//	if(myrank==0) cout << "scan 2.2\n";
	f_vals = new A[dma->localCols];
	for(int j = 0; j < dma->localCols; j++){
		f_vals[j] = dmb->at(0, j);
	}
	r_vals = new A[dma->localCols];
	for(int stage = 1; stage < dma->bir; stage <<= 1){
		int target = ((dma->myrank/dma->bic)^stage)*dma->bic + dma->myrank%dma->bic;
		if (((dma->myrank/dma->bic)^stage) >= dma->bir) continue;
		//	if (target >= blocks) continue;
		SKEL_MPI_Request reqs, reqr;
		if(Serializer<A>::needsSerializing()){
			unsigned char* sbuf = dma->sendSerializingData(target, t_vals, dma->localCols, TAG_SCAN22s, TAG_SCAN22, reqs);
			dma->recvSerializingData(target, r_vals, dma->localCols, TAG_SCAN22s, TAG_SCAN22);
			SKEL_MPI_Wait(reqs);	
			delete [] sbuf;
		} else {
			SKEL_MPI_Isend(t_vals, dma->localCols, dma->mpitype, target, TAG_SCAN22, dma->comm, reqs);
			SKEL_MPI_Irecv(r_vals, dma->localCols, dma->mpitype, target, TAG_SCAN22, dma->comm, reqr);
			SKEL_MPI_Wait(reqs);
			SKEL_MPI_Wait(reqr);
		}

		if( target > dma->myrank ) {
			for(int j = 0; j < dma->localCols; j++){
				A tmp;
				oplus(t_vals[j], r_vals[j], &tmp);
				t_vals[j] = tmp;
			}
		} else {
			for(int j = 0; j < dma->localCols; j++){
				A tmp;
				oplus(r_vals[j], f_vals[j], &tmp);
				f_vals[j] = tmp;
			}
			for(int j = 0; j < dma->localCols; j++){
				A tmp;
				oplus(r_vals[j], t_vals[j], &tmp);
				t_vals[j] = tmp;
			}
		}
	}
	/***
	 * 2.3. local scan(row)
	 */
	//if(myrank==0) cout << "scan 2.3\n";
	for (int j = 0; j < dma->localCols; j++){
		dmb->at(0, j) = f_vals[j];
		for(int i = 1; i < dma->localRows; i++){
			A tmp;
			oplus(dmb->at(i-1, j), dmb->at(i, j), &tmp);
			dmb->at(i, j) = tmp;
		}
	}
	delete [] r_vals;
	delete [] f_vals;
	delete [] t_vals;
		
		
}

template <typename A, typename F, typename G>
dist_matrix<A>*
matrix_skeletons::scan(const F &oplus, const G &otimes, const dist_matrix<A>* dma)
{
	dist_matrix<A>* dmb = new dist_matrix<A>();
	dmb->shapeCopy(dma);
	scan(oplus, otimes, dma, dmb);
	return dmb;
}


template <typename A, typename F, typename G>
void
matrix_skeletons::scanr(const F &oplus, const G &otimes, const dist_matrix<A>* dma, dist_matrix<A>* dmb)
{
	/***
	 * We should check
	 *  whether dma and dmb have the same communicators and shapes.
	 */
	/***
	 * this function consists of two sets of three steps:
	 * 1. row-wise
	 *  1.1. local reduce
	 *  1.2. communication
	 *  1.3. local scan
	 * 2. col-wise
	 *  2.1. local reduce
	 *  2.2. communication
	 *  2.3. local scan
	 */
	if(dma->myrank >= dma->blocks) return;

	int bicfloor = (dma->myrank/dma->bic)*dma->bic;
	A *t_vals;
	A *f_vals;
	A *r_vals;
		
	/***
	 * 1.1. local reduce(row)
	 */
	//if(myrank==0) cout << "scan 1.1\n";
	t_vals = new A[dma->localRows];
	for(int i = 0; i < dma->localRows; i++){
		A t_val = dma->at(i, dma->localCols-1);
		for(int j = dma->localCols-2; j >=0; j--){
			A tmp;
			otimes(dma->at(i, j), t_val, &tmp);
			t_val = tmp;
		}
		t_vals[i] = t_val;
	}
	/***
	 * 1.2. communication (row) (h-cube)
	 */
	//if(myrank==0) cout << "scan 1.2\n";
	f_vals = new A[dma->localRows];
	for(int i = 0; i < dma->localRows; i++){
		f_vals[i] = dma->at(i, dma->localCols-1);
	}
	r_vals = new A[dma->localRows];
	int vmyrank = dma->bic - (dma->myrank%dma->bic) - 1;
	for(int stage = 1; stage < dma->bic; stage <<= 1){
		int vtarget = (vmyrank^stage);
		int target = dma->bic - vtarget -1;
		if(target < 0) continue;
		target += bicfloor;
		SKEL_MPI_Request reqs, reqr;
			
		if(Serializer<A>::needsSerializing()){
			unsigned char* sbuf = dma->sendSerializingData(target, t_vals, dma->localRows, TAG_SCAN12s, TAG_SCAN12, reqs);
			dma->recvSerializingData(target, r_vals, dma->localRows, TAG_SCAN12s, TAG_SCAN12);
			SKEL_MPI_Wait(reqs);	
			delete [] sbuf;
		} else {
			SKEL_MPI_Isend(t_vals, dma->localRows, dma->mpitype, target, TAG_SCAN12, dma->comm, reqs);
			SKEL_MPI_Irecv(r_vals, dma->localRows, dma->mpitype, target, TAG_SCAN12, dma->comm, reqr);
			SKEL_MPI_Wait(reqs);
			SKEL_MPI_Wait(reqr);
		}
		if( vtarget > vmyrank ) {
			for(int i = 0; i < dma->localRows; i++){
				A tmp;
				otimes(r_vals[i], t_vals[i], &tmp);
				t_vals[i] = tmp;
			}
		} else {
			for(int i = 0; i < dma->localRows; i++){
				A tmp;
				otimes(f_vals[i], r_vals[i], &tmp);
				f_vals[i] = tmp;
			}
			for(int i = 0; i < dma->localRows; i++){
				A tmp;
				otimes(t_vals[i], r_vals[i], &tmp);
				t_vals[i] = tmp;
			}
		}
	}
	/***
	 * 1.3. local scan(row)
	 */
	//if(myrank==0) cout << "scan 1.3\n";
	for(int i = 0; i < dma->localRows; i++){
		dmb->at(i, dma->localCols-1) = f_vals[i];
		for (int j = dma->localCols-2; j >= 0; j--){
			otimes(dma->at(i, j), dmb->at(i, j+1), &dmb->at(i, j));
		}
	}
	delete [] r_vals;
	delete [] f_vals;
	delete [] t_vals;
	/* 		return; */
	/* 	LLABEL: */
	  
	/***
	 * 2.1. local reduce(col)
	 */
	//if(myrank==0) cout << "scan 2.1\n";
	t_vals = new A[dma->localCols];
	for(int j = 0; j < dma->localCols; j++){
		A t_val = dmb->at(dma->localRows-1, j);
		for(int i = dma->localRows-2; i >= 0; i--){
			A tmp;
			oplus(dmb->at(i, j), t_val, &tmp);
			t_val = tmp;
		}
		t_vals[j] = t_val;
	}
	/***
	 * 2.2. communication (row) (h-cube)
	 */
	//	if(myrank==0) cout << "scan 2.2\n";
	f_vals = new A[dma->localCols];
	for(int j = 0; j < dma->localCols; j++){
		f_vals[j] = dmb->at(dma->localRows-1, j);
	}
	vmyrank = dma->bir - dma->myrank/dma->bic - 1;
	r_vals = new A[dma->localCols];
	for(int stage = 1; stage < dma->bir; stage <<= 1){
		int vtarget = (vmyrank^stage);
		int target = dma->bir - vtarget -1;
		if(target < 0) continue;
		target = target * dma->bic + dma->myrank%dma->bic;
			
		//	if (target >= blocks) continue;
		SKEL_MPI_Request reqs, reqr;
		if(Serializer<A>::needsSerializing()){
			unsigned char* sbuf = dma->sendSerializingData(target, t_vals, dma->localCols, TAG_SCAN22s, TAG_SCAN22, reqs);
			dma->recvSerializingData(target, r_vals, dma->localCols, TAG_SCAN22s, TAG_SCAN22);
			SKEL_MPI_Wait(reqs);	
			delete [] sbuf;
		} else {
			SKEL_MPI_Isend(t_vals, dma->localCols, dma->mpitype, target, TAG_SCAN22, dma->comm, reqs);
			SKEL_MPI_Irecv(r_vals, dma->localCols, dma->mpitype, target, TAG_SCAN22, dma->comm, reqr);
			SKEL_MPI_Wait(reqs);
			SKEL_MPI_Wait(reqr);
		}

		if( vtarget > vmyrank ) {
			for(int j = 0; j < dma->localCols; j++){
				A tmp;
				oplus(r_vals[j], t_vals[j], &tmp);
				t_vals[j] = tmp;
			}
		} else {
			for(int j = 0; j < dma->localCols; j++){
				A tmp;
				oplus(f_vals[j], r_vals[j], &tmp);
				f_vals[j] = tmp;
			}
			for(int j = 0; j < dma->localCols; j++){
				A tmp;
				oplus(t_vals[j], r_vals[j], &tmp);
				t_vals[j] = tmp;
			}
		}
	}
	/***
	 * 2.3. local scan(row)
	 */
	//if(myrank==0) cout << "scan 2.3\n";
	for (int j = 0; j < dma->localCols; j++){
		dmb->at(dma->localRows-1, j) = f_vals[j];
		for(int i = dma->localRows-2; i>=0; i--){
			A tmp;
			oplus(dmb->at(i, j), dmb->at(i+1, j), &tmp);
			dmb->at(i, j) = tmp;
		}
	}
	delete [] r_vals;
	delete [] f_vals;
	delete [] t_vals;
		
		
	
}
template <typename A, typename F, typename G>
dist_matrix<A>*
matrix_skeletons::scanr(const F &oplus, const G &otimes, const dist_matrix<A>* dma)
{
	dist_matrix<A>* dmb = new dist_matrix<A>();
	dmb->shapeCopy(dma);
	scanr(oplus, otimes, dma, dmb);
	return dmb;
}




////////////////////////////////////////////////////////////////////////
// Data parallel Skeletons (shifting)

template<typename A>
void
matrix_skeletons::rotateRowsConst(int s, dist_matrix<A> *mat)
{
	if(mat->myrank >= mat->blocks) return;
	int receiver = rotRow(s, mat->posY, mat) * mat->bic + mat->posX;
	int sender = rotRow(-s, mat->posY, mat) * mat->bic + mat->posX;
	permute(sender, receiver, mat);
}

template<typename A, typename F>
void
matrix_skeletons::rotateRows(const F &f, dist_matrix<A> *mat)
{
	rotateRowsConst(f(mat->posX), mat);
}

template<typename A>
void
matrix_skeletons::rotateColsConst(int s, dist_matrix<A> *mat)
{
	if(mat->myrank >= mat->blocks) return;
	int receiver = mat->posY * mat->bic + rotCol(s, mat->posX, mat);
	int sender = mat->posY * mat->bic + rotCol(-s, mat->posX, mat);
	permute(sender, receiver, mat);
}

template<typename A, typename F>
void
matrix_skeletons::rotateCols(const F &f, dist_matrix<A> *mat)
{
	rotateColsConst(f(mat->posY), mat);
}



////////////////////////////////////////////////////////////////////////
// Private Utilities
template<typename A>
int
matrix_skeletons::rotRow(int s, int y, dist_matrix<A> *mat)
{
	int newY = y+s;
	while(newY >= mat->bir) newY-=mat->bir;
	while(newY < 0) newY+=mat->bir;
	return newY;
}

template<typename A>
int
matrix_skeletons::rotCol(int s, int x, dist_matrix<A> *mat)
{
	int newX = x+s;
	while(newX >= mat->bic) newX-=mat->bic;
	while(newX < 0) newX+=mat->bic;
	return newX;
}

template<typename A>
void
matrix_skeletons::permute(int sender, int receiver, dist_matrix<A> *mat)
{
	if(receiver==mat->myrank && mat->myrank==sender) return;
	matrixStructure ms, msr;
	ms.idxRow = mat->idxRow;
	ms.idxCol = mat->idxCol;
	ms.localRows = mat->localRows;
	ms.localCols = mat->localCols;
	SKEL_MPI_Send(&ms, 4, MPI_INT, receiver, TAG_PERMUTE1, mat->comm);
	SKEL_MPI_Recv(&msr, 4, MPI_INT, sender, TAG_PERMUTE1, mat->comm);
	SKEL_MPI_Request reqs, reqr;
	if(Serializer<A>::needsSerializing()){
		SKEL_MPI_Request reqs;	
		unsigned char *sbuf = mat->sendSerializingData(receiver, TAG_PERMUTE3, TAG_PERMUTE2, reqs);
		mat->idxRow = msr.idxRow;
		mat->idxCol = msr.idxCol;
		mat->localRows = msr.localRows;
		mat->localCols = msr.localCols;
		mat->recvSerializingData(sender, TAG_PERMUTE3, TAG_PERMUTE2);
		SKEL_MPI_Wait(reqs);
		delete [] sbuf;
	} else {
		int size = mat->rows*mat->cols;
		A *buf = new A[size];
		SKEL_MPI_Isend(mat->ele, size, mat->mpitype, receiver, TAG_PERMUTE2, mat->comm, reqs);
		mat->idxRow = msr.idxRow;
		mat->idxCol = msr.idxCol;
		mat->localRows = msr.localRows;
		mat->localCols = msr.localCols;
		SKEL_MPI_Irecv(buf, size, mat->mpitype, sender, TAG_PERMUTE2, mat->comm, reqr);
		SKEL_MPI_Wait(reqs);
		SKEL_MPI_Wait(reqr);
		A *p = buf;
		A *q = mat->ele;
		for(int i = 0; i < size; i++){
			*q++ = *p++;
		}
		delete [] buf;
	}
}

template<typename A>
dist_matrix< matrix< A > >*
matrix_skeletons::all_rows(const dist_matrix<A> *mat){
	dist_matrix< matrix<A> >* dmb = new dist_matrix< matrix<A> >();
	dmb->shapeCopy(mat);
	if(mat->myrank >= mat->blocks) return dmb;
	dist_matrix<A>* dmc = new dist_matrix<A>();
	dmc->shapeCopy(mat);
	for(int i = 0; i < mat->localRows; i++){
		dmb->at(i, 0).setSize(1, mat->getCols());
		for(int j = 0; j < mat->localCols; j++){
			dmc->at(i, j) = mat->at(i, j);
		}
	}
	for(int l = 0; l < mat->bic; l++){
		for(int i = 0; i < dmb->getLocalRows(); i++){
			int idxc = dmc->getIdxCol();
			for(int k = 0; k < dmc->getLocalCols(); k++){
				dmb->at(i,0).at(0, idxc++) = dmc->at(i, k);
			}
		}
		matrix_skeletons::rotateColsConst(1, dmc);
	}
	for(int i = 0; i < mat->localRows; i++){
		for (int j = 1; j < mat->localCols; j++){
			Shadower<matrix<A > >::shadow(dmb->at(i, 0),&dmb->at(i, j));
		}
	}
	delete dmc;
	return dmb;
}
template<typename A>
dist_matrix< matrix< A > >*
matrix_skeletons::all_cols(const dist_matrix<A> *mat){
	dist_matrix< matrix<A> >* dmb = new dist_matrix< matrix<A> >();
	dmb->shapeCopy(mat);
	if(mat->myrank >= mat->blocks) return dmb;
	dist_matrix<A>* dmc = new dist_matrix<A>();
	dmc->shapeCopy(mat);
	for(int i = 0; i < mat->localRows; i++){
		for(int j = 0; j < mat->localCols; j++){
			dmc->at(i, j) = mat->at(i, j);
		}
	}
	for(int j = 0; j < mat->localCols; j++){
		dmb->at(0, j).setSize(mat->getRows(), 1);
	}
	for(int l = 0; l < mat->bir; l++){
		for(int j = 0; j < dmb->getLocalCols(); j++){
			int idxc = dmc->getIdxRow();
			for(int k = 0; k < dmc->getLocalRows(); k++){
				dmb->at(0, j).at(idxc++, 0) = dmc->at(k, j);
			}
		}
		matrix_skeletons::rotateRowsConst(1, dmc);
	}
	for(int i = 1; i < mat->localRows; i++){
		for (int j = 0; j < mat->localCols; j++){
			Shadower<matrix<A> >::shadow(dmb->at(0, j),&dmb->at(i, j));
		}
	}
	delete dmc;
	return dmb;
}

template <typename A, typename F>
dist_matrix< A >*
matrix_skeletons::allreduce_row(const F &otimes, const dist_matrix<A> *mat){
	dist_matrix< A >* dmb = new dist_matrix< A >();
	dmb->shapeCopy(mat);
	if(mat->myrank >= mat->blocks) return dmb;

	const int localCols = mat->localCols;
	const int localRows = mat->localRows;
	//	const int bir = mat->bir;
	const int bic = mat->bic;
	const int myrank = mat->myrank;

	int bicfloor = (myrank/bic)*bic;
	A *t_vals;
	A *r_vals;
	
	t_vals = new A[localRows];
	for(int i = 0; i < localRows; i++){
		A t_val = mat->at(i, 0);
		for(int j = 1; j < localCols; j++){
			A tmp;
			otimes(t_val, mat->at(i, j), &tmp);
			t_val = tmp;
		}
		t_vals[i] = t_val;
	}
	r_vals = new A[localRows];
	/**
	 * butterfly (up)
	 */
	int stage;
	int cnt = 0;
	for(stage = 1; stage < bic; stage <<= 1){
		if(bic&stage) cnt++;
		int target = ((myrank%bic)^stage) + bicfloor;
		if (((myrank%bic)^stage) >= bic) continue;
		SKEL_MPI_Request reqs, reqr;
		
		if(Serializer<A>::needsSerializing()){
			unsigned char* sbuf = mat->sendSerializingData(target, t_vals, localRows, TAG_ALLREDUCEROW2, TAG_ALLREDUCEROW1, reqs);
			mat->recvSerializingData(target, r_vals, localRows, TAG_ALLREDUCEROW2, TAG_ALLREDUCEROW1);
			SKEL_MPI_Wait(reqs);	
			delete [] sbuf;
		} else {
			SKEL_MPI_Isend(t_vals, localRows, mat->mpitype, target, TAG_ALLREDUCEROW1, mat->comm, reqs);
			SKEL_MPI_Irecv(r_vals, localRows, mat->mpitype, target, TAG_ALLREDUCEROW1, mat->comm, reqr);
			SKEL_MPI_Wait(reqs);
			SKEL_MPI_Wait(reqr);
		}
		if( target > myrank ) {
			for(int i = 0; i < localRows; i++){
				A tmp;
				otimes(t_vals[i], r_vals[i], &tmp);
				t_vals[i] = tmp;
			}
		} else {
			for(int i = 0; i < localRows; i++){
				A tmp;
				otimes(r_vals[i], t_vals[i], &tmp);
				t_vals[i] = tmp;
			}
		}
	}
	delete [] r_vals;
	if(cnt > 1){
		/**
		 * tree (down), omitted when bic = 2^k
		 */
		for(; cnt > 0; cnt--) stage>>=1;
		int base = 1;
		for(;base&bic==0;base<<=1);
		unsigned char* sbuf=NULL;
		int bufSize = 0;
		for(; stage >= base; stage >>= 1){
			if((myrank%bic)&((stage-1)&(~(base-1)))!=0) continue;
			int target = ((myrank%bic)^stage) + bicfloor;
			if (((myrank%bic)^stage) >= bic) continue;
			SKEL_MPI_Request reqs, reqr;
			if( target > myrank ) {
				if(Serializer<A>::needsSerializing()){
					if(sbuf==NULL){
						bufSize = 0;
						for(int i = 0; i < localRows; i++){
							bufSize+=Serializer<A>::getSerializedSize(t_vals[i]);
						}
					}
					SKEL_MPI_Send(&bufSize, 1, MPI_INT, target, TAG_ALLREDUCEROW4, mat->comm);
					if(sbuf==NULL){
						unsigned char *buf = new unsigned char[bufSize];
						sbuf = buf;
						for(int i = 0; i < localRows; i++){
							buf=Serializer<A>::serialize(t_vals[i], buf);
						}
					}
					SKEL_MPI_Isend(sbuf, bufSize, BYTE_TYPE, target, TAG_ALLREDUCEROW3, mat->comm, reqs);
					SKEL_MPI_Wait(reqs);	
				} else {
					SKEL_MPI_Isend(t_vals, localRows, mat->mpitype, target, TAG_ALLREDUCEROW3, mat->comm, reqs);
					SKEL_MPI_Wait(reqs);
				}
			} else {
				if(Serializer<A>::needsSerializing()){
					SKEL_MPI_Recv(&bufSize, 1, MPI_INT, target, TAG_ALLREDUCEROW4, mat->comm);
					unsigned char *buf = new unsigned char[bufSize];
					sbuf = buf;
					SKEL_MPI_Irecv(buf, bufSize, BYTE_TYPE, target, TAG_ALLREDUCEROW3, mat->comm, reqr);
					SKEL_MPI_Wait(reqr);
					for(int i = 0; i < localRows; i++){
						buf = Serializer<A>::construct(&t_vals[i], buf);
					}
				} else {
					SKEL_MPI_Irecv(t_vals, localRows, mat->mpitype, target, TAG_ALLREDUCEROW3, mat->comm, reqr);
					SKEL_MPI_Wait(reqr);
				}		
			}
		}
		if(sbuf!=NULL) delete [] sbuf;
	}
	/**
	 * shadowing
	 */
	for(int i = 0; i < localRows; i++){
		dmb->at(i, 0) = t_vals[i];
		for (int j = 1; j < localCols; j++){
			Shadower<A>::shadow(dmb->at(i, 0),&dmb->at(i, j));
		}
	}
	delete [] t_vals;
	return dmb;
}

template <typename A, typename F>
dist_matrix< A >*
matrix_skeletons::allreduce_col(const F &oplus, const dist_matrix<A> *mat){
	dist_matrix< A >* dmb = new dist_matrix< A >();
	dmb->shapeCopy(mat);
	if(mat->myrank >= mat->blocks) return dmb;
	
	A* t_vals;
	A* r_vals;

	const int localCols = mat->localCols;
	const int localRows = mat->localRows;
	const int bir = mat->bir;
	const int bic = mat->bic;
	const int myrank = mat->myrank;
	
	t_vals = new A[localCols];
	for(int j = 0; j < localCols; j++){
		A t_val = mat->at(0, j);
		for(int i = 1; i < localRows; i++){
			A tmp;
			oplus(t_val, mat->at(i, j), &tmp);
			t_val = tmp;
		}
		t_vals[j] = t_val;
	}
	r_vals = new A[localCols];
	/**
	 * butterfly (up)
	 */
	int stage;
	int cnt = 0;
	for(stage = 1; stage < bir; stage <<= 1){
		if(bir&stage) cnt++;
		int target = ((myrank/bic)^stage)*bic + myrank%bic;
		if (((myrank/bic)^stage) >= bir) continue;
		SKEL_MPI_Request reqs, reqr;
		if(Serializer<A>::needsSerializing()){
			unsigned char* sbuf = mat->sendSerializingData(target, t_vals, localCols, TAG_ALLREDUCECOL2, TAG_ALLREDUCECOL1, reqs);
			mat->recvSerializingData(target, r_vals, localCols, TAG_ALLREDUCECOL2, TAG_ALLREDUCECOL1);
			SKEL_MPI_Wait(reqs);	
			delete [] sbuf;
		} else {
			SKEL_MPI_Isend(t_vals, localCols, mat->mpitype, target, TAG_ALLREDUCECOL1, mat->comm, reqs);
			SKEL_MPI_Irecv(r_vals, localCols, mat->mpitype, target, TAG_ALLREDUCECOL1, mat->comm, reqr);
			SKEL_MPI_Wait(reqs);
			SKEL_MPI_Wait(reqr);
		}
		if( target > myrank ) {
			for(int j = 0; j < localCols; j++){
				A tmp;
				oplus(t_vals[j], r_vals[j], &tmp);
				t_vals[j] = tmp;
			}
		} else {
			for(int j = 0; j < localCols; j++){
				A tmp;
				oplus(r_vals[j], t_vals[j], &tmp);
				t_vals[j] = tmp;
			}
		}
	}
	delete [] r_vals;
	if(cnt > 1){
		/**
		 * tree (down), omitted when bir = 2^k
		 */
		for(; cnt > 0; cnt--) stage>>=1;
		int base = 1;
		for(;base&bir==0;base<<=1);
		unsigned char* sbuf=NULL;
		int bufSize = 0;
		for(; stage >= base; stage >>= 1){
			if((myrank/bic)&((stage-1)&(~(base-1)))!=0) continue;
			int target = ((myrank/bic)^stage)*bic + myrank%bic;
			if (((myrank/bic)^stage) >= bir) continue;
			SKEL_MPI_Request reqs, reqr;
			if( target > myrank ) {
				if(Serializer<A>::needsSerializing()){
					if(sbuf==NULL){
						bufSize = 0;
						for(int i = 0; i < localCols; i++){
							bufSize+=Serializer<A>::getSerializedSize(t_vals[i]);
						}
					}
					SKEL_MPI_Send(&bufSize, 1, MPI_INT, target, TAG_ALLREDUCECOL4, mat->comm);
					if(sbuf==NULL){						
						unsigned char *buf = new unsigned char[bufSize];
						sbuf = buf;
						for(int i = 0; i < localCols; i++){
							buf=Serializer<A>::serialize(t_vals[i], buf);
						}
					}
					SKEL_MPI_Isend(sbuf, bufSize, BYTE_TYPE, target, TAG_ALLREDUCECOL3, mat->comm, reqs);
					
					SKEL_MPI_Wait(reqs);	
				} else {
					SKEL_MPI_Isend(t_vals, localCols, mat->mpitype, target, TAG_ALLREDUCECOL3, mat->comm, reqs);
					SKEL_MPI_Wait(reqs);
				}
			} else {
				if(Serializer<A>::needsSerializing()){
					//std::cout << "rank = " << myrank << ", Rcv target = " << target << etd::endl;
					SKEL_MPI_Recv(&bufSize, 1, MPI_INT, target, TAG_ALLREDUCECOL4, mat->comm);
					//std::cout << "rank = " << myrank << ", Rcv bufSize = " << bufSize << etd::endl;
					unsigned char *buf = new unsigned char[bufSize];
					sbuf = buf;
					SKEL_MPI_Irecv(buf, bufSize, BYTE_TYPE, target, TAG_ALLREDUCECOL3, mat->comm, reqr);
					SKEL_MPI_Wait(reqr);
					//std::cout << "rank = " << myrank << ", Rcv data" << etd::endl;
					for(int i = 0; i < localCols; i++){
						buf = Serializer<A>::construct(&t_vals[i], buf);
					}
				} else {
						SKEL_MPI_Irecv(t_vals, localCols, mat->mpitype, target, TAG_ALLREDUCECOL3, mat->comm, reqr);
						SKEL_MPI_Wait(reqr);
				}
			}
		}
		if(sbuf!=NULL) delete [] sbuf;
	}
	/**
	 * shadowing
	 */
	for (int j = 0; j < localCols; j++){
		dmb->at(0, j) = t_vals[j];
		for(int i = 1; i < localRows; i++){
			Shadower<A>::shadow(dmb->at(0, j),&dmb->at(i, j));
			}
	}
	delete [] t_vals;
	return dmb;
}


template <typename A, typename F, typename G>
void
matrix_skeletons::reduceDCN(const F &oplus, const G &otimes, const dist_matrix<A>* mat, A *res)
{
	if(mat->myrank >= mat->blocks) return;
	dist_matrix<A> *tmp = new dist_matrix<A>(*mat);
	{
		const int n = tmp->localRows;
		const int m = tmp->localCols;
		const int nm = n > m ? n : m;
		for(int stage = 1; stage < nm; stage <<=1){
			const int stage2 = stage << 1;
			for(int i = 0; i < n; i+=stage2){
				for(int j = 0; j < m; j+=stage2){
					const int i2 = i ^ stage;
					const int j2 = j ^ stage;
					if(i2 < n){
						A t = tmp->at(i, j);
						oplus(t, tmp->at(i2, j), &tmp->at(i,j));
						if(j2 < m){
							A u = tmp->at(i, j2);
							oplus(u, tmp->at(i2, j2), &tmp->at(i,j2));
						}
					}
					if(j2 < m){
						A t = tmp->at(i, j);
						otimes(t, tmp->at(i, j2), &tmp->at(i,j));
					}
				}
			}
		}
	}
	A val = tmp->at(0,0);
	delete tmp;

	int maxstage = mat->bir;
	if(maxstage < mat->bic) maxstage = mat->bic;
	int stage2 = 1;
	int sn = 0;
	for(; stage2 < maxstage; stage2 <<= 1, sn++);
	stage2>>=1;
	SKEL_MPI_INTRACOMM coms[sn*2+1];
	coms[sn*2] = mat->comm;
	
	for(int k = sn*2; stage2 > 0; stage2>>=1){
		SKEL_MPI_Comm_split(coms[k], ((mat->myrank/mat->bic) & stage2) ? 1 : 0, 0, coms[k-1]);
		k--;
		SKEL_MPI_Comm_split(coms[k], ((mat->myrank%mat->bic) & stage2) ? 1 : 0, 0, coms[k-1]);
		k--;
	}

	//std::cout << "mat->rank = " << mat->myrank << ", mat->bic = " << mat->bic << ", mat->bir = " << mat->bir << std::endl;
	int bic = 1;
	int bir = 1;
	for(int stage = 1, k = 1; stage < maxstage; stage <<= 1){
		bic = ((mat->myrank%mat->bic) < (mat->bic&(~(stage*2-1)))) ? stage*2 :  mat->bic&((stage*2-1));
		int myrank = 0;
		int mysize = 0;
		SKEL_MPI_Comm_rank(coms[k], myrank);
		SKEL_MPI_Comm_size(coms[k], mysize);
		if(stage < mat->bic){
			A rval;
			bool calcflag = true;
			if(((myrank%bic)&(stage-1))==0){
				int target = ((myrank%bic)^stage) + (myrank/bic)*bic;
				//std::cout << "J mat->rank = " << mat->myrank << ", myrank = " << myrank << ", target = " << target  << ", bic = " << bic << ", bir = " << bir  << ",stage="<<stage<<", size="<<mysize<< std::endl;
				if( target < myrank ) {
					SKEL_MPI_Request reqs;
					if(Serializer<A>::needsSerializing()){
						unsigned char* sbuf = mat->sendSerializingData(target, &val, 1, TAG_REDUCE3, TAG_REDUCE1, reqs, coms[k]);
						SKEL_MPI_Wait(reqs);	
						delete [] sbuf;
					} else {
						SKEL_MPI_Isend(&val, 1, mat->mpitype, target, TAG_REDUCE1, coms[k], reqs);
						SKEL_MPI_Wait(reqs);	
					}
					//break;
				} else if (((myrank%bic)^stage) < bic ) {
					if(Serializer<A>::needsSerializing()){
						mat->recvSerializingData(target, &rval, 1, TAG_REDUCE3, TAG_REDUCE1, coms[k]);
					} else {
						SKEL_MPI_Request reqr;
						SKEL_MPI_Irecv(&rval, 1, mat->mpitype, target, TAG_REDUCE1, coms[k], reqr);
						SKEL_MPI_Wait(reqr);
					}
					// ׻ʬ if γ˽ФơߤʤǼ¹Ԥ
					
				} else {
					calcflag = false;
				}
			}
			if(calcflag){
				//std::cout << "J mat->rank = " << mat->myrank << ", myrank = " << myrank << ", calc otimes!"<<", size="<<mysize<< std::endl;
				A tmp = val;
				otimes(tmp, rval, &val, coms[k]);
// 				if(mysize==1) {if(myrank==0)otimes(tmp, rval, &val);}
// 				else          otimes(tmp, rval, &val, coms[k]);
			}
		}
		bir = ((mat->myrank/mat->bic) < (mat->bir&(~(stage*2-1)))) ? stage*2 :  mat->bir&((stage*2-1));
		k++;
		SKEL_MPI_Comm_rank(coms[k], myrank);
		SKEL_MPI_Comm_size(coms[k], mysize);
		//std::cout << "K mat->rank = " << mat->myrank << ", myrank = " << myrank << ", (mat->myrank/mat->bic) = " << (mat->myrank/mat->bic) << ", (!(stage*2-1))= " << (~(stage*2-1))<< ", mat->bir = " << mat->bir << ",mat->bir&((stage*2-1))="<<(mat->bir&((stage*2-1)))<<std::endl;
		if(stage < mat->bir){
			A rval;
			bool calcflag = true;
			if(((myrank/bic)&(stage-1))==0){
				int target = ((myrank/bic)^stage)*bic+(myrank%bic);
				//std::cout << "K mat->rank = " << mat->myrank << ", myrank = " << myrank << ", target = " << target << ", bic = " << bic << ", bir = " << bir << ",stage="<<stage<<", size="<<mysize<<std::endl;
				if( target < myrank ) {	
					SKEL_MPI_Request reqs;
					if(Serializer<A>::needsSerializing()){
						unsigned char* sbuf = mat->sendSerializingData(target, &val, 1, TAG_REDUCE4, TAG_REDUCE2, reqs, coms[k]);
						SKEL_MPI_Wait(reqs);	
						delete [] sbuf;
					} else {
						SKEL_MPI_Isend(&val, 1, mat->mpitype, target, TAG_REDUCE2, coms[k], reqs);
						SKEL_MPI_Wait(reqs);	
					}
					//break;
				} else if (((myrank/bic)^stage) < bir ) {
					if(Serializer<A>::needsSerializing()){
						mat->recvSerializingData(target, &rval, 1, TAG_REDUCE4, TAG_REDUCE2, coms[k]);
					} else {
						SKEL_MPI_Request reqr;
						SKEL_MPI_Irecv(&rval, 1, mat->mpitype, target, TAG_REDUCE2, coms[k], reqr);
						SKEL_MPI_Wait(reqr);
					}
				} else {
					calcflag = false;
				}
			}
			if(calcflag){
				//std::cout << "K mat->rank = " << mat->myrank << ", myrank = " << myrank << ", calc oplus!"<<", size="<<mysize<< std::endl;
				A tmp = val;
				oplus(tmp, rval, &val,coms[k]);
// 				if(mysize==1) {if(myrank==0) oplus(tmp, rval, &val);}
// 				else          oplus(tmp, rval, &val,coms[k]);
			}
		}
		k++;
	}
	for(int k = 0; k < sn*2; k++){
		SKEL_MPI_Comm_free(coms[k]);
	}
	*res = val; 
}


template <typename A, typename F, typename G>
A*
matrix_skeletons::reduceDCN(const F &oplus, const G &otimes, const dist_matrix<A>* mat)
{
	A* res = new A;
	reduceDCN(oplus, otimes, mat, res);
	return res;		
}



template <typename A, typename F, typename G>
void
matrix_skeletons::reduceDCN2(const F &oplus, const G &otimes, const dist_matrix<A>* mat, A *res)
{
	if(mat->myrank >= mat->blocks) return;
	dist_matrix<A> *tmp = new dist_matrix<A>();
	tmp->shapeCopy(mat);
	{
		const int n = tmp->localRows;
		const int m = tmp->localCols;
		const int nm = n > m ? n : m;
// 		for(int stage = 1; stage < nm; stage <<=1){
// 			const int stage2 = stage << 1;
// 			for(int i = 0; i < n; i+=stage2){
// 				for(int j = 0; j < m; j+=stage2){
// 					const int i2 = i ^ stage;
// 					const int j2 = j ^ stage;
// 					if(i2 < n){
// 						if(j2 < m){
// 							A t;
// 							oplus(tmp->at(i, j), tmp->at(i2, j), &t);
// 							A u;
// 							oplus(tmp->at(i, j2), tmp->at(i2, j2), &u);
// 							otimes(t, u, &tmp->at(i,j));
// 						} else {
// 							A t = tmp->at(i, j);
// 							oplus(t, tmp->at(i2, j), &tmp->at(i,j));
// 						}
// 					} else {
// 						if(j2 < m){
// 							A t = tmp->at(i, j);
// 							otimes(t, tmp->at(i, j2), &tmp->at(i,j));
// 						}
// 					}
// 				}
// 			}
// 		}
		{
			int stage = 1;
			const int stage2 = stage << 1;
			for(int i = 0; i < n; i+=stage2){
				for(int j = 0; j < m; j+=stage2){
					const int i2 = i ^ stage;
					const int j2 = j ^ stage;
					if(i2 < n){
						if(j2 < m){
							A t;
							oplus(mat->at(i, j), mat->at(i2, j), &t);
							A u;
							oplus(mat->at(i, j2), mat->at(i2, j2), &u);
							otimes(t, u, &tmp->at(i,j));
						} else {
							oplus(mat->at(i,j), mat->at(i2, j), &tmp->at(i,j));
						}
					} else {
						if(j2 < m){
							otimes(mat->at(i,j), mat->at(i, j2), &tmp->at(i,j));
						} else {
							tmp->at(i, j) = mat->at(i,j);
						}
					}
				}
			}
		}
		for(int stage = 2; stage < nm; stage <<=1){
			const int stage2 = stage << 1;
			for(int i = 0; i < n; i+=stage2){
				for(int j = 0; j < m; j+=stage2){
					const int i2 = i ^ stage;
					const int j2 = j ^ stage;
					if(i2 < n){
						if(j2 < m){
							A t;
							oplus(tmp->at(i, j), tmp->at(i2, j), &t);
							A u;
							oplus(tmp->at(i, j2), tmp->at(i2, j2), &u);
							otimes(t, u, &tmp->at(i,j));
						} else {
							A t = tmp->at(i, j);
							oplus(t, tmp->at(i2, j), &tmp->at(i,j));
						}
					} else {
						if(j2 < m){
							A t = tmp->at(i, j);
							otimes(t, tmp->at(i, j2), &tmp->at(i,j));
						}
					}
				}
			}
		}
	}
	A val = tmp->at(0,0);
	delete tmp;

	int maxstage = mat->bir;
	if(maxstage < mat->bic) maxstage = mat->bic;
	int stage2 = 1;
	int sn = 0;
	for(; stage2 < maxstage; stage2 <<= 1, sn++);
	stage2>>=1;
	SKEL_MPI_INTRACOMM coms[sn*2+1];
	coms[sn*2] = mat->comm;
	
	for(int k = sn*2; stage2 > 0; stage2>>=1){
		SKEL_MPI_Comm_split(coms[k], ((mat->myrank/mat->bic) & stage2) ? 1 : 0, 0, coms[k-1]);
		k--;
		SKEL_MPI_Comm_split(coms[k], ((mat->myrank%mat->bic) & stage2) ? 1 : 0, 0, coms[k-1]);
		k--;
	}

	//std::cout << "mat->rank = " << mat->myrank << ", mat->bic = " << mat->bic << ", mat->bir = " << mat->bir << std::endl;
	int bic = 1;
	int bir = 1;
	for(int stage = 1, k = 1; stage < maxstage; stage <<= 1){
		bic = ((mat->myrank%mat->bic) < (mat->bic&(~(stage*2-1)))) ? stage*2 :  mat->bic&((stage*2-1));
		int myrank = 0;
		int mysize = 0;
		SKEL_MPI_Comm_rank(coms[k], myrank);
		SKEL_MPI_Comm_size(coms[k], mysize);
		if(stage < mat->bic){
			A rval;
			bool calcflag = true;
			if(((myrank%bic)&(stage-1))==0){
				int target = ((myrank%bic)^stage) + (myrank/bic)*bic;
				//std::cout << "J mat->rank = " << mat->myrank << ", myrank = " << myrank << ", target = " << target  << ", bic = " << bic << ", bir = " << bir  << ",stage="<<stage<<", size="<<mysize<< std::endl;
				if( target < myrank ) {
					SKEL_MPI_Request reqs;
					if(Serializer<A>::needsSerializing()){
						unsigned char* sbuf = mat->sendSerializingData(target, &val, 1, TAG_REDUCE3, TAG_REDUCE1, reqs, coms[k]);
						SKEL_MPI_Wait(reqs);	
						delete [] sbuf;
					} else {
						SKEL_MPI_Isend(&val, 1, mat->mpitype, target, TAG_REDUCE1, coms[k], reqs);
						SKEL_MPI_Wait(reqs);	
					}
					//break;
				} else if (((myrank%bic)^stage) < bic ) {
					if(Serializer<A>::needsSerializing()){
						mat->recvSerializingData(target, &rval, 1, TAG_REDUCE3, TAG_REDUCE1, coms[k]);
					} else {
						SKEL_MPI_Request reqr;
						SKEL_MPI_Irecv(&rval, 1, mat->mpitype, target, TAG_REDUCE1, coms[k], reqr);
						SKEL_MPI_Wait(reqr);
					}
					// ׻ʬ if γ˽ФơߤʤǼ¹Ԥ
					
				} else {
					calcflag = false;
				}
			}
			if(calcflag){
				//std::cout << "J mat->rank = " << mat->myrank << ", myrank = " << myrank << ", calc otimes!"<<", size="<<mysize<< std::endl;
				
				A tmp = val;
				dist_matrix<A>::bcast(&tmp, coms[k]);
				dist_matrix<A>::bcast(&rval, coms[k]);
				otimes(tmp, rval, &val, coms[k]);
// 				if(mysize==1) {if(myrank==0)otimes(tmp, rval, &val);}
// 				else          otimes(tmp, rval, &val, coms[k]);
			}
		}
		bir = ((mat->myrank/mat->bic) < (mat->bir&(~(stage*2-1)))) ? stage*2 :  mat->bir&((stage*2-1));
		k++;
		SKEL_MPI_Comm_rank(coms[k], myrank);
		SKEL_MPI_Comm_size(coms[k], mysize);
		//std::cout << "K mat->rank = " << mat->myrank << ", myrank = " << myrank << ", (mat->myrank/mat->bic) = " << (mat->myrank/mat->bic) << ", (!(stage*2-1))= " << (~(stage*2-1))<< ", mat->bir = " << mat->bir << ",mat->bir&((stage*2-1))="<<(mat->bir&((stage*2-1)))<<std::endl;
		if(stage < mat->bir){
			A rval;
			bool calcflag = true;
			if(((myrank/bic)&(stage-1))==0){
				int target = ((myrank/bic)^stage)*bic+(myrank%bic);
				//std::cout << "K mat->rank = " << mat->myrank << ", myrank = " << myrank << ", target = " << target << ", bic = " << bic << ", bir = " << bir << ",stage="<<stage<<", size="<<mysize<<std::endl;
				if( target < myrank ) {	
					SKEL_MPI_Request reqs;
					if(Serializer<A>::needsSerializing()){
						unsigned char* sbuf = mat->sendSerializingData(target, &val, 1, TAG_REDUCE4, TAG_REDUCE2, reqs, coms[k]);
						SKEL_MPI_Wait(reqs);	
						delete [] sbuf;
					} else {
						SKEL_MPI_Isend(&val, 1, mat->mpitype, target, TAG_REDUCE2, coms[k], reqs);
						SKEL_MPI_Wait(reqs);	
					}
					//break;
				} else if (((myrank/bic)^stage) < bir ) {
					if(Serializer<A>::needsSerializing()){
						mat->recvSerializingData(target, &rval, 1, TAG_REDUCE4, TAG_REDUCE2, coms[k]);
					} else {
						SKEL_MPI_Request reqr;
						SKEL_MPI_Irecv(&rval, 1, mat->mpitype, target, TAG_REDUCE2, coms[k], reqr);
						SKEL_MPI_Wait(reqr);
					}
				} else {
					calcflag = false;
				}
			}
			if(calcflag){
				//std::cout << "K mat->rank = " << mat->myrank << ", myrank = " << myrank << ", calc oplus!"<<", size="<<mysize<< std::endl;
				A tmp = val;
				dist_matrix<A>::bcast(&tmp, coms[k]);
				dist_matrix<A>::bcast(&rval, coms[k]);
				oplus(tmp, rval, &val,coms[k]);
// 				if(mysize==1) {if(myrank==0) oplus(tmp, rval, &val);}
// 				else          oplus(tmp, rval, &val,coms[k]);
			}
		}
		k++;
	}
	for(int k = 0; k < sn*2; k++){
		SKEL_MPI_Comm_free(coms[k]);
	}
	*res = val; 
}


template <typename A, typename F, typename G>
A*
matrix_skeletons::reduceDCN2(const F &oplus, const G &otimes, const dist_matrix<A>* mat)
{
	A* res = new A;
	reduceDCN2(oplus, otimes, mat, res);
	return res;		
}


template <typename A, typename F, typename G, typename H>
void
matrix_skeletons::cataDCN2(const F &oplus, const G &otimes, const H &f, const dist_matrix<A>* mat, typename H::result_type *res)
{
	if(mat->myrank >= mat->blocks) return;
	dist_matrix<typename H::result_type> *tmp = new dist_matrix<typename H::result_type>();
	tmp->shapeCopy(mat);
	{
		const int n = tmp->localRows;
		const int m = tmp->localCols;
		const int nm = n > m ? n : m;
		{
			int stage = 1;
			const int stage2 = stage << 1;
			for(int i = 0; i < n; i+=stage2){
				for(int j = 0; j < m; j+=stage2){
					const int i2 = i ^ stage;
					const int j2 = j ^ stage;
					if(i2 < n){
						if(j2 < m){
							typename H::result_type a, b, t, u;
							f(mat->at(i,j), &a);
							f(mat->at(i2,j), &b);
							oplus(a, b, &t);
							f(mat->at(i,j2), &a);
							f(mat->at(i2,j2), &b);
							oplus(a, b, &u);
							otimes(t, u, &tmp->at(i,j));
						} else {
							typename H::result_type a, b;
							f(mat->at(i,j), &a);
							f(mat->at(i2,j), &b);
							oplus(a, b, &tmp->at(i,j));
						}
					} else {
						if(j2 < m){
							typename H::result_type a, b;
							f(mat->at(i,j), &a);
							f(mat->at(i,j2), &b);
							otimes(a, b, &tmp->at(i,j));
						} else {
							f(mat->at(i, j), &tmp->at(i, j));
						}
					}
				}
			}
		}
		for(int stage = 2; stage < nm; stage <<=1){
			const int stage2 = stage << 1;
			for(int i = 0; i < n; i+=stage2){
				for(int j = 0; j < m; j+=stage2){
					const int i2 = i ^ stage;
					const int j2 = j ^ stage;
					if(i2 < n){
						if(j2 < m){
							typename H::result_type t;
							oplus(tmp->at(i, j), tmp->at(i2, j), &t);
							typename H::result_type u;
							oplus(tmp->at(i, j2), tmp->at(i2, j2), &u);
							otimes(t, u, &tmp->at(i,j));
						} else {
							typename H::result_type t = tmp->at(i, j);
							oplus(t, tmp->at(i2, j), &tmp->at(i,j));
						}
					} else {
						if(j2 < m){
							typename H::result_type t = tmp->at(i, j);
							otimes(t, tmp->at(i, j2), &tmp->at(i,j));
						}
					}
				}
			}
		}
	}
	typename H::result_type val = tmp->at(0,0);
	delete tmp;

	int maxstage = mat->bir;
	if(maxstage < mat->bic) maxstage = mat->bic;
	int stage2 = 1;
	int sn = 0;
	for(; stage2 < maxstage; stage2 <<= 1, sn++);
	stage2>>=1;
	SKEL_MPI_INTRACOMM coms[sn*2+1];
	coms[sn*2] = mat->comm;
	
	for(int k = sn*2; stage2 > 0; stage2>>=1){
		SKEL_MPI_Comm_split(coms[k], ((mat->myrank/mat->bic) & stage2) ? 1 : 0, 0, coms[k-1]);
		k--;
		SKEL_MPI_Comm_split(coms[k], ((mat->myrank%mat->bic) & stage2) ? 1 : 0, 0, coms[k-1]);
		k--;
	}

	//std::cout << "mat->rank = " << mat->myrank << ", mat->bic = " << mat->bic << ", mat->bir = " << mat->bir << std::endl;
	int bic = 1;
	int bir = 1;
	for(int stage = 1, k = 1; stage < maxstage; stage <<= 1){
		bic = ((mat->myrank%mat->bic) < (mat->bic&(~(stage*2-1)))) ? stage*2 :  mat->bic&((stage*2-1));
		int myrank = 0;
		int mysize = 0;
		SKEL_MPI_Comm_rank(coms[k], myrank);
		SKEL_MPI_Comm_size(coms[k], mysize);
		if(stage < mat->bic){
			typename H::result_type rval;
			bool calcflag = true;
			if(((myrank%bic)&(stage-1))==0){
				int target = ((myrank%bic)^stage) + (myrank/bic)*bic;
				//std::cout << "J mat->rank = " << mat->myrank << ", myrank = " << myrank << ", target = " << target  << ", bic = " << bic << ", bir = " << bir  << ",stage="<<stage<<", size="<<mysize<< std::endl;
				if( target < myrank ) {
					SKEL_MPI_Request reqs;
					if(Serializer<typename H::result_type>::needsSerializing()){
						unsigned char* sbuf = mat->sendSerializingData(target, &val, 1, TAG_REDUCE3, TAG_REDUCE1, reqs, coms[k]);
						SKEL_MPI_Wait(reqs);	
						delete [] sbuf;
					} else {
						SKEL_MPI_Isend(&val, 1, mat->mpitype, target, TAG_REDUCE1, coms[k], reqs);
						SKEL_MPI_Wait(reqs);	
					}
					//break;
				} else if (((myrank%bic)^stage) < bic ) {
					if(Serializer<typename H::result_type>::needsSerializing()){
						mat->recvSerializingData(target, &rval, 1, TAG_REDUCE3, TAG_REDUCE1, coms[k]);
					} else {
						SKEL_MPI_Request reqr;
						SKEL_MPI_Irecv(&rval, 1, mat->mpitype, target, TAG_REDUCE1, coms[k], reqr);
						SKEL_MPI_Wait(reqr);
					}
					// ׻ʬ if γ˽ФơߤʤǼ¹Ԥ
					
				} else {
					calcflag = false;
				}
			}
			if(calcflag){
				//std::cout << "J mat->rank = " << mat->myrank << ", myrank = " << myrank << ", calc otimes!"<<", size="<<mysize<< std::endl;
				
				typename H::result_type tmp = val;
				dist_matrix<typename H::result_type>::bcast(&tmp, coms[k]);
				dist_matrix<typename H::result_type>::bcast(&rval, coms[k]);
				otimes(tmp, rval, &val, coms[k]);
// 				if(mysize==1) {if(myrank==0)otimes(tmp, rval, &val);}
// 				else          otimes(tmp, rval, &val, coms[k]);
			}
		}
		bir = ((mat->myrank/mat->bic) < (mat->bir&(~(stage*2-1)))) ? stage*2 :  mat->bir&((stage*2-1));
		k++;
		SKEL_MPI_Comm_rank(coms[k], myrank);
		SKEL_MPI_Comm_size(coms[k], mysize);
		//std::cout << "K mat->rank = " << mat->myrank << ", myrank = " << myrank << ", (mat->myrank/mat->bic) = " << (mat->myrank/mat->bic) << ", (!(stage*2-1))= " << (~(stage*2-1))<< ", mat->bir = " << mat->bir << ",mat->bir&((stage*2-1))="<<(mat->bir&((stage*2-1)))<<std::endl;
		if(stage < mat->bir){
			typename H::result_type rval;
			bool calcflag = true;
			if(((myrank/bic)&(stage-1))==0){
				int target = ((myrank/bic)^stage)*bic+(myrank%bic);
				//std::cout << "K mat->rank = " << mat->myrank << ", myrank = " << myrank << ", target = " << target << ", bic = " << bic << ", bir = " << bir << ",stage="<<stage<<", size="<<mysize<<std::endl;
				if( target < myrank ) {	
					SKEL_MPI_Request reqs;
					if(Serializer<typename H::result_type>::needsSerializing()){
						unsigned char* sbuf = mat->sendSerializingData(target, &val, 1, TAG_REDUCE4, TAG_REDUCE2, reqs, coms[k]);
						SKEL_MPI_Wait(reqs);	
						delete [] sbuf;
					} else {
						SKEL_MPI_Isend(&val, 1, mat->mpitype, target, TAG_REDUCE2, coms[k], reqs);
						SKEL_MPI_Wait(reqs);	
					}
					//break;
				} else if (((myrank/bic)^stage) < bir ) {
					if(Serializer<typename H::result_type>::needsSerializing()){
						mat->recvSerializingData(target, &rval, 1, TAG_REDUCE4, TAG_REDUCE2, coms[k]);
					} else {
						SKEL_MPI_Request reqr;
						SKEL_MPI_Irecv(&rval, 1, mat->mpitype, target, TAG_REDUCE2, coms[k], reqr);
						SKEL_MPI_Wait(reqr);
					}
				} else {
					calcflag = false;
				}
			}
			if(calcflag){
				//std::cout << "K mat->rank = " << mat->myrank << ", myrank = " << myrank << ", calc oplus!"<<", size="<<mysize<< std::endl;
				typename H::result_type tmp = val;
				dist_matrix<typename H::result_type>::bcast(&tmp, coms[k]);
				dist_matrix<typename H::result_type>::bcast(&rval, coms[k]);
				oplus(tmp, rval, &val,coms[k]);
// 				if(mysize==1) {if(myrank==0) oplus(tmp, rval, &val);}
// 				else          oplus(tmp, rval, &val,coms[k]);
			}
		}
		k++;
	}
	for(int k = 0; k < sn*2; k++){
		SKEL_MPI_Comm_free(coms[k]);
	}
	*res = val; 
}


template <typename A, typename F, typename G, typename H>
typename H::result_type *
matrix_skeletons::cataDCN2(const F &oplus, const G &otimes, const H &f, const dist_matrix<A>* mat)
{
	typename H::result_type * res = new typename H::result_type;
	cataDCN2(oplus, otimes, f, mat, res);
	return res;		
}



template <typename A, typename F, typename G, typename H>
void
matrix_skeletons::cata(const F &oplus, const G &otimes, const H &f, const dist_matrix<A>* mat, typename H::result_type *res)
{
	if(mat->myrank >= mat->blocks) return;
	//if(mat->myrank==0) std::cout << "reduce 0\n";
	int bicfloor = (mat->myrank/mat->bic)*mat->bic;
	typename H::result_type *vals = new typename H::result_type[mat->localRows];
	for(int i = 0; i < mat->localRows; i++){
		typename H::result_type val;
		f(mat->at(i, 0), &val);
		for(int j = 1; j < mat->localCols; j++){
			//if(mat->myrank==0) std::cout << "i = " << i << ", j = " << j << std::endl;
			typename H::result_type tmp = val;
			typename H::result_type tmp2;
			f(mat->at(i, j), &tmp2);
			
			otimes(tmp, tmp2, &val);
		}
		vals[i] = val;
	}
	//if(mat->myrank==0) std::cout << "reduce 0.5\n";
	typename H::result_type val = vals[0];
	for(int i = 1; i < mat->localRows; i++){
		typename H::result_type tmp = val;
		oplus(tmp, vals[i], &val);
	}
	delete [] vals;
	//if(mat->myrank==0) std::cout << "reduce 1\n";
	// tree communication (row)
	for(int stage = 1; stage < mat->bic; stage <<= 1){
		int target = ((mat->myrank%mat->bic)^stage) + bicfloor;
		//cout << "rank = " << myrank << " target : " << target << endl;
		if( target < mat->myrank ) {
			SKEL_MPI_Request reqs;
			if(Serializer<typename H::result_type>::needsSerializing()){
				unsigned char* sbuf = mat->sendSerializingData(target, &val, 1, TAG_REDUCE3, TAG_REDUCE1, reqs);
				SKEL_MPI_Wait(reqs);	
				delete [] sbuf;
				//cout << "rank = " << myrank << " sent" << endl;
			} else {
				SKEL_MPI_Isend(&val, 1, mat->mpitype, target, TAG_REDUCE1, mat->comm, reqs);
				SKEL_MPI_Wait(reqs);	
			}
			break;
		} else if (((mat->myrank%mat->bic)^stage) < mat->bic ) {
			typename H::result_type rval;
			if(Serializer<typename H::result_type>::needsSerializing()){
				//cout << "rank = " << myrank << " ready to receive" << endl;
				mat->recvSerializingData(target, &rval, 1, TAG_REDUCE3, TAG_REDUCE1);
				//cout << "rank = " << myrank << " received" << endl;
			} else {
				SKEL_MPI_Request reqr;
				SKEL_MPI_Irecv(&rval, 1, mat->mpitype, target, TAG_REDUCE1, mat->comm, reqr);
				SKEL_MPI_Wait(reqr);
			}
			typename H::result_type tmp = val;
			otimes(tmp, rval, &val);
		}
	}
	//if(myrank==0) cout << "reduce 2\n";
	if(mat->myrank%mat->bic==0){
		for(int stage = 1; stage < mat->bir; stage <<= 1){
			int target = ((mat->myrank/mat->bic)^stage)*mat->bic;
			//cout << "rank = " << myrank << " target : " << target << endl;
			if( target < mat->myrank ) {	
				SKEL_MPI_Request reqs;
				if(Serializer<typename H::result_type>::needsSerializing()){
					unsigned char* sbuf = mat->sendSerializingData(target, &val, 1, TAG_REDUCE4, TAG_REDUCE2, reqs);
					SKEL_MPI_Wait(reqs);	
					delete [] sbuf;
					//cout << "rank = " << myrank << " sent" << endl;
				} else {
					SKEL_MPI_Isend(&val, 1, mat->mpitype, target, TAG_REDUCE2, mat->comm, reqs);
					SKEL_MPI_Wait(reqs);	
				}
				break;
				//} else if ( target < blocks ) {
			} else if (((mat->myrank/mat->bic)^stage) < mat->bir ) {
				typename H::result_type rval;
				if(Serializer<typename H::result_type>::needsSerializing()){
					//cout << "rank = " << myrank << " ready to receive" << endl;
					mat->recvSerializingData(target, &rval, 1, TAG_REDUCE4, TAG_REDUCE2);
					//cout << "rank = " << myrank << " received" << endl;
				} else {
					SKEL_MPI_Request reqr;
					SKEL_MPI_Irecv(&rval, 1, mat->mpitype, target, TAG_REDUCE2, mat->comm, reqr);
					SKEL_MPI_Wait(reqr);
				}
				typename H::result_type tmp = val;
				oplus(tmp, rval, &val);
			}
		}
	}
	*res = val; 
}


template <typename A, typename F, typename G, typename H>
typename H::result_type*
matrix_skeletons::cata(const F &oplus, const G &otimes, const H &f, const dist_matrix<A>* mat)
{
	typename H::result_type* res = new typename H::result_type;
	cata(oplus, otimes, f, mat, res);
	return res;		
}


template <typename A, typename OP2, typename OT2, typename P, typename OP, typename OT, typename Q>
void
matrix_skeletons::acc(const OP2 &oplus2, const OT2 &otimes2, const P &p, const OP &oplus, const OT &otimes, const Q&q, const dist_matrix<A>* mat, typename P::result_type *res)
{
	// lessen the amount of local memory use
	// local map & reduction(3 way)
	// global HC-like communication
	// local scan & local map & reduction
	
	if(mat->myrank >= mat->blocks) return;

	typename Q::result_type *r_t_vals;    // for global scan of row-direction
	typename Q::result_type *r_t_valsR;
	typename Q::result_type *r_f_vals;
	typename Q::result_type *c_t_vals;    // for global scan of column-direction
	typename Q::result_type *c_t_valsR; 
	typename Q::result_type *c_f_vals;
	typename Q::result_type *b_nw_val;    // for global 2D scan 
	typename Q::result_type *b_n_val;
	typename Q::result_type *b_w_val;
	typename Q::result_type *b_c_val;
	typename Q::result_type *b_n_valR;
	typename Q::result_type *b_w_valR;
	typename Q::result_type *b_c_valR;

	MPI_Datatype mpitypeQ = template_mpitype< typename Q::result_type>::type();
	MPI_Datatype mpitypeP = template_mpitype< typename P::result_type>::type();
	/****
	 * memory image: [r_t_vals, b_{n,w,c}_val, c_t_vals]
	 * if the communication is Row-direction, send [r_t_vals, b_{n,w,c}_val].
	 * if the communication is Col-direction, send [b_{n,w,c}_val, c_t_vals].
	 * 
	 */
	r_t_vals = new typename Q::result_type[mat->localRows+mat->localCols+3];
	b_n_val = r_t_vals + mat->localRows;
	b_w_val = b_n_val + 1;
	b_c_val = b_w_val + 1;
	c_t_vals = b_c_val + 1;
	r_t_valsR = new typename Q::result_type[mat->localRows+mat->localCols+3];
	b_n_valR = r_t_valsR + mat->localRows;
	b_w_valR = b_n_valR + 1;
	b_c_valR = b_w_valR + 1;
	c_t_valsR = b_c_valR + 1;
	r_f_vals = new typename Q::result_type[mat->localRows+mat->localCols+1];
	b_nw_val = r_f_vals + mat->localRows;
	c_f_vals = b_nw_val + 1;
	const int sendRecvSizeCol = mat->localCols+3;
	const int sendRecvSizeRow = mat->localRows+3;
	const typename Q::result_type* sendBufCol = b_n_val;
	const typename Q::result_type* sendBufRow = r_t_vals;
	typename Q::result_type* recvBufCol = b_n_valR;
	typename Q::result_type* recvBufRow = r_t_valsR;


	//std::cout << "mat->localRows = " << mat->localRows << ", mat->localCols = " << mat->localCols << std::endl;
	// local reduction(3 way)
	//std::cout << "local reduction(3 way)-1" << std::endl;
	typename Q::result_type tmp, tmp2;
	for(int i = 0; i < 1; i++){
		//std::cout << "local reduction(3 way)-1-1" << std::endl;
		q(mat->at(0,0), &c_t_vals[0]);
		//std::cout << "local reduction(3 way)-1-2" << std::endl;
		r_t_vals[0] = c_t_vals[0];
		for(int j = 1; j < mat->localCols; j++){
			//std::cout << "local reduction(3 way)-1-3-1" << std::endl;
			//c_t_vals[j].printStat(std::cout);
			q(mat->at(0,j), &c_t_vals[j]);
			//std::cout << "local reduction(3 way)-1-3-2" << std::endl;
			otimes(r_t_vals[0], c_t_vals[j], &tmp);
			//std::cout << "local reduction(3 way)-1-3-3" << std::endl;
			r_t_vals[0] = tmp;
		}
	}
	//std::cout << "local reduction(3 way)-2" << std::endl;
	for(int i = 1; i < mat->localRows; i++){
		q(mat->at(i,0), &tmp2);
		oplus(c_t_vals[0], tmp2, &tmp);
		c_t_vals[0] = tmp;
		r_t_vals[i] = tmp2;
		for(int j = 1; j < mat->localCols; j++){
			q(mat->at(i,j), &tmp2);
			oplus(c_t_vals[j], tmp2, &tmp);
			c_t_vals[j] = tmp;
			otimes(r_t_vals[i], tmp2, &tmp);
			r_t_vals[i] = tmp;
		}
	}
	//std::cout << "local reduction(3 way)-3" << std::endl;
	*b_c_val = r_t_vals[0];
	for(int i = 1; i < mat->localRows; i++){
		oplus(*b_c_val, r_t_vals[i], &tmp);
		*b_c_val = tmp;
	}

	const int bicfloor = (mat->myrank/mat->bic)*mat->bic;
	const int birfloor = (mat->myrank%mat->bic);
	const int id_col = mat->myrank%mat->bic;
	const int id_row = mat->myrank/mat->bic;
	int maxstage = mat->bir;
	if(maxstage < mat->bic) maxstage = mat->bic;
	
	bool flag_first_col = true;
	bool flag_first_row = true;

	//std::cout << "global HC-like communication" << std::endl;
	// global HC-like communication
	for(int stage = 1; stage < maxstage; stage <<= 1){
		//std::cout << "stage = " << stage << std::endl;
		if(stage < mat->bic && (id_col^stage) < mat->bic){  // row-direction
			int target = (id_col^stage) + bicfloor;
			//std::cout << "row-direction: " << mat->myrank << " to " << target << std::endl;

			SKEL_MPI_Request reqs, reqr;
			if(Serializer<typename Q::result_type>::needsSerializing()){
				//std::cout << "serializing & send" << std::endl;
				unsigned char* sbuf = mat->sendSerializingData(target, sendBufRow, sendRecvSizeRow, TAG_SCAN12s, TAG_SCAN12, reqs);
				//std::cout << "receive?" << std::endl;
				mat->recvSerializingData(target, recvBufRow, sendRecvSizeRow, TAG_SCAN12s, TAG_SCAN12);
				//std::cout << "waiting?" << std::endl;
				SKEL_MPI_Wait(reqs);	
				delete [] sbuf;
			} else {
				SKEL_MPI_Isend(sendBufRow, sendRecvSizeRow, mpitypeQ, target, TAG_SCAN12, mat->comm, reqs);
				SKEL_MPI_Irecv(recvBufRow, sendRecvSizeRow, mpitypeQ, target, TAG_SCAN12, mat->comm, reqr);
				SKEL_MPI_Wait(reqs);
				SKEL_MPI_Wait(reqr);
			}
			//std::cout << "send/recv done."  <<mat->myrank << " to " << target<< std::endl;
			if( target > mat->myrank ) {
				for(int i = 0; i < mat->localRows; i++){
					otimes(r_t_vals[i], r_t_valsR[i], &tmp);
					r_t_vals[i] = tmp;
				}
				otimes(*b_c_val, *b_c_valR, &tmp);
				*b_c_val = tmp;
				if(!flag_first_col){
					otimes(*b_n_val, *b_n_valR, &tmp);
					*b_n_val = tmp;
				}
			} else {
				if(flag_first_row){
					// this is inefficient...
					// receiving directly to r_f_vals is more efficient.
					for(int i = 0; i < mat->localRows; i++){
						r_f_vals[i] = r_t_valsR[i];
					}
					*b_w_val = *b_c_valR;
					if(!flag_first_col){
						*b_nw_val = *b_n_valR;
					}
					flag_first_row = false;
				} else {
					for(int i = 0; i < mat->localRows; i++){
						otimes(r_t_valsR[i], r_f_vals[i], &tmp);
						r_f_vals[i] = tmp;
					}
					otimes(*b_c_valR, *b_w_val, &tmp);
					*b_w_val = tmp;
					if(!flag_first_col){
						otimes(*b_n_valR, *b_nw_val, &tmp);
						*b_nw_val = tmp;
					}
				}
				for(int i = 0; i < mat->localRows; i++){
					otimes(r_t_valsR[i], r_t_vals[i], &tmp);
					r_t_vals[i] = tmp;
				}
				otimes(*b_c_valR, *b_c_val, &tmp);
				*b_c_val = tmp;
				if(!flag_first_col){
					otimes(*b_n_valR, *b_n_val, &tmp);
					*b_n_val = tmp;
				}
			}
		}
		///////////////////////////////////////////////////////////////////////////
		if(stage < mat->bir && (id_row^stage) < mat->bir){  // col-direction
			int target = (id_row^stage)*mat->bic + birfloor;
			//std::cout << "col-direction: " << mat->myrank << " to " << target << std::endl;

			SKEL_MPI_Request reqs, reqr;
			if(Serializer<typename Q::result_type>::needsSerializing()){
				//std::cout << "serializing & send" << std::endl;
				unsigned char* sbuf = mat->sendSerializingData(target, sendBufCol, sendRecvSizeCol, TAG_SCAN12s, TAG_SCAN12, reqs);
				//std::cout << "receive?" << std::endl;
				mat->recvSerializingData(target, recvBufCol, sendRecvSizeCol, TAG_SCAN12s, TAG_SCAN12);
				//std::cout << "waiting?" << std::endl;
				SKEL_MPI_Wait(reqs);	
				delete [] sbuf;
			} else {
				SKEL_MPI_Isend(sendBufCol, sendRecvSizeCol, mpitypeQ, target, TAG_SCAN12, mat->comm, reqs);
				SKEL_MPI_Irecv(recvBufCol, sendRecvSizeCol, mpitypeQ, target, TAG_SCAN12, mat->comm, reqr);
				SKEL_MPI_Wait(reqs);
				SKEL_MPI_Wait(reqr);
			}
			//std::cout << "send/recv done." << mat->myrank << " to " << target<< std::endl;
			if( target > mat->myrank ) {
				//std::cout << "local calc-1" << std::endl;
				for(int i = 0; i < mat->localCols; i++){
					oplus(c_t_vals[i], c_t_valsR[i], &tmp);
					c_t_vals[i] = tmp;
				}
				oplus(*b_c_val, *b_c_valR, &tmp);
				*b_c_val = tmp;
				if(!flag_first_row){
					oplus(*b_w_val, *b_w_valR, &tmp);
					*b_w_val = tmp;
				}
			} else {
				if(flag_first_col){
					// this is inefficient...
					// receiving directly to c_f_vals is more efficient.
					for(int i = 0; i < mat->localCols; i++){
						c_f_vals[i] = c_t_valsR[i];
					}
					*b_n_val = *b_c_valR;
					if(!flag_first_row){
						*b_nw_val = *b_w_valR;
					}
					flag_first_col = false;
				} else {
					for(int i = 0; i < mat->localCols; i++){
						oplus(c_t_valsR[i], c_f_vals[i], &tmp);
						c_f_vals[i] = tmp;
					}
					oplus(*b_c_valR, *b_n_val, &tmp);
					*b_n_val = tmp;
					if(!flag_first_row){
						oplus(*b_w_valR, *b_nw_val, &tmp);
						*b_nw_val = tmp;
					}
				}
				for(int i = 0; i < mat->localCols; i++){
					oplus(c_t_valsR[i], c_t_vals[i], &tmp);
					c_t_vals[i] = tmp;
				}
				oplus(*b_c_valR, *b_c_val, &tmp);
				*b_c_val = tmp;
				if(!flag_first_row){
					oplus(*b_w_valR, *b_w_val, &tmp);
					*b_w_val = tmp;
				}
			}
		}
	}
	//std::cout << "local scan" << std::endl;
	// local scan
	// flag_first_row == true -> r_f_vals is empty, *b_nw_val is empty
	// flag_first_col == true -> c_f_vals is empty, *b_nw_val is empty
	//std::cout << "local scan-1" << mat->myrank << ", "<<flag_first_row <<flag_first_col << std::endl;
	//std::cout << "--" << mat->myrank << "*b_w_val = " << *b_w_val << "*b_n_val = " << *b_n_val << "*b_c_val = " << *b_c_val << "...." << std::endl;
	//std::cout << "--" << mat->myrank << "*b_nw_val = " << *b_nw_val << "..." <<std::endl;
	//std::cout << "--" << mat->myrank << "c_f_vals = ";
	//for(int i = 0; i < mat->localCols; i++) std::cout << c_f_vals[i] << ", ";
	//std::cout << ".." <<std::endl;
	//std::cout << "--" << mat->myrank << "r_f_vals = ";
	//for(int i = 0; i < mat->localRows; i++) std::cout << r_f_vals[i] << ", ";
	//std::cout << "." <<std::endl;
	// switching of *res and tmp3 is more efficient...
	// substituting tmp3 to *res is much inefficient
	typename P::result_type tmp3, tmp4, tmp5;
	typename Q::result_type ac, atmp;

	for(int i = 0; i < 1; i++){
		if(!flag_first_col){
			q(mat->at(i,0), &tmp2);
			oplus(c_f_vals[0], tmp2, &tmp);
			c_f_vals[0] = tmp;
			if(!flag_first_row ){
				oplus(*b_nw_val, r_f_vals[i], &tmp);
				r_f_vals[i] = tmp;
				otimes(r_f_vals[i], c_f_vals[0], &ac);
			} else {
				 ac = c_f_vals[0];
			}
			p(mat->at(i,0), ac, &tmp5);
			for(int j = 1; j < mat->localCols; j++){
				q(mat->at(i,j), &tmp2);
				oplus(c_f_vals[j], tmp2, &tmp);
				c_f_vals[j] = tmp;
				otimes(ac, tmp, &atmp);
				ac = atmp;
				p(mat->at(i,j), ac, &tmp3);
				otimes2(tmp5, tmp3, &tmp4);
				tmp5 = tmp4;
			}
		} else {
			q(mat->at(i,0), &tmp2);
			c_f_vals[0] = tmp2;
			if(!flag_first_row){
				otimes(r_f_vals[i], tmp2, &ac);
			} else {
				ac = tmp2;
			}
			p(mat->at(i,0), ac, &tmp5);
			for(int j = 1; j < mat->localCols; j++){
				q(mat->at(i,j), &tmp2);
				c_f_vals[j] = tmp2;
				otimes(ac, tmp2, &atmp);
				ac = atmp;
				p(mat->at(i,j), ac, &tmp3);
				otimes2(tmp5, tmp3, &tmp4);
				tmp5 = tmp4;
			}
		}
	}
	*res = tmp5;
	// now, c_f_vals is not empty
	//std::cout << "local scan-2" << mat->myrank << ", "<<flag_first_row <<flag_first_col << std::endl;
	for(int i = 1; i < mat->localRows; i++){
		q(mat->at(i,0), &tmp2);
		oplus(c_f_vals[0], tmp2, &tmp);
		c_f_vals[0] = tmp;
		if(!flag_first_row ){
			oplus(r_f_vals[i-1], r_f_vals[i], &tmp);
			r_f_vals[i] = tmp;
			otimes(r_f_vals[i], c_f_vals[0], &ac);
		} else {
			ac = c_f_vals[0];
		}
		p(mat->at(i,0), ac, &tmp5);
		for(int j = 1; j < mat->localCols; j++){
			q(mat->at(i,j), &tmp2);
			oplus(c_f_vals[j], tmp2, &tmp);
			c_f_vals[j] = tmp;
			otimes(ac, tmp, &atmp);
			ac = atmp;
			p(mat->at(i,j), ac, &tmp3);
			otimes2(tmp5, tmp3, &tmp4);
			tmp5 = tmp4;
		}
		oplus2(*res, tmp5, &tmp4);
		*res = tmp4;
	}
	delete [] r_t_vals;
	delete [] r_t_valsR;
	delete [] r_f_vals;


	// final global reduce

	for(int stage = 1; stage < maxstage; stage <<= 1){
		//std::cout << "stage = " << stage << std::endl;
		if(stage < mat->bic && (id_col^stage) < mat->bic){  // row-direction
			int target = (id_col^stage) + bicfloor;
			//std::cout << "row-direction: " << mat->myrank << " to " << target << std::endl;

			if( target > mat->myrank ) { // recv
				
				SKEL_MPI_Request reqr;
				if(Serializer<typename P::result_type>::needsSerializing()){
					//std::cout << "receive?" << std::endl;
					mat->recvSerializingData(target, &tmp3, 1, TAG_SCAN12s, TAG_SCAN12);
				} else {
					SKEL_MPI_Irecv(&tmp3, 1, mpitypeP, target, TAG_SCAN12, mat->comm, reqr);
					SKEL_MPI_Wait(reqr);
				}
				//std::cout << "send/recv done."  <<mat->myrank << " to " << target<< std::endl;

				otimes2(*res, tmp3, &tmp4);
				*res = tmp4;
			} else {  // send
				SKEL_MPI_Request reqs;
				if(Serializer<typename P::result_type>::needsSerializing()){
					//std::cout << "serializing & send" << std::endl;
					unsigned char* sbuf = mat->sendSerializingData(target, res, 1, TAG_SCAN12s, TAG_SCAN12, reqs);
					//std::cout << "waiting?" << std::endl;
					SKEL_MPI_Wait(reqs);	
					delete [] sbuf;
				} else {
					SKEL_MPI_Isend(res, 1, mpitypeP, target, TAG_SCAN12, mat->comm, reqs);
					SKEL_MPI_Wait(reqs);
				}
				//std::cout << "send/recv done."  <<mat->myrank << " to " << target<< std::endl;
				break;
			}
		}
		///////////////////////////////////////////////////////////////////////////
		if(stage < mat->bir && (id_row^stage) < mat->bir){  // col-direction
			int target = (id_row^stage)*mat->bic + birfloor;
			//std::cout << "col-direction: " << mat->myrank << " to " << target << std::endl;

			if( target > mat->myrank ) { // recv
				SKEL_MPI_Request reqr;
				if(Serializer<typename P::result_type>::needsSerializing()){
				//std::cout << "receive?" << std::endl;
				mat->recvSerializingData(target, &tmp5, 1, TAG_SCAN12s, TAG_SCAN12);
				} else {
					SKEL_MPI_Irecv(&tmp5, 1, mpitypeP, target, TAG_SCAN12, mat->comm, reqr);
					SKEL_MPI_Wait(reqr);
				}
				//std::cout << "send/recv done." << mat->myrank << " to " << target<< std::endl;
				oplus2(*res, tmp5, &tmp4);
				*res = tmp4;
			} else { // send
				SKEL_MPI_Request reqs;
				if(Serializer<typename P::result_type>::needsSerializing()){
					//std::cout << "serializing & send" << std::endl;
					unsigned char* sbuf = mat->sendSerializingData(target, res, 1, TAG_SCAN12s, TAG_SCAN12, reqs);
					//std::cout << "waiting?" << std::endl;
					SKEL_MPI_Wait(reqs);	
					delete [] sbuf;
				} else {
					SKEL_MPI_Isend(res, 1, mpitypeP, target, TAG_SCAN12, mat->comm, reqs);
					SKEL_MPI_Wait(reqs);
				}
				//std::cout << "send/recv done." << mat->myrank << " to " << target<< std::endl;
				break;
			}
		}
	}

	// done
	
}

template <typename A, typename OP2, typename OT2, typename P, typename OP, typename OT, typename Q>
typename P::result_type *
matrix_skeletons::acc(const OP2 &oplus2, const OT2 &otimes2, const P &p, const OP &oplus, const OT &otimes, const Q&q, const dist_matrix<A>* mat)
{
	typename P::result_type* res = new typename P::result_type;
	acc(oplus2, otimes2, p, oplus, otimes, q, mat, res);
	return res;		
}



template <typename A, typename F, typename G>
void
matrix_skeletons::scan2(const F &oplus, const G &otimes, const dist_matrix<A>* mat, dist_matrix<A>* mat2)
{
	// lessen the amount of local memory use
	// local reduction(3 way)
	// global HC-like communication
	// local scan
	
	if(mat->myrank >= mat->blocks) return;

	A *r_t_vals;    // for global scan of row-direction
	A *r_t_valsR;
	A *r_f_vals;
	A *c_t_vals;    // for global scan of column-direction
	A *c_t_valsR; 
	A *c_f_vals;
	A *b_nw_val;    // for global 2D scan 
	A *b_n_val;
	A *b_w_val;
	A *b_c_val;
	A *b_n_valR;
	A *b_w_valR;
	A *b_c_valR;


	/****
	 * memory image: [r_t_vals, b_{n,w,c}_val, c_t_vals]
	 * if the communication is Row-direction, send [r_t_vals, b_{n,w,c}_val].
	 * if the communication is Col-direction, send [b_{n,w,c}_val, c_t_vals].
	 * 
	 */
	r_t_vals = new A[mat->localRows+mat->localCols+3];
	b_n_val = r_t_vals + mat->localRows;
	b_w_val = b_n_val + 1;
	b_c_val = b_w_val + 1;
	c_t_vals = b_c_val + 1;
	r_t_valsR = new A[mat->localRows+mat->localCols+3];
	b_n_valR = r_t_valsR + mat->localRows;
	b_w_valR = b_n_valR + 1;
	b_c_valR = b_w_valR + 1;
	c_t_valsR = b_c_valR + 1;
	r_f_vals = new A[mat->localRows+mat->localCols+1];
	b_nw_val = r_f_vals + mat->localRows;
	c_f_vals = b_nw_val + 1;
	const int sendRecvSizeCol = mat->localCols+3;
	const int sendRecvSizeRow = mat->localRows+3;
	const A* sendBufCol = b_n_val;
	const A* sendBufRow = r_t_vals;
	A* recvBufCol = b_n_valR;
	A* recvBufRow = r_t_valsR;


	//std::cout << "mat->localRows = " << mat->localRows << ", mat->localCols = " << mat->localCols << std::endl;
	// local reduction(3 way)
	//std::cout << "local reduction(3 way)-1" << std::endl;
	A tmp;
	for(int i = 0; i < 1; i++){
		//std::cout << "local reduction(3 way)-1-1" << std::endl;
		c_t_vals[0] = mat->at(0,0);
		//std::cout << "local reduction(3 way)-1-2" << std::endl;
		r_t_vals[0] = mat->at(0,0);
		for(int j = 1; j < mat->localCols; j++){
			//std::cout << "local reduction(3 way)-1-3-1" << std::endl;
			//c_t_vals[j].printStat(std::cout);
			c_t_vals[j] = mat->at(0,j);
			//std::cout << "local reduction(3 way)-1-3-2" << std::endl;
			otimes(r_t_vals[0], mat->at(0, j), &tmp);
			//std::cout << "local reduction(3 way)-1-3-3" << std::endl;
			r_t_vals[0] = tmp;
		}
	}
	//std::cout << "local reduction(3 way)-2" << std::endl;
	for(int i = 1; i < mat->localRows; i++){
		oplus(c_t_vals[0], mat->at(i, 0), &tmp);
		c_t_vals[0] = tmp;
		r_t_vals[i] = mat->at(i,0);
		for(int j = 1; j < mat->localCols; j++){
			oplus(c_t_vals[j], mat->at(i, j), &tmp);
			c_t_vals[j] = tmp;
			otimes(r_t_vals[i], mat->at(i, j), &tmp);
			r_t_vals[i] = tmp;
		}
	}
	//std::cout << "local reduction(3 way)-3" << std::endl;
	*b_c_val = r_t_vals[0];
	for(int i = 1; i < mat->localRows; i++){
		oplus(*b_c_val, r_t_vals[i], &tmp);
		*b_c_val = tmp;
	}

	const int bicfloor = (mat->myrank/mat->bic)*mat->bic;
	const int birfloor = (mat->myrank%mat->bic);
	const int id_col = mat->myrank%mat->bic;
	const int id_row = mat->myrank/mat->bic;
	int maxstage = mat->bir;
	if(maxstage < mat->bic) maxstage = mat->bic;
	
	bool flag_first_col = true;
	bool flag_first_row = true;

	//std::cout << "global HC-like communication" << std::endl;
	// global HC-like communication
	for(int stage = 1; stage < maxstage; stage <<= 1){
		//std::cout << "stage = " << stage << std::endl;
		if(stage < mat->bic && (id_col^stage) < mat->bic){  // row-direction
			int target = (id_col^stage) + bicfloor;
			//std::cout << "row-direction: " << mat->myrank << " to " << target << std::endl;

			SKEL_MPI_Request reqs, reqr;
			if(Serializer<A>::needsSerializing()){
				//std::cout << "serializing & send" << std::endl;
				unsigned char* sbuf = mat->sendSerializingData(target, sendBufRow, sendRecvSizeRow, TAG_SCAN12s, TAG_SCAN12, reqs);
				//std::cout << "receive?" << std::endl;
				mat->recvSerializingData(target, recvBufRow, sendRecvSizeRow, TAG_SCAN12s, TAG_SCAN12);
				//std::cout << "waiting?" << std::endl;
				SKEL_MPI_Wait(reqs);	
				delete [] sbuf;
			} else {
				SKEL_MPI_Isend(sendBufRow, sendRecvSizeRow, mat->mpitype, target, TAG_SCAN12, mat->comm, reqs);
				SKEL_MPI_Irecv(recvBufRow, sendRecvSizeRow, mat->mpitype, target, TAG_SCAN12, mat->comm, reqr);
				SKEL_MPI_Wait(reqs);
				SKEL_MPI_Wait(reqr);
			}
			//std::cout << "send/recv done."  <<mat->myrank << " to " << target<< std::endl;
			if( target > mat->myrank ) {
				for(int i = 0; i < mat->localRows; i++){
					otimes(r_t_vals[i], r_t_valsR[i], &tmp);
					r_t_vals[i] = tmp;
				}
				otimes(*b_c_val, *b_c_valR, &tmp);
				*b_c_val = tmp;
				if(!flag_first_col){
					otimes(*b_n_val, *b_n_valR, &tmp);
					*b_n_val = tmp;
				}
			} else {
				if(flag_first_row){
					// this is inefficient...
					// receiving directly to r_f_vals is more efficient.
					for(int i = 0; i < mat->localRows; i++){
						r_f_vals[i] = r_t_valsR[i];
					}
					*b_w_val = *b_c_valR;
					if(!flag_first_col){
						*b_nw_val = *b_n_valR;
					}
					flag_first_row = false;
				} else {
					for(int i = 0; i < mat->localRows; i++){
						otimes(r_t_valsR[i], r_f_vals[i], &tmp);
						r_f_vals[i] = tmp;
					}
					otimes(*b_c_valR, *b_w_val, &tmp);
					*b_w_val = tmp;
					if(!flag_first_col){
						otimes(*b_n_valR, *b_nw_val, &tmp);
						*b_nw_val = tmp;
					}
				}
				for(int i = 0; i < mat->localRows; i++){
					otimes(r_t_valsR[i], r_t_vals[i], &tmp);
					r_t_vals[i] = tmp;
				}
				otimes(*b_c_valR, *b_c_val, &tmp);
				*b_c_val = tmp;
				if(!flag_first_col){
					otimes(*b_n_valR, *b_n_val, &tmp);
					*b_n_val = tmp;
				}
			}
		}
		///////////////////////////////////////////////////////////////////////////
		if(stage < mat->bir && (id_row^stage) < mat->bir){  // col-direction
			int target = (id_row^stage)*mat->bic + birfloor;
			//std::cout << "col-direction: " << mat->myrank << " to " << target << std::endl;

			SKEL_MPI_Request reqs, reqr;
			if(Serializer<A>::needsSerializing()){
				//std::cout << "serializing & send" << std::endl;
				unsigned char* sbuf = mat->sendSerializingData(target, sendBufCol, sendRecvSizeCol, TAG_SCAN12s, TAG_SCAN12, reqs);
				//std::cout << "receive?" << std::endl;
				mat->recvSerializingData(target, recvBufCol, sendRecvSizeCol, TAG_SCAN12s, TAG_SCAN12);
				//std::cout << "waiting?" << std::endl;
				SKEL_MPI_Wait(reqs);	
				delete [] sbuf;
			} else {
				SKEL_MPI_Isend(sendBufCol, sendRecvSizeCol, mat->mpitype, target, TAG_SCAN12, mat->comm, reqs);
				SKEL_MPI_Irecv(recvBufCol, sendRecvSizeCol, mat->mpitype, target, TAG_SCAN12, mat->comm, reqr);
				SKEL_MPI_Wait(reqs);
				SKEL_MPI_Wait(reqr);
			}
			//std::cout << "send/recv done." << mat->myrank << " to " << target<< std::endl;
			if( target > mat->myrank ) {
				//std::cout << "local calc-1" << std::endl;
				for(int i = 0; i < mat->localCols; i++){
					oplus(c_t_vals[i], c_t_valsR[i], &tmp);
					c_t_vals[i] = tmp;
				}
				oplus(*b_c_val, *b_c_valR, &tmp);
				*b_c_val = tmp;
				if(!flag_first_row){
					oplus(*b_w_val, *b_w_valR, &tmp);
					*b_w_val = tmp;
				}
			} else {
				if(flag_first_col){
					// this is inefficient...
					// receiving directly to c_f_vals is more efficient.
					for(int i = 0; i < mat->localCols; i++){
						c_f_vals[i] = c_t_valsR[i];
					}
					*b_n_val = *b_c_valR;
					if(!flag_first_row){
						*b_nw_val = *b_w_valR;
					}
					flag_first_col = false;
				} else {
					for(int i = 0; i < mat->localCols; i++){
						oplus(c_t_valsR[i], c_f_vals[i], &tmp);
						c_f_vals[i] = tmp;
					}
					oplus(*b_c_valR, *b_n_val, &tmp);
					*b_n_val = tmp;
					if(!flag_first_row){
						oplus(*b_w_valR, *b_nw_val, &tmp);
						*b_nw_val = tmp;
					}
				}
				for(int i = 0; i < mat->localCols; i++){
					oplus(c_t_valsR[i], c_t_vals[i], &tmp);
					c_t_vals[i] = tmp;
				}
				oplus(*b_c_valR, *b_c_val, &tmp);
				*b_c_val = tmp;
				if(!flag_first_row){
					oplus(*b_w_valR, *b_w_val, &tmp);
					*b_w_val = tmp;
				}
			}
		}
	}
	//std::cout << "local scan" << std::endl;
	// local scan
	// flag_first_row == true -> r_f_vals is empty, *b_nw_val is empty
	// flag_first_col == true -> c_f_vals is empty, *b_nw_val is empty
	//std::cout << "local scan-1" << mat->myrank << ", "<<flag_first_row <<flag_first_col << std::endl;
	//std::cout << "--" << mat->myrank << "*b_w_val = " << *b_w_val << "*b_n_val = " << *b_n_val << "*b_c_val = " << *b_c_val << "...." << std::endl;
	//std::cout << "--" << mat->myrank << "*b_nw_val = " << *b_nw_val << "..." <<std::endl;
	//std::cout << "--" << mat->myrank << "c_f_vals = ";
	//for(int i = 0; i < mat->localCols; i++) std::cout << c_f_vals[i] << ", ";
	//std::cout << ".." <<std::endl;
	//std::cout << "--" << mat->myrank << "r_f_vals = ";
	//for(int i = 0; i < mat->localRows; i++) std::cout << r_f_vals[i] << ", ";
	//std::cout << "." <<std::endl;
	for(int i = 0; i < 1; i++){
		if(!flag_first_col){
			oplus(c_f_vals[0], mat->at(i, 0), &tmp);
			c_f_vals[0] = tmp;
			if(!flag_first_row ){
				oplus(*b_nw_val, r_f_vals[i], &tmp);
				r_f_vals[i] = tmp;
				otimes(r_f_vals[i], c_f_vals[0], &mat2->at(i, 0));
			} else {
				mat2->at(i, 0) = c_f_vals[0];
			}
			for(int j = 1; j < mat->localCols; j++){
				oplus(c_f_vals[j], mat->at(i, j), &tmp);
				c_f_vals[j] = tmp;
				otimes(mat2->at(i, j-1), tmp, &mat2->at(i, j));
			}
		} else {
			c_f_vals[0] = mat->at(i, 0);
			if(!flag_first_row){
				otimes(r_f_vals[i], mat->at(i, 0), &mat2->at(i, 0));
			} else {
				mat2->at(i, 0) = mat->at(i, 0);
			}
			for(int j = 1; j < mat->localCols; j++){
				c_f_vals[j] = mat->at(i, j);
				otimes(mat2->at(i, j-1), mat->at(i, j), &mat2->at(i, j));
			}
		}
	}
	// now, c_f_vals is not empty
	//std::cout << "local scan-2" << mat->myrank << ", "<<flag_first_row <<flag_first_col << std::endl;
	for(int i = 1; i < mat->localRows; i++){
		oplus(c_f_vals[0], mat->at(i, 0), &tmp);
		c_f_vals[0] = tmp;
		if(!flag_first_row ){
			oplus(r_f_vals[i-1], r_f_vals[i], &tmp);
			r_f_vals[i] = tmp;
			otimes(r_f_vals[i], c_f_vals[0], &mat2->at(i, 0));
		} else {
			mat2->at(i, 0) = c_f_vals[0];
		}
		for(int j = 1; j < mat->localCols; j++){
			oplus(c_f_vals[j], mat->at(i, j), &tmp);
			c_f_vals[j] = tmp;
			otimes(mat2->at(i, j-1), tmp, &mat2->at(i, j));
		}
	}
	// done
	delete [] r_t_vals;
	delete [] r_t_valsR;
	delete [] r_f_vals;
}

template <typename A, typename F, typename G>
dist_matrix<A>*
matrix_skeletons::scan2(const F &oplus, const G &otimes, const dist_matrix<A>* mat)
{
	dist_matrix<A>* mat2 = new dist_matrix<A>();
	mat2->shapeCopy(mat);
	scan2(oplus, otimes, mat, mat2);
	return mat2;
}


#endif // __MATRIX_SKELETONS_TPP__
#endif // __MATRIX_SKELETONS_H__
