/***
 * Copyright (c) 2005, SkeTo Project
 * All rights reserved.
 */
/***
 * @file fnorm.cpp
 * @brief Nested skeletons test.  
 * @author Kento EMOTO
 */
#include <iostream>
#include "matrix_skeletons.h"
#include "primitive_functions.h"
using namespace primitive_functions;


struct Neg { int operator()(const int i) const { return -i; }};

/***
 * multiplication of local block-matrices.
 */
template <class C, class A, class B>
inline void MatrixMultiplyLocal(dist_matrix<C>& Z, const dist_matrix<A>& X, const dist_matrix<B>& Y)
{
	int rows = X.getLocalRows();
	int cols = Y.getLocalCols();
	int inner = X.getLocalCols();
	for(int i = 0; i < rows; i++){
		for(int k = 0; k < inner; k++){
			for(int j = 0; j < cols; j++){
				Z.at(i, j) = Z.at(i, j) + X.get(i, k) * Y.get(k, j);
			}
		}
	}
}

/***
 * multiplication of matrices.
 */
template <class C, class A, class B>
inline void MatrixMultiply(dist_matrix<C>& Z, dist_matrix<A>& X, dist_matrix<B>& Y)
{
	matrix_skeletons::rotateCols(Neg(), &X);          // Cannon's algorithm
	matrix_skeletons::rotateRows(Neg(), &Y);
	int q = X.getBlocksInRow();
	for(int i = 0; i < q; i++){
		MatrixMultiplyLocal(Z, X, Y);
		matrix_skeletons::rotateColsConst(-1, &X);
		matrix_skeletons::rotateRowsConst(-1, &Y);
	}
}



struct FMat : public skeleton::unary_function<double, matrix<double> >{
	int n;
	FMat(int _n) : n(_n) {}
	void operator()(const double&, matrix<double> *mat) const{
		mat->setSize(n, n);
		mat->eye(1.0, 0.0);
		for(int j = 0; j < n-1; j++){
			mat->at(j,j+1) = 1.0;
		}
	}
};

struct Oplus : public skeleton::binary_function<matrix<double>,matrix<double>,matrix<double> > {
	void operator()(const matrix<double> &x, const matrix<double> &y, matrix<double> *z) const {
		const int n = x.getRows();
		for(int i = 0; i < n; i++){
			for(int k = 0; k < n; k++){
				z->at(i,k) = 0;
			}
			for(int j = 0; j < n; j++){
				for(int k = 0; k < n; k++){
					z->at(i,k) = z->at(i,k) + x.at(i,j) * y.at(j,k);
				}
			}
		}
	}
	void operator()(const matrix<double> &x, const matrix<double> &y, matrix<double> *z, const SKEL_MPI_INTRACOMM &com) const {
		int mysize;
		SKEL_MPI_Comm_size(com, mysize);
		int myrank;
		SKEL_MPI_Comm_rank(com, myrank);
		int q = 0;
		for(q = 1; q*q <= mysize; q++);
		q--;
		if(q==1){
			//			std::cout << "MM with q = " << q << std::endl;
			if(myrank==0) operator()(x,y,z);
			return;
		}
		dist_matrix<double> X(&x,com,q,q);
		dist_matrix<double> Y(&y,com,q,q);
		dist_matrix<double> Z(x.getRows(), y.getCols(), GenZero<double>(),com,q,q);
		//std::cout << "MMP!!! with q = " << q << ", myrank = "<<myrank<< std::endl;
		if(myrank < q*q) MatrixMultiply(Z,X,Y);
		//std::cout << "MMP done with q = " << q << std::endl;
		Z.gather(z);
	}
};

int SketoMain( int argc, char **argv )
{
	int n = 100;
	if(argc > 1){
		n = atoi(argv[1]);
	}
	int m = 10;
	if(argc > 2){
		m = atoi(argv[2]);
	}
	int debug = 0;
	if(argc > 3){
		debug = atoi(argv[3]);
	}
	dist_matrix<int> dma(n,n);
	//	matrix<double> mat(dma.getBlocksInRow(), dma.getBlocksInCol());
	matrix<double> mat(m,m);
	dist_matrix<double> dmb(&mat);
	dist_matrix<matrix<double> > *dmc = matrix_skeletons::map(FMat(n), &dmb);
	skeleton::time::initialize( );
 	matrix<double> *res = matrix_skeletons::reduceDCN(Oplus(), Oplus(), dmc);
 	skeleton::time::set_time( "after computation DCN" );
	matrix<double> *res2 = matrix_skeletons::reduceDC(Oplus(), Oplus(), dmc);
	skeleton::time::set_time( "after computation DC" );
	matrix<double> *res1 = matrix_skeletons::reduce(Oplus(), Oplus(), dmc);
	skeleton::time::set_time( "after computation" );
	skeleton::time::print( );
	if(skeleton::rank == 0){
		std::cout << "(" << m << " x " << m << ") of (" << n << " x " << n << ") matrix in ("<<dma.getBlocksInRow() << " x "<< dma.getBlocksInCol() << ") procs" <<std::endl;
		if(debug){
			std::cout << "result = " << std::endl;
 			std::cout << *res << std::endl;
 			std::cout << *res1 << std::endl;
 			std::cout << *res2 << std::endl;
		}
	}
 	delete res2;
 	delete res1;
 	delete res;
	delete dmc;
	return 0;
}
