/***
 * Copyright (c) 2005, SkeTo Project
 * All rights reserved.
 */
/**
 * @file test_dist_rose_tree.cpp
 * @brief Test code for dist_rose_tree.h and dist_rose_tree.tpp
 * @author Kiminori MATSUZAKI
 */

#include <assert.h>
#include <iostream>

#include <skeleton.h>
#include <rose_tree_skeletons.h>

using namespace std;

struct func_const_one : public skeleton::unary_function< double, int > {
  int operator()( const double & ) const {
    return 1;
  }
};

struct func_add : public skeleton::binary_function< double, double, double > {
  double operator()( const double &lhs, const double &rhs ) const {
    return lhs + rhs;
  }
};
double unit_add = 0.0;

struct func_mult : public skeleton::binary_function< double, double, double > {
  double operator()( const double &lhs, const double &rhs ) const {
    return lhs * rhs;
  }
};
double unit_mult = 1.0;

struct func_add_different_types
  : public skeleton::binary_function< double, int, char > {
  char operator()( const double &rhs, const int &lhs ) const {
    return static_cast< char >( '0'+ ( static_cast< int >( rhs ) + lhs ) % 10 );
  }
};

/**
 * function objects for testing reduce (extended distributivity)
 *
 * These function objects generates a string-notation of trees...
 * From tree of the shape
 *      1
 *    / | \
 *   2  3  4
 *  /|
 * 5 6
 * these functions generate the following code 
 *   1<2<5<>6<>>3<>4<>
 * This corresponds to XML serialization problem that is difficult
 * at the following points
 *   - concat operation is associative but is not commutative.
 *   - tag-enclosing operation is not distributive over concat operation.
 */

/**
 * @brief structure for the strings.
 * @warning Functions in this structure is not safe for general purposes.
 */
struct x2s_t {
  static const int MAX_LENGTH = 1024;
  char str[ MAX_LENGTH ];
  x2s_t( ) {
    str[ 0 ] = '\0';
  }
  x2s_t( const char *str_ ) {
    strcpy( str, str_ );
  }
  x2s_t( double val ) {
    sprintf( str, "%d", static_cast< int >( val ) );
  }
  x2s_t( bool ) { // for internal implementation.
    str[ 0 ] = '\0';
  }

  x2s_t &append( const char *append_str ) {
    strcat( str, append_str );
    return *this;
  }
  x2s_t &append( double val ) {
    char buf[ MAX_LENGTH ];
    sprintf( buf, "%d", static_cast< int >( val ) );
    this->append( buf );
    return *this;
  }
  x2s_t &append( const x2s_t &append_str ) {
    this->append( append_str.str );
    return *this;
  }
};

std::ostream& operator<<( std::ostream& sout, const x2s_t& str )
{
  return sout << str.str;
}

struct func_x2s_tags
  : skeleton::unary_function< double, std::pair< x2s_t, x2s_t > > {
  std::pair< x2s_t, x2s_t > operator()( const double &val ) const {
    x2s_t retfst, retsnd;
    retfst.append( val ).append( "<" );
    retsnd.append( ">" );
    return std::pair< x2s_t, x2s_t >( retfst, retsnd );
  }
};

struct func_x2s_oplus
  : skeleton::binary_function< std::pair< x2s_t, x2s_t >, x2s_t, x2s_t > {
  x2s_t operator()( const std::pair< x2s_t, x2s_t > &lhs,
		    const x2s_t &rhs ) const {
    x2s_t retval;
    retval.append( lhs.first ).append( rhs ).append( lhs.second );
    return retval;
  }
};

std::pair< x2s_t, x2s_t > func_x2s_oplus_unit;

struct func_x2s_otimes
  : skeleton::binary_function< x2s_t, x2s_t, x2s_t > {
  x2s_t operator()( const x2s_t &lhs,
		    const x2s_t &rhs ) const {
    x2s_t retval;
    retval.append( lhs ).append( rhs );
    return retval;
  }
};

const x2s_t func_x2s_otimes_unit;

struct func_x2s_p1 {
  std::pair< x2s_t, x2s_t > operator()( const std::pair< x2s_t, x2s_t > &se1,
					const x2s_t &ta1,
					const x2s_t &tb1,
					const std::pair< x2s_t, x2s_t > &se2,
					const x2s_t &ta2,
					const x2s_t &tb2 ) const {
    x2s_t retfst, retsnd;
    retfst.append( se1.first ).append( ta1 ).append( se2.first ).append( ta2 );
    retsnd.append( tb2 ).append( se2.second ).append( tb1 ).append( se1.second );
    return std::pair< x2s_t, x2s_t >( retfst, retsnd );
  }
};

struct func_x2s_p2 {
  x2s_t operator()( const std::pair< x2s_t, x2s_t > &,
		    const x2s_t &,
		    const x2s_t &,
		    const std::pair< x2s_t, x2s_t > &,
		    const x2s_t &,
		    const x2s_t & ) const {
    return x2s_t( );
  }
};

struct func_x2s_p3 {
  x2s_t operator()( const std::pair< x2s_t, x2s_t > &,
		    const x2s_t &,
		    const x2s_t &,
		    const std::pair< x2s_t, x2s_t > &,
		    const x2s_t &,
		    const x2s_t & ) const {
    return x2s_t( );
  }
};

////////////////////////////////////////////////////////////////////////
// test code

void test_rose_tree_skeletons_map( )
{
  const char filename_in[]  = "dist_rose_tree_data.txt";
  const char filename_out[] = "result_test_rose_tree_skeletons_map.txt";
  
  dist_rose_tree< double > *rtree1
    = dist_rose_tree< double >::read_from_file( filename_in );
  assert( rtree1 );

  dist_rose_tree< int > *rtree2
    = rose_tree_skeletons::map( func_const_one( ), rtree1 );
  assert( rtree2 );

  assert( rtree2->write_to_file( filename_out ) );
  if ( skeleton::rank == 0 ) {
    std::cerr << "Check the output of map: " << filename_out
	      << std::endl;
  }

  if ( rtree2 ) delete rtree2;
  if ( rtree1 ) delete rtree1;
}

void test_rose_tree_skeletons_zipwith( )
{
  const char filename_in[]  = "dist_rose_tree_data.txt";
  const char filename_out[] = "result_test_rose_tree_skeletons_zipwith.txt";
  
  dist_rose_tree< double > *rtree1
    = dist_rose_tree< double >::read_from_file( filename_in );
  assert( rtree1 );

  dist_rose_tree< int > *rtree2
    = rose_tree_skeletons::map( func_const_one( ), rtree1 );
  assert( rtree2 );

  dist_rose_tree< char > *rtree3
    = rose_tree_skeletons::zipwith( func_add_different_types( ),
				    rtree1, rtree2 );
  
  assert( rtree3->write_to_file( filename_out ) );
  if ( skeleton::rank == 0 ) {
    std::cerr << "Check the output of zipwith: " << filename_out
	      << std::endl;
  }

  if ( rtree3 ) delete rtree3;
  if ( rtree2 ) delete rtree2;
  if ( rtree1 ) delete rtree1;
}

void test_rose_tree_skeletons_reduce_ring( )
{
  const char filename_in[]  = "dist_rose_tree_data.txt";
  //  const char filename_out[] = "result_test_rose_tree_skeletons_uAcc_assoc.txt";
  
  dist_rose_tree< double > *rtree1
    = dist_rose_tree< double >::read_from_file( filename_in );
  assert( rtree1 );

  double result
    = rose_tree_skeletons::reduce_ring( func_mult( ), unit_mult, 
					func_add( ), unit_add, rtree1 );

  if ( skeleton::rank == 0 ) {
    std::cerr << "The result of reduce (ring): " << result 
	      << std::endl;
  }

  if ( rtree1 ) delete rtree1;
}

void test_rose_tree_skeletons_reduce_exdist( )
{
  const char filename_in[]  = "dist_rose_tree_data.txt";
  //  const char filename_out[] = "result_test_rose_tree_skeletons_uAcc_assoc.txt";
  
  dist_rose_tree< double > *rtree1
    = dist_rose_tree< double >::read_from_file( filename_in );
  assert( rtree1 );

  dist_rose_tree< std::pair< x2s_t, x2s_t > > *rtree2
    = rose_tree_skeletons::map( func_x2s_tags( ), rtree1 );
  assert( rtree2 );

  x2s_t result
    = rose_tree_skeletons::reduce_exdist( func_x2s_oplus( ),
					  func_x2s_oplus_unit,
					  func_x2s_otimes( ),
					  func_x2s_otimes_unit,
					  func_x2s_p1( ),
					  func_x2s_p2( ),
					  func_x2s_p3( ),
					  rtree2 );

  if ( skeleton::rank == 0 ) {
    std::cerr << "The result of reduce (exdist): " << result 
	      << std::endl;
  }

  if ( rtree2 ) delete rtree2;
  if ( rtree1 ) delete rtree1;
}

void test_rose_tree_skeletons_reduce_assoc( )
{
  const char filename_in[]  = "dist_rose_tree_data.txt";
  //  const char filename_out[] = "result_test_rose_tree_skeletons_uAcc_assoc.txt";
  
  dist_rose_tree< double > *rtree1
    = dist_rose_tree< double >::read_from_file( filename_in );
  assert( rtree1 );

  double result
    = rose_tree_skeletons::reduce_assoc( func_add( ), unit_add, rtree1 );

  if ( skeleton::rank == 0 ) {
    std::cerr << "The result of reduce (assoc): " << result 
	      << std::endl;
  }

  if ( rtree1 ) delete rtree1;
}
  
void test_rose_tree_skeletons_uAcc_ring( )
{
  const char filename_in[]  = "dist_rose_tree_data.txt";
  const char filename_out[] = "result_test_rose_tree_skeletons_uAcc_ring.txt";
  
  dist_rose_tree< double > *rtree1
    = dist_rose_tree< double >::read_from_file( filename_in );
  assert( rtree1 );

  dist_rose_tree< double > *rtree2
    = rose_tree_skeletons::uAcc_ring( func_mult( ), unit_mult, 
					func_add( ), unit_add, rtree1 );
  assert( rtree2 );
  
  assert( rtree2->write_to_file( filename_out ) );
  if ( skeleton::rank == 0 ) {
    std::cerr << "Check the result of uAcc (ring): " << filename_out 
	      << std::endl;
  }

  if ( rtree2 ) delete rtree2;
  if ( rtree1 ) delete rtree1;
}

void test_rose_tree_skeletons_uAcc_exdist( )
{
  const char filename_in[]  = "dist_rose_tree_data.txt";
  const char filename_out[] = "result_test_rose_tree_skeletons_uAcc_exdist.txt";
  
  dist_rose_tree< double > *rtree1
    = dist_rose_tree< double >::read_from_file( filename_in );
  assert( rtree1 );

  dist_rose_tree< std::pair< x2s_t, x2s_t > > *rtree2
    = rose_tree_skeletons::map( func_x2s_tags( ), rtree1 );
  assert( rtree2 );

  dist_rose_tree< x2s_t > *rtree3
    = rose_tree_skeletons::uAcc_exdist( func_x2s_oplus( ),
					func_x2s_oplus_unit,
					func_x2s_otimes( ),
					func_x2s_otimes_unit,
					func_x2s_p1( ),
					func_x2s_p2( ),
					func_x2s_p3( ),
					rtree2 );
  assert( rtree3 );

  assert( rtree3->write_to_file( filename_out ) );
  if ( skeleton::rank == 0 ) {
    std::cerr << "Check the result of uAcc (exdist): " << filename_out
	      << std::endl;
  }

  if ( rtree3 ) delete rtree3;
  if ( rtree2 ) delete rtree2;
  if ( rtree1 ) delete rtree1;
}

void test_rose_tree_skeletons_uAcc_assoc( )
{
  const char filename_in[]  = "dist_rose_tree_data.txt";
  const char filename_out[] = "result_test_rose_tree_skeletons_uAcc_assoc.txt";
  
  dist_rose_tree< double > *rtree1
    = dist_rose_tree< double >::read_from_file( filename_in );
  assert( rtree1 );

  dist_rose_tree< double > *rtree2
    = rose_tree_skeletons::uAcc_assoc( func_add( ), unit_add, rtree1 );
  assert( rtree2 );

  assert( rtree2->write_to_file( filename_out ) );
  if ( skeleton::rank == 0 ) {
    std::cerr << "Check the output of uAcc (assoc): " << filename_out
	      << std::endl;
  }

  if ( rtree2 ) delete rtree2;
  if ( rtree1 ) delete rtree1;
}
  
void test_rose_tree_skeletons_dAcc( )
{
  const char filename_in[]  = "dist_rose_tree_data.txt";
  const char filename_out[] = "result_test_rose_tree_skeletons_dAcc.txt";
  
  dist_rose_tree< double > *rtree1
    = dist_rose_tree< double >::read_from_file( filename_in );
  assert( rtree1 );

  dist_rose_tree< double > *rtree2
    = rose_tree_skeletons::dAcc( func_add( ), unit_add, rtree1 );
  assert( rtree2 );

  assert( rtree2->write_to_file( filename_out ) );
  if ( skeleton::rank == 0 ) {
    std::cerr << "Check the output of dAcc: " << filename_out
	      << std::endl;
  }

  if ( rtree2 ) delete rtree2;
  if ( rtree1 ) delete rtree1;
}

void test_rose_tree_skeletons_rAcc( )
{
  const char filename_in[]  = "dist_rose_tree_data.txt";
  const char filename_out[] = "result_test_rose_tree_skeletons_rAcc.txt";
  
  dist_rose_tree< double > *rtree1
    = dist_rose_tree< double >::read_from_file( filename_in );
  assert( rtree1 );

  dist_rose_tree< double > *rtree2
    = rose_tree_skeletons::rAcc( func_add( ), unit_add, rtree1 );
  assert( rtree2 );

  assert( rtree2->write_to_file( filename_out ) );
  if ( skeleton::rank == 0 ) {
    std::cerr << "Check the output of rAcc: " << filename_out
	      << std::endl;
  }

  if ( rtree2 ) delete rtree2;
  if ( rtree1 ) delete rtree1;
}

void test_rose_tree_skeletons_lAcc( )
{
  const char filename_in[]  = "dist_rose_tree_data.txt";
  const char filename_out[] = "result_test_rose_tree_skeletons_lAcc.txt";
  
  dist_rose_tree< double > *rtree1
    = dist_rose_tree< double >::read_from_file( filename_in );
  assert( rtree1 );

  dist_rose_tree< double > *rtree2
    = rose_tree_skeletons::lAcc( func_add( ), unit_add, rtree1 );
  assert( rtree2 );

  assert( rtree2->write_to_file( filename_out ) );
  if ( skeleton::rank == 0 ) {
    std::cerr << "Check the output of lAcc: " << filename_out
	      << std::endl;
  }

  if ( rtree2 ) delete rtree2;
  if ( rtree1 ) delete rtree1;
}

int SketoMain( int, char ** )
{
  if ( skeleton::rank == 0 ) {
    cout << "Testing " __FILE__ << endl;
  }

  test_rose_tree_skeletons_map( );
  test_rose_tree_skeletons_zipwith( );
  test_rose_tree_skeletons_reduce_ring( );
  test_rose_tree_skeletons_reduce_exdist( );
  test_rose_tree_skeletons_reduce_assoc( );
  test_rose_tree_skeletons_uAcc_ring( );
  test_rose_tree_skeletons_uAcc_exdist( );
  test_rose_tree_skeletons_uAcc_assoc( );
  test_rose_tree_skeletons_dAcc( );
  test_rose_tree_skeletons_rAcc( );
  test_rose_tree_skeletons_lAcc( );
  
  if ( skeleton::rank == 0 ) {
    cout << "   ---> OK " __FILE__ << endl;
  }
  return 0;
}
