/***
 * Copyright (c) 2005, SkeTo Project
 * All rights reserved.
 */
/**
 * @file test_tree_skeletons.cpp
 * @brief Test code for tree_skeletons.h and tree_skeletons.tpp
 * @author Kiminori MATSUZAKI
 */

#include <skeleton.h>
#include <tree_skeletons.h>

using namespace std;

////////////////////////////////////////////////////////////////////////
// Functions for testing skeletons

// function objects for testing map. 
struct FuncDB : skeleton::unary_function< double, int > {
  int operator()( const double& x ) const {
    return static_cast< int >( 2.0 * x );
  }
};

struct FuncSQ : skeleton::unary_function< double, int > {
  int operator()( const double& x ) const {
    return static_cast< int >( x * x );
  }
};

// function objects for testing zip. 
struct Func_zip_k1 : skeleton::binary_function< double, int, char > {
  char operator()( const double& a, const int& b ) const {
    return ( a == ( double )b ) ? 'T' : 'F';
  }
};

struct Func_zip_k2 : skeleton::binary_function< double, int, char > {
  char operator()( const double& a, const int& b ) const {
    return ( a <= ( double )b ) ? 'S' : 'L';
  }
};

// function objects for testing reduce and uAcc. 
struct Func_reduce_k1 : skeleton::unary_function< double, double > {
  double operator()( const double& a ) const {
    return a;
  }
};

struct Func_reduce_k2 : skeleton::ternary_function< double, double, double, double > {
  double operator()( const double& b, const double& l, const double& r ) const {
    return b+l+r;
  }
};

struct Func_reduce_psi : skeleton::unary_function< double, double > {
  double operator()( const double& b ) const {
    return b;
  }
};

struct Func_reduce_phiL : skeleton::ternary_function< double, double, double, double > {
  double operator()( const double& p, const double& l, const double& r ) const {
    return p+l+r;
  }
};

struct Func_reduce_phiR : skeleton::ternary_function< double, double, double, double > {
  double operator()( const double& p, const double& r, const double& l ) const {
    return p+r+l;
  }
};

struct Func_reduce_G : skeleton::ternary_function< double, double, double, double > {
  double operator()( const double& n, const double& l, const double& r ) const {
    return n+l+r;
  }
};

struct Func_dAcc_oplus : skeleton::binary_function< int, int, int > {
  int operator()( const int &lhs, const int &rhs ) const {
    return lhs + rhs;
  }
};

const int dAcc_e_oplus = 0;

struct Func_dAcc_gl : skeleton::unary_function< double, int > {
  int operator()( const double b ) const {
    return static_cast< int >( b );
  }
};

struct Func_dAcc_gr : skeleton::unary_function< double, int > {
  int operator()( const double b ) const {
    return static_cast< int >( -b );
  }
};

////////////////////////////////////////////////////////////////////////
// test codes

void test_tree_skeletons_map( )
{
  const char filename_in[]  = "dist_tree_data.txt";
  const char filename_out[] = "result_test_tree_skeletons_map.txt";
  
  dist_tree< double >* tree1
    = dist_tree< double >::read_from_file( filename_in );
  assert( tree1 );
  
  dist_tree< int > *tree2
    = tree_skeletons::map( FuncDB( ), FuncSQ( ), tree1 );
  tree2->write_to_file( filename_out );
  
  if ( skeleton::rank == 0 ) {
    std::cerr << "Check the result of map( db, sq ): " << filename_out
	      << std::endl;
  }

  if ( tree2 ) delete tree2;
  if ( tree1 ) delete tree1;
}

void test_tree_skeletons_zip( )
{
  const char filename_in[]  = "dist_tree_data.txt";
  const char filename_out[] = "result_test_tree_skeletons_zip.txt";
  
  dist_tree< double >* tree_double
    = dist_tree< double >::read_from_file( filename_in );    
  dist_tree< int >* tree_int
    = tree_skeletons::map( FuncDB( ), FuncSQ( ), tree_double );
  assert( tree_double );
  assert( tree_int );
  
  dist_tree< char > *tree_char
    = tree_skeletons::zipwith( Func_zip_k1( ), Func_zip_k2( ),
				       tree_double, tree_int );
  assert( tree_char );

  tree_char->write_to_file( filename_out );
  if ( skeleton::rank == 0 ) {
    std::cerr << "Check the result of zip: " << filename_out << std::endl;
  }

  if ( tree_char ) delete tree_char;
  if ( tree_int ) delete tree_int;
  if ( tree_double ) delete tree_double;
}

void test_tree_skeletons_reduce( )
{
  const char filename[] = "dist_tree_data.txt";
  dist_tree< double > *tree
    = dist_tree< double >::read_from_file( filename );
  assert( tree );
  
  double sum
    = tree_skeletons::reduce( Func_reduce_k1( ),
			      Func_reduce_k2( ),
			      Func_reduce_psi( ),
			      Func_reduce_phiL( ),
			      Func_reduce_phiR( ),
			      Func_reduce_G( ),
			      tree );

  if ( skeleton::rank == 0 ) {
    std::cerr << "result of reduce is " << sum << std::endl;
  }

  if ( tree ) delete tree;
}

void test_tree_skeletons_uAcc( )
{
  const char filename_in[]  = "dist_tree_data.txt";
  const char filename_out[] = "result_test_tree_skeletons_uAcc.txt";

  dist_tree< double > *tree1
    = dist_tree< double >::read_from_file( filename_in );
  assert( tree1 );

  dist_tree< double > *tree2
    = tree_skeletons::uAcc( Func_reduce_k1( ),
			    Func_reduce_k2( ),
			    Func_reduce_psi( ),
			    Func_reduce_phiL( ),
			    Func_reduce_phiR( ),
			    Func_reduce_G( ),
			    tree1 );
  tree2->write_to_file( filename_out );

  if ( skeleton::rank == 0 ) {
    std::cerr << "Check the result of uAcc: " << filename_out
	      << std::endl;
  }
  
  if ( tree2 ) delete tree2;
  if ( tree1 ) delete tree1;
}

void test_tree_skeletons_dAcc( )
{
  const char filename_in[]  = "dist_tree_data.txt";
  const char filename_out[] = "result_test_tree_skeletons_dAcc.txt";

  dist_tree< double > *tree1
    = dist_tree< double >::read_from_file( filename_in );
  assert( tree1 );

  dist_tree< int > *tree2
    = tree_skeletons::dAcc( Func_dAcc_oplus( ), dAcc_e_oplus,
			    Func_dAcc_gl( ), Func_dAcc_gr( ),
			    100, tree1 );
  tree2->write_to_file( filename_out );

  if ( skeleton::rank == 0 ) {
    std::cerr << "Check the result of dAcc: " << filename_out
	      << std::endl;
  }

  if ( tree2 ) delete tree2;
  if ( tree1 ) delete tree1;
}

void test_tree_skeletons_distribute_gather( )
{
  const char filename_in[]  = "dist_tree_data.txt";
  const char filename_out[]
    = "result_test_tree_skeletons_distribute_gather.txt";

  dist_tree< double >* tree1
    = dist_tree< double>::read_from_file( filename_in );
  assert( tree1 );

  node< double >* tree2 = tree_skeletons::gather( tree1 );
  if ( skeleton::rank == 0 ) {
    assert( tree2 );
  }

//   if ( skeleton::rank == 0 ) {
//     assert( tree_util::write_to_file( tree2, filename_out ) );
//   }

  dist_tree< double >* tree3 = tree_skeletons::distribute2( tree2 );
//  dist_tree< double >* tree3 = tree_skeletons::distribute( tree2 );
  assert( tree3 );
  assert( tree3->write_to_file( filename_out ) );
  if ( skeleton::rank == 0 ) {
    std::cerr << "Check the result of distribute and gather: " << filename_out
	      << std::endl;
  }
  
  delete tree3;
  tree_util::free_all_tree( tree2 );
  delete tree1;
}

void test_tree_skeletons_shiftp( )
{
  const char filename_in[]  = "dist_tree_data.txt";
  const char filename_out[]
    = "result_test_tree_skeletons_shiftp.txt";

  dist_tree< double >* tree1
    = dist_tree< double >::read_from_file( filename_in );
  assert( tree1 );

  dist_tree< double >* tree2
    = tree_skeletons::shiftp( 1.2345, tree1 );
  assert( tree2 );

  assert( tree2->write_to_file( filename_out ) );
  if ( skeleton::rank == 0 ) {
    std::cerr << "Check the result of shiftp: " << filename_out
	      << std::endl;
  }
  
  delete tree2;
  delete tree1;
}

void test_tree_skeletons_shiftl( )
{
  const char filename_in[]  = "dist_tree_data.txt";
  const char filename_out[]
    = "result_test_tree_skeletons_shiftl.txt";

  dist_tree< double >* tree1
    = dist_tree< double >::read_from_file( filename_in );
  assert( tree1 );

  dist_tree< double >* tree2
    = tree_skeletons::shiftl( 1.2345, tree1 );
  assert( tree2 );

  assert( tree2->write_to_file( filename_out ) );
  if ( skeleton::rank == 0 ) {
    std::cerr << "Check the result of shiftl: " << filename_out
	      << std::endl;
  }
  
  delete tree2;
  delete tree1;
}

void test_tree_skeletons_shiftr( )
{
  const char filename_in[]  = "dist_tree_data.txt";
  const char filename_out[]
    = "result_test_tree_skeletons_shiftr.txt";

  dist_tree< double >* tree1
    = dist_tree< double >::read_from_file( filename_in );
  assert( tree1 );

  dist_tree< double >* tree2
    = tree_skeletons::shiftr( 1.2345, tree1 );
  assert( tree2 );

  assert( tree2->write_to_file( filename_out ) );
  if ( skeleton::rank == 0 ) {
    std::cerr << "Check the result of shiftr: " << filename_out
	      << std::endl;
  }
  
  delete tree2;
  delete tree1;
}

int SketoMain( int, char *[] )
{
  skeleton::cout << "Testing " __FILE__ << endl;
  
  test_tree_skeletons_map( );
  test_tree_skeletons_zip( );
  test_tree_skeletons_reduce( );
  test_tree_skeletons_uAcc( );
  test_tree_skeletons_dAcc( );

  test_tree_skeletons_distribute_gather( );

  test_tree_skeletons_shiftp( );
  test_tree_skeletons_shiftl( );
  test_tree_skeletons_shiftr( );
  
  skeleton::cout << "   ---> OK " __FILE__ << endl;
  return 0;
}
