/***
 * Copyright (c) 2005, SkeTo Project
 * All rights reserved.
 */
/**
 * @file compose.h
 * @brief The header for compose functions.
 * @author Yoshiki AKASHI (yoshiki@ipl.cs.uec.ac.jp)
 */

#ifndef __COMPOSE_H__
#define __COMPOSE_H__

#include "functions.h"

namespace skeleton {
  template< typename F, typename G >
  class composite_unary_function
    : public skeleton::unary_function< typename G::argument_type,
				       typename F::result_type >
  {
    F f;
    G g;
  public:
    composite_unary_function( const F &f_, const G &g_ ) : f( f_ ), g( g_ )
    { }
    typename F::result_type
    operator()( typename G::argument_type x ) const
    {
      return f( g( x ) );
    }
  };
  
  template< typename F, typename G >
  class composite_unary_ptr_function
    : public skeleton::unary_ptr_function< typename F::first_argument_type,
					   typename G::second_argument_type >
  {
    F f;
    G g;
  public:
    composite_unary_ptr_function( const F &f_, const G &g_ ) : f( f_ ), g( g_ )
    { }
    void
    operator()( typename F::first_argument_type        *r,
		const typename G::second_argument_type *x ) const
    {
      typename G::first_argument_type a;
      g( &a, x );
      f( r, &a );
    }
  };
  
  template< typename F, typename G >
  class composite_binary_function
    : public skeleton::binary_function< typename G::first_argument_type,
					typename G::second_argument_type,
					typename F::result_type >
  {
    F f;
    G g;
  public:
    composite_binary_function( const F &f_, const G &g_ ) : f( f_ ), g( g_ )
    {}
    typename F::result_type
    operator()( typename G::first_argument_type  x,
		typename G::second_argument_type y ) const
    {
      return f( g( x, y ) );
    }
  };
  
  template< typename F, typename G >
  class composite_binary_ptr_function
    : public skeleton::binary_function< typename F::first_argument_type,
					typename G::second_argument_type,
					typename G::third_argument_type >
  {
    F f;
    G g;
  public:
    composite_binary_ptr_function( const F &f_, const G &g_ ) : f( f_ ), g( g_ )
    {}
    void
    operator()( typename F::first_argument_type        *r,
		const typename G::second_argument_type *x,
		const typename G::third_argument_type  *y ) const
    {
      typename F::first_argument_type a;
      g( &a, x, y );
      f( r, &a );
    }
  };
  
  template< typename F, typename G, typename H >
  class composite_binary_two_function
    : public skeleton::binary_function< typename G::argument_type,
					typename H::argument_type,
					typename F::result_type >
  {
    F f;
    G g;
    H h;
  public:
    composite_binary_two_function( const F &f_, const G &g_, const H &h_ ) : f( f_ ), g( g_ ), h( h_ )
    {}
    typename F::result_type
    operator()( typename G::argument_type x,
		typename H::argument_type y ) const
    {
      return f( g( x ), h( y ) );
    }
  };
  
  template< typename F, typename G, typename H >
  class composite_binary_two_ptr_function
    : public skeleton::binary_function< typename F::first_argument_type,
					typename G::second_argument_type,
					typename H::second_argument_type >
  {
    F f;
    G g;
    H h;
  public:
    composite_binary_two_ptr_function( const F &f_, const G &g_, const H &h_ ) : f( f_ ), g( g_ ), h( h_ )
    {}
    void
    operator()( typename F::first_argument_type        *r,
		const typename G::second_argument_type *x,
		const typename H::second_argument_type *y ) const
    {
      typename G::first_argument_type a;
      typename H::first_argument_type b;
      g( &a, x );
      h( &b, y );
      f( r, &a, &b );
    }
  };
  
  template< typename F, typename G >
  inline skeleton::composite_unary_function< F, G >
  compose_unary_functions( const F &f, const G &g )
  {
    return skeleton::composite_unary_function< F, G >( f, g );
  }
  
  template< typename F, typename G >
  inline skeleton::composite_unary_ptr_function< F, G >
  compose_unary_ptr_functions( const F &f, const G &g )
  {
    return skeleton::composite_unary_ptr_function< F, G >( f, g );
  }
  
  template< typename F, typename G >
  inline skeleton::composite_binary_function< F, G >
  compose_binary_functions( const F &f, const G &g )
  {
    return skeleton::composite_binary_function< F, G >( f, g );
  }
  
  template< typename F, typename G >
  inline skeleton::composite_binary_ptr_function< F, G >
  compose_binary_ptr_functions( const F &f, const G &g )
  {
    return skeleton::composite_binary_ptr_function< F, G >( f, g );
  }
  
  template< typename F, typename G, typename H >
  inline skeleton::composite_binary_two_function< F, G, H >
  compose_binary_two_functions( const F &f, const G &g, const H &h )
  {
    return skeleton::composite_binary_two_function< F, G, H >( f, g, h );
  }
  
  template< typename F, typename G, typename H >
  inline skeleton::composite_binary_two_ptr_function< F, G, H >
  compose_binary_two_ptr_functions( const F &f, const G &g, const H &h )
  {
    return skeleton::composite_binary_two_ptr_function< F, G, H >( f, g, h );
  }
}

#endif // __COMPOSE_H__
