/***
 * Copyright (c) 2005, SkeTo Project
 * All rights reserved.
 */
/**
 * @mainpage
 *
 * SkeTo (Skeletons in Tokyo) library provides many parallel skeletons
 * for lists, matrices, and trees. A parallel skeleton is a
 * computation pattern which is often used in parallel programming,
 * and is implemented in parallel efficiently. These parallel
 * skeletons gives a more straightforward way to compose parallel
 * programs.
 *
 * Quick links.
 *
 * - List skeletons:
 *  - @a dist_list
 *  - @a list_skeletons
 * - Matrix skeletons:
 *  - @a dist_matrix
 *  - @a matrix_skeletons
 * - Tree skeletons:
 *  - @a dist_tree
 *  - @a dist_rose_tree
 *  - @a tree_skeletons
 *  - @a rose_tree_skeletons
 * - Utilities
 *  - @a skeleton.h
 *  - @a functions.h
 *
 * Copyright (c) 2005, SkeTo Project.  All rights reserved.
 */
 
/**
 * @file skeleton.h
 * @brief The skeleton common file.
 * @author Yoshiki AKASHI (yoshiki@ipl.cs.uec.ac.jp)
 */

#ifndef __SKELETON_H__
#define __SKELETON_H__

#include <string>
#include <cstdio>
#include <iostream>
#include <fstream>
#include <cstdarg>
#include <vector>
#include <utility>
#include <mpi.h>
#include "compose.h"
#include "produce.h"
#include "resolve.h"
#include "exceptions.h"

namespace skeleton {
  extern int rank;  // MPI rank
  extern int procs; // Number of processors

  void init(int &argc, char **&argv);
  void finalize();

  bool is_master( );
  void mprintf( const char* format, ... );
  template< class A >
  inline void safe_delete( A *a )
  {
    if ( a ) delete a;
  }

  template< class A >
  inline void safe_delete_array( A *a )
  {
    if ( a ) delete[] a;
  }

  namespace time
  {
    void initialize( );
    void set_time( const char* name, bool barrier = true );
    void print( FILE *fp = stdout );
  };
  class sketoOstream : public std::ofstream {
   public:
    // for iomanip
 #if __GNUC__ > 2
   sketoOstream&
    operator<<(__ostream_type& (*__pf)(__ostream_type&)){
      if(rank==0) std::cout << __pf;
      return *this;
    }
    sketoOstream&
    operator<<(__ios_type& (*__pf)(__ios_type&)){
      if(rank==0) std::cout << __pf;
      return *this;
    }
    sketoOstream&
      operator<<(ios_base& (*__pf) (ios_base&)){
      if(rank==0) std::cout << __pf;
      return *this;
    }
#else
    sketoOstream&
      operator<<(ios& (*__pf) (ios&)){
      if(rank==0) std::cout << __pf;
      return *this;
    }
#endif
    // for other types
    template <typename C>
    sketoOstream& operator<<(const C&v){
      if(rank==0) std::cout << v;
      return *this;
    }
  };
  extern sketoOstream cout;
};

// class Main{
// public:
//   static int main(int argc, char **argv);
// };
int SketoMain( int argc, char **argv );

#endif // __SKELETON_H__
