/***
 * Copyright (c) 2005, SkeTo Project
 * All rights reserved.
 */
 
/**
 * @file skeleton.h
 * @brief The skeleton common file.
 * @author Yoshiki AKASHI (yoshiki@ipl.cs.uec.ac.jp)
 */

#include "skeleton.h"

namespace skeleton {
  int rank;  // MPI rank
  int procs; // Number of processors
  sketoOstream cout;

  void init(int &argc, char **&argv)
  {
    MPI_Init(&argc, &argv);
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    MPI_Comm_size(MPI_COMM_WORLD, &procs);
  }

  void finalize()
  {
    MPI_Finalize();
  }

  bool is_master( )
  {
    return ( rank == 0 );
  }

  void mprintf( const char* format, ... )
  {
    if ( !is_master( ) ) return;
    va_list args;
    va_start( args, format );
    vprintf( format, args );
    va_end( args );
  }
  namespace time
  {
    std::vector< std::pair< double, std::string > > time_table;
    void initialize( )
    {
      if ( is_master( ) ) time_table.clear( );
      MPI_Barrier( MPI_COMM_WORLD );
      if ( is_master( ) ) time_table.push_back( std::pair< double, std::string >( MPI_Wtime( ), std::string( "" ) ) );
    }

    void set_time( const char* name, bool barrier)
    {
      if ( barrier ) MPI_Barrier( MPI_COMM_WORLD );
      if ( is_master( ) ) time_table.push_back( std::pair< double, std::string >( MPI_Wtime( ), std::string( name ) ) );
    }

    void print( FILE *fp )
    {
      if ( !is_master( ) ) return;
      fprintf(fp, "%10s %10s %s\n", "total", "difference", "comment");
      for (unsigned int i = 1; i < time_table.size(); i++) {
	fprintf(fp, "%10.6f %10.6f %s\n",
		time_table[i].first - time_table[0].first,
		time_table[i].first - time_table[i-1].first,
		time_table[i].second.c_str());
      }
    }
  }
};

int main(int argc, char **argv)
{
  skeleton::init(argc, argv);

  int retval = 0;
  try {
    retval = SketoMain( argc, argv );
  }
  catch (skeleton::bad_size e) {
    std::cerr << "bad_size exception: (" << e.size_a << ", " << e.size_b << ")" << std::endl;
    retval = -1;
  }
  catch (skeleton::error e) {
    std::cerr << "Error: " << e.msg << std::endl;
    retval = e._errno;
  }     

  skeleton::finalize();
  return retval;
}

#include "../matrix/matrix_skeletons.h"
int matrix_skeletons::distributionPolicy = matrix_skeletons::DPOLICY_RECTANGLE;
void
matrix_skeletons::calcDivision(int &r, int &c, const SKEL_MPI_INTRACOMM &com)
{
	int np;
    SKEL_MPI_Comm_size(com, np);
	switch(distributionPolicy){
		case DPOLICY_SQUARE:{
			int k;
			for(k = 1; k*k <= np; k++);
			k--;
			r = k;
			c = k;}
			break;
		case DPOLICY_ROW:
			r = np;
			c = 1;
			break;
		case DPOLICY_COLUMN:
			r = 1;
			c = np;
			break;
		case DPOLICY_RECTANGLE:{
			int k;
			for(k = 1; k*k <= np; k++);
			k--;
			int l;
			for(l = k; l > 0; l--){
				if((np % l) == 0){
					break;
				}
			}
			r = np / l;
			c = l;}
			break;
		default:
		break;
	}
}

int
matrix_skeletons::setDistributionPolicy(int p)
{
	int o = distributionPolicy;
	distributionPolicy = p;
	return o;
}

int
matrix_skeletons::getDistributionPolicy()
{
	return distributionPolicy;
}


