/***
 * Copyright (c) 2005, SkeTo Project
 * All rights reserved.
 */
/**
 * @file anova.cpp
 * @brief This program compute ANOVA.
 * @author: Hideo NISHIMURA (nish@ipl.cs.uec.ac.jp)
 */

//-----------------------------//
//         How to use          //
//-----------------------------//
//
// This is One-way ANOVA program which outputs anova table.
// ( If you want to know about ANOVA, ask it of GOOGLE. )
// 
//  Argument of console command :  %s size number
//     "size" is the number of elemnts of each groups
//     "number" is the number of groups
//

//-----------------------------//
//           Compile           //
//-----------------------------//
//
//  put C++ parallel library and "anova.cpp" in the same directory
//
//     $ mpiCC anova.cpp
//
//  to execute
//     $ mpirun -np "N" a.out "size" "number"

#include "list_skeletons.h"
#include <cstdlib>

//---------------------------------//
//  functions using with Skeleton  //
//---------------------------------//

//--- function that generates proper data
struct Gen : public skeleton::unary_function< int, double > {
  double operator() ( int index ) const {
    //--- return [0,99] data generated at random
    return static_cast< double >( ( rand( ) % 100 ) );
  }
} generator;

//--- dyadic ADD operator function
template < typename A >
struct Add : public skeleton::binary_function< A, A, A > {
  A operator() ( A x, A y ) const { return x + y; }
};

//--- function that squares an argument
template < typename A >
struct Square : public skeleton::unary_function< A, A > {
  A operator() ( A x ) const { return x * x; }
};

//--- function that subtracts member variable ( given when constructed ) 
//    from an argument
struct Sub_ave : public skeleton::unary_function< double, double > {
  double ave_;
  // constructer with an argument to give a variable
  Sub_ave( double a ) { ave_ = a; }
  // main content of function
  double operator() ( double x ) const { return x - ave_; }
};

//------------------------//
//     main function      //
//------------------------//
int SketoMain(int argc, char **argv)
{ 
  if ( argc < 3 ) {
    skeleton::cout << "usage: " << argv[ 0 ] << " size number" << std::endl;
    return 0;
  }

  const int size = std::atoi( argv[ 1 ] );   // "size" is the number of elements of each group
  const int number = std::atoi( argv[ 2 ] ); // "number" is the number of groups

  skeleton::time::initialize( ); // for time ekeeping

  double ave = 0.0;
  Add< double > add;                    // make an instance of an add operator object
  const double zero = 0.0;              // constant variable which means zero
  double *ave_a = new double[ number ]; // store average of each groups in "ave_a"
  // for each dist_list "*a[ i ]", generates datas
  dist_list< double > **a = new dist_list< double >*[ number ];

  for ( int i = 0; i < number; i++ ) {
    a[ i ] = new dist_list< double >( generator, size );
    ave_a[ i ] = list_skeletons::reduce( add, zero, a[ i ] ) / size;
    ave += ave_a[ i ]; // store sum of averages of each groups in ave
  }

  // divides by number to store average of all elements in "ave"
  ave /= number;

  double sm = 0.0;
  double st = 0.0;
  Sub_ave sub_ave( ave );  // make an instance of sub "ave" operator object
  Square< double > square; // make an instance of square operator object
  for ( int i = 0; i < number; i++ ) {
    double a_ave = ave_a[ i ] - ave;
    sm += a_ave * a_ave;
    // below, get variations of each group
    list_skeletons::map_ow( sub_ave, a[ i ] );
    list_skeletons::map_ow( square, a[ i ] );
    // store sum of square of each group in "st" temporary
    st += list_skeletons::reduce( add, zero, a[ i ] );
  }

  // multiple by size to store between group sum of squares in "sm"
  sm *= size;

  for ( int i = 0; i < number; i++ ) {
    delete a[i];
  }
  delete [] a;
  delete [] ave_a;

  // store within group sum of squares in "se"
  double se = st - sm;

  // store
  //   total degrees of freedom in "dft"
  //   between group degrees of freedom in "dfm"
  //   within group degrees of freedom in "dfe"
  int dft = number * size - 1;
  int dfm = number - 1;
  int dfe = number * ( size - 1 );

  // store
  //   between group mean square in "vm"
  //   within group mean square in "ve"
  double vm = sm / dfm;
  double ve = se / dfe;

  skeleton::time::set_time( "calculate" ); // for time ekeeping

  // output anova table
  skeleton::cout << "factor" << "\t" << "SS" << "\t" << "DF" << "\t" << "MS" << "\t" << "F"   << std::endl
		 << "source" << "\t" << sm   << "\t" << dfm  << "\t" << vm   << "\t" << vm/ve << std::endl
		 << "error"  << "\t" << se   << "\t" << dfe  << "\t" << ve                    << std::endl
		 << "total"  << "\t" << st   << "\t" << dft  << "\t"                          << std::endl;

  skeleton::time::print( ); // show the result of time ekeeping

  return 0;
}
