/***
 * Copyright (c) 2005, SkeTo Project
 * All rights reserved.
 */
/**
 * @file bracket_matching.cpp
 * @brief This program check bracket matching.
 * @author Yoshiki AKASHI (yoshiki@ipl.cs.uec.ac.jp)
 * @attention: If brackets does not match, this program may freeze.
 */

#include <iostream>
#include <cstdlib>
#include <ctime>
#include <stack>
#include "list_skeletons.h"

const int KIND = 4;
const int DEPTH = 10;
const int MAXDEPTH = 10;
const char openbr[] = "({[<";
const char closebr[] = ")}]>";

//---------------------------------------------------------------------
// Stack
class Stack {
public:
  char in;
  char out;
  char data[ MAXDEPTH ];
  Stack( ): in( 0 ), out( 0 ) { }
  Stack( const Stack &s ): in( s.in ), out( s.out ) {
    for ( int i = 0; i < in; i++ ) {
      data[ i ] = s.data[ i ];
    }
  } 
  void push( char c ) {
    data[ in++ ] = c;
  }
  void pop( ){
    ( in > 0 ) ? in-- :  out++;
  }
  char top( ) const {
    return ( in > 0 ) ? data[ in-1 ] : 0;
  }
  int isempty( ) const {
    return ( in == 0 ) && ( out == 0 );
  }
  Stack& operator=( const Stack &s ) {
    if ( this != &s ) {
      in = s.in;
      out = s.out;
      for ( int i = 0; i < in; i++ ) {
	data[ i ] = s.data[ i ];
      }
    }
    return *this;
  }
};

template< >
class template_mpitype< Stack > {
public:
  static const MPI_Datatype type( ) {
    MPI_Datatype newtype;
    MPI_Type_contiguous( MAXDEPTH+2, MPI_CHAR, &newtype );
    MPI_Type_commit( &newtype );
    return newtype;
  }
};

//---------------------------------------------------------------------
// function declaration
int ran( int n );
void makestr( char *array, int size, int kind, int maxdepth );
int isopen( char c );
int isclose( char c );

//---------------------------------------------------------------------
// function definition
inline int ran( int n ) {
  return rand( ) % n;
}

inline void makestr( char *array, int size, int kind, int maxdepth ) {
  int n = 0;
  int *s = new int[ maxdepth+1 ];

  srand( time( NULL ) );

  while ( size > ( n + 2 ) ) {
    switch ( ( n == 0 ) ? ran( 2 ) : ran( 3 ) ) {
    case 0: // open
      if ( n >= maxdepth ) {
	continue;
      }
      *s = ran( kind );
      *array++ = openbr[ *s++ ];
      n++;
      break;
    case 1:
      *array++ = 'X';
      break;
    case 2: // close
      *array++ = closebr[ *--s ];
      n--;
      break;
    }
    size--;
  }
  while ( n-- > 0 ) {
    *array++ = closebr[ *--s ];
    size--;
  }
  while ( size-- > 1 ) {
    *array++ = 'X';
  }
  *array = '\0';
  delete [] s;
}

inline int isopen( char c ) {
  return ( c == '(' ) || ( c == '{' ) || ( c == '[' ) || ( c == '<' );
}

inline int isclose( char c ) {
  return ( c == ')' ) || ( c == '}' ) || ( c == ']' ) || ( c == '>' );
}

inline int match( char a, char b ) {
  if ( a == ')' ) return ( b == '(' );
  if ( a == '}' ) return ( b == '{' );
  if ( a == ']' ) return ( b == '[' );
  if ( a == '>' ) return ( b == '<' );
  return 0;
}

struct G : public skeleton::unary_ptr_function< int, Stack > {
  void operator()( int *b, const Stack *s ) const {
    *b = s->isempty( );
  }
} g;

struct P : public skeleton::binary_ptr_function< int, char, Stack > {
  void operator()( int *b, const char *c, const Stack *s ) const {
    *b = isclose( *c ) ? ( !s->isempty( ) && match( *c, s->top( ) ) ) : 1;
  }
} p;

struct Oplus : public skeleton::binary_ptr_function< int, int, int > {
  void operator()( int *r, const int *x, const int *y ) const {
    *r = *x && *y;
  }
} oplus;

struct Q : public skeleton::unary_ptr_function< Stack, char > {
  void operator()( Stack *s, const char *c ) const {
    if ( isopen( *c ) ) {
      s->push(*c);
    }
    else if ( isclose( *c ) ) {
      s->pop( );
    }
  }
} q;

struct Otimes : public skeleton::binary_ptr_function< Stack, Stack, Stack > {
  void operator()( Stack *c, const Stack *a, const Stack *b ) const {
    if ( a->in <= b->out ) {
      *c = *b;
      c->out += a->out - a->in;
    }
    else {
      *c = *a;
      for ( int i = 0; i < b->in; i++ ) {
	c->data[ a->in - b->out + i ] = b->data[ i ];
      }
      c->in += b->in - b->out;
    }
  }
} otimes;

int SketoMain( int argc, char **argv )
{
  if ( argc < 3 ) {
    skeleton::cout << "usage: " << argv[ 0 ] << " size loop" << std::endl;
    return 0;
  }

  const int size = std::atoi( argv[ 1 ] );
  const int loop = std::atoi( argv[ 2 ] );

  skeleton::time::initialize( ); // for time ekeeping

  for ( int i = 0; i < loop; i++ ) {
    // generate string include brackets
    char *str = new char[ size ];
    makestr( str, size, KIND, DEPTH );
    
    // output string
    skeleton::cout << str << std::endl;
    
    // distribute data
    dist_list< char > *as = new dist_list< char >( size, str );
    
    // calculate bracket matching
    Stack st;
    int result;
    list_skeletons::accumulate( g, p, oplus, q, otimes, &st, as, &result );
    
    delete as;
    delete [] str;
    
    // output result
    skeleton::cout << ( ( result == 1 ) ? "match" : "no match" ) << std::endl;
  }

  skeleton::time::set_time( "calculate" ); // for time ekeeping
  skeleton::time::print( ); // show the result of time ekeeping

  return 0;
}
