/***
 * Copyright (c) 2005, SkeTo Project
 * All rights reserved.
 */
/**
 * @file correlation.cpp
 * @brief This program compute correlation coefficient.
 * @author Yoshiki AKASHI (yoshiki@ipl.cs.uec.ac.jp)
 */

#include <cstdlib>
#include <cmath>
#include <ctime>
#include "list_skeletons.h"

struct Gen : public skeleton::unary_function< int, double > {
  double operator()( int index ) const {
    return static_cast< double >( std::rand( ) % 100 );
  }
} gen;

struct Add : public skeleton::binary_function< double, double, double > {
  double operator()( double x, double y ) const { return x + y; }
} add;
const double add_unit = 0.0;

struct Sqr : public skeleton::unary_function< double, double > {
  double operator()( double x ) const { return x * x; }
} sqr;

struct Sub : public skeleton::unary_function< double, double > {
  double val;
  Sub( double val_ ) : val( val_ ) { }
  double operator()( double x ) const { return x - val; }
};

struct Mul : public skeleton::binary_function< double, double, double > {
  double operator()( double x, double y ) const { return x * y; }
} mul;

int SketoMain( int argc, char ** argv )
{
  if ( argc < 3 ) {
    skeleton::cout << "usage: " << argv[ 0 ] << " size loop" << std::endl;
    return 0;
  }

  const int size = std::atoi( argv[ 1 ] );
  const int loop = std::atoi( argv[ 2 ] );

  std::srand( std::time( NULL ) ); // initialize random seed

  skeleton::time::initialize( ); // for time ekeeping

  for ( int i = 0; i < loop; i++ ) {
    // generate data
    dist_list< double > *x = new dist_list< double >( gen, size ); 
    dist_list< double > *y = new dist_list< double >( gen, size );
    
    // calculate average
    double ave_x = list_skeletons::reduce( add, add_unit, x ) / size;
    double ave_y = list_skeletons::reduce( add, add_unit, y ) / size;
    
    // x - ave_x
    Sub sub_ave_x( ave_x );
    list_skeletons::map_ow( sub_ave_x, x );
    // y - ave_y
    Sub sub_ave_y( ave_y );
    list_skeletons::map_ow( sub_ave_y, y );
    
    // sigma (x - ave_x)^2
    dist_list< double > *sqr_x = list_skeletons::map( sqr, x );
    double sum_x_x = list_skeletons::reduce( add, add_unit, sqr_x );
    delete sqr_x;
    
    // sigma (y - ave_y)^2
    dist_list< double > *sqr_y = list_skeletons::map( sqr, y );
    double sum_y_y = list_skeletons::reduce( add, add_unit, sqr_y );
    delete sqr_y;
    
    // sigma (x - ave_x)(y - ave_y)
    dist_list< double > *mul_x_y = list_skeletons::zipwith( mul, x, y );
    double sum_x_y = list_skeletons::reduce( add, add_unit, mul_x_y );
    delete x;
    delete y;
    delete mul_x_y;
    
    // calculate correlation coefficient
    double r = sum_x_y / std::sqrt( sum_x_x * sum_y_y );

    // output result    
    skeleton::cout << "correlation coefficient is " << r << std::endl;
  }

  skeleton::time::set_time( "calculate" ); // for time ekeeping
  skeleton::time::print( ); // show the result of time ekeeping
  
  return 0;
}
