/***
 * Copyright (c) 2005, SkeTo Project
 * All rights reserved.
 */
/**
 * @file MM.cpp
 * @brief Matrix Multiplication by Data Parallel Skeletons. 
 * @author Kento EMOTO
 *
 */ 
#include <iostream>
#include <iomanip>
#include <mpi.h>
#include <cmath>
#include <cstdlib>
#include "matrix_skeletons.h"

#include "primitive_functions.h"

using namespace primitive_functions;


template <typename A>
class Iprod2  : public skeleton::binary_function< matrix<A>, matrix<A> , A>{
public:
    void operator()(const matrix <A> &a, const matrix <A> &b, A *res) const{
		A r = 0;
		int n = a.getCols();
		for(int i = 0; i < n; i++){
			r+=a.at(0,i) * b.at(0, i);
		}
		*res = r;
	}
};

template <typename A>
class Tri {
public:
  	Tri(){}
    void operator()(A *r) const{
		r->tr();
	}
};

template <typename A>
dist_matrix <matrix <A> > * all_cols2(const dist_matrix <A> *a){
	dist_matrix<matrix <A> >*x = new dist_matrix<matrix <A> >();
	x->shapeCopy(a);	
	matrix_skeletons::map(Singleton<A>(), a, x);
	dist_matrix < matrix <A> > *y = matrix_skeletons::allreduce_col(Abv<A>(), x);
	delete x;
	return y;
}
template <typename A>
dist_matrix <matrix <A> > * all_rows2(const dist_matrix <A> *a){
	dist_matrix<matrix <A> >*x = new dist_matrix<matrix <A> >();
	x->shapeCopy(a);
	matrix_skeletons::map(Singleton<A>(), a, x);
	dist_matrix < matrix <A> > *y = matrix_skeletons::allreduce_row(Bsd<A>(), x);
	delete x;
	return y;
}




template <class C, class A, class B>
inline void mmP4(dist_matrix<C> &Z2, const dist_matrix<A> &X2, const dist_matrix<B> &Y2)
{
	dist_matrix < matrix < A > > *A2;
	dist_matrix < matrix < B > > *B2;
	A2 = all_rows2(&X2);
	B2 = all_cols2(&Y2);
	matrix_skeletons::map_i(Tri< matrix <B> >(), B2);
	matrix_skeletons::zipwith(Iprod2<C>(), A2, B2, &Z2);
	delete B2;
	delete A2;
}

void mult(const matrix<int> &x, const matrix<int> &y, matrix<int> *z){
	const int n = x.getRows();
	for(int i = 0; i < n; i++){
		for(int k = 0; k < n; k++){
			z->at(i,k) = 0;
		}
		for(int j = 0; j < n; j++){
			for(int k = 0; k < n; k++){
				z->at(i,k) = z->at(i,k) + x.at(i,j) * y.at(j,k);
			}
		}
	}
}
/***
 * used by rotations in Cannon's algorithm
 */
class Neg {
public:
    int operator()(const int i) const { return -i; }
};



/***
 * multiplication of local block-matrices.
 */
template <class C, class A, class B>
inline void MatrixMultiplyLocal(dist_matrix<C>& Z, const dist_matrix<A>& X, const dist_matrix<B>& Y)
{
	int rows = X.getLocalRows();
	int cols = Y.getLocalCols();
	int inner = X.getLocalCols();
	for(int i = 0; i < rows; i++){
		for(int k = 0; k < inner; k++){
			for(int j = 0; j < cols; j++){
				Z.at(i, j) = Z.at(i, j) + X.get(i, k) * Y.get(k, j);
			}
		}
	}
}

/***
 * multiplication of matrices.
 */
template <class C, class A, class B>
inline void MatrixMultiply(dist_matrix<C>& Z, dist_matrix<A>& X, dist_matrix<B>& Y)
{
	matrix_skeletons::rotateCols(Neg(), &X);          // Cannon's algorithm
	matrix_skeletons::rotateRows(Neg(), &Y);
	int q = X.getBlocksInRow();
	for(int i = 0; i < q; i++){
		MatrixMultiplyLocal(Z, X, Y);
		matrix_skeletons::rotateColsConst(-1, &X);
		matrix_skeletons::rotateRowsConst(-1, &Y);
	}
}

/***
 * main routine.
 * command line arguments: n
 *   n : #rows of A
 */

int SketoMain( int argc, char **argv )
{
	int n = 1000;
	int debug = 0;
	if(argc > 1){
		n = atoi(argv[1]);
	}
	if(argc > 2){
		debug = atoi(argv[2]);
	}

	int nproc = skeleton::procs;
	
	int q = (int)(sqrt(nproc)+0.1);
	
	int r = q;
	for(int k = 1; k <= q; k++){
		if((nproc % k)==0) r = k;
	}
	q = nproc / r;
	skeleton::cout << "A * B : (" << n << " x " << n << ") * (" << n << " x " << n << ")\n";
	skeleton::cout << " total prosess = " << nproc << "\n";
	skeleton::cout << " using prosess = " << q << " x "<< r << "\n";

	skeleton::time::initialize( );
	dist_matrix<int> A(n, n, GenAdd<int>(), DEFAULT_COMM, q, r);
	dist_matrix<int> B(n, n, GenSub<int>(), DEFAULT_COMM, q, r);
	dist_matrix<int> mmA(n, n, GenAdd<int>(), DEFAULT_COMM, q, r);
	dist_matrix<int> mmB(n, n, GenSub<int>(), DEFAULT_COMM, q, r);
	matrix<int> mA(n, n);
	mA.generate(GenAdd<int>());
	matrix<int> mB(n, n);
	mB.generate(GenSub<int>());
	skeleton::cout << "A = \n";
	if(debug){
		A.print(std::cout);
	}
	skeleton::cout << "B = \n";
	if(debug){
		B.print(std::cout);
	}
	dist_matrix<int> C(n, n, GenZero<int>(), DEFAULT_COMM, q, r);
	dist_matrix<int> mmC(n, n, GenZero<int>(), DEFAULT_COMM, q, r);
	matrix<int> mC(n, n);
	mC.generate(GenZero<int>());
	skeleton::time::set_time( "after distribute" );
	mmP4(C,A,B);
	skeleton::time::set_time( "after computation" );
	mult(mA,mB,&mC);
	skeleton::time::set_time( "after computation" );
	MatrixMultiply(mmC, mmA,mmB);
	skeleton::time::set_time( "after computation" );
	skeleton::time::print( );
	skeleton::cout << "C = A * B = \n";
	if(debug){
		C.print(std::cout);
	}
	return 0;
}

