/***
 * Copyright (c) 2005, SkeTo Project
 * All rights reserved.
 */
/**
 * @file MSSN.cpp
 * @brief Maximum Subarray Sum (parallel version, nested parallelism)
 * This program reads an input matrix from stdin of the root process.
 * A sample of the input is given in "MSS-matrix.txt"(size and elements).
 * Its answer is 75.
 * @author Kento EMOTO
 */

#include <iostream>
#include <iomanip>
#include <cmath>
#include <cstdlib>
#include "matrix_skeletons.h"

#include "primitive_functions.h"

using namespace primitive_functions;


/***
 * used by rotations in Cannon's algorithm
 */
class Neg {
public:
    int operator()(const int i) const { return -i; }
};

/***
 * generalized multiplication of local block-matrices.
 */
template <class C, class A, class B>
inline void GeneralizedMatrixMultiplyLocal(dist_matrix<C>& Z, const dist_matrix<A>& X, const dist_matrix<B>& Y, void (&oplus)( const C&, const C&, C*), void (&otimes)(const A&, const B&, C*))
{

	int rows = X.getLocalRows();
	int cols = Y.getLocalCols();
	int inner = X.getLocalCols();
	// 	for(int i = 0; i < rows; i++){
	// 		for(int j = 0; j < cols; j++){
	// 			C v = Z.get(i, j);
	// 			for(int k = 0; k < inner; k++){
	// 				C u;
	// 				otimes(X.get(i, k), Y.get(k, j), &u);
	// 				C tmp = v;
	// 				oplus(tmp, u, &v);
	// 			}
	// 			Z.set(i, j, v);
	// 		}
	// 	}
	for(int i = 0; i < rows; i++){
		for(int k = 0; k < inner; k++){
			for(int j = 0; j < cols; j++){
				C u;
				otimes(X.get(i, k), Y.get(k, j), &u);
				C tmp = Z.at(i, j);
				oplus(tmp, u, &Z.at(i, j));
			}
		}
	}
}

/***
 * generalized multiplication of matrices.
 */
template <class C, class A, class B>
inline void GeneralizedMatrixMultiply(dist_matrix<C>& Z, dist_matrix<A>& X, dist_matrix<B>& Y, void (&oplus)(const C&, const C&, C*), void (&otimes)(const A&, const B&, C*))
{
	matrix_skeletons::rotateCols(Neg(), &X);          // Cannon's algorithm
	matrix_skeletons::rotateRows(Neg(), &Y);
	int q = X.getBlocksInRow();
	for(int i = 0; i < q; i++){
		GeneralizedMatrixMultiplyLocal(Z, X, Y, oplus, otimes);
		matrix_skeletons::rotateColsConst(-1, &X);
		matrix_skeletons::rotateRowsConst(-1, &Y);
	}
}


template <class C, class A, class B>
inline void gemm(matrix<C>& Z, const matrix<A>& X, const matrix<B>& Y, void (&oplus)(const C&, const C&, C*), void (&otimes)(const A&, const B&, C*), const SKEL_MPI_INTRACOMM &com)
{
	int mysize;
	SKEL_MPI_Comm_size(com, mysize);
	int myrank;
	SKEL_MPI_Comm_rank(com, myrank);
	int q = 0;
	for(q = 1; q*q <= mysize; q++);
	q--;
	if(q==1){
		if(myrank==0) gemm(Z,X,Y,oplus,otimes);
		return;
	}
	dist_matrix<A> DX(&X,com,q,q);
	dist_matrix<B> DY(&Y,com,q,q);
	dist_matrix<C> DZ(&Z,com,q,q);
	if(myrank < q*q) GeneralizedMatrixMultiply(DZ,DX,DY,oplus,otimes);
	//std::cout << "MMP done with q = " << q << std::endl;
	DZ.gatherB(&Z);
}
	


#define MINIMUM 0x80000000

template <typename A>
A max(const A a, const A b)
{
	return (a > b) ? a : b;
}
void imax(const int &a, const int &b, int *c) {
	*c = (a > b) ? a : b;
}
void add(const int &a, const int &b, int *c) {
	*c = a + b;
}


class FN {
public:
	void operator()(const int &n1, const int &c1, const int &n2, int *r) const {
		*r = max( n1, c1 + n2);
	}
};

class FS{
public:
	void operator()(const int &s1, const int &c2, const int &s2, int *r) const {
		*r = max( s1 + c2, s2);
	}
};

class FE {
public:
	void operator()(const int &e1, const int &r2, const int &e2, int *r) const {
		*r = max( e1 + r2, e2);
	}
};

class FW{
public:
	void operator()(const int &w1, const int &r1, const int &w2, int *r) const {
		*r = max( w1, r1 + w2);
	}
};

class ZipwithPlus  : public skeleton::unary_function <matrix<int> , matrix<int> >{
	matrix <int> *withmat;
public:
	ZipwithPlus(matrix <int> *w){
		withmat = w;
	}
public:
	void operator()(const matrix <int> &cur, matrix <int> *res) const {
		m_skeletons::zipwith(Add<int>(), withmat, &cur, res);
	}
};

template <class C, class A, class B>
inline void gemm(matrix<C>& Z, const matrix<A>& X, const matrix<B>& Y, void (&oplus)(const C&, const C&, C*), void (&otimes)(const A&, const B&, C*))
{
	int rows = X.getRows();
	int cols = Y.getCols();
	int inner = X.getCols();	
	
	/*	for(int i = 0; i < rows; i++){
		for(int j = 0; j < cols; j++){
		C v = Z.get(i, j);
		for(int k = 0; k < inner; k++){
		C u;
		otimes(X.get(i, k), Y.get(k, j), &u);
		C tmp = v;
		oplus(tmp, u, &v);
		}
		Z.set(i, j, v);
		}
		}*/
	for(int i = 0; i < rows; i++){
		for(int k = 0; k < inner; k++){
			for(int j = 0; j < cols; j++){
				C u;
				otimes(X.get(i, k), Y.get(k, j), &u);
				C tmp = Z.at(i, j);
				oplus(tmp, u, &Z.at(i, j));
			}
		}
	}
}




class tuple {
public:
	int b;
	matrix<int> N;
	matrix<int> S;
	matrix<int> E;
	matrix<int> W;
	matrix<int> NE;
	matrix<int> NW;
	matrix<int> SE;
	matrix<int> SW;
	matrix<int> C;
	matrix<int> R;
	static const bool needsSerializing(){ return true; }
	tuple(){}
	tuple(int v){
		set(v);
	}
	void set(int v){
		b = v;
		N.singleton(v);
		S.singleton(v);
		E.singleton(v);
		W.singleton(v);
		NE.singleton(v);
		NW.singleton(v);
		SE.singleton(v);
		SW.singleton(v);
		C.singleton(v);
		R.singleton(v);
	}

	/******************* for serialization  ****************************/
	/**
	 * @brief construct a matrix from serialized data
	 */
public:
	unsigned char* construct(unsigned char *buf){
		buf=Serializer<int>::construct(&b, buf);
		buf=Serializer<matrix<int> >::construct(&N, buf);
		buf=Serializer<matrix<int> >::construct(&S, buf);
		buf=Serializer<matrix<int> >::construct(&E, buf);
		buf=Serializer<matrix<int> >::construct(&W, buf);
		buf=Serializer<matrix<int> >::construct(&NE, buf);
		buf=Serializer<matrix<int> >::construct(&NW, buf);
		buf=Serializer<matrix<int> >::construct(&SE, buf);
		buf=Serializer<matrix<int> >::construct(&SW, buf);
		buf=Serializer<matrix<int> >::construct(&C, buf);
		buf=Serializer<matrix<int> >::construct(&R, buf);
		return buf;
	}
	/**
	 * @brief serialize
	 */
public:
	unsigned char* serialize(unsigned char *buf) const{
		buf=Serializer<int>::serialize(b, buf);
		buf=Serializer<matrix<int> >::serialize(N, buf);
		buf=Serializer<matrix<int> >::serialize(S, buf);
		buf=Serializer<matrix<int> >::serialize(E, buf);
		buf=Serializer<matrix<int> >::serialize(W, buf);
		buf=Serializer<matrix<int> >::serialize(NE, buf);
		buf=Serializer<matrix<int> >::serialize(NW, buf);
		buf=Serializer<matrix<int> >::serialize(SE, buf);
		buf=Serializer<matrix<int> >::serialize(SW, buf);
		buf=Serializer<matrix<int> >::serialize(C, buf);
		buf=Serializer<matrix<int> >::serialize(R, buf);
		return buf;
	}
	/**
	 * @brief returns serialized size
	 */
public:
	int getSerializedSize() const{
		int size = 0;
		size+=Serializer<int>::getSerializedSize(b);
		size+=Serializer<matrix<int> >::getSerializedSize(N);
		size+=Serializer<matrix<int> >::getSerializedSize(S);
		size+=Serializer<matrix<int> >::getSerializedSize(E);
		size+=Serializer<matrix<int> >::getSerializedSize(W);
		size+=Serializer<matrix<int> >::getSerializedSize(NE);
		size+=Serializer<matrix<int> >::getSerializedSize(NW);
		size+=Serializer<matrix<int> >::getSerializedSize(SE);
		size+=Serializer<matrix<int> >::getSerializedSize(NW);
		size+=Serializer<matrix<int> >::getSerializedSize(C);
		size+=Serializer<matrix<int> >::getSerializedSize(R);
		return size;
	}
public:
	void print(std::ostream& os) const{
		os << "b = " << b << ", ";
		os << "N = " << N << ", ";
		os << "S = " << S << ", ";
		os << "E = " << E << ", ";
		os << "W = " << W << ", ";
		os << "NE = " << NE << ", ";
		os << "NW = " << NW << ", ";
		os << "SE = " << SE << ", ";
		os << "SW = " << SW << ", ";
		os << "C = " << C << ", ";
		os << "R = " << R << ", ";
		os << std::endl;
	}
public:
	friend std::ostream &operator<<(std::ostream& stream, const tuple &tup){
		tup.print(stream);
		return stream;
	}
};


class Oplus{
public:
	void operator()(const tuple &x, const tuple &y, tuple *res) const{
		matrix<int> *tmp = m_skeletons::zipwith<int>(Add<int>(), &x.S, &y.N);
		int b;
		m_skeletons::reduce(Max<int>(), Max<int>(), tmp, &b);
		res->b = max(max(x.b, y.b), b);           // b
		delete tmp;
		
		m_skeletons::zipwith(FN(), &x.N, &x.C, &y.N, &res->N);       // N
		m_skeletons::zipwith(FS(), &x.S, &y.C, &y.S, &res->S);       // S
		
		matrix<int> etmp(y.NE);
		etmp.tr();
		matrix<int> etmp2(x.SE.getRows(), etmp.getCols());
		etmp2.fill(MINIMUM);
		gemm(etmp2, x.SE, etmp, imax, add);
		matrix<int> etmp3;
		matrix<int> etmp4;
		m_skeletons::bsd(&x.E, &etmp2, &etmp3);
		etmp2.tr();
		etmp2.fill(0);
		m_skeletons::bsd(&etmp2, &y.E, &etmp4);
		m_skeletons::abv(&etmp3, &etmp4, &res->E);                 // E
		
		matrix<int> wtmp(y.NW);
		wtmp.tr();
		matrix<int> wtmp2(x.SW.getRows(), wtmp.getCols());
		wtmp2.fill(MINIMUM);
		gemm(wtmp2, x.SW, wtmp, imax, add);
		matrix<int> wtmp3;
		matrix<int> wtmp4;
		m_skeletons::bsd(&x.W, &wtmp2, &wtmp3);
		wtmp2.tr();
		wtmp2.fill(0);
		m_skeletons::bsd(&wtmp2, &y.W, &wtmp4);
		m_skeletons::abv(&wtmp3, &wtmp4, &res->W);                 // W

		matrix<int> *tmpne = m_skeletons::bottom(&x.NE);
		matrix<int> *tmpne2 = m_skeletons::mapr<int>(ZipwithPlus(tmpne), &y.NE);
		m_skeletons::abv(&x.NE, tmpne2, &res->NE);               // NE
		delete tmpne;
		delete tmpne2;
		
		matrix<int> *tmpnw = m_skeletons::bottom(&x.NW);
		matrix<int> *tmpnw2 = m_skeletons::mapr<int>(ZipwithPlus(tmpnw), &y.NW);
		m_skeletons::abv(&x.NW, tmpnw2, &res->NW);               // NW
		delete tmpnw;
		delete tmpnw2;
		
		matrix<int> *tmpse = m_skeletons::top(&y.SE);
		matrix<int> *tmpse2 = m_skeletons::mapr<int>(ZipwithPlus(tmpse), &x.SE);
		m_skeletons::abv(tmpse2, &y.SE, &res->SE);               // SE
		delete tmpse;
		delete tmpse2;
		
		matrix<int> *tmpsw = m_skeletons::top(&y.SW);
		matrix<int> *tmpsw2 = m_skeletons::mapr<int>(ZipwithPlus(tmpsw), &x.SW);
		m_skeletons::abv(tmpsw2, &y.SW, &res->SW);               // SW
		delete tmpsw;
		delete tmpsw2;
		
		m_skeletons::zipwith(Add<int>(), &x.C, &y.C, &res->C);            // C
		
		matrix<int> *tmpr1 = m_skeletons::right(&x.R);
		matrix<int> *tmpr2 = m_skeletons::top(&y.R);
		matrix<int> tmpr3(x.R.getRows(), y.R.getCols());
		matrix<int> tmpr4, tmpr5;
		tmpr3.fill(MINIMUM);
		gemm(tmpr3, *tmpr1, *tmpr2, imax, add);
		m_skeletons::bsd(&x.R, &tmpr3, &tmpr4);
		tmpr3.tr();
		tmpr3.fill(0);
		m_skeletons::bsd(&tmpr3, &y.R, &tmpr5);
		m_skeletons::abv(&tmpr4, &tmpr5, &res->R);                 // R
		delete tmpr1;
		delete tmpr2;
	}
	// for nested parallelism (DIRTY)
	void operator()(const tuple &x, const tuple &y, tuple *res, const SKEL_MPI_INTRACOMM &com) const{
		matrix<int> *tmp = m_skeletons::zipwith<int>(Add<int>(), &x.S, &y.N);
		int b;
		m_skeletons::reduce(Max<int>(), Max<int>(), tmp, &b);
		res->b = max(max(x.b, y.b), b);           // b
		delete tmp;
		
		m_skeletons::zipwith(FN(), &x.N, &x.C, &y.N, &res->N);       // N
		m_skeletons::zipwith(FS(), &x.S, &y.C, &y.S, &res->S);       // S
			
		matrix<int> etmp(y.NE);
		etmp.tr();
		matrix<int> etmp2(x.SE.getRows(), etmp.getCols());
		etmp2.fill(MINIMUM);
		gemm(etmp2, x.SE, etmp, imax, add, com);
		//gemm(etmp2, x.SE, etmp, imax, add);
		matrix<int> etmp3;
		matrix<int> etmp4;
		m_skeletons::bsd(&x.E, &etmp2, &etmp3);
		etmp2.tr();
		etmp2.fill(0);
		m_skeletons::bsd(&etmp2, &y.E, &etmp4);
		m_skeletons::abv(&etmp3, &etmp4, &res->E);                 // E
			
		matrix<int> wtmp(y.NW);
		wtmp.tr();
		matrix<int> wtmp2(x.SW.getRows(), wtmp.getCols());
		wtmp2.fill(MINIMUM);
		gemm(wtmp2, x.SW, wtmp, imax, add, com);
		//gemm(wtmp2, x.SW, wtmp, imax, add);
		matrix<int> wtmp3;
		matrix<int> wtmp4;
		m_skeletons::bsd(&x.W, &wtmp2, &wtmp3);
		wtmp2.tr();
		wtmp2.fill(0);
		m_skeletons::bsd(&wtmp2, &y.W, &wtmp4);
		m_skeletons::abv(&wtmp3, &wtmp4, &res->W);                 // W
			
		matrix<int> *tmpne = m_skeletons::bottom(&x.NE);
		matrix<int> *tmpne2 = m_skeletons::mapr<int>(ZipwithPlus(tmpne), &y.NE);
		m_skeletons::abv(&x.NE, tmpne2, &res->NE);               // NE
		delete tmpne;
		delete tmpne2;
			
		matrix<int> *tmpnw = m_skeletons::bottom(&x.NW);
		matrix<int> *tmpnw2 = m_skeletons::mapr<int>(ZipwithPlus(tmpnw), &y.NW);
		m_skeletons::abv(&x.NW, tmpnw2, &res->NW);               // NW
		delete tmpnw;
		delete tmpnw2;
			
		matrix<int> *tmpse = m_skeletons::top(&y.SE);
		matrix<int> *tmpse2 = m_skeletons::mapr<int>(ZipwithPlus(tmpse), &x.SE);
		m_skeletons::abv(tmpse2, &y.SE, &res->SE);               // SE
		delete tmpse;
		delete tmpse2;
			
		matrix<int> *tmpsw = m_skeletons::top(&y.SW);
		matrix<int> *tmpsw2 = m_skeletons::mapr<int>(ZipwithPlus(tmpsw), &x.SW);
		m_skeletons::abv(tmpsw2, &y.SW, &res->SW);               // SW
		delete tmpsw;
		delete tmpsw2;
			
		m_skeletons::zipwith(Add<int>(), &x.C, &y.C, &res->C);            // C
			
		matrix<int> *tmpr1 = m_skeletons::right(&x.R);
		matrix<int> *tmpr2 = m_skeletons::top(&y.R);
		matrix<int> tmpr3(x.R.getRows(), y.R.getCols());
		matrix<int> tmpr4, tmpr5;
		tmpr3.fill(MINIMUM);
		gemm(tmpr3, *tmpr1, *tmpr2, imax, add,com);
		//gemm(tmpr3, *tmpr1, *tmpr2, imax, add);
		m_skeletons::bsd(&x.R, &tmpr3, &tmpr4);
		tmpr3.tr();
		tmpr3.fill(0);
		m_skeletons::bsd(&tmpr3, &y.R, &tmpr5);
		m_skeletons::abv(&tmpr4, &tmpr5, &res->R);                 // R
		delete tmpr1;
		delete tmpr2;
	}
};


class Otimes {
public:
	void operator()(const tuple &x, const tuple &y, tuple *res) const {
		matrix<int> *tmp = m_skeletons::zipwith<int>(Add<int>(), &x.E, &y.W);
		int b;
		m_skeletons::reduce(Max<int>(), Max<int>(), tmp, &b);
		res->b = max(max(x.b, y.b), b);           // b
		m_skeletons::zipwith(FE(), &x.E, &y.R, &y.E, &res->E);       // E
		m_skeletons::zipwith(FW(), &x.W, &x.R, &y.W, &res->W);       // W
		delete tmp;
		
		matrix<int> ntmp(x.NE);
		ntmp.tr();
		matrix<int> ntmp2(ntmp.getRows(), y.NW.getCols());
		ntmp2.fill(MINIMUM);
		gemm(ntmp2, ntmp, y.NW, imax, add);
		matrix<int> ntmp3;
		matrix<int> ntmp4;
		m_skeletons::bsd(&x.N, &ntmp2, &ntmp3);
		ntmp2.tr();
		ntmp2.fill(0);
		m_skeletons::bsd(&ntmp2, &y.N, &ntmp4);
		m_skeletons::abv(&ntmp3, &ntmp4, &res->N);                 // N
		
		matrix<int> stmp(x.SE);
		stmp.tr();
		matrix<int> stmp2(stmp.getRows(), y.SW.getCols());
		stmp2.fill(MINIMUM);
		gemm(stmp2, stmp, y.SW, imax, add);
		matrix<int> stmp3;
		matrix<int> stmp4;
		m_skeletons::bsd(&x.S, &stmp2, &stmp3);
		stmp2.tr();
		stmp2.fill(0);
		m_skeletons::bsd(&stmp2, &y.S, &stmp4);
		m_skeletons::abv(&stmp3, &stmp4, &res->S);                 // S
		
		matrix<int> *tmpne = m_skeletons::left(&y.NE);
		matrix<int> *tmpne2 = m_skeletons::mapc<int>(ZipwithPlus(tmpne), &x.NE);
		m_skeletons::bsd(tmpne2, &y.NE, &res->NE);               // NE
		delete tmpne;
		delete tmpne2;
		
		
		matrix<int> *tmpnw = m_skeletons::right(&x. NW);
		matrix<int> *tmpnw2 = m_skeletons::mapc<int>(ZipwithPlus(tmpnw), &y.NW);
		m_skeletons::bsd(&x.NW, tmpnw2, &res->NW);               // NW
		delete tmpnw;
		delete tmpnw2;
		
		matrix<int> *tmpse = m_skeletons::left(&y.SE);
		matrix<int> *tmpse2 = m_skeletons::mapc<int>(ZipwithPlus(tmpse), &x.SE);
		m_skeletons::bsd(tmpse2, &y.SE, &res->SE);               // SE
		delete tmpse;
		delete tmpse2;
		
		matrix<int> *tmpsw = m_skeletons::right(&x.SW);
		matrix<int> *tmpsw2 = m_skeletons::mapc<int>(ZipwithPlus(tmpsw), &y.SW);
		m_skeletons::bsd(&x.SW, tmpsw2, &res->SW);               // SW	
		delete tmpsw;
		delete tmpsw2;
		
		m_skeletons::zipwith(Add<int>(), &x.R, &y.R, &res->R);            // R
		
		matrix<int> *tmpc1 = m_skeletons::right(&x.C);
		matrix<int> *tmpc2 = m_skeletons::top(&y.C);
		matrix<int> tmpc3(x.C.getRows(), y.C.getCols());
		matrix<int> tmpc4, tmpc5;
		tmpc3.fill(MINIMUM);
		gemm(tmpc3, *tmpc1, *tmpc2, imax, add);
		m_skeletons::bsd(&x.C, &tmpc3, &tmpc4);
		tmpc3.tr();
		tmpc3.fill(0);
		m_skeletons::bsd(&tmpc3, &y.C, &tmpc5);
		m_skeletons::abv(&tmpc4, &tmpc5, &res->C);                 // C	
		delete tmpc1;
		delete tmpc2;
	}
	// for nested parallelism (DIRTY)
	void operator()(const tuple &x, const tuple &y, tuple *res, const SKEL_MPI_INTRACOMM &com) const {
		matrix<int> *tmp = m_skeletons::zipwith<int>(Add<int>(), &x.E, &y.W);
		int b;
		m_skeletons::reduce(Max<int>(), Max<int>(), tmp, &b);
		res->b = max(max(x.b, y.b), b);           // b
		m_skeletons::zipwith(FE(), &x.E, &y.R, &y.E, &res->E);       // E
		m_skeletons::zipwith(FW(), &x.W, &x.R, &y.W, &res->W);       // W
		delete tmp;
			
		matrix<int> ntmp(x.NE);
		ntmp.tr();
		matrix<int> ntmp2(ntmp.getRows(), y.NW.getCols());
		ntmp2.fill(MINIMUM);
		gemm(ntmp2, ntmp, y.NW, imax, add,com);
		//gemm(ntmp2, ntmp, y.NW, imax, add);
		matrix<int> ntmp3;
		matrix<int> ntmp4;
		m_skeletons::bsd(&x.N, &ntmp2, &ntmp3);
		ntmp2.tr();
		ntmp2.fill(0);
		m_skeletons::bsd(&ntmp2, &y.N, &ntmp4);
		m_skeletons::abv(&ntmp3, &ntmp4, &res->N);                 // N
			
		matrix<int> stmp(x.SE);
		stmp.tr();
		matrix<int> stmp2(stmp.getRows(), y.SW.getCols());
		stmp2.fill(MINIMUM);
		gemm(stmp2, stmp, y.SW, imax, add,com);
		//gemm(stmp2, stmp, y.SW, imax, add);
		matrix<int> stmp3;
		matrix<int> stmp4;
		m_skeletons::bsd(&x.S, &stmp2, &stmp3);
		stmp2.tr();
		stmp2.fill(0);
		m_skeletons::bsd(&stmp2, &y.S, &stmp4);
		m_skeletons::abv(&stmp3, &stmp4, &res->S);                 // S
			
		matrix<int> *tmpne = m_skeletons::left(&y.NE);
		matrix<int> *tmpne2 = m_skeletons::mapc<int>(ZipwithPlus(tmpne), &x.NE);
		m_skeletons::bsd(tmpne2, &y.NE, &res->NE);               // NE
		delete tmpne;
		delete tmpne2;
		
			
		matrix<int> *tmpnw = m_skeletons::right(&x. NW);
		matrix<int> *tmpnw2 = m_skeletons::mapc<int>(ZipwithPlus(tmpnw), &y.NW);
		m_skeletons::bsd(&x.NW, tmpnw2, &res->NW);               // NW
		delete tmpnw;
		delete tmpnw2;
			
		matrix<int> *tmpse = m_skeletons::left(&y.SE);
		matrix<int> *tmpse2 = m_skeletons::mapc<int>(ZipwithPlus(tmpse), &x.SE);
		m_skeletons::bsd(tmpse2, &y.SE, &res->SE);               // SE
		delete tmpse;
		delete tmpse2;
			
		matrix<int> *tmpsw = m_skeletons::right(&x.SW);
		matrix<int> *tmpsw2 = m_skeletons::mapc<int>(ZipwithPlus(tmpsw), &y.SW);
		m_skeletons::bsd(&x.SW, tmpsw2, &res->SW);               // SW	
		delete tmpsw;
		delete tmpsw2;
			
		m_skeletons::zipwith(Add<int>(), &x.R, &y.R, &res->R);            // R
			
		matrix<int> *tmpc1 = m_skeletons::right(&x.C);
		matrix<int> *tmpc2 = m_skeletons::top(&y.C);
		matrix<int> tmpc3(x.C.getRows(), y.C.getCols());
		matrix<int> tmpc4, tmpc5;
		tmpc3.fill(MINIMUM);
		gemm(tmpc3, *tmpc1, *tmpc2, imax, add,com);
		//gemm(tmpc3, *tmpc1, *tmpc2, imax, add);
		m_skeletons::bsd(&x.C, &tmpc3, &tmpc4);
		tmpc3.tr();
		tmpc3.fill(0);
		m_skeletons::bsd(&tmpc3, &y.C, &tmpc5);
		m_skeletons::abv(&tmpc4, &tmpc5, &res->C);                 // C	
		delete tmpc1;
		delete tmpc2;

	}
};

class Finit : public skeleton::unary_function <int, tuple >{
public:
	void operator()(const int &v, tuple *tup) const{
		tup->set(v);
	}
};

template<>
class template_mpitype< tuple > {
public:
    static const MPI_Datatype type() { return MPI_BYTE;}
};

int SketoMain( int argc, char **argv ) {
	int myrank = skeleton::rank;
	int nproc = skeleton::procs;

	int q = (int)(sqrt(nproc)+0.1);
	
	int r = q;
	for(int k = 1; k <= q; k++){
		if((nproc % k)==0) r = k;
	}
	q = nproc / r;
	int dbg = 0;
	int m = 0;
	int n = 0;
//	int nest = 0;
//	if(argc > 1){
//		nest = atoi(argv[1]);
//		argc--;
//		argv++;
//	}
	
	if(argc > 1){
		dbg = atoi(argv[1]);
	}
	if(argc > 2){
		q = atoi(argv[2]);
	}
	if(argc > 3){
		r = atoi(argv[3]);
	}

	int **es = NULL;
	int *ebuf = NULL;
	if(myrank==0){
		std::cin >> m >> n;
		es = new int*[m];
		ebuf = new int[m*n];
		for(int i = 0; i < m; i++){
			es[i] = &ebuf[i*n];
		}
		std::cout << m << " x " << n << " in " << q << " x " << r << " prosessors." << std::endl;
		for(int i = 0; i < m; i++){
			for(int j = 0; j < n; j++){
				std::cin >> es[i][j];
			}
		}
		std::cout << m << " x " << n << std::endl;
		if(dbg){
			std::cout << "[";
			for(int i = 0; i < m; i++){
				if(i!=0) std::cout << ",";
				std::cout << "[";
				for(int j = 0; j < n; j++){
					if(j!=0) std::cout << ",";
					std::cout << es[i][j];
				}
				std::cout << "]"<<std::endl;
			}
			std::cout << "]\n";
		}
	}


	matrix<int> mat(m, n, es);

	skeleton::time::initialize( );

	//dist_matrix<int> dmat(&mat);
	dist_matrix<int> dmat(&mat, DEFAULT_COMM, q, r);

	skeleton::time::set_time( "after distribute" );

	tuple *res;
	res = matrix_skeletons::cataDCN2(Oplus(), Otimes(), Finit(), &dmat);
	skeleton::time::set_time( "after computation" );
	skeleton::time::print( );

	//std::cout << "res = " << *res << std::endl;
	skeleton::cout << "res = " << res->b << std::endl;
	delete res;
	return 0;
}

