/***
 * Copyright (c) 2005, SkeTo Project
 * All rights reserved.
 */
/**
 * @file qrwise.cpp
 * @brief QR Factorization (experimental)
 * using sequential householder transformation.
 * @author Kento EMOTO
 */ 

#include <iostream>
#include <iomanip>
#include <mpi.h>
#include <cmath>
#include <cstdlib>
#include "matrix_skeletons.h"


#include "primitive_functions.h"

using namespace primitive_functions;

int sqrproc, nproc;
int myrank;
int threhh;

class Eye {
public:
	void operator()(int i, int j, double *r) const {
		*r = (i==j) ? 1 : 0;
	}
};

template <typename A>
class Neg {
public:
	void operator()(A *r) const {
		*r = - *r;
	}
};

template <typename A>
class DivByi : public skeleton::unary_function < A, A >{
	A d;
public:
	DivByi(A dd){
		d = dd;
	}
	void operator()(A *r) const {
		*r = *r / d;
	}
};

template <typename A>
class Abs {
public:
	void operator()(A *r) const {
		*r = (*r < 0) ? -*r : *r;
	}
};


class NegI {
public:
	int operator()(int r) const {
		return - r;
	}
};


template <class C, class A, class B>
inline void mm(matrix<C>& Z, const matrix<A>& X, const matrix<B>& Y)
{
	int rows = X.getRows();
	int cols = Y.getCols();
	int inner = X.getCols();
	for(int i = 0; i < rows; i++){
		for(int j = 0; j < cols; j++){
			C v = Z.get(i, j);
			for(int k = 0; k < inner; k++){
				v += X.get(i, k)*Y.get(k, j);
			}
			Z.set(i, j, v);
		}
	}
}
template <class C, class A, class B>
inline void mmPLocal(dist_matrix<C>& Z, const dist_matrix<A>& X, const dist_matrix<B>& Y)
{
	int rows = X.getLocalRows();
	int cols = Y.getLocalCols();
	int inner = X.getLocalCols();
	for(int i = 0; i < rows; i++){
		for(int j = 0; j < cols; j++){
			C v = Z.get(i, j);
			for(int k = 0; k < inner; k++){
				v += X.get(i, k) * Y.get(k, j);
			}
			Z.set(i, j, v);
		}
	}
}

template <class C, class A, class B>
inline void mmP(matrix<C>& Z, const matrix<A>& X, const matrix<B>& Y)
{
	dist_matrix<A> X2(&X, DEFAULT_COMM, sqrproc, sqrproc);
	dist_matrix<B> Y2(&Y, DEFAULT_COMM, sqrproc, sqrproc);
	dist_matrix<C> Z2(Z.getRows(), Z.getCols(), GenZero<C>(), DEFAULT_COMM, sqrproc, sqrproc);

	matrix_skeletons::rotateCols(NegI(), &X2);
	matrix_skeletons::rotateRows(NegI(), &Y2);
	for(int i = 0; i < sqrproc; i++){
		mmPLocal(Z2, X2, Y2);
		matrix_skeletons::rotateColsConst(-1, &X2);
		matrix_skeletons::rotateRowsConst(-1, &Y2);
	}
	//Z2.gatherB(Z);
	Z2.gather(&Z);
}

template <typename A>
class GenQh {
	const matrix<A> &e;
public:
	GenQh(const matrix<A> &ee) : e(ee) {
	}
	void operator()(int i, int j, A*r) const {
		*r = - e.at(i,0) * e.at(j,0);
		if(i==j) *r+=1;
	}
};

void householder(const matrix<double>& v, matrix<double>& Q, matrix<double>& R)
{
	//if(myrank==0)std::cout << "hh\n";
	double v0 = v.at(0,0); 
	double e0 = 1;
	double vnorm = 0;
	matrix<double> v2;
	m_skeletons::zipwith(Mul<double>(), &v, &v, &v2);
	m_skeletons::reduce(Add<double>(), Add<double>(), &v2, &vnorm);
	e0 = sqrt(vnorm);

	if(e0==0){
		Q.setSize(v.getRows(), v.getRows());
		Q.eye(1, 0);
		R.setSize(v.getRows(), 1);
		R.fill(0);
		return;
	}
	if(fabs(v0-e0)<e0){
		e0 = - e0;
	}
	matrix<double> e = v;
	e.at(0,0) -= e0;
	double a = 0;
	matrix<double> e2;
	m_skeletons::zipwith(Mul<double>(), &e, &e, &e2);
	m_skeletons::reduce(Add<double>(), Add<double>(), &e2, &a);
	a = sqrt(a/2);
	m_skeletons::map_i(DivByi<double>(a), &e);
// 	matrix<double> et = e;
// 	et.tr();
// 	matrix<double> I(e.getRows(), et.getCols()), tmp(e.getRows(), et.getCols());
// 	tmp.fill(0);
// 	mm(tmp, e, et);
// 	I.eye(1, 0);
//    	Q.setSize(e.getRows(), et.getCols());
// 	m_skeletons::zipwith(Sub<double>(), I, tmp, Q);
	Q.setSize(e.getRows(), e.getRows());
	Q.generate(GenQh<double>(e));
	
	R.setSize(e.getRows(), 1);
	R.fill(0);
	R.at(0,0) = e0;
}

// parallel version of hh transformation... slower than sequencial
void householderP(const matrix<double>& v, matrix<double>& Q, matrix<double>& R)
{
	if(v.getRows() < threhh) {
		if(myrank==0) return householder(v, Q, R);
		return;
	}
	double v0 = 0;
	if(myrank==0) v0 = v.at(0,0); 
	double e0 = 1;
	double vnorm = 0;
	dist_matrix<double> vd(&v, DEFAULT_COMM, nproc, 1);
	dist_matrix<double> v2;
	v2.shapeCopy(&vd);
	matrix_skeletons::zipwith(Mul<double>(), &vd, &vd, &v2);
	matrix_skeletons::reduceB(Add<double>(), Add<double>(), &v2, &vnorm);
	e0 = sqrt(vnorm);

	if(e0==0){
		Q.setSize(v.getRows(), v.getRows());
		Q.eye(1, 0);
		R.setSize(v.getRows(), 1);
		R.fill(0);
		return;
	}
	if(fabs(v0-e0)<e0){
		e0 = - e0;
	}
	
	matrix<double> e = v;
	if(myrank==0) e.at(0,0) -= e0;
	double a = 0;
	
	dist_matrix<double> ed(&e, DEFAULT_COMM, nproc, 1);
	dist_matrix<double> e2;
	e2.shapeCopy(&ed);
	matrix_skeletons::zipwith(Mul<double>(), &ed, &ed, &e2);
	matrix_skeletons::reduceB(Add<double>(), Add<double>(), &e2, &a);
	a = sqrt(a/2);

	matrix_skeletons::map_i(DivByi<double>(a), &ed);
	ed.gather(&e);
	//ed.gatherB(e);
// 	matrix<double> et = e;
// 	et.tr();
// 	matrix<double> tmp(e.getRows(), et.getCols());
// 	//tmp.fill(0);
// 	//mm(tmp, e, et);
// 	mmP(tmp, e, et);
// 	dist_matrix<double> I(tmp.getRows(), tmp.getCols(), nproc, 1, Eye());
// 	dist_matrix<double> tmp2(tmp, nproc, 1), Q2;
//    	Q2.shapeCopy(I);
	
// 	matrix_skeletons::zipwith(Sub<double>(), I, tmp2, Q2);
// 	//Q2.gatherB(Q);
// 	Q2.gather(Q);
// 	if(myrank!=0){
// 		Q.setSize(v.getRows(), v.getRows());
// 	}

	Q.setSize(e.getRows(), e.getRows());
	Q.generate(GenQh<double>(e));
	R.setSize(e.getRows(), 1);
	R.fill(0);
	R.at(0,0) = e0;
}
template <typename A>
class EmbedEye {
	const matrix<A> &q3;
	const int n, m;
public:
	EmbedEye(int mm, int nn,const matrix<A> &qq3) : q3(qq3), n(nn), m(mm) {
	}
	void operator()( int i, int j, A*r) const {
		if ((i >= m) && (j >= n)){
			*r = q3.get(i-m, j-n);
		} else if (i == j){
			*r = 1;
		} else {
			*r = 0;
		}
	}
};

void qrfact(const matrix<double>& A, matrix<double>& Q, matrix<double>& R)
{
	if(A.getCols()==1){
		//householderP(A, Q, R);
		householder(A, Q, R);
		return;
	}
	matrix<double> A1, A2;
	int n = (A.getCols() + 1)/2;

	A.splitv(n, &A1, &A2);
	matrix<double> q1, r1, r11, z;
	qrfact(A1, q1, r1);
	r1.splith(n, &r11, &z);
	matrix<double> a(A2.getRows(), A2.getCols()), qt;
	qt = q1;
	qt.tr();
	a.fill(0);
	mm(a, qt, A2);
	matrix<double> r12, A22;
	a.splith(n, &r12, &A22);
	matrix<double> q2, r22;
	qrfact(A22, q2, r22);
	matrix<double> q3(A2.getRows(), A2.getRows());
	q3.generate(EmbedEye<double>(A2.getRows() - q2.getRows(), A2.getRows() - q2.getRows(), q2));
	// 	q3.eye(1, 0);
// 	q3.embed(A2.getRows() - q2.getRows(), A2.getRows() - q2.getRows(), q2);
	Q.setSize(A2.getRows(), A2.getRows());
	Q.fill(0);
	mm(Q, q1, q3);
	m_skeletons::uppertriangular(&r11, &r12, &r22, 0.0, &R);
}


int thre;
void qrfactP(const matrix<double>& A, matrix<double>& Q, matrix<double>& R)
{
	if(A.getCols()==1){
		householder(A, Q, R);
		//householderP(A, Q, R);
		return;
	}
	if(A.getRows() < thre) {
		if(myrank==0){
			return qrfact(A,Q,R);
		}
		Q.setSize(A.getRows(), A.getRows());
		R.setSize(A.getRows(), A.getCols());
		
		return;
	}
	matrix<double> A1, A2;
	int n = (A.getCols() + 1)/2;
	A.splitv(n, &A1, &A2);
	matrix<double> q1, r1, r11, z;
	qrfactP(A1, q1, r1);
	r1.splith(n, &r11, &z);
	matrix<double> a(A2.getRows(), A2.getCols()), qt;
	qt = q1;
	qt.tr();
	a.fill(0);
	mm(a, qt, A2);
	//mmP(a, qt, A2);
	matrix<double> r12, A22;
	a.splith(n, &r12, &A22);
	matrix<double> q2, r22;
	qrfactP(A22, q2, r22);
	matrix<double> q3(A2.getRows(), A2.getRows());
	q3.generate(EmbedEye<double>(A2.getRows() - q2.getRows(), A2.getRows() - q2.getRows(), q2));
// 	q3.eye(1, 0);
// 	q3.embed(A2.getRows() - q2.getRows(), A2.getRows() - q2.getRows(), q2);
	Q.setSize(A2.getRows(), A2.getRows());
	//Q.fill(0);
	mmP(Q, q1, q3);
	m_skeletons::uppertriangular(&r11, &r12, &r22, 0.0, &R);
}

// return max(abs(A-Q*R))
double getErr(const matrix<double> &A, const matrix<double> &Q, const matrix<double> &R)
{
	matrix<double> tmp(Q.getRows(), R.getCols());
	tmp.fill(0);
 	mm(tmp,Q,R);
	matrix<double> tmp2(A.getRows(), A.getCols());
	m_skeletons::zipwith(Sub<double>(), &A, &tmp, &tmp2);
	m_skeletons::map_i(Abs<double>(), &tmp2);
	double merr = 0;
	m_skeletons::reduce(Max<double>(), Max<double>(), &tmp2, &merr);
	return merr;
}


int SketoMain( int argc, char **argv )
{
	myrank= skeleton::rank;
	nproc = skeleton::procs;
	sqrproc = (int)(sqrt((double)nproc)+0.1);
	int m = 5;
	int n = 4;
	int db = 0;
	if(argc > 1){
		m = atoi(argv[1]);
	}
	if(argc > 2){
		n = atoi(argv[2]);
	}
	if(argc > 3){
		db = atoi(argv[3]);
	}
	thre = 0;
	if(argc > 4){
		thre = atoi(argv[4]);
	}

	threhh = 0;
	if(argc > 5){
		threhh = atoi(argv[5]);
	}

	matrix< double > a(m, n);
	for(int i = 0; i < m; i++){
		for(int j = 0; j < n; j++){
			a.at(i, j) = (i+j);
		}
	}
	//std::cout << "QR factorization of ( " << m << " x " << n << " ) matrix " << myrank << "\n";
	skeleton::cout << "QR factorization of ( " << m << " x " << n << " ) matrix\n";
	if(db) skeleton::cout << "a = \n" << a << std::endl;
	
	matrix<double> q, r;
	skeleton::time::initialize( );
	qrfactP(a, q, r);
	skeleton::time::set_time( "after computation" );
	skeleton::time::print( );

	if(myrank==0){
		if(db){
			std::cout << "q = \n" << q << std::endl;
			std::cout << "r = \n" << r << std::endl;
		}
		matrix<double> qt = q;
		qt.tr();
		matrix<double> eye(q.getRows(), q.getCols());
		eye.eye(1,0);
		double merr = getErr(eye, q, qt);
		std::cout << "max(abs(I-Q'*Q)) = " << merr << std::endl;
		double merr2 = getErr(a, q, r);
		std::cout << "max(abs(A-Q*R)) = " << merr2 << std::endl;

	}
	return 0;
}
