/***
 * Copyright (c) 2005, SkeTo Project
 * All rights reserved.
 */
/**
 * @file height_bin.cpp
 * @brief A sample program that compute the height of the input tree
 *        with map, downwards accumulate and reduce.
 * @author Kiminori MATSUZAKI
 */

#include <cassert>
#include <iostream>
#include <algorithm>
#include <climits>

#include <mpi.h>

#include <skeleton.h>
#include <tree_skeletons.h>


////////////////////////////////////////////////////////////////////////
// function for map skeleton

struct func_one : skeleton::unary_function< char, int > {
  int operator()( const char& ) const {
    return 1;
  }
};

////////////////////////////////////////////////////////////////////////
// functions for dAcc skeleton

struct func_plus : skeleton::binary_function< int, int, int > {
  int operator()( const int& lhs, const int& rhs ) const {
    return lhs + rhs;
  }
};

struct func_id : skeleton::unary_function< int, int > {
  int operator()( const int& x ) const {
    return x;
  }
};

////////////////////////////////////////////////////////////////////////
// functions for uAcc skeleton

struct func_max_lr : skeleton::ternary_function< int, int, int, int > {
  int operator()( const int&, const int& l, const int& r ) const {
    return std::max( l, r );
  }
};

struct func_max_lr_psi : skeleton::unary_function< int, int > {
  int operator()( const int& ) const {
    return INT_MIN;
  }
};

struct func_max_lr_phiL : skeleton::ternary_function< int, int, int, int > {
  int operator()( const int& , const int& l, const int& ) const {
    return l;
  }
};

struct func_max_lr_phiR : skeleton::ternary_function< int, int, int, int > {
  int operator()( const int&, const int& r, const int& ) const {
    return r;
  }
};

struct func_max_lr_G : skeleton::ternary_function< int, int, int, int > {
  int operator()( const int& p, const int& l, const int& r ) const {
    return std::max( p, std::max( l, r ) );
  }
};

////////////////////////////////////////////////////////////////////////
// main function

int SketoMain( int argc, char *argv[] ) {
  if ( argc <= 0 + 1 ) {
    throw skeleton::error( 1, "usage: height1 <input file>" );
  }

  const char* filename_in = argv[ 1 ];

  dist_tree< char > *tree1 = dist_tree< char >::read_from_file( filename_in );
  assert( tree1 );

  /* tree2 = map one one tree1
   *         where one x = 1
   */
  dist_tree< int > *tree2 = tree_skeletons::map( func_one( ), func_one( ),
						 tree1 );
  
  /* tree3 = dAcc (+) id id 0 tree2
   *         where id x = x
   */
  dist_tree< int > *tree3 = tree_skeletons::dAcc( func_plus( ), 0,
						  func_id( ), func_id( ),
						  0, tree2 );

  /* result = reduce id max_lr tree3
   *          where max_lr n l r = l `max` r
   */
  int result = tree_skeletons::reduce( func_id( ), func_max_lr( ),
				       func_max_lr_psi( ),
				       func_max_lr_phiL( ), func_max_lr_phiR( ),
				       func_max_lr_G( ), tree3 );

  skeleton::cout << "height of the input tree is " << result << std::endl;

  skeleton::safe_delete( tree3 );
  skeleton::safe_delete( tree2 );
  skeleton::safe_delete( tree1 );

  return 0;
}
